#include <stdlib.h>
#include <string.h>
#include "net.h"
#include "bridgewin.h"
#include "xsw.h"  


static void BW_SET_INPUT_AREA(
	xsw_bridge_win_struct *bw,
	int btnpos_num,
	int x, int y,
	unsigned int width, unsigned int height
);

void BridgeWinScannerScaleValueChangedCB(
	scale_bar_struct *scale, float value, void *data
);
void BridgeWinDrawMessagesCB(void *widget, void *data);
void BridgeWinSBValueChangedCB(
	scroll_bar_struct *sb, float x, float y, void *data
);

void BridgeWinSetTitle(
	xsw_bridge_win_struct *bw, 
	const char *address,  
	const char *title,  
	const char *player_name  
);

xsw_bridge_win_struct *BridgeWinNew(void);
void BridgeWinResize(xsw_bridge_win_struct *bw);
void BridgeWinResizePreset(xsw_bridge_win_struct *bw, int step);
void BridgeWinMap(xsw_bridge_win_struct *bw);
void BridgeWinUnmap(xsw_bridge_win_struct *bw);
void BridgeWinDelete(xsw_bridge_win_struct *bw);


#define RADIANS_TO_DEGREES(r) ((r) * 180 / PI)
			

/*
 *	Bridge window sizes:
 */
/* Standard bridge window size */
#define BW_DEF_WIDTH			800
#define BW_DEF_HEIGHT			600

/* Size of each console panel */
#define BRIDGE_PANEL_WIDTH		150
#define BRIDGE_PANEL_HEIGHT		130
/* Height of last (lowest) console panel */
#define BRIDGE_PANEL_LAST_HEIGHT	300

/* Message box height */
#define BW_MESGBOX_HEIGHT		(BW_DEF_HEIGHT - (BRIDGE_PANEL_HEIGHT * 4))

/* Viewscreen minimum size */
#define BW_VS_MIN_WIDTH			100
#define BW_VS_MIN_HEIGHT		100

/* Bridge window minimum size */
#define BW_MIN_WIDTH			((2 * BRIDGE_PANEL_WIDTH) + BW_VS_MIN_WIDTH)
#define BW_MIN_HEIGHT			(BW_VS_MIN_HEIGHT + BW_MESGBOX_HEIGHT)

/* Bridge window maximum size */
#define BW_MAX_WIDTH			(gui->display_width)
#define BW_MAX_HEIGHT			(gui->display_height)


#define FTOI_RNDOFF(x)	((int)((x) + 0.5f))

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Sets the position values for the button positiion number.
 */
static void BW_SET_INPUT_AREA(
	xsw_bridge_win_struct *bw, int btnpos_num,
	int x, int y,
	unsigned int width, unsigned int height
)
{
	xsw_input_area_struct *bp;

	if((btnpos_num < 0) || (btnpos_num >= bw->ninput_areas) ||
	   (bw->input_area == NULL)
	)
	    return;

	bp = bw->input_area[btnpos_num];
	if(bp == NULL)
	    return;

	bp->x = x;
	bp->y = y;
	bp->width = width;
	bp->height = height;
}


/*
 *	Scanner scale bar value changed callback.
 */
void BridgeWinScannerScaleValueChangedCB(
        scale_bar_struct *scale, float value, void *data
)
{
	const time_t	cur_ms = cur_millitime,
			cur_sec = cur_systime;
	const float tc = time_compensation;
        int player_obj_num;
        xsw_bridge_win_struct *bw = XSW_BRIDGE_WIN(data);
        if((scale == NULL) || (bw == NULL))
            return;

        player_obj_num = net_parms.player_obj_num;

        /* Get new zoom */
        bw->scanner_zoom = CLIP(
	    ((float)value / 100.0f),
	    0.001f, 1.0f
	);

        /* Recreate scanner labels */
        ScannerUpdateLabels(bw, player_obj_num);

        /* Redraw scanner */  
        ScannerDraw(
            bw, player_obj_num,
            bw->scanner,
            bw->scanner_image,
            0,
            0,
            bw->scanner_zoom,
	    cur_ms, cur_sec, tc
        );
}

/*
 *	Bridge window draw messages callback.
 */
void BridgeWinDrawMessagesCB(void *widget, void *data)
{
	xsw_bridge_win_struct *bw = (xsw_bridge_win_struct *)data;
	if(bw == NULL)
	    return;

	if(bw->map_state)
	    BridgeDrawMessages(bw);
}

/*
 *	Bridge window scroll bar value changed callback.
 */
void BridgeWinSBValueChangedCB(
	scroll_bar_struct *sb, float x, float y, void *data
)
{
	xsw_bridge_win_struct *bw = (xsw_bridge_win_struct *)data;
	if(bw == NULL)
	    return;

	if(&bw->mesg_box_sb == sb)
	{
	    if(bw->map_state)
		BridgeQueueDrawMessages(bw);
	}
}


/*
 *	Sets the bridge window's title.
 */
void BridgeWinSetTitle(
	xsw_bridge_win_struct *bw, 
	const char *address,  
	const char *title,  
	const char *player_name  
)
{
	char *s;
	win_t w;

	if(bw == NULL)
	    return;

	w = bw->toplevel;
	if(w == 0)
	    return;

	if(!STRISEMPTY(address) || !STRISEMPTY(title) ||
	   !STRISEMPTY(player_name)
	)
	{
	    if(!STRISEMPTY(title))
	    {
		if(!STRISEMPTY(player_name))
		{
		    s = (char *)malloc(
			strlen(PROG_NAME_FULL) + strlen(player_name) +
			strlen(title) + 80
		    );
		    sprintf(
			s,
			"%s: %s - %s",
			PROG_NAME_FULL, player_name, title
		    );
		}
		else
		{
		    s = (char *)malloc(
			strlen(PROG_NAME_FULL) + strlen(title) + 80
		    );
		    sprintf(
			s,
			"%s: %s",
			PROG_NAME_FULL, title
		    );
		}
	    }
	    else if(!STRISEMPTY(address))
	    {
		if(!STRISEMPTY(player_name))
		{
		    s = (char *)malloc(
			strlen(PROG_NAME_FULL) + strlen(player_name) +
			strlen(address) + 80  
		    );
		    sprintf(
			s,
			"%s: %s - %s",
			PROG_NAME_FULL, player_name, address
		    );
		}
		else
		{
		    s = (char *)malloc(
			strlen(PROG_NAME_FULL) + strlen(address) + 80  
		    );
		    sprintf(
			s,
			"%s: %s",
			PROG_NAME_FULL, address
		    );
		}
	    }
	    else
	    {
		s = (char *)malloc(strlen(PROG_NAME_FULL) + 80);
		sprintf(
		    s,
		    "%s: Untitled",
		    PROG_NAME_FULL
		);
	    }
	}
	else
	{
	    s = (char *)malloc(strlen(PROG_NAME_FULL) + 80);
	    sprintf(
		s,
		"%s: Untitled",
		PROG_NAME_FULL
	    );
	}

	OSWSetWindowTitle(w, s);
	free(s);
}


/*
 *	Creates a new bridge window.
 */
xsw_bridge_win_struct *BridgeWinNew(void)
{
	int i, x, y, cwidth, ascent, descent, font_height;
	unsigned int width, height;
	font_t *font;
	pixmap_t pixmap;
	image_t *img;
	shared_image_t *simg;
	prompt_struct *prompt;
	scale_bar_struct *scale;
	scroll_bar_struct *sb;
	osw_gui_struct *gui = &osw_gui[0];
	win_t w, parent = gui->root_win;
	xsw_option_struct *opt = &option;
	xsw_bridge_win_struct *bw = XSW_BRIDGE_WIN(calloc(
	    1, sizeof(xsw_bridge_win_struct)
	));
	if(bw == NULL)
	    return(NULL);

	bw->map_state = False;
	bw->has_focus = False;
	if(gui->def_geometry_set)
	{
	    bw->x = gui->def_toplevel_x;
	    bw->y = gui->def_toplevel_y;
	    bw->width = MAX(
		gui->def_toplevel_width,
	        BW_MIN_WIDTH
	    );
	    bw->height = MAX(
		gui->def_toplevel_height,
		BW_MIN_HEIGHT
	    );
	}
	else
	{
	    bw->x = opt->bridge_x;
	    bw->y = opt->bridge_y;
	    bw->width = opt->bridge_width;
	    bw->height = opt->bridge_height;
	    if(bw->width < BW_MIN_WIDTH)
		bw->width = BW_DEF_WIDTH;
	    if(bw->height < BW_MIN_HEIGHT)
		bw->height = BW_DEF_HEIGHT;
	}

	bw->viewscreen_map_state = False;

	bw->visibility_state = VisibilityFullyObscured;
	bw->viewscreen_width = (unsigned int)MAX(
	    ((int)bw->width - (2 * BRIDGE_PANEL_WIDTH)),
	    2
	);
	bw->viewscreen_height = (unsigned int)MAX(
	    ((int)bw->height - BW_MESGBOX_HEIGHT),
	    2
	);

	bw->preset_zoom_code = 0;

	bw->viewscreen_zoom = XSW_VS_ZOOM_MIN;
	bw->viewscreen_gamma = 0.0f;	/* Start gamma at 0 */

	bw->vs_weapon_width = 200;
	bw->vs_weapon_height = 50;

	bw->net_stats_width = 400;
	bw->net_stats_height = 20;

	bw->pan_p1_width = BRIDGE_PANEL_WIDTH;
	bw->pan_p1_height = BRIDGE_PANEL_HEIGHT;
	bw->pan_p2_width = BRIDGE_PANEL_WIDTH;
	bw->pan_p2_height = BRIDGE_PANEL_HEIGHT;
	bw->pan_p3_width = BRIDGE_PANEL_WIDTH;
	bw->pan_p3_height = BRIDGE_PANEL_HEIGHT;
	bw->pan_p4_width = BRIDGE_PANEL_WIDTH;
	bw->pan_p4_height = BRIDGE_PANEL_LAST_HEIGHT;

	bw->pan_p2_current_frame = 0;

	bw->scanner_width = BRIDGE_PANEL_WIDTH;
	bw->scanner_height = BRIDGE_PANEL_HEIGHT;
	bw->scanner_zoom = 0.8f;
	bw->scanner_orient = opt->def_scanner_orient;

	bw->pan_s1_width = BRIDGE_PANEL_WIDTH;
	bw->pan_s1_height = BRIDGE_PANEL_HEIGHT;
	bw->pan_s2_width = BRIDGE_PANEL_WIDTH;
	bw->pan_s2_height = BRIDGE_PANEL_HEIGHT;
	bw->pan_s3_width = BRIDGE_PANEL_WIDTH;
	bw->pan_s3_height = BRIDGE_PANEL_LAST_HEIGHT;

	bw->pan_s1_current_frame = 0;   

	bw->mesg_box_x = 0;
	bw->mesg_box_y = (int)bw->height - BW_MESGBOX_HEIGHT;
	bw->mesg_box_width = bw->width;
	bw->mesg_box_height = BW_MESGBOX_HEIGHT;

	bw->mesg_box_font = font = xsw_font.con_message;
	OSWGetCharBounds(
	    font, 'X',
	    NULL, NULL, &cwidth, &ascent, &descent
	);
	font_height = ascent + descent;
	bw->mesg_box_line_spacing = MAX(font_height, 1);
	bw->mesg_box_char_width = MAX(cwidth, 1);
	bw->mesg_box_longest_line = 0;

	bw->mesg_box_sel_in_progress = False; 
	bw->mesg_box_sel_start_x = 0; 
	bw->mesg_box_sel_start_y = 0; 

	/* Set up input areas */
	bw->ninput_areas = BRIDGE_INPUT_AREAS;
	bw->input_area = NULL;
	if(bw->ninput_areas > 0)
	{
	    bw->input_area = (xsw_input_area_struct **)realloc(
		bw->input_area,
		bw->ninput_areas * sizeof(xsw_input_area_struct *)
	    );
	    if(bw->input_area == NULL)
	    {
		bw->ninput_areas = 0;
		return(bw);
	    }
	}
	for(i = 0; i < bw->ninput_areas; i++)
	{
	    bw->input_area[i] = (xsw_input_area_struct *)calloc(
		1, sizeof(xsw_input_area_struct)
	    );
	}

	/* Set button positions */
	x = 5;
	y = 25;
	width = 145;
	height = 6;
	BW_SET_INPUT_AREA(
	    bw, BRIDGE_INPUT_AREA_PSHIELDFREQ_UP,
	    x, y, width, height
	);
	y += (int)height;
	BW_SET_INPUT_AREA(
	    bw, BRIDGE_INPUT_AREA_PSHIELDFREQ_DOWN,
	    x, y, width, height
	);

	x = 5;
	y = 37;
	BW_SET_INPUT_AREA( 
	    bw, BRIDGE_INPUT_AREA_PWEAPONFREQ_UP,
	    x, y, width, height
	);
	y += (int)height;
	BW_SET_INPUT_AREA(
	    bw, BRIDGE_INPUT_AREA_PWEAPONFREQ_DOWN,
	    x, y, width, height
	);

	x = 5;
	y = 49;
	BW_SET_INPUT_AREA(
	    bw, BRIDGE_INPUT_AREA_PCOMCHANNEL_UP,
	    x, y, width, height
	);
	y += (int)height;
	BW_SET_INPUT_AREA(
	    bw, BRIDGE_INPUT_AREA_PCOMCHANNEL_DOWN,
	    x, y, width, height
	);

	BW_SET_INPUT_AREA(
	    bw, BRIDGE_INPUT_AREA_PSHIELDS,
	    6, 62, 8, 50
	);
	BW_SET_INPUT_AREA(
	    bw, BRIDGE_INPUT_AREA_PCLOAK,
	    136, 62, 8, 50
	);
	BW_SET_INPUT_AREA(
	    bw, BRIDGE_INPUT_AREA_PDMGCTL,
	    26, 122, 43, 8
	);

	BW_SET_INPUT_AREA(
	    bw, BRIDGE_INPUT_AREA_PINTERCEPT,
	    3, 0, 144, 12
	);
	BW_SET_INPUT_AREA(
	    bw, BRIDGE_INPUT_AREA_PLOCK,
	    3, 14, 144, 12
	);
	BW_SET_INPUT_AREA(			/* On scanner readout console 1 */
	    bw, BRIDGE_INPUT_AREA_PLOCKNEXT,
	    0, 0, 150, 150
	);

	BW_SET_INPUT_AREA(
	    bw, BRIDGE_INPUT_AREA_PWEPYIELD,
	    3, 28, 144, 14
	);
	x = 7;
	y = 42;
	width = 144;
	height = 12;
	BW_SET_INPUT_AREA(
	    bw, BRIDGE_INPUT_AREA_PSELWEP1,
	    x, y, width, height
	);
	y += (int)height;
	BW_SET_INPUT_AREA(
	    bw, BRIDGE_INPUT_AREA_PSELWEP2,
	    x, y, width, height
	);
	y += (int)height;
	BW_SET_INPUT_AREA(
	    bw, BRIDGE_INPUT_AREA_PSELWEP3,
	    x, y, width, height
	);
	y += (int)height;
	BW_SET_INPUT_AREA(
	    bw, BRIDGE_INPUT_AREA_PSELWEP4,
	    x, y, width, height
	);
	y += (int)height;
	BW_SET_INPUT_AREA(
	    bw, BRIDGE_INPUT_AREA_PSELWEP5,
	    x, y, width, height
	);
	y += (int)height;
	BW_SET_INPUT_AREA(
	    bw, BRIDGE_INPUT_AREA_PSELWEP6,
	    x, y, width, height
	);
	y += (int)height;
	BW_SET_INPUT_AREA(
	    bw, BRIDGE_INPUT_AREA_PSELWEP7,
	    x, y, width, height
	);      
	y += (int)height;
	BW_SET_INPUT_AREA(
	    bw, BRIDGE_INPUT_AREA_PSELWEP8,
	    x, y, width, height
	);
	y += (int)height;
	BW_SET_INPUT_AREA(
	    bw, BRIDGE_INPUT_AREA_PSELWEP9,
	    x, y, width, height
	);

	BW_SET_INPUT_AREA(
	    bw, BRIDGE_INPUT_AREA_PENGINESTATE,
	    75, 92, 55, 30
	);

	bw->target_string = osw_atom.target_string;
	bw->dnd_action = None;
	bw->dnd_tar_w = None;
	bw->dnd_x = 0;
	bw->dnd_y = 0;   


	/* Create toplevel */
	x = bw->x;
	y = bw->y;
	width = bw->width;
	height = bw->height;
	OSWCreateWindow(&w, parent, x, y, width, height);
	bw->toplevel = w;
	OSWSetWindowInput(
	    w,
	    StructureNotifyMask |
	    KeyPressMask | KeyReleaseMask |
	    FocusChangeMask
	);

	/* WM properties */
	if(ImgIsAllocated(XSW_IMG_ICON_XSW))
	    pixmap = OSWCreatePixmapFromImage(
		xsw_image[XSW_IMG_ICON_XSW]->image
	    );
	else
	    pixmap = widget_global.std_icon_pm;
	OSWSetWindowWMProperties(
	    w,
	    PROG_NAME_FULL,
	    PROG_NAME,			/* Icon title */
	    pixmap,			/* Icon */
	    False,			/* Let WM set coordinates? */
	    x, y,			/* Position */
	    BW_MIN_WIDTH, BW_MIN_HEIGHT,	/* Min size */
	    BW_MAX_WIDTH, BW_MAX_HEIGHT,	/* Max size */
	    OSW_WINDOW_FRAME_STANDARD,
	    0, NULL
	);
	OSWDNDSetWindowTarget(w);
	parent = w;

	/* Create the viewscreen */
	x = BRIDGE_PANEL_WIDTH;
	y = 0;
	width = bw->viewscreen_width;
	height = bw->viewscreen_height;
	OSWCreateWindow(&w, parent, x, y, width, height);
	bw->viewscreen = w;

	OSWCreateSharedImage(&simg, width, height);
	bw->viewscreen_image = simg;
	bw->viewscreen_width = simg->width;
	bw->viewscreen_height = simg->height;


	/* Create the weapon stats image and pixmap */
	width = bw->vs_weapon_width;
	height = bw->vs_weapon_height;
	OSWCreateImage(&img, width, height);
	bw->vs_weapon_image = img;
	bw->vs_weapon_width = width = img->width;
	bw->vs_weapon_height = height = img->height;
	OSWCreatePixmap(&pixmap, width, height);
	bw->vs_weapon_buf = pixmap;

	/* Create the netstats image and pixmap */
	width = bw->net_stats_width;
	height = bw->net_stats_height;
	OSWCreateImage(&img, width, height);
	bw->net_stats_image = img;
	bw->net_stats_width = width = img->width;
	bw->net_stats_height = height = img->height;
 	OSWCreatePixmap(&pixmap, width, height);
	bw->net_stats_buf = pixmap;

	/* Console 2 image */
	width = bw->pan_p2_width;
	height = bw->pan_p2_height;
	OSWCreateSharedImage(&simg, width, height);
	bw->pan_p2_img = simg;

	/* Console 3 image */
	width = bw->pan_p3_width;
	height = bw->pan_p3_height;
	OSWCreateSharedImage(&simg, width, height);
	bw->pan_p3_img = simg;

	/* Subject console 1 image */
	width = bw->pan_s1_width;
	height = bw->pan_s1_height;
	OSWCreateSharedImage(&simg, width, height);
	bw->pan_s1_img = simg;

	/* Subject console 2 image */
	width = bw->pan_s2_width;
	height = bw->pan_s2_height;
	OSWCreateSharedImage(&simg, width, height);
	bw->pan_s2_img = simg;


	/* Player console 1 */
	x = 0;
	y = 0;
	width = BRIDGE_PANEL_WIDTH;
	height = BRIDGE_PANEL_HEIGHT;
	OSWCreateWindow(&w, parent, x, y, width, height);
	bw->pan_p1 = w;
	OSWSetWindowInput(
	    w,
	    ExposureMask | 
	    ButtonPressMask | ButtonReleaseMask  
	);

	/* Player console 2 */
	y += (int)height;
	OSWCreateWindow(&w, parent, x, y, width, height);
	bw->pan_p2 = w;
	OSWSetWindowInput(
	    w,
	    ExposureMask | 
	    ButtonPressMask | ButtonReleaseMask  
	);

	/* Player console 3 */
	y += (int)height;
	OSWCreateWindow(&w, parent, x, y, width, height);
	bw->pan_p3 = w;
	OSWSetWindowInput(
	    w,
	    ExposureMask | 
	    ButtonPressMask | ButtonReleaseMask  
	);

	/* Player console 4 */
	y += (int)height;
	bw->pan_p4_height = height = (unsigned int)MAX(
	    ((int)bw->height - y - (int)bw->mesg_box_height),
	    1
	);
	OSWCreateWindow(&w, parent, x, y, width, height);
	bw->pan_p4 = w;
	OSWSetWindowInput(
	    w,
	    ExposureMask |
	    ButtonPressMask | ButtonReleaseMask
	);


	/* Reset weapon yield scale input window */
	bw->weapon_yield_iwin = None;

	/* Create the scanner */
	x = MAX(((int)bw->width - BRIDGE_PANEL_WIDTH), 0);
	y = 0;
	width = BRIDGE_PANEL_WIDTH;
	height = BRIDGE_PANEL_HEIGHT;
	OSWCreateWindow(&w, parent, x, y, width, height);
	bw->scanner = w;
	OSWSetWindowInput(
	    w,
	    ExposureMask |
	    ButtonPressMask | ButtonReleaseMask |
	    VisibilityChangeMask
	);
	WidgetSetWindowCursor(w, xsw_cursor.scanner_lock);

	OSWCreateSharedImage(&simg, width, height);
	bw->scanner_image = simg;
	bw->scanner_width = simg->width;
	bw->scanner_height = simg->height;

	/* Scanner scale bar widget */
	scale = &bw->scanner_sb;
	ScaleBarInit(
	    scale, bw->scanner,
	    (int)width - SCALEBAR_BAR_FLUSHED_WIDTH,
	    0,
	    height,			/* Length */
	    SCALEBAR_STYLE_FLUSHED,	/* Style */
	    5,				/* Ticks */
	    SCALEBAR_ORIENT_VERTICAL,	/* Orientation */
	    0, 100, 80,			/* Pos min, max, start */
	    True			/* Flipped */
	);
	scale->value_changed_cb = BridgeWinScannerScaleValueChangedCB;
	scale->value_changed_data = bw;


	/* Subject console 1 */
	y += (int)height;
	OSWCreateWindow(&w, parent, x, y, width, height);
	bw->pan_s1 = w;
	OSWSetWindowInput(
	    w,
	    ExposureMask |
	    ButtonPressMask | ButtonReleaseMask
	);

	/* Subject console 2 */
	y += (int)height;
	OSWCreateWindow(&w, parent, x, y, width, height);
	bw->pan_s2 = w;
	OSWSetWindowInput(
	    w,
	    ExposureMask |
	    ButtonPressMask | ButtonReleaseMask
	);

	/* Subject console 3 */
	y += (int)height;
	bw->pan_s3_height = height = (unsigned int)MAX(
	    ((int)bw->height - y - (int)bw->mesg_box_height),
	    1
	);
	OSWCreateWindow(&w, parent, x, y, width, height);
	bw->pan_s3 = w;
	OSWSetWindowInput(
	    w,
	    ExposureMask |
	    ButtonPressMask | ButtonReleaseMask
	);

	/* Create the console buffer (used for all bridge panel
	 * consoles)
	 */
	width = MAX(bw->pan_s3_width, BRIDGE_PANEL_WIDTH);
	height = MAX(bw->pan_s3_height, BRIDGE_PANEL_HEIGHT);
	OSWCreatePixmap(&pixmap, width, height);
	bw->pan_buf = pixmap;
	OSWClearPixmap(pixmap, width, height, gui->black_pix);


	/* Create the message box */
	x = bw->mesg_box_x;
	y = bw->mesg_box_y;
	width = bw->mesg_box_width;
	height = bw->mesg_box_height;
	OSWCreateWindow(&w, parent, x, y, width, height);
	bw->mesg_box = w;
	OSWSetWindowInput(
	    w,
	    ExposureMask |
	    ButtonPressMask | ButtonReleaseMask |
	    PointerMotionMask |
	    EnterWindowMask | LeaveWindowMask
	);
	WidgetSetWindowCursor(w, xsw_cursor.text);

	OSWCreatePixmap(&pixmap, width, height);
	bw->mesg_box_buf = pixmap;

	/* Message box scroll bar */
	sb = &bw->mesg_box_sb;
	SBarInit(
	    sb, bw->mesg_box,
	    width, height,
	    width,
	    bw->mesg_box_line_spacing * (total_messages + 1)
	);
	sb->x_step_inc = FTOI_RNDOFF((float)width * 0.10f);
	sb->y_step_inc = bw->mesg_box_line_spacing;
	SBarSetPositionX(sb, 0.0f);
	SBarSetPositionY(sb, (float)sb->total_height);
	sb->value_changed_cb = BridgeWinSBValueChangedCB;
	sb->value_changed_data = bw;
	

	/* Prompt */
	prompt = &bw->prompt;
	x = bw->pan_p1_width + 10;
	y = MAX(
	    ((int)bw->viewscreen_height - 70),
	    0
	);
	width = (unsigned int)MAX(
	    (int)bw->viewscreen_width - (2 * 10),
	    100
	);
	height = 30;
	PromptInit(
	    prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_RAISED,
	    "Command:",
	    XSW_DEF_PROMPT_BUF_LEN, XSW_DEF_PROMPT_HIST_BUFS
	);


	/* Initialize pages */
	PageInit(
	    &bw->main_page,
	    bw->viewscreen,
	    bw->viewscreen_image
	);

	PageInit(
	    &bw->destroyed_page,
	    bw->viewscreen,
	    bw->viewscreen_image
	);

	/* Select the main page */
	bw->cur_page = &bw->main_page;


	/* Force/trick unmap both the viewscreen and the current page
	 *
	 * This is to initialize certain resources on them (ie the
	 * viewscreen input event selection and cursor)
	 */
	bw->viewscreen_map_state = True;
	VSUnmap(bw);

	bw->cur_page->map_state = True;
	PageUnmap(
	    bw->cur_page,
	    bw->viewscreen,
	    bw->viewscreen_image
	);

	/* Set title */
	BridgeWinSetTitle(bw, NULL, NULL, NULL);

	return(bw);
}

/*
 *	Resizes the bridge window.
 */
void BridgeWinResize(xsw_bridge_win_struct *bw)
{
	int x, y;
	unsigned int width, height;
	pixmap_t pixmap;
	page_struct *page;
	prompt_struct *prompt;
	scale_bar_struct *scale;
	scroll_bar_struct *sb;
	osw_gui_struct *gui = &osw_gui[0];

#define DESTROY_PIXMAP		OSWDestroyPixmap
#define DESTROY_SHARED_IMAGE	OSWDestroySharedImage

	if(bw == NULL)
	    return;

	OSWGetWindowGeometry(
	    bw->toplevel, &x, &y, &width, &height, NULL
	);

	bw->x = x;
	bw->y = y;

	/* No change in size? */
	if((bw->width == width) && (bw->height == height))
	    return;

	bw->width = width;
	bw->height = height;


	/* Resize viewscreen */

	/* Wait for any draw procedures to finish */
	OSWSyncSharedImage(
	    bw->viewscreen_image,
	    bw->viewscreen
	);

	x = BRIDGE_PANEL_WIDTH;
	y = 0;
	width = (unsigned int)MAX(
	    (int)bw->width - (2 * BRIDGE_PANEL_WIDTH),
	    1
	);
	height = (unsigned int)MAX(
	    (int)bw->height - (int)bw->mesg_box_height,
	    1
	);
	OSWMoveResizeWindow(bw->viewscreen, x, y, width, height);
	bw->viewscreen_width = width;
	bw->viewscreen_height = height;

	/* Recreate the viewscreen's shared image */
	DESTROY_SHARED_IMAGE(&bw->viewscreen_image);
	OSWCreateSharedImage(&bw->viewscreen_image, width, height);


	/* Resize the current page viewscreen */
	page = bw->cur_page;
	if(page != NULL)
	{
	    PageResize(page, bw->viewscreen, bw->viewscreen_image);
	    if(page->map_state)
		PageDraw(
		    page,
		    bw->viewscreen, bw->viewscreen_image,
		    PAGE_DRAW_AMOUNT_COMPLETE,
		    True		/* Put to window */
		);
	}

	/* Player consoles */
	x = 0;
	y = 0;
	width = BRIDGE_PANEL_WIDTH;
	height = BRIDGE_PANEL_HEIGHT;
	OSWMoveResizeWindow(bw->pan_p1, x, y, width, height);

	y += (int)height;
	OSWMoveResizeWindow(bw->pan_p2, x, y, width, height);

	y += (int)height;
	OSWMoveResizeWindow(bw->pan_p3, x, y, width, height);

	y += (int)height;
	bw->pan_p4_height = height = (unsigned int)MAX(
	    ((int)bw->height - y - (int)bw->mesg_box_height),
	    1
	);
	OSWMoveResizeWindow(bw->pan_p4, x, y, width, height);


	/* Scanner and subject consoles */
	x = MAX(((int)bw->width - BRIDGE_PANEL_WIDTH), 0);
	y = 0;
	width = BRIDGE_PANEL_WIDTH;
	height = BRIDGE_PANEL_HEIGHT;
	OSWMoveResizeWindow(bw->scanner, x, y, width, height);

	scale = &bw->scanner_sb;
	OSWMoveResizeWindow(
	    scale->toplevel,
	    (int)width - SCALEBAR_BAR_FLUSHED_WIDTH,
	    0,
	    SCALEBAR_BAR_FLUSHED_WIDTH,
	    height
	);
	WidgetQueueResize(scale);

	y += (int)height;
	OSWMoveResizeWindow(bw->pan_s1, x, y, width, height);

	y += (int)height;
	OSWMoveResizeWindow(bw->pan_s2, x, y, width, height);

	y += (int)height;
	bw->pan_s3_height = height = (unsigned int)MAX(
	    ((int)bw->height - y - (int)bw->mesg_box_height),
	    1
	);
	OSWMoveResizeWindow(bw->pan_s3, x, y, width, height);


	/* Recreate the console buffer (used for all bridge panel
	 * consoles)
	 */
	DESTROY_PIXMAP(&bw->pan_buf);
	width = MAX(bw->pan_s3_width, BRIDGE_PANEL_WIDTH);
	height = MAX(bw->pan_s3_height, BRIDGE_PANEL_HEIGHT);
	OSWCreatePixmap(&pixmap, width, height);
	bw->pan_buf = pixmap;
	OSWClearPixmap(pixmap, width, height, gui->black_pix);


	/* Message box */
	x = 0;
	y = bw->viewscreen_height;
	width = bw->width;
	height = bw->mesg_box_height;
	OSWMoveResizeWindow(bw->mesg_box, x, y, width, height);
	bw->mesg_box_x = x;
	bw->mesg_box_y = y;
	bw->mesg_box_width = width;
	bw->mesg_box_height = height;
	DESTROY_PIXMAP(&bw->mesg_box_buf);
	OSWCreatePixmap(&bw->mesg_box_buf, width, height);

	/* Resize the message box's scroll bars */
	sb = &bw->mesg_box_sb;
	SBarQueueResize(
	    sb,
	    width, height,
	    width,
/*
	    (bw->mesg_box_longest_line * bw->mesg_box_char_width) +
		SCROLLBAR_YBAR_WIDTH + (2 * 10),
 */
	    bw->mesg_box_line_spacing * (total_messages + 1)
	);
	/* Scroll to end */
	SBarSetPositionY(sb, (float)sb->total_height);
	BridgeQueueDrawMessages(bw);


	/* Prompt */
	prompt = &bw->prompt;
	x = BRIDGE_PANEL_WIDTH + 10;
	y = MAX(
	    ((int)bw->viewscreen_height - 70),
	    0
	);
	width = (unsigned int)MAX(
	    (int)bw->viewscreen_width - (2 * 10),
	    100
	);
	height = 30;
	OSWMoveResizeWindow(prompt->toplevel, x, y, width, height);
	WidgetQueueResize(&bw->prompt);

#undef DESTROY_PIXMAP
#undef DESTROY_SHARED_IMAGE
}

/*
 *	Changes the size of the bridge window to one of
 *	the `preset/preffered' sizes, then calls
 *	BridgeWinResize().
 *
 *	Updates the preset_zoom_code to value of step on
 *	bridge_win structure.
 */
void BridgeWinResizePreset(xsw_bridge_win_struct *bw, int step)
{
/*	const float aspect = 800.0f / 600.0f; */
	const float aspect = 1.0f;
	const unsigned int	min_width = BW_VS_MIN_WIDTH +
	    (BRIDGE_PANEL_WIDTH * 2),
				min_height = BW_VS_MIN_HEIGHT +
	    BW_MESGBOX_HEIGHT;
	int x, y;
	unsigned int width, height, prev_width, prev_height;

	if(bw == NULL)
	    return;

	/* Sanitize step code */
	if(step < -2)
	    step = -2;
	if(step > 0)
	    step = 0;

#if 0
	/* Check if there was no change */
	if(bw->preset_zoom_code == step)
	    return;
#endif

	/* Calculate size by the step code */
	switch(step)
	{
	  case -2:
	    height = MAX(
		(unsigned int)(
		    (2 * (int)bw->pan_p1_height) +
		    (int)bw->mesg_box_height
		),
		min_height
	    );
	    width = MAX(
		(unsigned int)(
		    (((int)height - (int)bw->mesg_box_height) * aspect) +
		    (2 * bw->pan_p1_width)
		),
		min_width
	    );
	    break;

	  case -1:
	    height = MAX(
		(unsigned int)(
		    (3 * (int)bw->pan_p1_height) +
		    (int)bw->mesg_box_height
		),
		min_height
	    );
	    width = MAX(
		(unsigned int)(
		    (((int)height - (int)bw->mesg_box_height) * aspect) +
		    (2 * bw->pan_p1_width)
		),
		min_width
	    );
	    break;

	  default:	/* Default to 0, nominal/standard */
	    width = BW_DEF_WIDTH;
	    height = BW_DEF_HEIGHT;
	    break;
	}

	/* Get the current position and size of the toplevel window */
	OSWGetWindowGeometry(
	    bw->toplevel, NULL, NULL, &prev_width, &prev_height, NULL
	);
	if(OSWGetWindowRootOrigin(bw->toplevel, &x, &y))
	{
	    /* Move and resize to the new geometry */
	    x += ((int)prev_width - (int)width) / 2;
	    y += ((int)prev_height - (int)height) / 2;
	    OSWMoveResizeWindow(bw->toplevel, x, y, width, height);
	}
	else
	{
	    /* Resize to the new geometry */
	    OSWResizeWindow(bw->toplevel, width, height);
	}

	/* Update zoom code */
	bw->preset_zoom_code = step;

	/* Reposition windows and recreate buffers */
	BridgeWinResize(bw);

	BridgeWinDrawAll(bw);
}

/*
 *	Maps the bridge window.
 */
void BridgeWinMap(xsw_bridge_win_struct *bw)
{
	int tar_obj_num;
	const int obj_num = net_parms.player_obj_num;
	shared_image_t *simg;
	xsw_object_struct *obj = net_parms.player_obj_ptr;

	if(bw == NULL)
	    return;

	/* Map */
	OSWMapRaised(bw->toplevel);

	/* If already mapped then just raise the toplevel */
	if(bw->map_state)
	    return;

	OSWMapWindow(bw->pan_p1);
	OSWMapWindow(bw->pan_p2);
	OSWMapWindow(bw->pan_p3);
	OSWMapWindow(bw->pan_p4);
	OSWMapWindow(bw->pan_s1);
	OSWMapWindow(bw->pan_s2);
	OSWMapWindow(bw->pan_s3);

	bw->scanner_map_state = True;
	simg = bw->scanner_image;
	if(simg != NULL)
	    simg->in_progress = False;
	OSWMapWindow(bw->scanner);

	OSWMapWindow(bw->viewscreen);
	/* Do not set the viewscreen's map state here because it is for
	 * switching out of menus
	 */

	OSWMapWindow(bw->mesg_box);


	/* Scanner scale bar */
	WidgetMap(&bw->scanner_sb);
 
	/* Prompt window */
	WidgetUnmap(&bw->prompt);


	/* Redraw all panels */
	BridgeDrawPanel(obj_num, BRIDGE_PANEL_P1);
	BridgeDrawPanel(obj_num, BRIDGE_PANEL_P2);
	BridgeDrawPanel(obj_num, BRIDGE_PANEL_P3);
	BridgeDrawPanel(obj_num, BRIDGE_PANEL_P4);
	tar_obj_num = (obj != NULL) ? obj->locked_on : -1;
	BridgeDrawPanel(tar_obj_num, BRIDGE_PANEL_S1);
	BridgeDrawPanel(tar_obj_num, BRIDGE_PANEL_S2);
	BridgeDrawPanel(tar_obj_num, BRIDGE_PANEL_S3);

	/* Grab focus */
	OSWGrabFocus(bw->toplevel);
}

/*
 *	Unmaps the bridge window.
 */
void BridgeWinUnmap(xsw_bridge_win_struct *bw)
{
	if(bw == NULL)
	    return;

	OSWUnmapWindow(bw->toplevel);
}

/*
 *	Deletes the bridge window.
 */
void BridgeWinDelete(xsw_bridge_win_struct *bw)
{
	int i;

	if(bw == NULL)
	    return;

	/* Destroy all the widgets */
	WidgetDestroy(&bw->prompt);
	WidgetDestroy(&bw->mesg_box_sb);
	WidgetDestroy(&bw->scanner_sb);

	/* Destroy all the pages */
	PageDestroy(
	    &bw->main_page,
	    bw->viewscreen,
	    bw->viewscreen_image  
	);
	PageDestroy(
	    &bw->destroyed_page,
	    bw->viewscreen,
	    bw->viewscreen_image
	);

#define DESTROY_WINDOW		OSWDestroyWindow
#define DESTROY_PIXMAP		OSWDestroyPixmap
#define DESTROY_SHARED_IMAGE	OSWDestroySharedImage
  
	/* Delete the viewscreen & scanner image buffers */
	DESTROY_SHARED_IMAGE(&bw->viewscreen_image);
	DESTROY_SHARED_IMAGE(&bw->scanner_image);

	/* Delete the scanner label images */
	OSWDestroyImage(&bw->scanner_range_label);
	OSWDestroyImage(&bw->scanner_loc_label);


	/* Destroy Windows and Pixmaps */
 
	DESTROY_PIXMAP(&bw->mesg_box_buf);
	DESTROY_PIXMAP(&bw->pan_buf);

	DESTROY_SHARED_IMAGE(&bw->pan_p2_img);
	DESTROY_SHARED_IMAGE(&bw->pan_p3_img);
	DESTROY_SHARED_IMAGE(&bw->pan_s1_img);
	DESTROY_SHARED_IMAGE(&bw->pan_s2_img);

	DESTROY_PIXMAP(&bw->net_stats_buf);
	OSWDestroyImage(&bw->net_stats_image);

	DESTROY_PIXMAP(&bw->vs_weapon_buf);
	OSWDestroyImage(&bw->vs_weapon_image);

	DESTROY_WINDOW(&bw->mesg_box);

	DESTROY_WINDOW(&bw->weapon_yield_iwin);
	DESTROY_WINDOW(&bw->pan_p4);
	DESTROY_WINDOW(&bw->pan_p3);
	DESTROY_WINDOW(&bw->pan_p2);
	DESTROY_WINDOW(&bw->pan_p1);

	DESTROY_WINDOW(&bw->pan_s3);
	DESTROY_WINDOW(&bw->pan_s2);
	DESTROY_WINDOW(&bw->pan_s1);
	DESTROY_WINDOW(&bw->scanner);

	DESTROY_WINDOW(&bw->viewscreen);
	DESTROY_WINDOW(&bw->toplevel);

#undef DESTROY_WINDOW
#undef DESTROY_PIXMAP
#undef DESTROY_SHARED_IMAGE

	/* Delete the bridge panel button positions */
	for(i = 0; i < bw->ninput_areas; i++)
	    free(bw->input_area[i]);
	free(bw->input_area);
	bw->input_area = NULL;
	bw->ninput_areas = 0;

	free(bw);
}
