#include <stdlib.h>
#include <string.h>

#include "../include/reality.h"
#include "../include/unvmatch.h"
#include "../include/unvmath.h"
#include "../include/string.h"

#include "../include/swsoundcodes.h"

#include "keymaps.h"
#ifdef JS_SUPPORT
# include "jsmaps.h"
#endif
#include "gctl.h"
#include "sound.h"
#include "net.h"
#include "bridgewin.h"
#include "msgwin.h"
#include "vsmenu.h"
#include "xsw.h"
#include "xswaction.h"

#include "../include/oswkeyvals.h"


static bool_t BridgeIsCoordinatesInInputArea(
	xsw_bridge_win_struct *bw, const int input_area_num,
	const int x, const int y
);
static void BridgePlaySoundButtonClick(
	xsw_bridge_win_struct *bw,
	const float left_vol, const float right_vol
);
void BridgeMapPromptMode(
	xsw_bridge_win_struct *bw,
	const char *label, const char *val, const int mode
);

static void BridgeMessagesMarkCoordinates(
	xsw_bridge_win_struct *bw,
	int start_x, int start_y,
	int end_x, int end_y
);
static void BridgeMessagesScrollTimeoutCB(void *widget, void *data);
static void BridgeMessagesAddScrollTimeout(
	xsw_bridge_win_struct *bw, const time_t interval_ms
);
static void BridgeMessagesRemoveScrollTimeout(
	xsw_bridge_win_struct *bw
);

static void BridgePrintSubjectStats(
	xsw_bridge_win_struct *bw, const int obj_num
);

void BridgeWarnWeaponsOffline(
	xsw_bridge_win_struct *bw, const int obj_num
);

static int BridgeManagePrompt(
	xsw_bridge_win_struct *bw, event_t *event
);
int BridgeManage(
	xsw_bridge_win_struct *bw, event_t *event,
	const time_t cur_ms, const time_t cur_sec, const float tc
);


#define FTOI_RNDOFF(x)	((int)((x) + 0.5f))

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Checks if the specified coordinates fall within the input
 *	area specified by input_area_num.
 */
static bool_t BridgeIsCoordinatesInInputArea(
	xsw_bridge_win_struct *bw, const int input_area_num,
	const int x, const int y
)
{
	xsw_input_area_struct *a;

	if(bw == NULL)
	    return(False);

	if((input_area_num < 0) ||
	   (input_area_num >= bw->ninput_areas) ||
	   (bw->input_area == NULL)
	)
	    return(False);

	a = bw->input_area[input_area_num];
	if(a == NULL)
	    return(False);

	if((x >= a->x) &&
	   (y >= a->y) &&
	   (x < (a->x + (int)a->width)) &&
	   (y < (a->y + (int)a->height))
	)
	    return(True);
	else
	    return(False);
}


/*
 *	Plays the button press click sound.
 */
static void BridgePlaySoundButtonClick(
	xsw_bridge_win_struct *bw,
	const float left_vol, const float right_vol
)
{
	xsw_sound_struct *snd = &sound;
	xsw_option_struct *opt = &option;

	if(bw == NULL)
	    return;

	if(opt->sounds >= XSW_SOUNDS_EVENTS)
	    SoundPlay(
		snd, SW_SNDREF_BUTTON_PRESS,
		left_vol, right_vol,
		0, 0
	    );
}

/*
 *	Sets the Bridge Window's prompt mode and maps the prompt.
 */
void BridgeMapPromptMode(
	xsw_bridge_win_struct *bw,
	const char *label, const char *val, const int mode
)
{
	prompt_struct *prompt;

	if(bw == NULL)
	    return;

	prompt = &bw->prompt;

	if(label != NULL)
	    PromptSetLabel(prompt, label);

	PromptSetS(prompt, val);

	prompt_mode = mode;

	PromptMarkBuffer(prompt, PROMPT_POS_END);

	PromptMap(prompt);
	prompt->has_focus = True;
}


/*
 *	Marks the Bridge Window's messages with the specified
 *	window coordinates.
 *
 *	Scroll bar positions will be applied to the specified values.
 */
static void BridgeMessagesMarkCoordinates(
	xsw_bridge_win_struct *bw,
	int start_x, int start_y,
	int end_x, int end_y
)
{
	int	sb_x, sb_y,
		start_line, end_line,
		start_line_x, end_line_x,
		margin_x, margin_y,
		line_spacing, cwidth;
	scroll_bar_struct *sb;

	if(bw == NULL)
	    return;

	margin_x = 10;
	margin_y = 10;
	line_spacing = MAX(bw->mesg_box_line_spacing, 1);
	cwidth = MAX(bw->mesg_box_char_width, 1);
	sb = &bw->mesg_box_sb;
	sb_x = (int)SBarGetPositionX(sb);
	sb_y = (int)SBarGetPositionY(sb);

	/* Calculate start and end lines */
	start_line = CLIP(
	    (start_y - margin_y + sb_y) / line_spacing,
	    0, (total_messages - 1)
	);
	end_line = CLIP(
	    (end_y - margin_y + sb_y) / line_spacing,
	    0, (total_messages - 1)
	);

	/* Calculate the start and end line x positions (do not clip) */
	start_line_x = (start_x - margin_x + sb_x) / cwidth;
	end_line_x = (end_x - margin_x + sb_x) / cwidth;

#define MSG_SEL_SWAP_AS_NEEDED(_msg_)	{	\
 if((_msg_)->sel_start > (_msg_)->sel_end) {	\
  const int i = (_msg_)->sel_end;		\
  (_msg_)->sel_end = (_msg_)->sel_start;	\
  (_msg_)->sel_start = i;			\
} }

	/* Single line? */
	if(start_line == end_line)
	{
	    xsw_mesg_struct *msg = &mesg_list[start_line];
	    const int line_len = strlen(msg->message);
	    if(start_line_x < end_line_x)
	    {
		msg->sel_start = CLIP(start_line_x, 0, (line_len - 1));
		msg->sel_end = CLIP(end_line_x - 1, 0, (line_len - 1));
	    }
	    else if(start_line_x > end_line_x)
	    {
		msg->sel_start = CLIP(end_line_x, 0, (line_len - 1));
		msg->sel_end = CLIP(start_line_x - 1, 0, (line_len - 1));
	    }
	}
	/* Select multiple lines top to bottom? */
	else if(start_line < end_line)
	{
	    xsw_mesg_struct *msg = &mesg_list[start_line];
	    int i, line_len = strlen(msg->message);
	    if(start_line_x < line_len)
	    {
		msg->sel_start = CLIP(start_line_x, 0, (line_len - 1));
		msg->sel_end = line_len - 1;
		MSG_SEL_SWAP_AS_NEEDED(msg);
	    }

	    for(i = start_line + 1; i < end_line; i++)
	    {
		msg = &mesg_list[i];
		msg->sel_start = 0;
		msg->sel_end = MAX(strlen(msg->message) - 1, 0);
	    }

	    msg = &mesg_list[end_line];
	    line_len = strlen(msg->message);
	    if(end_line_x >= 0)
	    {
		msg->sel_start = 0;
		msg->sel_end = CLIP(end_line_x, 0, (line_len - 1));
		MSG_SEL_SWAP_AS_NEEDED(msg);
	    }
	}
	/* Select multiple lines bottom to top? */
	else if(end_line < start_line)
	{
	    xsw_mesg_struct *msg = &mesg_list[start_line];
	    int i, line_len = strlen(msg->message);
	    if(start_line_x >= 0)
	    {
		msg->sel_start = 0;
		msg->sel_end = CLIP(start_line_x - 1, 0, (line_len - 1));
		MSG_SEL_SWAP_AS_NEEDED(msg);
	    }

	    for(i = start_line - 1; i > end_line; i--)
	    {
		msg = &mesg_list[i];
		msg->sel_start = 0;
		msg->sel_end = MAX(strlen(msg->message) - 1, 0);
	    }

	    msg = &mesg_list[end_line];
	    line_len = strlen(msg->message);
	    if(end_line_x < line_len)
	    {
		msg->sel_start = CLIP(end_line_x, 0, (line_len - 1));
		msg->sel_end = MAX(strlen(msg->message) - 1, 0);
		MSG_SEL_SWAP_AS_NEEDED(msg);
	    }
	}
#undef MSG_SEL_SWAP_AS_NEEDED
}

/*
 *	Scroll messages timeout.
 */
static void BridgeMessagesScrollTimeoutCB(void *widget, void *data)
{
	int x, y;
	unsigned int width, height;
	win_t w;
	widget_global_struct *wglobal = &widget_global;
	xsw_bridge_win_struct *bw = XSW_BRIDGE_WIN(widget);
	if(bw == NULL)
	    return;

	w = bw->mesg_box;
	if(w == None)
	    return;

	OSWGetPointerCoords(w, NULL, NULL, &x, &y);
	OSWGetWindowGeometry(
	    w, NULL, NULL, &width, &height, NULL
	);

	if(y < 0)
	{
	    scroll_bar_struct *sb = &bw->mesg_box_sb;
	    const int	dy = 0 - y,
			prev_y_pos = (int)SBarGetPositionY(sb);
	    SBarSetPositionY(sb, (float)(prev_y_pos - dy));
	    bw->mesg_box_sel_start_y += prev_y_pos -
		(int)SBarGetPositionY(sb);
	    BridgeMessagesMarkCoordinates(
		bw,
		bw->mesg_box_sel_start_x,
		bw->mesg_box_sel_start_y,
		x, y
	    );
	}
	else if(y >= (int)height)
	{
	    scroll_bar_struct *sb = &bw->mesg_box_sb;
	    const int	dy = (y + 1 - (int)height),
			prev_y_pos = (int)SBarGetPositionY(sb);
	    SBarSetPositionY(sb, (float)(prev_y_pos + dy));
	    bw->mesg_box_sel_start_y += prev_y_pos - 
		(int)SBarGetPositionY(sb);
	    BridgeMessagesMarkCoordinates(
		bw,
		bw->mesg_box_sel_start_x,
		bw->mesg_box_sel_start_y,
		x, y
	    );
	}

	/* Schedual next timeout */
	BridgeMessagesAddScrollTimeout(
	    bw, wglobal->prompt_scroll_repeat_int
	);
}

/*
 *	Adds a scroll messages timeout.
 */
static void BridgeMessagesAddScrollTimeout(
	xsw_bridge_win_struct *bw, const time_t interval_ms
)
{
	if(bw == NULL)
	    return;

	WidgetTimeoutAdd(
	    BridgeMessagesScrollTimeoutCB,
	    bw, NULL,
	    interval_ms
	);
}

/*
 *	Removes any scroll messages timeouts.
 */
static void BridgeMessagesRemoveScrollTimeout(
	xsw_bridge_win_struct *bw
)
{
	if(bw == NULL)
	    return;

	WidgetTimeoutRemoveByFuncWidget(
	    BridgeMessagesScrollTimeoutCB, bw
	);
}


/*
 *	Prints stats of the subject object locked on scanner.
 */
static void BridgePrintSubjectStats(
	xsw_bridge_win_struct *bw, const int obj_num
)
{
	int i;
	char *buf, *name;
	xsw_object_struct *obj;
	xsw_weapon_struct *wep;
	xsw_net_parms_struct *net = &net_parms;
	xsw_option_struct *opt = &option;

	if((bw == NULL) || DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* Get object's name */
	switch(opt->show_formal_label)
	{
	  case XSW_SHOW_FORMAL_LABEL_ALWAYS:
	    name = STRDUP(DBGetFormalNameStr(obj_num));
	    break;
	  case XSW_SHOW_FORMAL_LABEL_AS_NEEDED:
	    name = (obj->owner == net->player_obj_num) ?
		STRDUP(DBGetFormalNameStr(obj_num)) :
		STRDUP(obj->name);
	    break;
	  default:	/* XSW_SHOW_FORMAL_LABEL_NEVER */
	    name = STRDUP(obj->name);
	    break;
	}

	/* Print name and description */
	buf = (char *)malloc(
	    STRLEN(name) + STRLEN(obj->description) + 80
	);
	if(obj->description != NULL)
	    sprintf(
		buf,
		"%s - %s",
		name, obj->description
	    );
	else
	    sprintf(
		buf,
		"%s",
		name
	    );
	XSWPrint(buf);
	free(buf);
	free(name);

	/* Is the object of type area? */
	if(obj->type == SW_OBJ_TYPE_AREA)
	{
	    sw_legend_struct *legend = &sw_legend;
	    sw_units_struct units;
	    const char *area_type_s;
	    char size_text[80];

	    float size = obj->size;		/* In screen units */
	    units.ru_to_au = legend->ru_to_au;
	    switch(option.units)
	    {
	      case XSW_UNITS_METRIC:
		size = SWConvertRUToAU(&units, size / 1000.0f);
		sprintf(size_text, "%.4f au", size);
		break;
	      case XSW_UNITS_ENGLISH:
		size = SWConvertRUToKiloMiles(&units, size / 1000.0f);
		sprintf(size_text, "%.1f kmiles", size);
		break;
	      case XSW_UNITS_PARSECS:
		size = SWConvertRUToMP(&units, size / 1000.0f);
		sprintf(size_text, "%.4f mp", size);
		break;
	      default:
		sprintf(size_text, "%.0f su", size);
		break;
	    }
	    switch(obj->loc_type)
	    {
	      case SW_LOC_TYPE_SPACE_TERRITORY:
		area_type_s = SW_LOC_TYPE_STR_SPACE_TERRITORY;
		break;
	      case SW_LOC_TYPE_NEBULA:
		area_type_s = SW_LOC_TYPE_STR_NEBULA;
		break;
	      case SW_LOC_TYPE_ASTEROID_FIELD:
		area_type_s = SW_LOC_TYPE_STR_ASTEROID_FIELD;
		break;
	      case SW_LOC_TYPE_PULSAR:
		area_type_s = SW_LOC_TYPE_STR_PULSAR;
		break;
	      default:	/* SW_LOC_TYPE_SPACE */
		area_type_s = SW_LOC_TYPE_STR_SPACE;
		break;
	    }
	    buf = (char *)malloc(
		STRLEN(area_type_s) + STRLEN(size_text) + 80
	    );
	    sprintf(
		buf,
"Area Type: %s  Size: %s",
		area_type_s,
		size_text
	    );
	}
	else
	{
	    buf = (char *)malloc(
		40 + 40 + 40 + 40 + 40 + 40 + 128
	    );
	    sprintf(
		buf,
"Size: %ld su  Density: %.2f(%.2f)  Energy: %.2f(%.2f)  Visibility: %.0f%%",
		obj->size,
		obj->hp, obj->hp_max,
		obj->power, obj->power_max,
		DBGetObjectVisibilityPtr(obj) * 100.0
	    );
	}
	XSWPrint(buf);
	free(buf);

	/* Weapons */
	for(i = 0; i < obj->total_weapons; i++)
	{
	    wep = obj->weapons[i];
	    if(wep == NULL)
		continue;

	    switch(wep->emission_type)
	    {
	      case SW_WEP_EMISSION_PULSE:
		buf = (char *)malloc(
		    STRLEN(wep->name) + 40 + 40 + 40 + 80
		);
		if(wep->create_power > 0.0)
		    sprintf(
			buf,
"%s: Damage Hull: %.2f  Damage Shields: %.2f  Usage Power: %.2f",
			wep->name,
			wep->damage_hull,
			wep->damage_shields,
			wep->create_power    
		    );
		else
		    sprintf(
			buf,
"%s: Damage Hull: %.2f  Damage Shields: %.2f",
			wep->name,
			wep->damage_hull,
			wep->damage_shields
		    );
		break;

	      case SW_WEP_EMISSION_PROJECTILE:
		if(wep->amount > 0)
		{
		    buf = (char *)malloc(
			STRLEN(wep->name) + 40 + 40 + 40 + 80
		    );
		    sprintf(
			buf,
"%s: Damage Hull: %.2f  Damage Shields: %.2f  Quantity: %d",
			wep->name,
			wep->damage_hull,
			wep->damage_shields,
			wep->amount
		    );
		}
		else
		{
		    buf = NULL;
		}
		break;

	      default:	/* SW_WEP_EMISSION_STREAM */
		buf = (char *)malloc(
		    STRLEN(wep->name) + 40 + 40 + 40 + 40 + 80
		);
		if(wep->create_power > 0.0)
		    sprintf(
			buf,
"%s: Damage Hull: %.2f  Damage Shields: %.2f  Usage Power: %.2f  Range: %ld su",
			wep->name, 
			wep->damage_hull,
			wep->damage_shields,
			wep->create_power,
			wep->range
		    );
		else
		    sprintf(
			buf,
"%s: Damage Hull: %.2f  Damage Shields: %.2f  Range: %ld su",
			wep->name,
			wep->damage_hull,
			wep->damage_shields,
			wep->range
		    );
		break;
	    }
	    if(buf != NULL)
	    {
		XSWPrint(buf);
		free(buf);
	    }
	}
}

/*
 *	Prints warning about weapons being offline and plays
 *	appropriate sound.
 */
void BridgeWarnWeaponsOffline(
	xsw_bridge_win_struct *bw, const int obj_num
)
{
	const char *key_name;
	char *s;
	xsw_option_struct *opt = &option;

	key_name = OSWGetKeyCodeName(
	    KeymapGetKeycode(XSW_KM_WEAPONS_ONLINE)
	);
	s = (char *)malloc(STRLEN(key_name) + 128);
	sprintf(
	    s,
"Weapons are offline, press \"%s\" to bring weapons online.",
	    key_name
	);
	XSWPrintWarning(s);
	free(s);

	if(opt->sounds >= XSW_SOUNDS_EVENTS)
	    SoundPlay(
		&sound, SW_SNDREF_STD_ERROR,
		1.0f, 1.0f,
		0, 0
	    );
}

/*
 *	Manages the bridge window's prompt.
 *
 *	Checks if event is for the bridge window's prompt, if it is
 *	then it performs the appropriate operation.
 *
 *	The event must be either KeyPress or KeyRelease and the
 *	bridge window must be in focus.
 *
 *	Returns 1 or greater if the event was for the bridge prompt
 *	and an execution was performed. Returns 0 on error or if
 *	event was not for the bridge prompt.
 */
static int BridgeManagePrompt(
	xsw_bridge_win_struct *bw, event_t *event
)
{
	bool_t is_press;
	int obj_num, events_handled = 0;
	char *value;
	keycode_t keycode;
	key_state_t key_state;
	keyval_t keyval;
	event_key_t *key;
	prompt_struct *prompt;
	xsw_object_struct *obj;
	xsw_player_values_struct *pvals = &player_values;
	xsw_net_parms_struct *net = &net_parms;

	if((bw == NULL) || (event == NULL))
	    return(events_handled);

	if(!bw->has_focus || (prompt_mode == XSW_PROMPT_NONE))
	    return(events_handled);

	/* The event type must be a KeyPress or KeyRelease */
	if((event->type != KeyPress) &&
	   (event->type != KeyRelease)
	)
	    return(events_handled);

	prompt = &bw->prompt;
	value = STRDUP(PromptGetS(prompt));
	if(value == NULL)
	    value = strdup("");
	is_press = (event->type == KeyPress) ? True : False;
	key = &event->xkey;
	keycode = key->keycode;
	key_state = key->state;
	keyval = OSWGetKeyValFromKeyCode(keycode, key_state);

	/* Cancel */
	if(keyval == OSW_KEY_Escape)
	{
	    if(is_press)
	    {
		/* Unmap the prompt and reset the prompt mode */
		WidgetUnmap(prompt);
		prompt_mode = XSW_PROMPT_NONE;
	    }
	    events_handled++;
	}
	/* Activate */ 
	else if((keyval == OSW_KEY_Return) ||
		(keyval == OSW_KEY_KP_Enter)
	)
	{
	    const int mode = prompt_mode;

	    /* Send this key event to the prompt's manager, it still
	     * needs to know about it
	     */
	    events_handled += WidgetManage(prompt, event);

	    /* Unmap prompt and reset prompt mode */
	    WidgetUnmap(prompt);
	    prompt_mode = XSW_PROMPT_NONE;

	    if(is_press)
	    {
		switch(mode)
		{
		  case XSW_PROMPT_CONNECT:
		    XSWConnect(value);
		    break;

		   case XSW_PROMPT_CLIENT_CMD:
		    CmdHandleInput(value);
		    break;

		   case XSW_PROMPT_SERVER_CMD:
		    if(XSW_CON_STATE_IS_CONNECTED(net))
		    {
			NetSendExec(value);
		    }
		    else
		    {
			XSWPrintWarning("Not connected.");
		    }
		    break;

		  case XSW_PROMPT_MESSAGE:
		    obj_num = net->player_obj_num;
		    obj = net->player_obj_ptr;
		    if(obj != NULL)
		    {
			NetSendComMessage(
			    obj_num,
			    -1,
			    obj->com_channel,
			    value
			);
		    }
		    else
		    {
			XSWPrintWarning("Not connected or not logged in.");
		    }
		    break;

		  case XSW_PROMPT_WEAPONS_FREQ:
		    obj_num = net->player_obj_num;
		    obj = net->player_obj_ptr;
		    if(obj != NULL)
		    {
			const float vf = CLIP(
			    ATOF(value), SW_FREQ_MIN, SW_FREQ_MAX
			);
			pvals->weapon_freq = vf;
			BridgeDrawPanel(
			    obj_num,
			    BRIDGE_PANEL_PWEAPONFREQ
			);
		    }
		    else
		    {
			XSWPrintWarning("Not connected or not logged in.");
		    }
		    break;

		  case XSW_PROMPT_SHIELD_FREQ:
		    obj_num = net->player_obj_num;
		    obj = net->player_obj_ptr;
		    if(obj != NULL)
		    {
			const float vf = CLIP(
			    ATOF(value), SW_FREQ_MIN, SW_FREQ_MAX
			);
			NetSendSetShields(
			    obj_num,
			    obj->shield_state,
			    vf
			);
			BridgeDrawPanel(
			    obj_num,
			    BRIDGE_PANEL_PSHIELDFREQ
			);
		    }
		    else
		    {
			XSWPrintWarning("Not connected or not logged in.");
		    }
		    break;

		  case XSW_PROMPT_INTERCEPT:
		    obj_num = net->player_obj_num;
		    obj = net->player_obj_ptr;
		    if(obj != NULL)
		    {
			NetSendIntercept(
			    obj_num,
			    (*value != '\0') ? value : "#off"
			);
		    }
		    else
		    {
			XSWPrintWarning("Not connected or not logged in.");
		    }
		    break;

		  case XSW_PROMPT_COM_CHANNEL:
		    obj_num = net->player_obj_num;
		    obj = net->player_obj_ptr;
		    if(obj != NULL)
		    {
		        const int vi = (int)(CLIP(
			    ATOF(value), SW_FREQ_MIN, SW_FREQ_MAX
			) * 100.0f);
			obj->com_channel = vi;
			NetSendSetChannel(
			    obj_num,
			    obj->com_channel
			);
			if(option.sounds >= XSW_SOUNDS_EVENTS)
			    SoundPlay(
				&sound, SW_SNDREF_CHANNEL_OPEN,
				1.0f, 1.0f,
				0, 0
			    );
		    }
		    else
		    {
			XSWPrintWarning("Not connected or not logged in.");
		    }
		    break;
		}
	    }
	    events_handled++;
	}
	/* All else send the key event to the prompt's manager */
	else
	{
	    events_handled += WidgetManage(prompt, event);
	}

	free(value);

	return(events_handled);
}


/*
 *	Manages the bridge window event;
 */
int BridgeManage(
	xsw_bridge_win_struct *bw, event_t *event,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int		events_handled = 0,
			x, y,
			root_x, root_y,
			obj_num;
	time_t t;
	atom_t action;
	keycode_t keycode;
	key_state_t key_state;
	keyval_t keyval;
	win_t w, src_w;
	event_button_t *button;
	event_motion_t *motion;
	event_key_t *key;
	event_visibility_t *visibility;
	xsw_object_struct *obj;
	osw_gui_struct *gui = &osw_gui[0];
/*	widget_global_struct *wglobal = &widget_global; */
	xsw_player_values_struct *pvals = &player_values;
	xsw_net_parms_struct *net = &net_parms;
	xsw_option_struct *opt = &option;

	if((bw == NULL) || (event == NULL))
	    return(events_handled);

#if 0
#if defined(HAVE_X) && defined(USE_XSHM)
	if(!bw->map_state &&
	   (event->type != MapNotify) &&
	   (event->type != (int)gui->shm_completion_event_code)
	)
	    return(events_handled);
#else
	if(!bw->map_state && (event->type != MapNotify))
	    return(events_handled);
#endif	/* defined(HAVE_X) && defined(USE_XSHM) */
#endif

	/* Get the event window */
	w = event->xany.window; 


#if defined(HAVE_X) && defined(USE_XSHM)
	/* Check for viewscreen shared image put completion */
	if(event->type == (int)gui->shm_completion_event_code)
	{
	    /* Viewscreen window? */
	    if(w == bw->viewscreen)
	    {
		shared_image_t *img = bw->viewscreen_image;
		if(img != NULL)
		    img->in_progress = False;
		events_handled++;
		return(events_handled);
	    }
	    /* Scanner window? */
	    else if(w == bw->scanner)
	    {
		shared_image_t *img = bw->scanner_image;
		if(img != NULL)
		    img->in_progress = False;
		events_handled++;
		return(events_handled);
	    }
	}
#endif	/* defined(HAVE_X) && defined(USE_XSHM) */


	/* Check if current page is mapped, if so then manage it */
	if(bw->cur_page != NULL)
	{
	    if(bw->cur_page->map_state)
	    {
		/* Allow main menu to manage this event */
		events_handled += PageManage(
		    bw->cur_page,
		    bw->viewscreen,
		    bw->viewscreen_image,
		    event,
		    cur_ms, cur_sec, tc
		);
		/* Return if event was handled */
		if(events_handled > 0)
		    return(events_handled);
	    }
	}


	/* If in a prompt mode then allow the prompt to manage event
	 * (it will handle KeyPress and KeyRelease events only) and
	 * Return if the prompt managed the event
	 */
	events_handled += BridgeManagePrompt(bw, event);
	if(events_handled > 0)
	    return(events_handled);


	/* All other event management */
	switch(event->type)
	{
	  case ConfigureNotify:
	    if(w == bw->toplevel)
	    {
		BridgeWinResize(bw);
		if(bw->map_state)
		    BridgeWinDrawAll(bw);
		events_handled++;
	    }
	    break;

	  case Expose:
	    /* Toplevel */
	    if(w == bw->toplevel)
	    {
		/* Nothing to draw */
		events_handled++;
	    }
	    /* Viewscreen */
	    else if(w == bw->viewscreen)
	    {
		VSDrawViewScreen(bw);
		events_handled++;
	    }
	    /* Message box */
	    else if(w == bw->mesg_box)
	    {
		BridgeQueueDrawMessages(bw);
		events_handled++;
	    }
	    /* Player object console panels */
	    else if(w == bw->pan_p1)
	    {
		BridgeDrawPanel(
		    net->player_obj_num,
		    BRIDGE_PANEL_P1
		);
		events_handled++;
	    } 
	    else if(w == bw->pan_p2)
	    {
		BridgeDrawPanel(
		    net->player_obj_num,
		    BRIDGE_PANEL_P2
		);
		events_handled++;
	    }
	    else if(w == bw->pan_p3)
	    {
		BridgeDrawPanel(
		    net->player_obj_num,
		    BRIDGE_PANEL_P3
		);
		events_handled++;
	    }
	    else if(w == bw->pan_p4)
	    {
		BridgeDrawPanel(
		    net->player_obj_num,
		    BRIDGE_PANEL_P4
		);
		events_handled++;
	    }
	    /* Scanner */
	    else if(w == bw->scanner)
	    {
		ScannerDraw(
		    bw, net->player_obj_num,
		    bw->scanner,
		    bw->scanner_image,
		    0,
		    0,
		    bw->scanner_zoom,
		    cur_ms, cur_sec, tc
		);
		WidgetQueueDraw(&bw->scanner_sb);
		events_handled++;
	    }
	    /* Subject object console panels */
	    else if(w == bw->pan_s1)
	    {
		xsw_object_struct *player_obj = net->player_obj_ptr;
		obj_num = (player_obj != NULL) ?
		    player_obj->locked_on : -1;
		BridgeDrawPanel(obj_num, BRIDGE_PANEL_S1);
		events_handled++;
	    }
	    else if(w == bw->pan_s2)
	    {
		xsw_object_struct *player_obj = net->player_obj_ptr;
		obj_num = (player_obj != NULL) ?
		    player_obj->locked_on : -1;
		BridgeDrawPanel(obj_num, BRIDGE_PANEL_S2);
		events_handled++;
	    }
	    else if(w == bw->pan_s3)
	    {
		xsw_object_struct *player_obj = net->player_obj_ptr;
		obj_num = (player_obj != NULL) ?
		    player_obj->locked_on : -1;
		BridgeDrawPanel(obj_num, BRIDGE_PANEL_S3);
		events_handled++;
	    }
	    break;

	  case KeyPress:
	    if(!bw->has_focus)
		return(events_handled);

	    key = &event->xkey;
	    keycode = key->keycode;
	    key_state = key->state;

	    /* Skip modifier keys */
	    if(OSWIsModifierKey(keycode))
		return(events_handled);

	    keyval = OSWGetKeyValFromKeyCode(keycode, key_state);

	    /* Help */
	    if(keycode == KeymapGetKeycode(XSW_KM_HELP))
	    {
		XSWPrintHelp();
		events_handled++; 
	    }
	    /* Exit */
	    else if(keycode == KeymapGetKeycode(XSW_KM_EXIT))
	    {
		XSWAction(
		    XSW_ACTION_EXIT_CONFIRM, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Toggle Vector Lights */
	    else if(keycode == KeymapGetKeycode(XSW_KM_LIGHTS_VECTOR))
	    {
		XSWAction(
		    XSW_ACTION_LIGHTS_VECTOR, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Toggle Strobes */
	    else if(keycode == KeymapGetKeycode(XSW_KM_LIGHTS_STROBE))
	    {
		XSWAction(
		    XSW_ACTION_LIGHTS_STROBE, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Toggle External Lumination */
	    else if(keycode == KeymapGetKeycode(XSW_KM_LIGHTS_LUMINATION))
	    {
		XSWAction(
		    XSW_ACTION_LIGHTS_LUMINATION, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Set Weapon Frequency */
	    else if(keycode == KeymapGetKeycode(XSW_KM_WEAPON_FREQ))
	    {
		XSWAction(
		    XSW_ACTION_WEAPON_FREQ, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Toggle Shields */
	    else if(keycode == KeymapGetKeycode(XSW_KM_SHIELD))
	    {
		XSWAction(
		    XSW_ACTION_SHIELD, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++; 
	    }
	    /* Set Shield Frequency */
	    else if(keycode == KeymapGetKeycode(XSW_KM_SHIELD_FREQ))
	    {
		XSWAction(
		    XSW_ACTION_SHIELD_FREQ, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Toggle Damage Control */ 
	    else if(keycode == KeymapGetKeycode(XSW_KM_DMGCTL))
	    {
		XSWAction(
		    XSW_ACTION_DMGCTL, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++; 
	    }
	    /* Toggle Cloak */
	    else if(keycode == KeymapGetKeycode(XSW_KM_CLOAK))
	    {
		XSWAction(
		    XSW_ACTION_CLOAK, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++; 
	    }
	    /* Set Intercept */
	    else if(keycode == KeymapGetKeycode(XSW_KM_INTERCEPT_SET))
	    {
		XSWAction(
		    XSW_ACTION_INTERCEPT_SET, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Hail */
	    else if(keycode == KeymapGetKeycode(XSW_KM_HAIL))
	    {
		XSWAction(
		    XSW_ACTION_HAIL, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Set Com Channel */
	    else if(keycode == KeymapGetKeycode(XSW_KM_COM_CHANNEL_SET))
	    {
		XSWAction(
		    XSW_ACTION_COM_CHANNEL_SET, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Send Message */
	    else if(keycode == KeymapGetKeycode(XSW_KM_SEND_MESSAGE))
	    {
		XSWAction(
		    XSW_ACTION_SEND_MESSAGE, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Ignore/Unignore Object */
	    else if(keycode == KeymapGetKeycode(XSW_KM_IGNORE_OBJECT))
	    {
		XSWAction(
		    XSW_ACTION_IGNORE_OBJECT, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Viewscreen Zoom In */
	    else if(keycode == KeymapGetKeycode(XSW_KM_VIEWSCREEN_ZOOM_IN))
	    {
		XSWAction(
		    XSW_ACTION_VIEWSCREEN_ZOOM_IN, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Viewscreen Zoom Out */
	    else if(keycode == KeymapGetKeycode(XSW_KM_VIEWSCREEN_ZOOM_OUT))
	    {
		XSWAction(
		    XSW_ACTION_VIEWSCREEN_ZOOM_OUT, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Viewscreen Zoom Auto */
	    else if(keycode == KeymapGetKeycode(XSW_KM_VIEWSCREEN_ZOOM_AUTO))
	    {
		XSWAction(
		    XSW_ACTION_VIEWSCREEN_ZOOM_AUTO, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);    
		events_handled++;
	    }
	    /* Scanner Zoom In */
	    else if(keycode == KeymapGetKeycode(XSW_KM_SCANNER_ZOOM_IN))
	    {
		XSWAction(
		    XSW_ACTION_SCANNER_ZOOM_IN, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Scanner Zoom Out */
	    else if(keycode == KeymapGetKeycode(XSW_KM_SCANNER_ZOOM_OUT))
	    {
		XSWAction(
		    XSW_ACTION_SCANNER_ZOOM_OUT, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Scanner Zoom Min */
	    else if(keycode == KeymapGetKeycode(XSW_KM_SCANNER_ZOOM_MIN))
	    {
		XSWAction(
		    XSW_ACTION_SCANNER_ZOOM_MIN, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Scanner Zoom Max */
	    else if(keycode == KeymapGetKeycode(XSW_KM_SCANNER_ZOOM_MAX))
	    {
		XSWAction(
		    XSW_ACTION_SCANNER_ZOOM_MAX, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Scanner Orient */
	    else if(keycode == KeymapGetKeycode(XSW_KM_SCANNER_ORIENT))
	    {
		XSWAction(
		    XSW_ACTION_SCANNER_ORIENT, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Engine State */
	    else if(keycode == KeymapGetKeycode(XSW_KM_ENGINE_STATE))
	    {
		XSWAction(
		    XSW_ACTION_ENGINE_STATE, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Throttle Mode */
	    else if(keycode == KeymapGetKeycode(XSW_KM_THROTTLE_MODE))
	    {
		XSWAction(
		    XSW_ACTION_THROTTLE_MODE, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Lock Next */
	    else if(keycode == KeymapGetKeycode(XSW_KM_LOCK_NEXT))
	    {
		XSWAction(
		    XSW_ACTION_LOCK_NEXT, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Lock Nearest */
	    else if(keycode == KeymapGetKeycode(XSW_KM_LOCK_NEAREST))
	    {
		XSWAction(
		    XSW_ACTION_LOCK_NEAREST, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Unlock */
	    else if(keycode == KeymapGetKeycode(XSW_KM_UNLOCK))
	    {
		XSWAction(
		    XSW_ACTION_UNLOCK, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Toggle Weapons Online/Offline */
	    else if(keycode == KeymapGetKeycode(XSW_KM_WEAPONS_ONLINE))
	    {
		XSWAction(
		    XSW_ACTION_WEAPONS_ONLINE, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Weapon Select Next */
	    else if(keycode == KeymapGetKeycode(XSW_KM_WEAPON_SELECT_NEXT))
	    {
		XSWAction(
		    XSW_ACTION_WEAPON_SELECT_NEXT, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Weapon Select Prev */
	    else if(keycode == KeymapGetKeycode(XSW_KM_WEAPON_SELECT_PREV))
	    {
		XSWAction(
		    XSW_ACTION_WEAPON_SELECT_PREV, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Select weapon 1 */
	    else if(keycode == KeymapGetKeycode(XSW_KM_WEAPON_SELECT_1))
	    {
		XSWAction(
		    XSW_ACTION_WEAPON_SELECT_1, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++; 
	    }
	    /* Select weapon 2 */
	    else if(keycode == KeymapGetKeycode(XSW_KM_WEAPON_SELECT_2))
	    {
		XSWAction(
		    XSW_ACTION_WEAPON_SELECT_2, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++; 
	    }
	    /* Select weapon 3 */
	    else if(keycode == KeymapGetKeycode(XSW_KM_WEAPON_SELECT_3))
	    {
		XSWAction(
		    XSW_ACTION_WEAPON_SELECT_3, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Select weapon 4 */
	    else if(keycode == KeymapGetKeycode(XSW_KM_WEAPON_SELECT_4))
	    {
		XSWAction(
		    XSW_ACTION_WEAPON_SELECT_4, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Select weapon 5 */
	    else if(keycode == KeymapGetKeycode(XSW_KM_WEAPON_SELECT_5))
	    {
		XSWAction(
		    XSW_ACTION_WEAPON_SELECT_5, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Select weapon 6 */
	    else if(keycode == KeymapGetKeycode(XSW_KM_WEAPON_SELECT_6))
	    {
		XSWAction(
		    XSW_ACTION_WEAPON_SELECT_6, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Select weapon 7 */
	    else if(keycode == KeymapGetKeycode(XSW_KM_WEAPON_SELECT_7))
	    {
		XSWAction(
		    XSW_ACTION_WEAPON_SELECT_7, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Select weapon 8 */
	    else if(keycode == KeymapGetKeycode(XSW_KM_WEAPON_SELECT_8))
	    {
		XSWAction(
		    XSW_ACTION_WEAPON_SELECT_8, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Select weapon 9 */
	    else if(keycode == KeymapGetKeycode(XSW_KM_WEAPON_SELECT_9))
	    {
		XSWAction(
		    XSW_ACTION_WEAPON_SELECT_9, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Select weapon 10 */
	    else if(keycode == KeymapGetKeycode(XSW_KM_WEAPON_SELECT_10))
	    {
		XSWAction(
		    XSW_ACTION_WEAPON_SELECT_10, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Tractor Beam Lock */
	    else if(keycode == KeymapGetKeycode(XSW_KM_TRACTOR_BEAM_LOCK))
	    {
		if(key->state & OSW_KEY_STATE_SHIFT)
		    XSWAction(
			XSW_ACTION_TRACTOR_BEAM_UNLOCK, NULL,
			bw,
			cur_ms, cur_sec, tc
		    );
		else
		    XSWAction(
			XSW_ACTION_TRACTOR_BEAM_LOCK, NULL,
			bw,
			cur_ms, cur_sec, tc
		    );
		events_handled++;
	    }
	    /* Auto Destruct */
	    else if(keycode == KeymapGetKeycode(XSW_KM_AUTO_DESTRUCT))
	    {
		XSWAction(
		    XSW_ACTION_AUTO_DESTRUCT, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Viewscreen Markings */
	    else if(keycode == KeymapGetKeycode(XSW_KM_VS_MARKINGS))
	    {
		XSWAction(
		    XSW_ACTION_VS_MARKINGS, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++; 
	    }
	    /* Viewscreen Labels */
	    else if(keycode == KeymapGetKeycode(XSW_KM_VS_LABELS))
	    {
		XSWAction(
		    XSW_ACTION_VS_LABELS, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++; 
	    }
	    /* Energy Saver Mode */
	    else if(keycode == KeymapGetKeycode(XSW_KM_ENERGY_SAVER_MODE))
	    {
		XSWAction(
		    XSW_ACTION_ENERGY_SAVER_MODE, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Net Interval Increase */
	    else if(keycode == KeymapGetKeycode(XSW_KM_NET_INTERVAL_INC))
	    {
		XSWAction(
		    XSW_ACTION_NET_INTERVAL_INC, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++; 
	    }
	    /* Net Interval Decrease */
	    else if(keycode == KeymapGetKeycode(XSW_KM_NET_INTERVAL_DEC))
	    {
		XSWAction(
		    XSW_ACTION_NET_INTERVAL_DEC, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Message Box Scroll Up */
	    else if(keycode == KeymapGetKeycode(XSW_KM_MSG_SCROLL_UP))
	    {
		scroll_bar_struct *sb = &bw->mesg_box_sb;
		const int y = (int)SBarGetPositionY(sb);
		SBarSetPositionY(sb, (float)y - (float)sb->y_step_inc);
		events_handled++;
	    }
	    /* Message Box Scroll Down */
	    else if(keycode == KeymapGetKeycode(XSW_KM_MSG_SCROLL_DOWN))
	    {
		scroll_bar_struct *sb = &bw->mesg_box_sb;
		const int y = (int)SBarGetPositionY(sb);
		SBarSetPositionY(sb, (float)y + (float)sb->y_step_inc);
		events_handled++; 
	    }
	    /* Message Box Scroll To End */
	    else if(keyval == OSW_KEY_End)
	    {
		scroll_bar_struct *sb = &bw->mesg_box_sb;
		SBarSetPositionY(sb, (float)sb->total_height);
		events_handled++;
	    }
	    /* Economy */
	    else if(keycode == KeymapGetKeycode(XSW_KM_ECONOMY))
	    {
		XSWAction(
		    XSW_ACTION_ECONOMY, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* StarChart */
	    else if(keycode == KeymapGetKeycode(XSW_KM_STARCHART))
	    {
		XSWAction(
		    XSW_ACTION_STARCHART, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Connect */
	    else if(keycode == KeymapGetKeycode(XSW_KM_CONNECT))
	    {
		char *url = (char *)malloc(
		    STRLEN(net->login_name) +
		    STRLEN(net->login_password) +
		    STRLEN(net->address) +
		    80
		);
		sprintf(
		    url,
		    "swserv://%s:%s@%s:%i",
		    (net->login_name != NULL) ?
			net->login_name : "",
		    (net->login_password != NULL) ?
			net->login_password : "",
		    (net->address != NULL) ?
			net->address : "",
		    net->port
		);
		BridgeMapPromptMode(
		    bw,
		    "Connect:",
		    url,
		    XSW_PROMPT_CONNECT
		);
		free(url);
		events_handled++;
	    }
	    /* Connect Last */
	    else if(keycode == KeymapGetKeycode(XSW_KM_CONNECT_LAST))
	    {
		XSWAction(
		    XSW_ACTION_CONNECT_LAST, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Disconnect */
	    else if(keycode == KeymapGetKeycode(XSW_KM_DISCONNECT))
	    {
		XSWAction(
		    XSW_ACTION_DISCONNECT, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++; 
	    }
	    /* Refresh */
	    else if(keycode == KeymapGetKeycode(XSW_KM_REFRESH))
	    {
		XSWAction(
		    XSW_ACTION_REFRESH, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++; 
	    }
	    /* Client command */
	    else if(keycode == KeymapGetKeycode(XSW_KM_CLIENT_CMD))
	    {
		XSWAction(
		    XSW_ACTION_CLIENT_CMD, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++;
	    }
	    /* Server command */
	    else if(keycode == KeymapGetKeycode(XSW_KM_SERVER_CMD))
	    {
		XSWAction(
		    XSW_ACTION_SERVER_CMD, NULL,
		    bw,
		    cur_ms, cur_sec, tc
		);
		events_handled++; 
	    }
	    /* Screen shot */
	    else if(keycode == KeymapGetKeycode(XSW_KM_SCREEN_SHOT))
	    {
		BridgeScreenShot(
		    bw,
		    dname.downloads		/* Save to directory */
		);
		events_handled++;
	    }

	    /* Unknown Command */
	    else
	    {
		char *arg = STRDUP(OSWGetKeyCodeName(KeymapGetKeycode(XSW_KM_HELP)));
		char *s = (char *)malloc(STRLEN(arg) + 80);
		strtoupper(arg);
		sprintf(
		    s,
"Unknown command, press %s for help.",
		    arg
		);
		free(arg);
		XSWPrintWarning(s);
		free(s);
	    }

	    /* Do not continure further and do not let widgets handle
	     * this key event if it was already handled above
	     */
	    if(events_handled > 0)
		return(events_handled);	

	    break;

	  /* *********************************************************** */
	  case KeyRelease:
	    key = &event->xkey;

	    /* Skip if not in focus */
	    if(!bw->has_focus)
		return(events_handled);

	    keycode = key->keycode;

	    /* Skip modifier keys */
	    if(OSWIsModifierKey(keycode))
		return(events_handled);


	    break;

	  /* *********************************************************** */
	  case ButtonPress:
	    button = &event->xbutton;

	    /* Grab focus as needed */
	    if(!bw->has_focus)
	    {
		if(OSWCheckWindowAncestory(bw->toplevel, w))
		    OSWGrabFocus(bw->toplevel);
	    }

	    /* Player console panel section 1 */
	    if(w == bw->pan_p1)
	    {
		obj = net->player_obj_ptr;
		obj_num = net->player_obj_num;

		/* Com channel inc */
		if(BridgeIsCoordinatesInInputArea(
		    bw,
		    BRIDGE_INPUT_AREA_PCOMCHANNEL_UP,
		    button->x,
		    button->y
		   ) &&
		   (obj != NULL)
		)
		{
		    obj->com_channel += (button->state & OSW_KEY_STATE_SHIFT) ? 100 : 5;
		    if(((float)obj->com_channel / 100) > SW_FREQ_MAX)
			obj->com_channel = static_cast<int>(SW_FREQ_MAX * 100);
		    if(((float)obj->com_channel / 100) < SW_FREQ_MIN)
			obj->com_channel = static_cast<int>(SW_FREQ_MIN * 100);

		    NetSendSetChannel(
			obj_num,
			obj->com_channel
		    );
		}

		/* Com channel dec */
		if(BridgeIsCoordinatesInInputArea(
		    bw,
		    BRIDGE_INPUT_AREA_PCOMCHANNEL_DOWN,
		    button->x,
		    button->y
		   ) &&
		   (obj != NULL)
		)
		{
		    obj->com_channel -= (button->state & OSW_KEY_STATE_SHIFT) ? 100 : 5;
		    if(((float)obj->com_channel / 100) > SW_FREQ_MAX)
			obj->com_channel = static_cast<int>(SW_FREQ_MAX * 100);
		    if(((float)obj->com_channel / 100) < SW_FREQ_MIN)
			obj->com_channel = static_cast<int>(SW_FREQ_MIN * 100);

		    NetSendSetChannel(
			obj_num,
			obj->com_channel
		    );
		}

		BridgePlaySoundButtonClick(bw, 1.0f, 0.2f);

		events_handled++;
	    }
	    /* Player console panel section 2 */
	    else if(w == bw->pan_p2)
	    {
		obj = net->player_obj_ptr;
		obj_num = net->player_obj_num;

		/* If it's button 2 then print stats of the player */
		if(button->button == Button2)
		{
		    BridgePrintSubjectStats(bw, obj_num);
		    BridgePlaySoundButtonClick(bw, 1.0f, 0.2f);
		}
		/* Shields toggle */
		else if(BridgeIsCoordinatesInInputArea(
		    bw,
		    BRIDGE_INPUT_AREA_PSHIELDS,
		    button->x,
		    button->y
		   ) &&
		   (obj != NULL)
		)
		{
                    XSWAction(
                        XSW_ACTION_SHIELD, NULL,
                        bw,
			cur_ms, cur_sec, tc
                    );
 		}
		/* Cloak */
		else if(BridgeIsCoordinatesInInputArea(
		    bw,
		    BRIDGE_INPUT_AREA_PCLOAK,
		    button->x,
		    button->y
		   ) &&
		   (obj != NULL)
		)
		{
		    XSWAction(
			XSW_ACTION_CLOAK, NULL,
			bw,
			cur_ms, cur_sec, tc
		    );
		}
		/* Damage control toggle */
		else if(BridgeIsCoordinatesInInputArea(
		    bw,
		    BRIDGE_INPUT_AREA_PDMGCTL,
		    button->x,
		    button->y
		   ) &&
		   (obj != NULL)
		)
		{
                    XSWAction(
                        XSW_ACTION_DMGCTL, NULL,
                        bw,
			cur_ms, cur_sec, tc
                    );
		}

		events_handled++;
	    }
	    /* Player console panel section 3 */
	    else if(w == bw->pan_p3)
	    {
		obj = net->player_obj_ptr;
		obj_num = net->player_obj_num;

		/* Engine state */
		if(BridgeIsCoordinatesInInputArea(
		    bw,
		    BRIDGE_INPUT_AREA_PENGINESTATE,
		    button->x,
		    button->y
		   ) &&
		   (obj != NULL)
		)
		{
                    XSWAction(
                        XSW_ACTION_ENGINE_STATE, NULL,
                        bw,
			cur_ms, cur_sec, tc
                    );
		}
		events_handled++;
	    }
	    /* Player console panel section 4 */
	    else if(w == bw->pan_p4)
	    {
		obj = net->player_obj_ptr;
		obj_num = net->player_obj_num;

		/* Weapon yield */
		if(BridgeIsCoordinatesInInputArea(
		    bw,
		    BRIDGE_INPUT_AREA_PWEPYIELD,
		    button->x,
		    button->y
		   ) &&
		   (obj != NULL)
		)
		{
		    const unsigned int bar_width = 80,
				       bar_height = 14;
		    const int	bar_x = 30,
				bar_y = 28;

		    /* Map weapon yield input window */
		    OSWDestroyWindow(&bw->weapon_yield_iwin);
		    if(!OSWCreateInputWindow(
			&bw->weapon_yield_iwin,
			bw->pan_p4,
			bar_x, bar_y,
			bar_width, bar_height
		    ))
		    {
			unsigned int width, height;
			win_t w = bw->weapon_yield_iwin;

			OSWMapRaised(w);
			OSWGrabPointer(
			    w,
			    True,
			    ButtonReleaseMask | PointerMotionMask,
			    GrabModeAsync, GrabModeAsync,
			    w,
			    None
			);

			width = bar_width;
			height = bar_height;

			/* Adjust weapon yield */
			if(width > 1)
			    pvals->weapon_yield = CLIP(
				((float)(button->x - 30) /
				    ((float)width - 1)),
				0.0f, 1.0f
			    );
			else
			    pvals->weapon_yield = 1.0f;

			obj = net->player_obj_ptr;
			obj_num = net->player_obj_num;
			if(obj != NULL)
			{
			    BridgeDrawPanel(
				obj_num,
				BRIDGE_PANEL_PWEAPONS
			    );
			}

		    }
		}

		/* Select weapon 1 */
		if(BridgeIsCoordinatesInInputArea(
		    bw,
		    BRIDGE_INPUT_AREA_PSELWEP1,
		    button->x,
		    button->y
		   ) &&
		   (obj != NULL)
		)
		{
		    NetSendSelectWeapon(obj_num, 0);
		}
		/* Select weapon 2 */
		if(BridgeIsCoordinatesInInputArea(
		    bw,
		    BRIDGE_INPUT_AREA_PSELWEP2,
		    button->x,
		    button->y
		   ) &&
		   (obj != NULL)
		)
		{
		    NetSendSelectWeapon(obj_num, 1);
		}
		/* Select weapon 3 */
		if(BridgeIsCoordinatesInInputArea(
		    bw,
		    BRIDGE_INPUT_AREA_PSELWEP3,
		    button->x,
		    button->y  
		   ) &&
		   (obj != NULL)
		)
		{
		    NetSendSelectWeapon(obj_num, 2);
		}
		/* Select weapon 4 */
		if(BridgeIsCoordinatesInInputArea(
		    bw,
		    BRIDGE_INPUT_AREA_PSELWEP4,
		    button->x,
		    button->y  
		   ) &&
		   (obj != NULL)
		)
		{
		    NetSendSelectWeapon(obj_num, 3);
		}
		/* Select weapon 5 */
		if(BridgeIsCoordinatesInInputArea(
		    bw,
		    BRIDGE_INPUT_AREA_PSELWEP5,
		    button->x,
		    button->y
		   ) &&
		   (obj != NULL)
		)
		{
		    NetSendSelectWeapon(obj_num, 4);
		}
		/* Select weapon 6 */
		if(BridgeIsCoordinatesInInputArea(
		    bw,
		    BRIDGE_INPUT_AREA_PSELWEP6,
		    button->x,
		    button->y
		   ) &&
		   (obj != NULL)
		)
		{
		    NetSendSelectWeapon(obj_num, 5);
		}
		/* Select weapon 7 */
		if(BridgeIsCoordinatesInInputArea(
		    bw,
		    BRIDGE_INPUT_AREA_PSELWEP7,
		    button->x,
		    button->y
		   ) &&
		   (obj != NULL)
		)
		{
		    NetSendSelectWeapon(obj_num, 6);
		}

		/* Intercept */
		if(BridgeIsCoordinatesInInputArea(
		    bw,
		    BRIDGE_INPUT_AREA_PINTERCEPT,
		    button->x,
		    button->y  
		   ) &&
		   (obj != NULL)
		)
		{
		    if(button->button == Button1)
		    {
			/* Query ETA */
			NetSendExec("eta");
		    }
		    else if(button->button == Button3)
		    {
			/* Disengage intercept */
			NetSendIntercept(obj_num, "#off");
		    }
		}

		BridgePlaySoundButtonClick(bw, 1.0, 0.2);
		events_handled++;
	    }
	    /* Weapon yield input window */
	    else if(w == bw->weapon_yield_iwin)
	    {
		/* Not suppose to recieve a ButtonPress in this window */
		OSWUngrabPointer();
		OSWDestroyWindow(&bw->weapon_yield_iwin);
		BridgePlaySoundButtonClick(bw, 1.0f, 0.2f);
		events_handled++;
	    }
	    /* Subject object console panel section 1 */
	    else if(w == bw->pan_s1)
	    {
		obj = net->player_obj_ptr;
		obj_num = net->player_obj_num;

		/* Lock next */
		if(BridgeIsCoordinatesInInputArea(
		    bw,
		    BRIDGE_INPUT_AREA_PLOCKNEXT,
		    button->x,
		    button->y
		   ) &&
		   (obj != NULL)
		)
		{
		    switch(button->button)
		    {
		      case Button3:
			NetSendWeaponsUnlock(obj_num);
			break;

		      case Button2:
			if(obj == NULL)
			    break;
			BridgePrintSubjectStats(bw, obj->locked_on);
			break;

		      default:
			NetSendWeaponsLock(obj_num, -2);
			break;
		    }
		}

		BridgePlaySoundButtonClick(bw, 0.2, 1.0);

		events_handled++;
	    }
	    /* Subject object console panel section 2 */
	    else if(w == bw->pan_s2)
	    {
		obj = net->player_obj_ptr;
		obj_num = net->player_obj_num;





		BridgePlaySoundButtonClick(bw, 0.2, 1.0);

		events_handled++;
	    }
	    /* Subject object console panel section 3 */
	    else if(w == bw->pan_s3)
	    {
		obj = net->player_obj_ptr;
		obj_num = net->player_obj_num;



		BridgePlaySoundButtonClick(bw, 0.2, 1.0);

		events_handled++;
	    }
	    /* Viewscreen (except Button3) */
	    else if((w == bw->viewscreen) &&
		    (button->button != Button3) &&
		    !WidgetMapped(&xsw_vsmenu->menu)
	    )
	    {
		obj = net->player_obj_ptr;
		switch(button->button)
		{
		  case Button1:
		    if(obj != NULL)
		    {
			obj_num = VSButtonMatch(bw, event);
			if(!DBIsObjectGarbage(obj_num))
			{
			    /* ButtonPress occured over an object */
			    if(button->state & OSW_KEY_STATE_SHIFT)
			    {
				/* Hail specific vessel */
				NetSendHail(
				    net->player_obj_num,
				    obj_num,		/* Hail this object */
				    obj->com_channel
				);
				BridgePlaySoundButtonClick(bw, 1.0f, 1.0f);
			    }
			    else
		            {
				/* Send scanner and weapons lock */
				NetSendWeaponsLock(
				    net->player_obj_num,
				    obj_num
				);
				BridgePlaySoundButtonClick(bw, 1.0f, 1.0f);
			    }
			}
		    }
		    break;

		  case Button2:
		    if(obj != NULL)
		    {
			obj_num = VSButtonMatch(bw, event);
			if(DBIsObjectGarbage(obj_num))
			{
			    /* ButtonPress did not occure over an object */
			    if(button->state & OSW_KEY_STATE_SHIFT)
			    {
				/* Send tractor beam unlock */
				NetSendTractorBeamLock(
				    net->player_obj_num,
				    -1
				);
				BridgePlaySoundButtonClick(bw, 1.0f, 1.0f);
			    }
			}
			else
			{
			    /* ButtonPress occured over an object */
 			    if(button->state & OSW_KEY_STATE_SHIFT)
			    {
				/* Send tractor beam lock */
				NetSendTractorBeamLock(
				    net->player_obj_num,
				    obj_num
				);
				BridgePlaySoundButtonClick(bw, 1.0f, 1.0f);
			    }
			    else
			    {
				/* IFF */
/* TODO */
			    }
			}
		    }
		    break;

		  case Button3:
		    /* Do not handle this and allow it to be handled
		     * further below
		     */
		    break;

		  case Button4:
		    /* Zoom out */
		    XSWAction(
			XSW_ACTION_VIEWSCREEN_ZOOM_OUT, NULL,
			bw,
			cur_ms, cur_sec, tc
		    );
		    break;

		  case Button5:
		    /* Zoom in */
		    XSWAction(
			XSW_ACTION_VIEWSCREEN_ZOOM_IN, NULL,
			bw,
			cur_ms, cur_sec, tc
		    );
		    break;
		}
		events_handled++;
	    }
	    /* Scanner */
	    else if(w == bw->scanner)
	    {
		const int	button_num = button->button,
				x = button->x,
				y = button->y;

		obj = net->player_obj_ptr;
		/* Zoom out */
		if(button_num == Button4)
		{
		    scale_bar_struct *scale = &bw->scanner_sb;

		    scale->pos += 3.0f *
			((button->state & OSW_KEY_STATE_SHIFT) ? 4 : 1) *
			MAX((scale->pos / 100.0f * 4.0f), 1.0f);
		    if(scale->pos > scale->pos_max)
			scale->pos = scale->pos_max;

		    BridgeWinScannerScaleValueChangedCB(scale, scale->pos, bw);
		    WidgetQueueDraw(scale);
		}
		/* Zoom in */
		else if(button_num == Button5)
		{
		    scale_bar_struct *scale = &bw->scanner_sb;

		    scale->pos -= 3.0f *
			((button->state & OSW_KEY_STATE_SHIFT) ? 4 : 1) *
			MAX((scale->pos / 100.0f * 4.0f), 1.0f);
		    if(scale->pos < scale->pos_min)
			scale->pos = scale->pos_min;

		    BridgeWinScannerScaleValueChangedCB(scale, scale->pos, bw);
		    WidgetQueueDraw(scale);
		}
		/* Select object */
		else if(obj != NULL)
		{
		    /* Match object on the scanner window */
		    if(obj->locked_on > -1)
		    {
			obj_num = ScannerButtonMatch(
			    bw, x, y,
			    obj->locked_on + 1
			);
			if(DBIsObjectGarbage(obj_num))
			    obj_num = ScannerButtonMatch(
				bw, x, y,
				0
			    );
		    }
		    else
		    {
			obj_num = ScannerButtonMatch(
			    bw, x, y,
			    0
			);
		    }
		    if(DBIsObjectGarbage(obj_num))
		    {
			/* No match */

			/* Disengage intercept if button pressed is
			 * Button2 on a failed object match
			 */
			if(button_num == Button2)
			{
			    /* Disengage intercept */
			    NetSendIntercept(
				net->player_obj_num,
				"#off"
			    );
 			}
		    }
		    else
		    {
			/* Matched object, only use obj_num for
			 * scanner/weapons lock, all else use
			 * obj->locked_on
			 */

			/* Button1 */
			if(button_num == Button1)
			{
			    if(button->state & OSW_KEY_STATE_SHIFT)
			    {
				/* Hail specific vessel */
				NetSendHail(
				    net->player_obj_num,
				    obj->locked_on,
				    obj->com_channel
				);
			    }
			    else
			    {
				/* Send scanner and weapons lock */
				NetSendWeaponsLock(
				    net->player_obj_num,
				    obj_num
				);
			    }
			}
			/* Button2 */
			else if(button_num == Button2)
			{
			    if((button->state & OSW_KEY_STATE_SHIFT) &&
			       !DBIsObjectGarbage(obj->locked_on)
			    )
			    {
				/* ??? */
			    }
			    else if(!DBIsObjectGarbage(obj->locked_on))
			    {
				const char *name;
				char *s;
				xsw_object_struct *tar_obj = xsw_object[obj->locked_on];

				/* Send set intercept */
				s = (char *)malloc(40);
				sprintf(s, "#%i", obj->locked_on);
				NetSendIntercept(
				    net->player_obj_num,
				    s
				);
				free(s);

				/* Print intercept message */
				name = (opt->show_formal_label == XSW_SHOW_FORMAL_LABEL_ALWAYS) ?
				    DBGetFormalNameStr(obj->locked_on) :
				    tar_obj->name;
				s = (char *)malloc(STRLEN(name) + 80);
				sprintf(
				    s,
"Setting course to `%s'.",
				    name
				);
				XSWPrint(s);
				free(s);
			    }
			}
			/* Button3 */
			else if(button_num == Button3)
			{
			    if((button->state & OSW_KEY_STATE_SHIFT) &&
			       !DBIsObjectGarbage(obj->locked_on)
			    )
			    {
				/* Disarm weapon */
				xsw_object_struct *tar_obj = xsw_object[obj->locked_on];
				const char *name = (opt->show_formal_label == XSW_SHOW_FORMAL_LABEL_ALWAYS) ?
				    DBGetFormalNameStr(obj->locked_on) :
				    tar_obj->name;
				char *s = (char *)malloc(STRLEN(name) + 80);
				NetSendWeaponDisarm(
				    net->player_obj_num,
				    obj->locked_on
				);
				sprintf(
				    s,
"Disarm weapon code sent to `%s'.",
				    name
				);
				XSWPrint(s);
				free(s);
			    }
			    else
			    {
				/* ??? */
			    }
			}

			BridgePlaySoundButtonClick(bw, 0.2, 1.0);
		    }
		}
		events_handled++;
	    }
	    /* Message box */
	    else if(w == bw->mesg_box)
	    {
		xsw_msg_win_struct *mw = xsw_msg_win;
		scroll_bar_struct *sb = &bw->mesg_box_sb;

		switch(button->button)
		{
		  case Button1:
		    MsgListUnmarkAll();
		    bw->mesg_box_sel_in_progress = True;
		    bw->mesg_box_sel_start_x = button->x;
		    bw->mesg_box_sel_start_y = button->y;
		    break;
		  case Button2:
		    break;
		  case Button3:
		    if(mw != NULL)
		    {
			if(mw->map_state)
			    MsgWinUnmap(mw);
			else
			    MsgWinMap(mw);
		    }
		    break;
		  case Button4:
		    SBarSetPositionY(
			sb,
			SBarGetPositionY(sb) - sb->y_step_inc
		    );
		    break;
		  case Button5:
		    SBarSetPositionY(
			sb,
			SBarGetPositionY(sb) + sb->y_step_inc
		    );
		    break;
		}
		events_handled++;
	    }
	    /* Quick menu (should be checked last) */
	    else if(button->button == Button3)
	    {
		/* Map quick menu */
		if((w == bw->toplevel) ||
		   (w == bw->viewscreen) ||
		   (w == bw->pan_p1) ||
		   (w == bw->pan_p2) ||
		   (w == bw->pan_p3) ||
		   (w == bw->pan_p4) ||
		   (w == bw->pan_s1) ||
		   (w == bw->pan_s2) ||
		   (w == bw->pan_s3)
		)
		{
		    VSMenuMap(xsw_vsmenu);
		    events_handled++;
		}
	    }
	    break;

	  case ButtonRelease:
	    button = &event->xbutton;

	    /* Weapon yield input window */
	    if(w == bw->weapon_yield_iwin)
	    {
		unsigned int width, height;
		win_t w = bw->weapon_yield_iwin;

		/* Adjust weapon yield */
		OSWGetWindowGeometry(w, NULL, NULL, &width, &height, NULL); 
		if(width > 1)
		    pvals->weapon_yield = CLIP(
			((float)button->x / ((float)width - 1)),
			0.0f, 1.0f
		    );
		else
		    pvals->weapon_yield = 1.0f;

		obj = net->player_obj_ptr;
		obj_num = net->player_obj_num;
		if(obj != NULL)
		{
		    BridgeDrawPanel(
			obj_num,
			BRIDGE_PANEL_PWEAPONS  
		    );
		}

		/* Ungrab pointer and destroy the weapon yield input
		 * window
		 */
		OSWUngrabPointer();
		OSWDestroyWindow(&bw->weapon_yield_iwin);
		events_handled++;
	    }
	    /* Message box */
	    else if(w == bw->mesg_box)
	    {
		if(bw->mesg_box_sel_in_progress)
		{
		    BridgeMessagesRemoveScrollTimeout(bw);
		    MsgListCopyToClipboard(bw->mesg_box);
		    bw->mesg_box_sel_in_progress = False;
		    MsgListQueueDrawAllWindows();
		}
		events_handled++;
	    }
	    break;

	  case MotionNotify:
	    motion = &event->xmotion;
	    /* Weapon yield input window */
	    if(w == bw->weapon_yield_iwin)
	    {
		unsigned int width, height;

		/* Adjust weapon yield */
		OSWGetWindowGeometry(w, NULL, NULL, &width, &height, NULL);
		if(width > 1)
		    pvals->weapon_yield = CLIP(
			((float)motion->x / ((float)width - 1)),
			0.0f, 1.0f
		    );
		else
		    pvals->weapon_yield = 1.0f;

		obj = net->player_obj_ptr;
		obj_num = net->player_obj_num;
		if(obj != NULL)
		    BridgeDrawPanel(
			obj_num,
			BRIDGE_PANEL_PWEAPONS
		    );
	    }
	    /* Message box */
	    else if(w == bw->mesg_box)
	    {
		if(bw->mesg_box_sel_in_progress)
		{
		    MsgListUnmarkAll();
		    BridgeMessagesMarkCoordinates(
			bw,
			bw->mesg_box_sel_start_x,
			bw->mesg_box_sel_start_y,
			motion->x,
			motion->y
		    );
		    BridgeQueueDrawMessages(bw);
		    events_handled++;
		}
	    }
	    break;

	  case LeaveNotify:
	    if(w == bw->mesg_box)
	    {
		if(bw->mesg_box_sel_in_progress)
		{
		    BridgeMessagesRemoveScrollTimeout(bw);
		    BridgeMessagesAddScrollTimeout(bw, 0l);
		}
		events_handled++;
	    }
	    break;

	  case FocusIn:
	    if(w == bw->toplevel)
	    {
		bw->has_focus = True;

#ifdef JS_SUPPORT
		/* Reinitialize all the joysticks when focused? */
		if((opt->controller == XSW_CONTROLLER_JOYSTICK) &&
		   opt->focus_out_js_close      
		)
		{
		    /* Reinitialize all the joysticks */
		    int i;
		    xsw_jsmap_struct *jsm;
		    for(i = 0; i < total_jsmaps; i++)
		    {
			jsm = xsw_jsmap[i];
			if(jsm == NULL)
			    continue;

			JSMapInitializeJoystick(
			    jsm,
			    jsm->device_name,
			    fname.js_calib
			);
		    }
		}
#endif	/* JS_SUPPORT */
		events_handled++;
	    }
	    break;

	  case FocusOut:
	    if(w == bw->toplevel)
	    {
		bw->has_focus = False;

#ifdef JS_SUPPORT
		/* Close all the joysticks when unfocused? */
		if((opt->controller == XSW_CONTROLLER_JOYSTICK) &&
		   opt->focus_out_js_close
		)
		{
		    /* Close all the joysticks */
		    int i;
		    xsw_jsmap_struct *jsm;
		    for(i = 0; i < total_jsmaps; i++)
		    {
			jsm = xsw_jsmap[i];
			if(jsm == NULL)
			    continue;

			JSMapShutdownJoystick(jsm);
		    }
		}
#endif	/* JS_SUPPORT */

		events_handled++;
	    }
	    break;

	  case MapNotify:
	    if(w == bw->toplevel)
	    {
		if(!bw->map_state)
		{
		    WidgetMap(&bw->scanner_sb);

		    bw->map_state = True;
		}
		events_handled++;
	    }
	    break;

	  case UnmapNotify:
	    if(w == bw->toplevel)
	    {
		if(bw->map_state)
		{
		    /* Record the window position */
		    OSWGetWindowRootOrigin(
			w,
			&opt->bridge_x, &opt->bridge_y
		    );
		    opt->bridge_width = bw->width;
		    opt->bridge_height = bw->height;

		    WidgetUnmap(&bw->prompt);
		    WidgetUnmap(&bw->scanner_sb);

		    bw->map_state = False;
		}
		events_handled++;
	    }
	    break;

	  case VisibilityNotify:
	    visibility = &event->xvisibility;
	    if(w == bw->toplevel)
	    {
		bw->visibility_state = visibility->state;
	    }
	    else if(w == bw->viewscreen)
	    {
		bw->viewscreen_vis_state = visibility->state;
	    }
	    else if(w == bw->scanner)
	    {
		bw->scanner_vis_state = visibility->state;
	    }
	    break;

	  case SelectionClear:
	    if(event->xselectionclear.window == bw->mesg_box)
	    {
		MsgListUnmarkAll();
		MsgListQueueDrawAllWindows();
	    }
	    break;

	  case ClientMessage:
	    /* WM Take Focus? */
	    if(OSWEventIsTakeFocus(bw->toplevel, event, &t))
	    {
		OSWSetInputFocus(bw->toplevel, RevertToParent, t);
		events_handled++;
	    }
	    /* WM Delete Window? */
	    else if(OSWEventIsDeleteWindow(bw->toplevel, event))
	    {
		/* Pop runlevel to begin shutting down */
		XSWMainPop(NULL);
		events_handled++;
	    }
	    /* DND Enter? */
	    else if(OSWEventIsDNDEnter(
		bw->toplevel, event,
		&src_w,
		&bw->target_string, 1
	    ))
	    {
		events_handled++;
	    }
	    /* DND Position? */
	    else if(OSWEventIsDNDPosition(
		bw->toplevel, event,
		&src_w, &w,
		&x, &y,
		&root_x, &root_y,
		&t, &action
	    ))
	    {
		win_t prev_dnd_tar_w = bw->dnd_tar_w;

		/* Viewscreen */
		if(w == bw->viewscreen)
		{
		    /* Report status as accepted */
		    bw->dnd_tar_w = w;
		    bw->dnd_action = osw_atom.dnd_action_copy;
		    bw->dnd_x = x;
		    bw->dnd_y = y;
		    if(bw->map_state)
			VSDrawViewScreen(bw);
		}
		/* Scanner */
		else if(w == bw->scanner)
		{
		    /* Report status as accepted */
		    bw->dnd_tar_w = w;
		    bw->dnd_action = osw_atom.dnd_action_copy;
		    bw->dnd_x = x;
		    bw->dnd_y = y;

		}
		/* Non-DND subwindow */
		else
		{
		    /* Report status as not accepted */
		    bw->dnd_action = None;
		    bw->dnd_tar_w = None;
		}
		/* Position now over a different DND target window? */
		if(prev_dnd_tar_w != bw->dnd_tar_w)
		{
		    if(bw->map_state)
			BridgeWinDrawAll(bw);
		}
		OSWDNDSendStatus(
		    src_w, bw->toplevel, bw->dnd_action, NULL
		);
		events_handled++;
	    }
	    /* DND Leave? */
	    else if(OSWEventIsDNDLeave(
		bw->toplevel, event,
		&src_w
	    ))
	    {
		if(bw->dnd_tar_w != None)
		{
		    bw->dnd_tar_w = None;
		    bw->dnd_action = None;
		    if(bw->map_state)
			BridgeWinDrawAll(bw);
		}
		events_handled++;
	    }
	    /* DND Drop? */
	    else if(OSWEventIsDNDDrop(
		bw->toplevel, event,
		&src_w, &t
	    ))
	    {
		int len;
		char *url = (char *)OSWDNDGetData(
		    bw->toplevel,
		    bw->target_string,
		    &len,
		    t
		);
		OSWDNDSendFinished(
		    src_w, bw->toplevel, 
		    True,
		    bw->dnd_action
		);
		if(url != NULL)
		{
		    XSWConnect(url);
		    free(url);
		}
		bw->dnd_tar_w = None;
		bw->dnd_action = None;
		if(bw->map_state)
		    BridgeWinDrawAll(bw);
		events_handled++;
	    }
	    break;
	}


	/* Manage widgets */

	/* Viewsscreen Menu */
	if(events_handled == 0)
	    events_handled += VSMenuManage(xsw_vsmenu, event);

	/* Prompt */
	if(events_handled == 0)
	    events_handled += WidgetManage(&bw->prompt, event);

	/* Scanner scale bar */
	if(events_handled == 0)
	    events_handled += WidgetManage(&bw->scanner_sb, event);

	/* Message box scroll bar */
	if(events_handled == 0)
	    events_handled += WidgetManage(&bw->mesg_box_sb, event);


	return(events_handled);
}
