#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "../include/string.h"
#include "../include/osw-x.h"
#include "../include/widget.h"

#include "cdlg.h"

#include "../include/oswkeyvals.h"


static void CDlgDrawCB(void *widget, void *data);
static void CDlgResizeCB(void *widget, void *data);

static void CDlgYesBtnCB(push_button_struct *btn, void *data);
static void CDlgYesToAllBtnCB(push_button_struct *btn, void *data);
static void CDlgNoBtnCB(push_button_struct *btn, void *data);
static void CDlgCancelBtnCB(push_button_struct *btn, void *data);

int CDlgGetResponse(
	cdlg_struct *d,
	const char *title,
	const char *msg,
	image_t *icon,
	unsigned int btn_flags,
	int map_rel, win_t toplevel,
	void (*main_push_func)(void *),
	void *main_push_data,
	void (*main_pop_func)(void *),
	void *main_pop_data
);

int CDlgInit(cdlg_struct *d);
void CDlgResize(cdlg_struct *d);
void CDlgQueueResize(cdlg_struct *d);
void CDlgDraw(cdlg_struct *d);
void CDlgQueueDraw(cdlg_struct *d);
int CDlgManage(cdlg_struct *d, event_t *event);
void CDlgMap(cdlg_struct *d);
void CDlgUnmap(cdlg_struct *d);
void CDlgDestroy(cdlg_struct *d);


#define CDLG_BUTTON_WIDTH		70
#define CDLG_BUTTON_HEIGHT		28


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Resize callback.
 */
static void CDlgResizeCB(void *widget, void *data)
{
	cdlg_struct *d = CDLG(widget);
	if(d == NULL)
	    return;

	CDlgResize(d);
}

/*
 *	Draw callback.
 */
static void CDlgDrawCB(void *widget, void *data)
{
	cdlg_struct *d = CDLG(widget);
	if(d == NULL)
	    return;

	if(d->map_state)
	    CDlgDraw(d);
}


/*
 *	Yes button callback.
 */
static void CDlgYesBtnCB(push_button_struct *btn, void *data)
{
	cdlg_struct *d = CDLG(data);
	if(d == NULL)
	    return;

	d->response = CDlgYes;
	CDlgUnmap(d);
	if(d->main_pop_func != NULL)
	    d->main_pop_func(d->main_pop_data);
}

/*
 *	Yes to all button callback.
 */
static void CDlgYesToAllBtnCB(push_button_struct *btn, void *data)
{
	cdlg_struct *d = CDLG(data);
	if(d == NULL)
	    return;

	d->response = CDlgYesToAll;
	CDlgUnmap(d);
	if(d->main_pop_func != NULL)
	    d->main_pop_func(d->main_pop_data);
}

/*
 *	No button callback.
 */
static void CDlgNoBtnCB(push_button_struct *btn, void *data)
{
	cdlg_struct *d = CDLG(data);
	if(d == NULL)
	    return;

	d->response = CDlgNo;
	CDlgUnmap(d);
	if(d->main_pop_func != NULL)
	    d->main_pop_func(d->main_pop_data);
}

/*
 *	Cancel button callback.
 */
static void CDlgCancelBtnCB(push_button_struct *btn, void *data)
{
	cdlg_struct *d = CDLG(data);
	if(d == NULL)
	    return;

	d->response = CDlgCancel;
	CDlgUnmap(d);
	if(d->main_pop_func != NULL)
	    d->main_pop_func(d->main_pop_data);
}


/*
 *	Maps the confirmation dialog and blocks until user response.
 *
 *	Returns one of CDlg* response codes.
 */
int CDlgGetResponse(
	cdlg_struct *d,
	const char *title,
	const char *msg,
	image_t *icon,
	unsigned int btn_flags,
	int map_rel, win_t toplevel,
	void (*main_push_func)(void *),
	void *main_push_data,
	void (*main_pop_func)(void *),
	void *main_pop_data
)
{
	win_t w;

	if(d == NULL)
	    return(CDlgError);

	w = d->toplevel;
	if((w == None) || d->map_state)
	    return(CDlgError);

	/* Cannot query if blocking input */
	if(OSWBlockInputLevel() > 0)
	    return(CDlgError);

	/* Set values */
	d->map_rel = map_rel;
	d->transient_for_w = toplevel;
	OSWSetTransientFor(d->toplevel, toplevel);

	OSWSetWindowTitle(w, (title != NULL) ? title : "Confirmation");
	free(d->msg);
	d->msg = STRDUP(msg);
	d->icon = icon;
	d->btn_flags = btn_flags;

	d->main_push_func = main_push_func;
	d->main_push_data = main_push_data;
	d->main_pop_func = main_pop_func;
	d->main_pop_data = main_pop_data;

	d->response = CDlgError;

	/* Map the dialog */
	CDlgMap(d);

	/* Push main until response */
	if(d->main_push_func != NULL)
	    d->main_push_func(d->main_push_data);

	return(d->response);
}


/*
 *	Initializes the confirmation dialog.
 */
int CDlgInit(cdlg_struct *d)
{
	int x, y;
	unsigned int width, height;
	keyval_t accel_keys[4];
	osw_gui_struct *gui = &osw_gui[0];
	win_t w, parent = gui->root_win;
	push_button_struct *btn;
	widget_global_struct *wglobal = &widget_global;

	if(d == NULL)
	    return(-2);

	d->map_state = False;
	d->has_focus = False;
	d->x = 0;
	d->y = 0;
	d->width = 100;
	d->height = 75;
	d->visibility_state = VisibilityFullyObscured;
	d->font = wglobal->std_font;

	d->map_rel = WIDGET_WINDOW_RELATIVE_POINTER;
	d->transient_for_w = None;

	d->msg = NULL;
	d->icon = None;
	d->btn_flags = 0;

	d->main_push_func = NULL;
	d->main_push_data = NULL;
	d->main_pop_func = NULL;
	d->main_pop_data = NULL;

	d->response = CDlgError;

	/* Create toplevel */
	x = d->x;
	y = d->y;
	width = d->width;
	height = d->height;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	d->toplevel = w;
	OSWSetWindowWMProperties(
	    w,
	    "Confirmation",
	    "Confirmation",
	    widget_global.std_icon_pm,		/* Icon */
	    False,				/* Let WM set coordinates? */
	    x, y,
	    width, height,
	    gui->display_width, gui->display_height,
	    OSW_WINDOW_FRAME_DIALOG,
	    0, NULL
	);
	OSWSetWindowInput(w, OSW_EVENTMASK_TOPLEVEL);

	parent = w;

	/* Buttons */
	btn = &d->yes_btn;
	width = CDLG_BUTTON_WIDTH;
	height = CDLG_BUTTON_HEIGHT;
	x = 0;
	y = 0;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Yes",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    CDlgYesBtnCB, d
	);
	accel_keys[0] = OSW_KEY_y;
	accel_keys[1] = OSW_KEY_Return;
	PBtnSetAccelKeys(btn, accel_keys, NULL, 2);

	btn = &d->no_btn;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "No",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    CDlgNoBtnCB, d
	);
	accel_keys[0] = OSW_KEY_n;
	accel_keys[1] = OSW_KEY_Escape;
	PBtnSetAccelKeys(btn, accel_keys, NULL, 2);

	btn = &d->yes_to_all_btn;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "All",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    CDlgYesToAllBtnCB, d
	);
	accel_keys[0] = OSW_KEY_a;
	PBtnSetAccelKeys(btn, accel_keys, NULL, 1);

	btn = &d->cancel_btn;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Cancel",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    CDlgCancelBtnCB, d
	);
	accel_keys[0] = OSW_KEY_c;
	accel_keys[1] = OSW_KEY_Escape;
	PBtnSetAccelKeys(btn, accel_keys, NULL, 2);

	return(0);
}

/*
 *	Resizes the confirmation dialog.
 */
void CDlgResize(cdlg_struct *d)
{
	const int	x_margin = 10, y_margin = 10;
	int x, y;
	unsigned int width, height, btn_flags;
	win_t w;
	push_button_struct *btn;

	if(d == NULL)
	    return;

	w = d->toplevel;
	if(w == None)
	    return;

	OSWGetWindowGeometry(w, &x, &y, &width, &height, NULL);
	d->x = x;
	d->y = y;
	d->width = width;
	d->height = height;

	OSWDestroyPixmap(&d->toplevel_buf);

	btn_flags = d->btn_flags;

	/* Map/Unmap/Move buttons */

	/* Yes button */
	y = (int)d->height - y_margin - CDLG_BUTTON_HEIGHT;
	btn = &d->yes_btn;
	if(btn_flags & CDlgBtnYes)
	{
	    x = (int)d->width - (int)CDLG_BUTTON_WIDTH - x_margin;
	    if(btn_flags & CDlgBtnYesToAll)
		x -= (int)CDLG_BUTTON_WIDTH + x_margin;
	    if(btn_flags & CDlgBtnNo)
		x -= (int)CDLG_BUTTON_WIDTH + x_margin;
	    if(btn_flags & CDlgBtnCancel)
		x -= (int)CDLG_BUTTON_WIDTH + x_margin;
	    OSWMoveWindow(btn->toplevel, x, y);
	    WidgetMap(btn);
	}
	else
	{
	    WidgetUnmap(btn);
	}

	/* Yes To All button */
	btn = &d->yes_to_all_btn;
	if(btn_flags & CDlgBtnYesToAll)
	{
	    x = (int)d->width - (int)CDLG_BUTTON_WIDTH - x_margin;
	    if(btn_flags & CDlgBtnNo)
		x -= (int)CDLG_BUTTON_WIDTH + x_margin;
	    if(btn_flags & CDlgBtnCancel)
		x -= (int)CDLG_BUTTON_WIDTH + x_margin;
	    OSWMoveWindow(btn->toplevel, x, y);
	    WidgetMap(btn);
	}
	else
	{   
	    WidgetUnmap(btn);
	}

	/* No button */
	btn = &d->no_btn;
	if(btn_flags & CDlgBtnNo)
	{
	    x = (int)d->width - (int)CDLG_BUTTON_WIDTH - x_margin;
	    if(btn_flags & CDlgBtnCancel)
		x -= (int)CDLG_BUTTON_WIDTH + x_margin;
	    OSWMoveWindow(btn->toplevel, x, y);
	    WidgetMap(btn);
	}
	else
	{
	    WidgetUnmap(btn);
	}

	/* Cancel button */
	btn = &d->cancel_btn;
	if(btn_flags & CDlgBtnCancel)
	{
	    x = (int)d->width - (int)CDLG_BUTTON_WIDTH - x_margin;
	    OSWMoveWindow(btn->toplevel, x, y);
	    WidgetMap(btn);
	}
	else
	{
	    WidgetUnmap(btn);
	}
}

/*
 *	Queues the confirmation dialog to resize.
 */
void CDlgQueueResize(cdlg_struct *d)
{
	if(d == NULL)
	    return;

	if(d->toplevel == None)
	    return;

	WidgetQueueResizeAction(CDlgResizeCB, d, NULL);
}

/*
 *	Draws the confirmation dialog.
 */
void CDlgDraw(cdlg_struct *d)
{
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if(d == NULL)
	    return;

	if(d->toplevel == None)
	    return;

	/* Map the dialog as needed */
	if(!d->map_state)
	{
	    const int	x_margin = 10, y_margin = 10;
	    const unsigned int btn_flags = d->btn_flags;
	    int swidth, ascent, descent,
		font_height, line_len, nbtns;
	    unsigned int	width, height,
				msg_width, msg_height,
				icon_width, icon_height,
				btns_width;
	    const char *s;
	    font_t *font = d->font;
	    win_t w = d->toplevel;
	    image_t *icon = d->icon;

	    /* Calculate the size of the message */
	    msg_width = 0;
	    msg_height = 0;
	    s = d->msg;
	    while(s != NULL)
	    {
		line_len = strlinelen(s);
		OSWGetTextBounds(
		    font, s, line_len,
		    NULL, NULL, &swidth, &ascent, &descent
		);
		font_height = ascent + descent;

		if(swidth > (int)msg_width)
		    msg_width = (unsigned int)swidth;
		msg_height += (unsigned int)font_height;

		s += line_len;
		if(*s == '\n')
		    s += 1; 
		if(*s == '\r')
		    s += 1;
		if(*s == '\0')
		    s = NULL;
	    }
	    d->msg_width = msg_width;
	    d->msg_height = msg_height;

	    /* Get the size of the icon */
	    if(icon != NULL)
	    {
		icon_width = icon->width + x_margin;
		icon_height = icon->height;
	    }
	    else
	    {
		icon_width = 0;
		icon_height = 0;
	    }

	    /* Calculate the width to display all the buttons */
	    btns_width = 0;
	    nbtns = 0;
	    if(btn_flags & CDlgBtnYes)
	    {
		btns_width += CDLG_BUTTON_WIDTH;
		nbtns++;
	    }
	    if(btn_flags & CDlgBtnYesToAll)
	    {
		btns_width += CDLG_BUTTON_WIDTH;
		nbtns++;
	    }
	    if(btn_flags & CDlgBtnNo)
	    {
		btns_width += CDLG_BUTTON_WIDTH;
		nbtns++;
	    }
	    if(btn_flags & CDlgBtnCancel)
	    {
		btns_width += CDLG_BUTTON_WIDTH;
		nbtns++;
	    }
	    btns_width += (unsigned int)(MAX(
		(nbtns - 1), 0) * x_margin
	    );

	    /* Calculate and set the size of the toplevel */
	    width = MAX(
		icon_width + msg_width,
		btns_width
	    ) + (unsigned int)(2 * x_margin);
	    height = (unsigned int)y_margin +
		MAX(icon_height, msg_height) +
		(2 * y_margin) + 2 +
		CDLG_BUTTON_HEIGHT +
		(unsigned int)y_margin;
	    OSWResizeWindow(w, width, height);

	    /* Set the initial position */
	    switch(d->map_rel)
	    {
	      case WIDGET_WINDOW_RELATIVE_PARENT:
		WidgetCenterWindowToWindow(w, d->transient_for_w);
		break;
	      case WIDGET_WINDOW_RELATIVE_ROOT:
	      case WIDGET_WINDOW_RELATIVE_POINTER:
		WidgetCenterWindow(w, d->map_rel);
		break;
	    }

	    OSWGetWindowGeometry(
		d->toplevel,
		&d->x, &d->y, &d->width, &d->height, NULL
	    );

	    /* Map toplevel */
	    OSWMapRaised(w);
	    d->map_state = True;
	}

	/* Recreate buffers as needed */
	if(d->toplevel_buf == None)
	    OSWCreatePixmap(&d->toplevel_buf, d->width, d->height);

	/* Draw the confirmation dialog */
	if(d->toplevel != None)
	{
	    const int	x_margin = 10,
			y_margin = 10;
	    int	x, y, line_len,
		lbearing, rbearing, swidth, ascent, descent,
		font_height;
	    unsigned int	width, height,
				icon_height,
				msg_height = d->msg_height;
	    const char *s;
	    font_t *font = d->font;
	    image_t *icon = d->icon;
	    win_t w = d->toplevel;
	    pixmap_t pixmap = d->toplevel_buf;

	    OSWGetWindowGeometry(
		w, &x, &y, &width, &height, NULL
	    );
	    icon_height = (icon != NULL) ? icon->height : 0;

	    /* Redraw background */
	    if(wglobal->force_mono)
	    {
		OSWClearPixmap(
		    pixmap, width, height,
		    gui->black_pix
		);
		OSWSetForeground(gui->white_pix);
	    }
	    else
	    { 
		WidgetPutImageTile(
		    pixmap,
		    wglobal->std_bkg_img,
		    width, height
		);
		OSWSetForeground(wglobal->normal_text_pix);
	    }

	    /* Draw messages */
	    OSWSetFont(font);
	    x = x_margin + ((icon != NULL) ?
		((int)icon->width + x_margin) : 0);
	    y = y_margin + ((msg_height < icon_height) ?
		(((int)icon_height - (int)msg_height) / 2) : 0);
	    s = d->msg;
	    while(s != NULL)
	    {
		line_len = strlinelen(s);
		OSWGetTextBounds(
		    font, s, line_len,
		    &lbearing, &rbearing, &swidth, &ascent, &descent
		);
		font_height = ascent + descent;
		OSWDrawStringLimited(
		    pixmap,
		    x - lbearing,
		    y + ascent,
		    s,
		    line_len
		);

		s += line_len;
		if(*s == '\n')
		    s += 1;
		if(*s == '\r')
		    s += 1;
		if(*s == '\0')
		    s = NULL;

		y += font_height;
	    }

	    /* Draw horizontal rule */
	    x = 0;
	    y = (int)d->height - y_margin - (int)CDLG_BUTTON_HEIGHT -
		y_margin - 2;
	    if(wglobal->force_mono)
	    {
		OSWSetForeground(gui->white_pix);
		OSWDrawLine(
		    pixmap,
		    0, y,
		    (int)d->width, y
		);
	    }
	    else
	    {
		OSWSetForeground(wglobal->surface_shadow_pix);
		OSWDrawLine(
		    pixmap,
		    0, y,
		    (int)d->width, y
		);
		OSWSetForeground(wglobal->surface_highlight_pix);
		OSWDrawLine(
		    pixmap,
		    0, y + 1,
		    (int)d->width, y + 1
		);
	    }

	    /* Icon */
	    if(icon != NULL)
	    {
		if(msg_height > icon->height)
		{
		    x = x_margin;
		    y = y_margin +
			(((int)msg_height - (int)icon->height) / 2);
		}
		else
		{
		    x = x_margin;
		    y = y_margin;
		}
		WidgetPutImageRaised(
		    pixmap, icon, x, y, 6
		);
	    }

	    OSWPutDrawableToWindow(w, pixmap);
	}
}

/*
 *	Queues the confirmation dialog to draw.
 */
void CDlgQueueDraw(cdlg_struct *d)
{
	if(d == NULL)
	    return;

	if(d->toplevel == None)
	    return;

	WidgetQueueDrawAction(CDlgDrawCB, d, NULL);
}

/*
 *	Manages the confirmation dialog.
 */
int CDlgManage(cdlg_struct *d, event_t *event)
{
	int events_handled = 0;
	time_t t;
	win_t w;
	event_visibility_t *visibility;

	if((d == NULL) || (event == NULL))
	    return(events_handled);

	if(d->toplevel == None)
	    return(events_handled);

	w = event->xany.window;

	switch(event->type)
	{
	  case ConfigureNotify:
	    if(w == d->toplevel)
	    {
		CDlgQueueResize(d);
		CDlgQueueDraw(d);
		events_handled++;
	    }
	    break;

	  case Expose:
	    if(w == d->toplevel)
	    {
		CDlgQueueDraw(d);
		events_handled++;
	    }
	    break;

          case KeyPress:
            if(!d->has_focus)
                return(events_handled);

	    break;

          case KeyRelease:
            if(!d->has_focus)
                return(events_handled);

	    break;

          case ButtonPress:
            /* Grab focus as needed */
            if(!d->has_focus)
            {
                if(OSWCheckWindowAncestory(d->toplevel, w))
                    OSWGrabFocus(d->toplevel);
            }
	    break;

	  case FocusIn:
	    if(w == d->toplevel)
	    {
		d->has_focus = True;
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case FocusOut:
	    if(w == d->toplevel)
	    {
		d->has_focus = False;
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case MapNotify:
	    if(w == d->toplevel)
	    {
		if(!d->map_state)
		    CDlgDraw(d);
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case UnmapNotify:
	    if(w == d->toplevel)
	    {
		if(d->map_state)
		{
		    WidgetUnmap(&d->no_btn);  
		    WidgetUnmap(&d->yes_btn); 
		    WidgetUnmap(&d->yes_to_all_btn); 
		    WidgetUnmap(&d->cancel_btn);
		    OSWDestroyPixmap(&d->toplevel_buf);
		    d->map_state = False;
		}
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case VisibilityNotify:
	    visibility = &event->xvisibility;
	    if(w == d->toplevel)
	    {
		d->visibility_state = visibility->state;
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case ClientMessage:
	    /* WM Take Focus? */
	    if(OSWEventIsTakeFocus(d->toplevel, event, &t))
	    {
		OSWGrabFocusTime(d->toplevel, t);
		events_handled++;
		return(events_handled);
	    }
	    /* WM Delete Window? */
	    else if(OSWEventIsDeleteWindow(d->toplevel, event))
	    {
		if(d->btn_flags & CDlgBtnCancel)
		    CDlgCancelBtnCB(&d->cancel_btn, d);
		else if(d->btn_flags & CDlgBtnNo)
		    CDlgNoBtnCB(&d->no_btn, d);
		events_handled++;
		return(events_handled);
	    }
	    break;
	}

	if(events_handled == 0)
	    events_handled += WidgetManage(&d->no_btn, event);

	if(events_handled == 0)
	    events_handled += WidgetManage(&d->yes_btn, event);

	if(events_handled == 0)
	    events_handled += WidgetManage(&d->yes_to_all_btn, event);

	if(events_handled == 0)  
	    events_handled += WidgetManage(&d->cancel_btn, event);

	return(events_handled);
}

/*
 *	Maps the confirmation dialog.
 */
void CDlgMap(cdlg_struct *d)
{ 
	if(d == NULL)
	    return;

	/* Map the dialog */
	d->map_state = False;
	CDlgDraw(d);

	OSWSetModal(d->toplevel);
	OSWGrabFocus(d->toplevel);
}

/*
 *	Unmaps the confirmation dialog.
 */
void CDlgUnmap(cdlg_struct *d)
{ 
	if(d == NULL)
	    return;

	OSWUnmapWindow(d->toplevel);
}

/*
 *	Destroys the confirmation dialog.
 */
void CDlgDestroy(cdlg_struct *d)
{
	if(d == NULL)
	    return;

	d->map_rel = 0;
	d->transient_for_w = None;

	free(d->msg);
	d->msg = NULL;
	d->icon = NULL;				/* Shared */
	d->btn_flags = 0;

	d->main_push_func = NULL;
	d->main_push_data = NULL;
	d->main_pop_func = NULL;
	d->main_pop_data = NULL;

	d->response = 0;

	WidgetDestroy(&d->no_btn);
	WidgetDestroy(&d->yes_btn);
	WidgetDestroy(&d->yes_to_all_btn);
	WidgetDestroy(&d->cancel_btn);

	OSWDestroyPixmap(&d->toplevel_buf);
	OSWDestroyWindow(&d->toplevel);
}
