#include <stdlib.h>
#include <string.h>

#include "../include/string.h"
#include "../include/unvmatch.h"
#include "../include/osw-x.h"
#include "../include/widget.h"

#include "net.h"
#include "xsw.h"
#include "ecowin.h"

#include "../include/oswkeyvals.h"


static char *EcoWinGetParsedName(const char *s);

static void EcoWinUnfocusAll(xsw_eco_win_struct *ew);

void EcoWinUpdateCredits(
	xsw_eco_win_struct *ew,
	const int obj_num
);
void EcoWinAddProduct(
	xsw_eco_win_struct *ew,
	const int proprietor_obj_num,		/* Can be -1 */
	xsw_ecoproduct_struct product
);
void EcoWinClearProducts(xsw_eco_win_struct *ew);

static void EcoWinResizeCB(void *widget, void *data);
static void EcoWinDrawCB(void *widget, void *data);
static void EcoWinInventoryCListDataDestroyCB(void *data);
static void EcoWinProprietorPromptCB(
	prompt_struct *prompt, const char *s, void *data
);
static void EcoWinRefreshPBCB(push_button_struct *btn, void *data);
static void EcoWinBuyPBCB(push_button_struct *btn, void *data);
static void EcoWinSellPBCB(push_button_struct *btn, void *data);
static void EcoWinClosePBCB(push_button_struct *btn, void *data);

xsw_eco_win_struct *EcoWinNew(void);
void EcoWinResize(xsw_eco_win_struct *ew);
void EcoWinQueueResize(xsw_eco_win_struct *ew);
void EcoWinDraw(xsw_eco_win_struct *ew);
void EcoWinQueueDraw(xsw_eco_win_struct *ew);
int EcoWinManage(xsw_eco_win_struct *ew, event_t *event);
void EcoWinMap(xsw_eco_win_struct *ew);
void EcoWinUnmap(xsw_eco_win_struct *ew);
void EcoWinDelete(xsw_eco_win_struct *ew);


#define EW_DEF_WIDTH			400
#define EW_DEF_HEIGHT			400

#define EW_MIN_WIDTH			100
#define EW_MIN_HEIGHT			100

#define EW_TITLE			"Buy or Sell"
#define EW_TITLE_ICON			"Buy or Sell"

#define EW_BTN_WIDTH			70
#define EW_BTN_HEIGHT			28

#define EW_NAME_PROMPT_NAME		"Proprietor:"
#define EW_VALUE_PROMPT_NAME		"Quantity:"


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : True)


/*
 *	Returns a copy of the simplified product name string.
 */
static char *EcoWinGetParsedName(const char *s)
{
	if(s == NULL)
	    return(NULL);

	while(ISBLANK(*s))
	    s++;

	/* Not a token string? */
	if(*s == '$')
	{
	    s++;
	    while((*s != ':') && (*s != '\0'))
		s++;
	    if(*s == ':')
		s++;

	    while(ISBLANK(*s))
		s++;

	    return(strdup(s));
	}

	return(strdup(s));
}


/*
 *	Unfocus all widgets.
 */
static void EcoWinUnfocusAll(xsw_eco_win_struct *ew)
{
	clist_struct *clist;

	if(ew == NULL)
	    return;

	ew->proprietor_prompt.has_focus = False;
	clist = &ew->inventory;
	WidgetSetFocus(clist, False);
	ew->amount_prompt.has_focus = False;
}


/*
 *	Updates the object's credits displayed.
 *
 *	The specified object must be either the customer object or
 *	the proprietor object, otherwise it is ignored.
 */
void EcoWinUpdateCredits(xsw_eco_win_struct *ew, const int obj_num)
{
	xsw_object_struct       *customer_obj = NULL,
				*proprietor_obj = NULL;
	xsw_net_parms_struct *net = &net_parms;

	if(ew == NULL)
	    return;

	if((obj_num != net->player_obj_num) &&
	   (obj_num != ew->proprietor_obj_num)
	)
	    return;

	if(!DBIsObjectGarbage(net->player_obj_num))
	    customer_obj = xsw_object[net->player_obj_num];
	if(!DBIsObjectGarbage(ew->proprietor_obj_num))
	    proprietor_obj = xsw_object[ew->proprietor_obj_num];

	/* Update the credits value label */
	if((customer_obj != NULL) &&
	   (proprietor_obj != NULL)
	)
	{
	    char *s = (char *)malloc(40 + 40 + 80);
	    sprintf(
		s,
"%.2f\n\
%.2f",
		(customer_obj->score != NULL) ?
		    customer_obj->score->credits : 0.0f,
		(proprietor_obj->score != NULL) ?
		    proprietor_obj->score->credits : 0.0f
	    );
	    LabelSetText(&ew->credits_value_label, s);
	    free(s);
	}
	else if(customer_obj != NULL)
	{
	    char *s = (char *)malloc(40 + 40 + 80);
	    sprintf(
		s,
"%.2f\n\
%.2f",
		(customer_obj->score != NULL) ?
		    customer_obj->score->credits : 0.0f,
		0.0f
	    );
	    LabelSetText(&ew->credits_value_label, s);
	    free(s);
	}
	else
	{
	    LabelSetText(
		&ew->credits_value_label,
"0.00\n\
0.00"
	    );
	}

	/* Update the taxes label */
	if((proprietor_obj != NULL) ?
	    (proprietor_obj->eco != NULL) : False
	)
	{
	    char *s = (char *)malloc(40 + 40 + 40 + 40 + (2 * 80));
	    xsw_ecodata_struct *eco = proprietor_obj->eco;
	    sprintf(
		s,
"%s\n\
Tax(F/G/H): %i/%i/%i%%",
		(eco->flags & ECO_FLAG_OPEN) ? "Opened" : "Closed",
		(int)((eco->tax_friend - 1.0f) * 100.0f),
		(int)((eco->tax_general - 1.0f) * 100.0f),
		(int)((eco->tax_hostile - 1.0f) * 100.0f)
	    );
	    LabelSetText(&ew->taxes_label, s);
	    free(s);
	}
}

/*
 *	Adds (or updates) a product on the inventory list.
 *
 *	If the proprietor object is not the same object currently
 *	set on the eco window then the inventory list will be cleared
 *	before adding the specified product.
 */
void EcoWinAddProduct(
	xsw_eco_win_struct *ew,
	const int proprietor_obj_num,	/* Can be -1 */
	xsw_ecoproduct_struct product
)
{
	int i;
	xsw_object_struct *proprietor_obj = NULL;
	clist_struct *clist;
	xsw_ecoproduct_struct *dprod;

	if(ew == NULL)
	    return;

	clist = &ew->inventory;

	if(!DBIsObjectGarbage(proprietor_obj_num))
	    proprietor_obj = xsw_object[proprietor_obj_num];

	/* Target object -1 or different from last? */
	if((proprietor_obj_num < 0) ||
	   (ew->proprietor_obj_num != proprietor_obj_num)
	)
	{
	    /* The proprietor has changed */

	    /* Clear the inventory list */
	    CListClear(clist);

	    /* Change name in proprietor prompt */
	    PromptSetS(
		&ew->proprietor_prompt,
		(proprietor_obj != NULL) ?
		    proprietor_obj->name : ""
	    );
	    PromptMarkBuffer(
		&ew->proprietor_prompt,
		PROMPT_POS_END
	    );
	}

	/* Do not continue if the proprietor object is -1 */
	if(proprietor_obj_num < 0)
	{
	    /* Set proprietor object on eco window to none */
	    ew->proprietor_obj_num = -1;

	    /* Redraw inventory list as needed */
	    if(ew->map_state)
		CListQueueDraw(clist, CLIST_DRAW_AMOUNT_COMPLETE);

	    return;
	}
	else
	{
	    /* Set the new proprietor object */
	    ew->proprietor_obj_num = proprietor_obj_num;
	}


	/* Check if the product is already in the list */
	for(i = 0; i < clist->total_rows; i++)
	{
	    dprod = (xsw_ecoproduct_struct *)CListRowGetData(clist, i);
	    if(dprod == NULL)
		continue;

	    if(dprod->name == NULL)
		continue;

	    if(!strcasecmp(dprod->name, product.name))
		break;
	}
	if(i < clist->total_rows)
	{
	    /* List item already exists, update it */
	    const char *v[5];
	    char	*name,
			sell_price_text[40],
			buy_price_text[40],
			amount_text[40],
			amount_max_text[40];
	    const int row_num = i;

	    /* Name (dynamically allocated) */
	    v[0] = name = EcoWinGetParsedName(product.name);

	    /* Sell price */
	    if(product.sell_price > 0.0f)
		sprintf(sell_price_text, "%.2f", product.sell_price);
	    else
		strcpy(sell_price_text, "-");
	    v[1] = sell_price_text;

	    /* Buy price */
	    if(product.buy_price > 0.0f)
		sprintf(buy_price_text, "%.2f", product.buy_price);
	    else
		strcpy(buy_price_text, "-");
	    v[2] = buy_price_text;

	    /* Amount */
	    if(product.amount_max >= 0.0f)
		sprintf(amount_text, "%.2f", product.amount);
	    else
		strcpy(amount_text, "Unlimited");
	    v[3] = amount_text;

	    /* Amount max */
	    if(product.amount_max >= 0.0f)
		sprintf(amount_max_text, "%.2f", product.amount_max);
	    else
		strcpy(amount_max_text, "Unlimited");
	    v[4] = amount_max_text;

	    CListSetRowTexts(
		clist, row_num,
		v, sizeof(v) / sizeof(char *)
	    );

	    /* Update the product */
	    dprod = (xsw_ecoproduct_struct *)CListRowGetData(
		clist, row_num
	    );
	    if(dprod != NULL)
	    {
		strncpy(dprod->name, product.name, sizeof(dprod->name));
		dprod->name[sizeof(dprod->name) - 1] = '\0';
		dprod->ocs_code = product.ocs_code;
		dprod->sell_price = product.sell_price;
		dprod->buy_price = product.buy_price;
		dprod->amount = product.amount;
		dprod->amount_max = product.amount_max;
	    }

	    CListAutosizeColumns(clist);

	    free(name);
	}
	else
	{
	    const char *v[5];
	    char        *name,
			sell_price_text[40],
			buy_price_text[40],
			amount_text[40],
			amount_max_text[40];

	    /* List item does not exist, append it */
	    const int row_num = CListRowAppend(clist);
	    if(row_num < 0)
	        return;

	    /* Name (dynamically allocated) */
	    v[0] = name = EcoWinGetParsedName(product.name);

	    /* Sell price */
	    if(product.sell_price > 0.0f)
		sprintf(sell_price_text, "%.2f", product.sell_price);
	    else
		strcpy(sell_price_text, "-");
	    v[1] = sell_price_text;

	    /* Buy price */
	    if(product.buy_price > 0.0f)
		sprintf(buy_price_text, "%.2f", product.buy_price);
	    else
		strcpy(buy_price_text, "-");
	    v[2] = buy_price_text;

	    /* Amount */
	    if(product.amount_max >= 0.0f)
		sprintf(amount_text, "%.2f", product.amount);
	    else
		strcpy(amount_text, "Unlimited");
	    v[3] = amount_text;

	    /* Amount max */
	    if(product.amount_max >= 0.0f)
		sprintf(amount_max_text, "%.2f", product.amount_max);
	    else
		strcpy(amount_max_text, "Unlimited");
	    v[4] = amount_max_text;

	    CListSetRowTexts(
		clist, row_num,
		v, sizeof(v) / sizeof(char *)
	    );

	    /* Create a new product data for this new row */
	    dprod = (xsw_ecoproduct_struct *)calloc(
		1, sizeof(xsw_ecoproduct_struct)
	    );
	    if(dprod != NULL)
	    {
		strncpy(dprod->name, product.name, sizeof(dprod->name));
		dprod->name[sizeof(dprod->name) - 1] = '\0';
		dprod->ocs_code = product.ocs_code;
		dprod->sell_price = product.sell_price;
		dprod->buy_price = product.buy_price;
		dprod->amount = product.amount;  
		dprod->amount_max = product.amount_max;

		/* Set the new product data for the new row */
		CListRowSetData(
		    clist,
		    row_num,
		    dprod,
		    EcoWinInventoryCListDataDestroyCB
		);
	    }

	    CListAutosizeColumns(clist);

	    free(name);
	}
}


/*
 *	Clears the economy window's inventory list.
 */
void EcoWinClearProducts(xsw_eco_win_struct *ew)
{
	xsw_ecoproduct_struct product;

	if(ew == NULL)
	    return;

	memset(&product, 0x00, sizeof(xsw_ecoproduct_struct));

	/* Clear inventory by passing the proprietor object number
	 * as -1
	 */
	EcoWinAddProduct(
	    ew,
	    -1,
	    product
	);
}


/*
 *	Resize callback.
 */
static void EcoWinResizeCB(void *widget, void *data)
{
	xsw_eco_win_struct *ew = (xsw_eco_win_struct *)widget;
	if(ew == NULL) 
	    return;

	EcoWinResize(ew);
}

/*
 *	Draw callback.
 */
static void EcoWinDrawCB(void *widget, void *data)
{
	xsw_eco_win_struct *ew = (xsw_eco_win_struct *)widget;
	if(ew == NULL) 
	    return;

	if(ew->map_state)
	    EcoWinDraw(ew);
}

/*
 *	Inventory clist data destroy callback.
 */
static void EcoWinInventoryCListDataDestroyCB(void *data)
{
	xsw_ecoproduct_struct *prod = (xsw_ecoproduct_struct *)data;
	if(prod == NULL)
	    return;

	free(prod);
}

/*
 *	Proprietor prompt callback.
 */
static void EcoWinProprietorPromptCB(
	prompt_struct *prompt, const char *s, void *data  
)
{
	xsw_eco_win_struct *ew = (xsw_eco_win_struct *)data;
	if(ew == NULL)
	    return;

	EcoWinRefreshPBCB(&ew->refresh_btn, ew);
}

/*
 *	Refresh callback.
 */
static void EcoWinRefreshPBCB(push_button_struct *btn, void *data)
{
	int proprietor_obj_num = -1;
	const char *name;
	xsw_eco_win_struct *ew = (xsw_eco_win_struct *)data; 
	if(ew == NULL)
	    return;

	name = PromptGetS(&ew->proprietor_prompt);
	if(name == NULL)
	    return;

	/* Match new proprietor object */
	if((*name == '\0') ||
	   !strcasecmp(name, "me")
	)
	    proprietor_obj_num = net_parms.player_obj_num;
	else
	    proprietor_obj_num = UNVMatchObjectByName(
		xsw_object, total_objects,
		name,
		-1			/* Any type */
	    );

	/* Unable to match proprietor object? */
	if(proprietor_obj_num < 0)
	{
	    char *s = (char *)malloc(STRLEN(name) + 80);
	    sprintf(
		s,
"Unable to find proprietor object:\n\n    `%s'",
		name
	    );
	    XSWDialog(
		"Unable To Find Object",
		XSW_IMG_DIALOG_ICON_WARNING,
		s,
		ew->toplevel
	    );
	    free(s);
	}
	else
	{
	    /* Delete all inventory list items */
	    EcoWinClearProducts(ew);

	    /* Request new economy values from server */
	    NetSendEcoReqValues(proprietor_obj_num);
	    if(!XSW_CON_IS_CONNECTED(&net_parms))
		return;
	}
}

/*
 *	Buy callback.
 */
static void EcoWinBuyPBCB(push_button_struct *btn, void *data)
{
	int i, row_num;
	clist_struct *clist;
	xsw_ecoproduct_struct *dprod, product;
	xsw_eco_win_struct *ew = (xsw_eco_win_struct *)data;
	if(ew == NULL)
	    return;

	if(ew->proprietor_obj_num < 0)
	    return;

	/* Get amount from amount prompt */
	memset(&product, 0x00, sizeof(xsw_ecoproduct_struct));
	product.amount = PromptGetF(&ew->amount_prompt);

	/* Iterate through the selected products list */
	clist = &ew->inventory;
	for(i = 0; i < clist->total_sel_rows; i++)
	{
	    row_num = clist->sel_row[i];
	    if((row_num < 0) || (row_num >= clist->total_rows))
		continue;

	    dprod = (xsw_ecoproduct_struct *)CListRowGetData(clist, row_num);
	    if(dprod == NULL)
		continue;

	    strncpy(product.name, dprod->name, sizeof(product.name));
	    product.name[sizeof(product.name) - 1] = '\0';
	    product.ocs_code = dprod->ocs_code;

	    /* Buy */
	    NetSendEcoBuy(
	        net_parms.player_obj_num,
	        ew->proprietor_obj_num,
	        &product
	    );
	    if(!XSW_CON_IS_CONNECTED(&net_parms))
		return;
	}
}

/*
 *	Sell callback.
 */
static void EcoWinSellPBCB(push_button_struct *btn, void *data)
{
	int i, row_num;
	clist_struct *clist;
	xsw_ecoproduct_struct *dprod, product;
	xsw_eco_win_struct *ew = (xsw_eco_win_struct *)data;
	if(ew == NULL)   
	    return;

	if(ew->proprietor_obj_num < 0)
	    return;

	/* Get amount from amount prompt */
	memset(&product, 0x00, sizeof(xsw_ecoproduct_struct));
	product.amount = PromptGetF(&ew->amount_prompt);

	/* Iterate through the selected products list */
	clist = &ew->inventory;
	for(i = 0; i < clist->total_sel_rows; i++)
	{
	    row_num = clist->sel_row[i]; 
	    if((row_num < 0) || (row_num >= clist->total_rows))
		continue;

	    dprod = (xsw_ecoproduct_struct *)CListRowGetData(clist, row_num);
	    if(dprod == NULL)
		continue;   

	    strncpy(product.name, dprod->name, sizeof(product.name));
	    product.name[sizeof(product.name) - 1] = '\0';
	    product.ocs_code = dprod->ocs_code;

	    /* Sell */
	    NetSendEcoSell(
		net_parms.player_obj_num,
		ew->proprietor_obj_num,
		&product
	    );
	    if(!XSW_CON_IS_CONNECTED(&net_parms))
		return;
	}
}

/*
 *	Close callback.
 */
static void EcoWinClosePBCB(push_button_struct *btn, void *data)
{
	xsw_eco_win_struct *ew = (xsw_eco_win_struct *)data;
	if(ew == NULL)
	    return;

	EcoWinUnmap(ew);
}


/*
 *	Creates a new Economy Window.
 */
xsw_eco_win_struct *EcoWinNew(void)
{
	int x, y, column_num;
	unsigned int width, height;
	keyval_t accel_keys[4];
	key_state_t accel_mods[4];
	pixel_t pix;
	font_t *font;
	pixmap_t pixmap;
	osw_gui_struct *gui = &osw_gui[0];
	win_t w, parent = gui->root_win;
	label_struct *label;
	push_button_struct *btn;
	prompt_struct *prompt;
	clist_struct *clist;
	xsw_option_struct *opt = &option;
	xsw_eco_win_struct *ew = XSW_ECO_WIN(calloc(
	    1, sizeof(xsw_eco_win_struct)
	));
	if(ew == NULL)
	    return(NULL);

	ew->map_state = False;
	ew->has_focus = False;
	ew->disabled = False;
	ew->freeze_count = 0;
	ew->busy_count = 0;
	ew->x = opt->ecow_x;
	ew->y = opt->ecow_y;
	ew->width = opt->ecow_width;
	ew->height = opt->ecow_height;
	if(ew->width < EW_MIN_WIDTH)
	    ew->width = EW_DEF_WIDTH;
	if(ew->height < EW_MIN_HEIGHT)
	    ew->height = EW_DEF_HEIGHT;
	if((ew->x + (int)ew->width) < 0)
	    ew->x = 0;
	if((ew->y + (int)ew->height) < 0)
	    ew->y = 0;
	if(ew->x > (int)gui->display_width)
	    ew->x = 0;
	if(ew->y > (int)gui->display_height)
	    ew->y = 0;
	ew->visibility_state = VisibilityFullyObscured;
	ew->proprietor_obj_num = -1;

	/* Create toplevel */
	x = ew->x;
	y = ew->y;
	width = ew->width;
	height = ew->height;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	ew->toplevel = w;
	OSWSetWindowInput(w, OSW_EVENTMASK_TOPLEVEL);

	/* WM properties */
	if(ImgIsAllocated(XSW_IMG_ICON_ECONOMY))
	    pixmap = OSWCreatePixmapFromImage( 
		xsw_image[XSW_IMG_ICON_ECONOMY]->image
	    );
	else
	    pixmap = widget_global.std_icon_pm;
	OSWSetWindowWMProperties(
	    w,
	    EW_TITLE,			/* Title */
	    EW_TITLE_ICON,		/* Icon title */
	    pixmap,			/* Icon */
	    False,			/* Let WM set coordinates? */
	    x, y,
	    EW_MIN_WIDTH, EW_MIN_HEIGHT,
	    gui->display_width, gui->display_height,
	    OSW_WINDOW_FRAME_STANDARD,
	    0, NULL
	);

	/* Set this window to be a transient for the bridge window */
	if(xsw_bridge_win != NULL)
	    OSWSetTransientFor(w, xsw_bridge_win->toplevel);

	parent = w;


	/* Credits labels */
	label = &ew->credits_name_label;
	x = 10;
	y = 10;
	width = 120;
	height = 35;
	LabelInit(
	    label, parent,
	    x, y, width, height,
"Your Credits:\n\
Proprietor Credits:"
	);
	LabelSetJustify(label, LABEL_JUSTIFY_RIGHT);

	label = &ew->credits_value_label;	
	x += (int)width + 2;
	width = 80;
	LabelInit(
	    label, parent,
	    x, y, width, height,
"0.00\n\
0.00"
	);
	LabelSetJustify(label, LABEL_JUSTIFY_LEFT);

	/* Taxes label */
	label = &ew->taxes_label;
	width = 160;
	x = (int)ew->width - 10 - (int)width;
	LabelInit(
	    label, parent,
	    x, y, width, height,
	    NULL
	);
	LabelSetJustify(label, LABEL_JUSTIFY_RIGHT);

	/* Proprietor prompt */
	prompt = &ew->proprietor_prompt;
	x = 0;
	y += (int)height + 5;
	width = (unsigned int)MAX(
	    ((int)ew->width - EW_BTN_WIDTH - 5 - 10),
	    0
	);
	height = 30;
	PromptInit(
	    prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Proprietor:",
	    SW_OBJ_NAME_MAX, 0
	);
	prompt->activate_cb = EcoWinProprietorPromptCB;
	prompt->activate_data = ew;

	/* Refresh button */
	btn = &ew->refresh_btn;
	width = EW_BTN_WIDTH;
	height = EW_BTN_HEIGHT;
	x = ew->width - (int)width - 10;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Refresh",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    EcoWinRefreshPBCB, ew
	);
	accel_keys[0] = OSW_KEY_r;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	accel_keys[1] = OSW_KEY_F5;
	accel_mods[1] = 0;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 2);

	/* Inventory clist */
	font = widget_global.std_font;
	pix = widget_global.normal_text_pix;
	clist = &ew->inventory;
	x = 0;
	y += (int)height + 5;
	width = ew->width;
	height = (unsigned int)MAX(
	    ((int)ew->height - y -
		(10 + EW_BTN_HEIGHT + 10 + 2 + 10 + 30 + 5)
	    ),
	    0
	);
	CListInit(
	    clist, parent,
	    x, y, width, height
	);
	clist->options |= CLIST_FLAG_ALLOW_MULTI_SELECT;
	/* Headings */
	column_num = CListColumnAppend(
	    clist,
	    "Product",
	    font, pix,
	    CLIST_COLUMN_FLAG_ALIGN_LEFT,
	    0
	);
	column_num = CListColumnAppend(
	    clist,
	    "Sell Price",
	    font, pix,
	    CLIST_COLUMN_FLAG_ALIGN_RIGHT,
	    0
	);
	column_num = CListColumnAppend(
	    clist,
	    "Buy Price",
	    font, pix,
	    CLIST_COLUMN_FLAG_ALIGN_RIGHT,
	    0
	);
	column_num = CListColumnAppend(
	    clist,
	    "Amount",
	    font, pix,
	    CLIST_COLUMN_FLAG_ALIGN_RIGHT,
	    0
	);
	column_num = CListColumnAppend(
	    clist,
	    "Total",
	    font, pix,
	    CLIST_COLUMN_FLAG_ALIGN_RIGHT,
	    0
	);
	CListAutosizeColumns(clist);

	/* Amount prompt */
	prompt = &ew->amount_prompt;
	x = 0;
	y = 60 + EW_BTN_HEIGHT + (int)ew->height - 180;
	width = ew->width;
	height = 30;
	PromptInit(
	    prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Quantity:",
	    64,			/* Buffer length */
	    0
	);
	PromptSetI(prompt, 0);

	/* Link prompts togeather */
	ew->proprietor_prompt.next = &ew->amount_prompt;
	ew->proprietor_prompt.prev = &ew->amount_prompt;

	ew->amount_prompt.next = &ew->proprietor_prompt;
	ew->amount_prompt.prev = &ew->proprietor_prompt;


	/* Buy button */
	btn = &ew->buy_btn;
	width = EW_BTN_WIDTH;
	height = EW_BTN_HEIGHT;
	x = 10;
	y = (int)ew->height - (int)height - 10;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Buy",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    EcoWinBuyPBCB, ew
	);
	accel_keys[0] = OSW_KEY_b;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 1);

	/* Sell button */
	btn = &ew->sell_btn;
	x += 10 + (int)width;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Sell",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    EcoWinSellPBCB, ew
	);
	accel_keys[0] = OSW_KEY_s;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 1);

	/* Close button */
	btn = &ew->close_btn;
	x = (int)ew->width - (int)width - 10;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Close",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    EcoWinClosePBCB, ew
	);
	accel_keys[0] = OSW_KEY_c;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	accel_keys[1] = OSW_KEY_w;
	accel_mods[1] = OSW_KEY_STATE_CTRL;
	accel_keys[2] = OSW_KEY_Escape;
	accel_mods[2] = 0;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 3);

	return(ew);
}

/*
 *	Resizes the economy window.
 */
void EcoWinResize(xsw_eco_win_struct *ew)
{
	int x, y;
	unsigned int width, height;
	win_t w;
	label_struct *label;
	push_button_struct *btn;
	prompt_struct *prompt;
	clist_struct *clist;

	if(ew == NULL)
	    return;

	w = ew->toplevel;
	OSWGetWindowGeometry(w, &x, &y, &width, &height, NULL);
	ew->x = x;
	ew->y = y;
	ew->width = width;
	ew->height = height;


	/* Credits labels */
	label = &ew->credits_name_label;
	w = label->toplevel;
	x = 10;
	y = 10;
	width = 120;
	height = 35;
	OSWMoveResizeWindow(w, x, y, width, height);
	WidgetQueueResize(label);

	label = &ew->credits_value_label;
	w = label->toplevel;
	x += 2 + (int)width;
	width = 80;
	OSWMoveResizeWindow(w, x, y, width, height);
	WidgetQueueResize(label);

	/* Taxes label */
	label = &ew->taxes_label;
	w = label->toplevel;
	width = 160;
	x = (int)ew->width - 10 - (int)width;
	OSWMoveResizeWindow(w, x, y, width, height);
	WidgetQueueResize(label);

	/* Refresh button */
	btn = &ew->refresh_btn;
	w = btn->toplevel;
	x = ew->width - (int)btn->width - 10;
	y += (int)height + 5;
	OSWMoveWindow(w, x, y);
	WidgetQueueResize(btn);

	/* Proprietor prompt */
	prompt = &ew->proprietor_prompt;   
	w = prompt->toplevel;
	x = 0;
	width = (unsigned int)MAX(
	    ((int)ew->width - (int)btn->width - 5 - 10),
	    0
	);
	height = 30;
	OSWMoveResizeWindow(w, x, y, width, height);   
	WidgetQueueResize(prompt);

	/* Inventory clist */
	clist = &ew->inventory;
	w = clist->toplevel;
	x = 0;
	y += (int)height + 5;
	width = ew->width;
	height = (unsigned int)MAX(
	    ((int)ew->height - y -
		(10 + EW_BTN_HEIGHT + 10 + 2 + 10 + 30 + 5)
	    ),
	    0
	);
	OSWMoveResizeWindow(w, x, y, width, height);
	WidgetQueueResize(clist);

	/* Amount prompt */
	prompt = &ew->amount_prompt;
	w = prompt->toplevel;
	x = 0;
	y = 60 + EW_BTN_HEIGHT + (int)ew->height - 180;
	width = ew->width;
	height = 30;
	OSWMoveResizeWindow(w, x, y, width, height);
	WidgetQueueResize(prompt);


	/* Buy button */
	btn = &ew->buy_btn;
	w = btn->toplevel;
	x = (0 * (int)btn->width) + (1 * 10);
	y = (int)ew->height - (int)btn->height - 10;
	OSWMoveWindow(w, x, y);
	WidgetQueueResize(btn);

	/* Sell button */
	btn = &ew->sell_btn;
	w = btn->toplevel;
	x = (1 * (int)btn->width) + (2 * 10);
	OSWMoveWindow(w, x, y);
	WidgetQueueResize(btn);

	/* Close button */
	btn = &ew->close_btn;
	w = btn->toplevel;
	x = (int)ew->width - (int)btn->width - 10;
	OSWMoveWindow(w, x, y);
	WidgetQueueResize(btn);
}

/*
 *	Queues the economy window to resize.
 */
void EcoWinQueueResize(xsw_eco_win_struct *ew)
{
	if(ew == NULL)
	    return;

	WidgetQueueResizeAction(EcoWinResizeCB, ew, NULL);
}

/*
 *	Draws the economy window.
 */
void EcoWinDraw(xsw_eco_win_struct *ew)
{
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if(ew == NULL)
	    return;

	if(ew->toplevel != None)
	{
	    unsigned int	width = ew->width,
				height = ew->height;
	    win_t w = ew->toplevel;

	    /* Draw the background */
	    if(wglobal->force_mono)
	    {
		int y;

		OSWSetForeground(gui->black_pix);
		OSWDrawSolidRectangle(w, 0, 0, width, height);

		y = (int)height - EW_BTN_HEIGHT - 25;
		OSWSetForeground(gui->white_pix);
		OSWDrawLine(
		    w,
		    0, y, (int)width, y
		);

		OSWSetForeground(gui->white_pix);
	    }
	    else
	    {
		int y;

		WidgetPutImageTile(
		    w, wglobal->std_bkg_img,
		    width, height
		);

		y = (int)height - EW_BTN_HEIGHT - 25;
		OSWSetForeground(wglobal->surface_shadow_pix);
		OSWDrawLine(
		    w,
		    0, y, (int)width, y
		);
		OSWSetForeground(wglobal->surface_highlight_pix);
		OSWDrawLine(
		    w,
		    0, y + 1, (int)width, y + 1
		);

		OSWSetForeground(wglobal->normal_text_pix);
	    }
	}
}

/*
 *	Queues the economy window to draw.
 */
void EcoWinQueueDraw(xsw_eco_win_struct *ew)
{
	if(ew == NULL)
	    return;

	WidgetQueueDrawAction(EcoWinDrawCB, ew, NULL);
}

/*
 *	Manages the economy window.
 */
int EcoWinManage(xsw_eco_win_struct *ew, event_t *event)
{
	int events_handled = 0;
	time_t t;
	win_t w;
	event_key_t *key;
	event_visibility_t *visibility;
        xsw_option_struct *opt = &option;

	if((ew == NULL) || (event == NULL))
	    return(events_handled);

	w = event->xany.window;

	switch(event->type)
	{
	  case ConfigureNotify:
	    if(w == ew->toplevel)
	    {
		EcoWinQueueResize(ew);
		EcoWinQueueDraw(ew);
		CListQueueDraw(&ew->inventory, CLIST_DRAW_AMOUNT_COMPLETE);
		events_handled++;
	    }
	    break;

	  case Expose:
	    if(w == ew->toplevel)
	    {
		EcoWinQueueDraw(ew);
		events_handled++;
	    }
	    break;

	  case ButtonPress:
            /* Grab focus as needed */
            if(!ew->has_focus)
            {
                if(OSWCheckWindowAncestory(ew->toplevel, w))
                    OSWGrabFocus(ew->toplevel);
            }
	    break;

	  case KeyPress:
	    key = &event->xkey;
	    if(!ew->has_focus)
	        return(events_handled);

	    break;

	  case KeyRelease:
	    key = &event->xkey;
	    if(!ew->has_focus)
		return(events_handled);

	    break;

	  case FocusIn:
	    if(w == ew->toplevel)
	    {
		ew->has_focus = True;
		events_handled++;
	    }
	    break;

	  case FocusOut:
	    if(w == ew->toplevel)
	    {
		EcoWinUnfocusAll(ew);
		ew->has_focus = False;
		events_handled++; 
	    }
	    break;

	  case MapNotify:
	    if(w == ew->toplevel)
	    {
		if(!ew->map_state)
		{
		    WidgetMap(&ew->credits_name_label);
		    WidgetMap(&ew->credits_value_label);
		    WidgetMap(&ew->taxes_label);
		    WidgetMap(&ew->proprietor_prompt);
		    WidgetMap(&ew->refresh_btn);
		    WidgetMap(&ew->inventory);
		    WidgetMap(&ew->amount_prompt);
		    WidgetMap(&ew->buy_btn);
		    WidgetMap(&ew->sell_btn);
		    WidgetMap(&ew->close_btn);

		    ew->map_state = True;
		}
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case UnmapNotify:
	    if(w == ew->toplevel)
	    {
		if(ew->map_state)
                { 
		    /* Record the window position */
		    OSWGetWindowRootOrigin(
			w,
			&opt->ecow_x, &opt->ecow_y
		    );
		    opt->ecow_width = ew->width;
		    opt->ecow_height = ew->height;

		    WidgetUnmap(&ew->credits_name_label);
		    WidgetUnmap(&ew->credits_value_label);
		    WidgetUnmap(&ew->taxes_label);
		    WidgetUnmap(&ew->proprietor_prompt);
		    WidgetUnmap(&ew->refresh_btn);
		    WidgetUnmap(&ew->inventory);
		    WidgetUnmap(&ew->amount_prompt);
		    WidgetUnmap(&ew->buy_btn);
		    WidgetUnmap(&ew->sell_btn);
		    WidgetUnmap(&ew->close_btn);

		    EcoWinClearProducts(ew);

		    ew->map_state = False;
		}
		events_handled++;
		return(events_handled);
	    }
	    break;

          case VisibilityNotify:
            visibility = &event->xvisibility;
            if(w == ew->toplevel)
            {
                ew->visibility_state = visibility->state;
            }
	    break;

	  case ClientMessage:
	    /* WM Take Focus? */
	    if(OSWEventIsTakeFocus(ew->toplevel, event, &t))
	    {
		OSWSetInputFocus(ew->toplevel, RevertToParent, t);
		events_handled++;
		return(events_handled);
	    }
	    /* WM Delete Window? */
	    else if(OSWEventIsDeleteWindow(ew->toplevel, event))
	    {
		EcoWinUnmap(ew);
		events_handled++;
		return(events_handled);
	    }
	    break;
	}

	if(events_handled == 0)
	    events_handled += WidgetManage(&ew->credits_name_label, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&ew->credits_value_label, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&ew->taxes_label, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&ew->refresh_btn, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&ew->buy_btn, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&ew->sell_btn, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&ew->close_btn, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&ew->inventory, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&ew->proprietor_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&ew->amount_prompt, event);

	return(events_handled);
}

/*
 *	Maps the economy window.
 */
void EcoWinMap(xsw_eco_win_struct *ew)
{
	if(ew == NULL)
	    return;

	EcoWinUnfocusAll(ew);

	/* Map */
	OSWMapRaised(ew->toplevel);

	/* Grab focus */
	OSWGrabFocus(ew->toplevel);
	ew->inventory.has_focus = True;
}

/*
 *	Unmaps the economy window.
 */
void EcoWinUnmap(xsw_eco_win_struct *ew)
{
	if(ew == NULL)
	    return;

	OSWUnmapWindow(ew->toplevel);
}

/*
 *	Deletes the Economy window.
 */
void EcoWinDelete(xsw_eco_win_struct *ew)
{
	if(ew == NULL)
	    return;

	CListClear(&ew->inventory);

	WidgetDestroy(&ew->credits_name_label);
	WidgetDestroy(&ew->credits_value_label);
	WidgetDestroy(&ew->taxes_label);
	WidgetDestroy(&ew->proprietor_prompt);
	WidgetDestroy(&ew->refresh_btn);
	WidgetDestroy(&ew->inventory);
	WidgetDestroy(&ew->amount_prompt);
	WidgetDestroy(&ew->buy_btn);
	WidgetDestroy(&ew->sell_btn);
	WidgetDestroy(&ew->close_btn);

	OSWDestroyWindow(&ew->toplevel);

	free(ew);
}
