#include <stdlib.h>
#include <string.h>

#include "../include/swsoundcodes.h"

#include "blitting.h"

#include "page.h"
#include "sound.h"
#include "net.h"
#include "xsw.h"
#include "xswaction.h"


static image_t *PageLoadImage(const char *path);
static int PageLabelGetLabelNumFromPosition(
	page_struct *p,
	win_t w, shared_image_t *image,
	int x, int y
);

int PageIsLabelAllocated(page_struct *p, int n);
int PageCreateLabel(page_struct *p);

void PageResize(
	page_struct *p, win_t w, shared_image_t *image
);
void PageDrawLabel(
	page_struct *p,
	win_t w, shared_image_t *image, int label_num,
	bool_t put_to_window
);
void PageDraw(
	page_struct *p,
	win_t w, shared_image_t *image, int amount,
	bool_t put_to_window
);
int PageManage(
	page_struct *p,
	win_t w, shared_image_t *image, event_t *event,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
void PageMap(
	page_struct *p,
	win_t w, shared_image_t *image
);
void PageUnmap(
	page_struct *p,
	win_t w, shared_image_t *image
);
void PageDestroy(
	page_struct *p,
	win_t w, shared_image_t *image
);

void PageDoAction(
	page_struct *p,
	win_t w, shared_image_t *image,
	const int action,
	const time_t cur_ms, const time_t cur_sec, const float tc
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


static image_t *PageLoadImage(const char *path)
{
	return(WidgetLoadImageFromTgaFile(path));
}


/*
 *	Macro to get the main menu label number from
 *	given coordinates.
 *
 *	If labels overlap, older entry gets priority.
 *
 *	Returns -1 on error or no match.
 */
static int PageLabelGetLabelNumFromPosition(
	page_struct *p,
	win_t w, shared_image_t *img,
	int x, int y 
)
{
	int i, n, m, tmp_x, tmp_y;
	image_t *label_img;
	xsw_imglabel_struct *imglabel;
	page_label_struct *plabel;

	if((p == NULL) || (w == 0) || (img == NULL))
	    return(-1);

	/* Iterate through each label on this page */
	for(i = 0; i < p->total_labels; i++)
	{
	    plabel = p->label[i];
	    if(plabel == NULL)
		continue;

	    /* Iterate through each image label */
	    m = (int)sizeof(plabel->imglabel) /
		(int)sizeof(xsw_imglabel_struct);
	    for(n = 0; n < m; n++)
	    {
		/* Get pointer to image label */
		imglabel = &plabel->imglabel[n];
		if(imglabel == NULL)
		    continue;

		/* Image label not loaded? */
		label_img = imglabel->image;
		if(label_img == NULL)
		    continue;

		/* Check coordinates */
		if(imglabel->pos_by_percent)
		{
		    tmp_x = imglabel->x * (int)img->width / 100;
		    tmp_y = imglabel->y * (int)img->height / 100; 
		}
		else
		{
		    tmp_x = imglabel->x;
		    tmp_y = imglabel->y;
		}
		if((x >= tmp_x) &&
		   (y >= tmp_y) &&
		   (x < (tmp_x + (int)label_img->width)) &&
		   (y < (tmp_y + (int)label_img->height))
		)
		    return(i);
	    }
	}

	return(-1);
}


/*
 *	Checks if the main menu label n is allocated on mm.
 */
int PageIsLabelAllocated(page_struct *p, int n)
{
	if(p == NULL)
	    return(0);
	else if((p->label == NULL) ||
		(n < 0) ||
		(n >= p->total_labels)
	)
	    return(0);
	else if(p->label[n] == NULL)
	    return(0);
	else
	    return(1);
}


/*
 *	Allocates a new label on the mm.
 *
 *	The new label is reset and not loaded.
 *	Returns the label's number or -1 on error.
 */
int PageCreateLabel(page_struct *p)
{
	int i, m, new_label;
	page_label_struct *label;

	if(p == NULL)
	    return(-1);;

	/* Sanitize total_labels */
	if(p->total_labels < 0)
	    p->total_labels = 0;

	/* Look for already allocated label pointer */
	for(i = 0; i < p->total_labels; i++)
	{
	    if(p->label[i] == NULL)
		break;
	}
	if(i < p->total_labels)
	{
	    new_label = i;
	}
	else
	{
	    new_label = p->total_labels;
	    p->total_labels++;

	    p->label = (page_label_struct **)realloc(
		p->label,
		p->total_labels * sizeof(page_label_struct *)
	    );
	    if(p->label == NULL)
	    {
		p->total_labels = 0;
		return(-3);
	    }
	}

	/* Allocate new main menu label structure */
	p->label[new_label] = label = (page_label_struct *)calloc(
	    1, sizeof(page_label_struct)
	);
	if(label == NULL)
	    return(-3);

	label->map_state = False;
	label->allow_transparency = 0;
	label->op_code = XSW_ACTION_NONE;

	/* Reset each xsw image label in the main menu label */
	m = (int)sizeof(label->imglabel) /
	    (int)sizeof(xsw_imglabel_struct);
	for(i = 0; i < m; i++)
	    memset(
		&(label->imglabel[i]),
		0x00,
		sizeof(xsw_imglabel_struct)
	    );

	return(new_label);
}

/*
 *	Initializes page.
 */
int PageInit(page_struct *p, win_t w, shared_image_t *image)
{
	if(p == NULL)
	    return(-1);

	memset(p, 0x00, sizeof(page_struct));

	return(0);
}

/*
 *	Resizes the main menu.
 */
void PageResize(
	page_struct *p,
	win_t w, shared_image_t *image
)
{
	unsigned int width, height;
	image_t	*ori_img = NULL,
		*new_img = NULL;

	if((p == NULL) || (w == 0) || (image == NULL))
	    return;

	/* Do not resize if not mapped */
	if(!p->map_state)
	    return;

	/* Get size of image */
	width = image->width;
	height = image->height;

	/* Load background as needed */
	if(p->bg_filename != NULL)
	{
	    if(p->bg_image == NULL)
	    {
	        p->bg_image = PageLoadImage(p->bg_filename);
		if(p->bg_image == NULL)
		    return;
	    }
	    ori_img = p->bg_image;

	    /* Resize only as needed */
	    if((width != (unsigned int)ori_img->width) ||
	       (height != (unsigned int)ori_img->height)
	    )
	    {
	        /* Create new sized background */
	        if(OSWCreateImage(&new_img, width, height))
	            return;

	        /* Resize background */
	        WidgetResizeImageBuffer(
	            new_img->bits_per_pixel,
	            (u_int8_t *)new_img->data,		/* Target */
		    (const u_int8_t *)ori_img->data,	/* Source */
	            new_img->width, new_img->height,
		    new_img->bytes_per_line,
	            ori_img->width, ori_img->height,
		    ori_img->bytes_per_line
	        );

	        /* Destroy original image and set new image */
	        OSWDestroyImage(&ori_img);

		/* Set new resized image for bg image */
		p->bg_image = new_img;
	    }
	}
}

/*
 *	Draws the label.
 *
 *	The label_num specifies the label to draw.
 */
void PageDrawLabel(
	page_struct *p,
	win_t w, shared_image_t *image, int label_num,
	bool_t put_to_window
)
{
	int x, y, label_state;
	unsigned int width, height;
	image_t *label_img;
	xsw_imglabel_struct *imglabel;
	page_label_struct *label;

	if((p == NULL) || (w == 0) || (image == NULL))
	    return;

	/* Do not draw if not mapped */
	if(!p->map_state)
	    return;

	/* Do not draw if no background image is available */
	if(p->bg_image == NULL)
	    return;

	if(!PageIsLabelAllocated(p, label_num))
	    return;

	label = p->label[label_num];

	/* Get position and sizes */
	x = 0;
	y = 0;
	width = 0;
	height = 0;

	/* Get the label's state */
	label_state = (p->sel_label == label_num) ?
	    p->sel_label_state : PAGE_LABEL_STATE_UNARMED;

	switch(label_state)
	{
	  case PAGE_LABEL_STATE_ARMED:
	    imglabel = &label->imglabel[PAGE_LABEL_STATE_ARMED];
	    if(imglabel->pos_by_percent)
	    {
		x = imglabel->x * (int)image->width / 100;
		y = imglabel->y * (int)image->height / 100;
	    }
	    else
	    {
		x = imglabel->x;
		y = imglabel->y;
	    }
	    label_img = imglabel->image;
	    if(label_img != NULL)
	    {
		width = label_img->width;
		height = label_img->height;
	    }
	    break;

	  case PAGE_LABEL_STATE_HIGHLIGHTED:
	    imglabel = &label->imglabel[PAGE_LABEL_STATE_HIGHLIGHTED];
	    if(imglabel->pos_by_percent)
	    {
		x = imglabel->x * (int)image->width / 100;
		y = imglabel->y * (int)image->height / 100;
	    }
	    else
	    {
		x = imglabel->x;
		y = imglabel->y;
	    }
	    label_img = imglabel->image;
	    if(label_img != NULL)
	    {
		width = label_img->width;
		height = label_img->height;
	    }
	    break;

	  default:	/* Default to unarmed */
	    imglabel = &label->imglabel[PAGE_LABEL_STATE_UNARMED];
	    if(imglabel->pos_by_percent)
	    {
		x = imglabel->x * (int)image->width / 100;
		y = imglabel->y * (int)image->height / 100;
	    }
	    else
	    {
		x = imglabel->x;
		y = imglabel->y;
	    }
	    label_img = imglabel->image;
	    if(label_img != NULL)
	    {
		width = label_img->width;
		height = label_img->height;
	    }
	    break;
	}

	/* Label not allocated? */
	if(label_img == NULL)
	    return;

	/* Blit portion of background to image */
	BlitBufAbsolute(
	    image->bytes_per_pixel, image->depth,
	    (u_int8_t *)image->data,
	    (u_int8_t *)p->bg_image->data,
	    x, y,			/* Target coordinates */
	    image->width, image->height,
	    image->bytes_per_line,
	    x, y,			/* Source coordinates */
	    p->bg_image->width, p->bg_image->height,
	    p->bg_image->bytes_per_line,
	    width, height,		/* Copy width and height */
	    1.0,			/* Zoom */
	    1.0				/* Magnification */
	);

	/* Blit label to image */
	if(label->allow_transparency)
	    BlitBufNormal(
		image->bytes_per_pixel, image->depth,
		(u_int8_t *)image->data,
		(u_int8_t *)label_img->data,
		x, y,			/* Target coordinates */ 
		image->width, image->height, image->bytes_per_line,
		0, 0,			/* Source coordinates */ 
		label_img->width, label_img->height, label_img->bytes_per_line,
		label_img->width, label_img->height,
		1.0,			/* Zoom */
		1.0,			/* Visibility */
		1.0			/* Magnificaition */
	    );
	else
	    BlitBufAbsolute(
		image->bytes_per_pixel, image->depth,
		(u_int8_t *)image->data,
		(u_int8_t *)label_img->data,
		x, y,			/* Target coordinates */
		image->width, image->height,
		image->bytes_per_line,
		0, 0,			/* Source coordinates */
		label_img->width, label_img->height,
		label_img->bytes_per_line,
		label_img->width, label_img->height,
		1.0,			/* Zoom */
		1.0			/* Magnification */
	    );


	/* Put to window? */
	if(put_to_window)
	    OSWPutSharedImageToDrawableSect(
		image, w,
		x, y,		/* Target */
		x, y,		/* Source */
		width, height
	    );
}



/*
 *	Redraws the main menu mm on window w.
 *
 *	If put_to_window is False, then only the image is updated,
 *	else both the image is updated and put to the window w.
 */
void PageDraw(
	page_struct *p,
	win_t w, shared_image_t *image,
	int amount,
	bool_t put_to_window
)
{
	int i, x, y, label_num;
	image_t *label_img;
	xsw_imglabel_struct *imglabel;
	page_label_struct *label;

	if((p == NULL) || (w == 0) || (image == NULL))
	    return;

	/* Map as needed */
	if(!p->map_state)
	{
	    p->map_state = True;

	    /* Map all labels */
	    for(i = 0; i < p->total_labels; i++)
	    {
		label = p->label[i];
		if(label == NULL)
		    continue;

		label->map_state = True;
	    }

	    /* Resize page on mapping */
	    PageResize(p, w, image);
	}


	/* Draw background */

	/* Load background as needed */
	if(p->bg_filename != NULL)
	{
	    if(p->bg_image == NULL)
	    {
		p->bg_image = PageLoadImage(p->bg_filename);
		if(p->bg_image == NULL)
		    return;
	    }

	    /* Draw the background image */
	    WidgetCopyImageToShared(image, p->bg_image);
	}


	/* ******************************************************** */
	/* Draw labels */

	for(label_num = 0;
	    label_num < p->total_labels;
	    label_num++
	)
	{
	    /* Get main menu label pointer */
	    label = p->label[label_num];
	    /* Cannot be NULL */
	    if(label == NULL)
		continue;

	    /* Skip if label is not mapped */
	    if(!label->map_state)
		continue;


	    /* Is this label selected? */
	    if(label_num == p->sel_label)
	    {
		x = 0;
		y = 0;
		/* Image label index corresponds with state */
		switch(p->sel_label_state)
		{
		  case PAGE_LABEL_STATE_ARMED:
		    imglabel = &label->imglabel[PAGE_LABEL_STATE_ARMED];
		    if(imglabel->pos_by_percent)
		    {
			x = imglabel->x * (int)image->width / 100;
			y = imglabel->y * (int)image->height / 100;
	            }
		    else
		    {
		        x = imglabel->x;
			y = imglabel->y;
		    }
		    label_img = imglabel->image;
		    break;

		  case PAGE_LABEL_STATE_HIGHLIGHTED:
		    imglabel = &label->imglabel[PAGE_LABEL_STATE_HIGHLIGHTED];
		    if(imglabel->pos_by_percent)
		    {
			x = imglabel->x * (int)image->width / 100;
			y = imglabel->y * (int)image->height / 100;
		    }
		    else
		    {
			x = imglabel->x;
			y = imglabel->y;
		    }
		    label_img = imglabel->image;
		    break;

		  default:	/* Default to unarmed */
		    imglabel = &label->imglabel[PAGE_LABEL_STATE_UNARMED];
		    if(imglabel->pos_by_percent)
		    {
			x = imglabel->x * (int)image->width / 100;
			y = imglabel->y * (int)image->height / 100;
		    }
		    else
		    {
			x = imglabel->x;
			y = imglabel->y;
		    }
		    label_img = imglabel->image;
		    break;
		}
		if(label_img == NULL)
		{
		    imglabel = &label->imglabel[PAGE_LABEL_STATE_UNARMED];
		    if(imglabel->pos_by_percent)
		    {
			x = imglabel->x * (int)image->width / 100;
			y = imglabel->y * (int)image->height / 100; 
		    }
		    else
		    {
			x = imglabel->x;
			y = imglabel->y;
		    }
		    label_img = imglabel->image;
		}
	    }
	    else
	    {
		/* Not selected, use unarmed image */
		imglabel = &label->imglabel[PAGE_LABEL_STATE_UNARMED];
		if(imglabel->pos_by_percent)
		{
		    x = imglabel->x * (int)image->width / 100;
		    y = imglabel->y * (int)image->height / 100;
		}
		else
		{
		    x = imglabel->x;
		    y = imglabel->y;
		}
		label_img = imglabel->image;
	    }
	    if(label_img == NULL)
		continue;

	    /* Blit label_img to image */
	    if(label->allow_transparency)
	    {
		BlitBufNormal(
		    image->bytes_per_pixel, image->depth,
		    (u_int8_t *)image->data,
		    (u_int8_t *)label_img->data,
		    x, y, 		/* Target coordinates */
		    image->width, image->height, image->bytes_per_line,
		    0, 0,		/* Source coordinates */
		    label_img->width, label_img->height, label_img->bytes_per_line,
		    label_img->width, label_img->height,
		    1.0f,	/* Zoom */
		    1.0f,	/* Visibility */
		    1.0f	/* Magnification */
		);
	    }
	    else
	    {
		BlitBufAbsolute(
		    image->bytes_per_pixel, image->depth,
		    (u_int8_t *)image->data,
		    (u_int8_t *)label_img->data,
		    x, y,		/* Target coordinates */
		    image->width, image->height,
		    image->bytes_per_line,
		    0, 0,		/* Source coordinates */
		    label_img->width, label_img->height,
		    label_img->bytes_per_line,
		    label_img->width, label_img->height,
		    1.0f,	/* Zoom */
		    1.0f	/* Magnification */
		);
	    }
	}

	/* Put image to window? */
	if(put_to_window)
	    OSWPutSharedImageToDrawable(image, w);
}

/*
 *	Manage the event for the specified page.
 */
int PageManage(
	page_struct *p,
	win_t w, shared_image_t *image, event_t *event,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int label_num, events_handled = 0;
	win_t event_w;
	event_key_t *key;
	event_button_t *button;
	event_motion_t *motion;
	xsw_sound_struct *snd = &sound;

	if((p == NULL) || (w == None) || (image == NULL) || (event == NULL))
	    return(events_handled);

	/* Do not handle the event if this page is not mapped, pages
	 * are mapped manually so pages also ignore map events all
	 * togeather
	 */
	if(!p->map_state)
	    return(events_handled);

	event_w = event->xany.window;

#define DRAW_PAGE_COMPLETE(_page_,_put_) {	\
 PageDraw(					\
  (_page_),					\
  w, image,					\
  PAGE_DRAW_AMOUNT_COMPLETE,			\
  (_put_)					\
 );						\
}

	switch(event->type)
	{
	  case Expose:
	    if(event_w == w)
	    {
		DRAW_PAGE_COMPLETE(p, True);
		events_handled++;
	    }
	    break;

	  case KeyPress:
	    key = &event->xkey;
	    if(!p->has_focus)
		break;

	    break;

	  case KeyRelease:
	    key = &event->xkey;
	    if(!p->has_focus)
		break;

	    break;

	  case ButtonPress:
	    button = &event->xbutton;

	    /* Is this event not for this window? */
	    if(event_w != w)
	    {
		/* Unfocus this page and break */
		p->has_focus = False;
		break;
	    }

	    /* Set this page into focus */
	    p->has_focus = True;

	    /* Find the label that this button press occured over */
	    label_num = PageLabelGetLabelNumFromPosition(
		p, w, image,
		(int)button->x, (int)button->y
	    );
	    if(PageIsLabelAllocated(p, label_num))
	    {
		page_label_struct *label = p->label[label_num];

		switch(button->button)
		{
		  case Button1:
		    /* Set this label as selected and armed */
		    p->sel_label = label_num;
		    p->sel_label_state = PAGE_LABEL_STATE_ARMED;

		    /* Redraw only this label */
		    PageDrawLabel(p, w, image, p->sel_label, True);

		    if(label->op_code != XSW_ACTION_NONE)
			events_handled++;
		    break;
		}
	    }
	    break;

	  case ButtonRelease:
	    button = &event->xbutton;

	    /* Is this event not for this window? */
	    if(event_w != w)
		break;

	    /* Find the label that this button press occured over */
	    label_num = PageLabelGetLabelNumFromPosition(
		p, w, image,
		(int)button->x, (int)button->y
	    );
	    if(PageIsLabelAllocated(p, label_num))
	    {
		page_label_struct *label = p->label[label_num];

		switch(button->button)
		{
		  case Button1:
		    /* Was label previously armed? */
		    if((label_num == p->sel_label) &&
		       (p->sel_label_state == PAGE_LABEL_STATE_ARMED)
		    )
		    {
			/* Unmap the Tooltips */
			TooltipsUnmap(NULL);

			/* Set label back to highlighted */
			p->sel_label_state = PAGE_LABEL_STATE_HIGHLIGHTED;

			/* Redraw only this label */
			PageDrawLabel(p, w, image, p->sel_label, True);
 
			/* Play the select sound */
			if((option.sounds > XSW_SOUNDS_NONE) &&
			   (label->op_code != XSW_ACTION_NONE)
			)
			    SoundPlay(
				snd,
				SW_SNDREF_MENU_SELECT,
				1.0f, 1.0f,
				0,
				0
			    );

			/* Perform the action defined by this label */
			PageDoAction(
			    p,
			    w, image,
			    label->op_code,
			    cur_ms, cur_sec, tc
			);
		    }
		    else
		    {
			/* The button was released on some other
			 * label, in which case select that label
			 * and set it to be highlighted, do not
			 * perform its action
			 */
			p->sel_label = label_num;
			p->sel_label_state = PAGE_LABEL_STATE_HIGHLIGHTED;

			/* Redraw only this label */
			PageDrawLabel(p, w, image, p->sel_label, True);
		    }
		    if(label->op_code != XSW_ACTION_NONE)
			events_handled++;
		    break;
		}
	    }
	    break;

	  case MotionNotify:
	    motion = &event->xmotion;
	    if(event_w == w)
	    {
		/* Find the label that the pointer moved over */
		label_num = PageLabelGetLabelNumFromPosition(
		    p, w, image,
		    (int)motion->x, (int)motion->y
		);
		if(PageIsLabelAllocated(p, label_num))
		{
		    page_label_struct *label = p->label[label_num];

		    /* Is this label the currently selected label
		     * and is it armed?
		     */
		    if((p->sel_label == label_num) &&
		       (p->sel_label_state == PAGE_LABEL_STATE_ARMED)
	            )
		    {
		        /* Do nothing */
		        events_handled++;
		    }
		    /* Is this label not the currently selected label? */
		    else if(p->sel_label != label_num)
		    {
			/* Pointer has moved over to a new label */
			const int prev_sel_label = p->sel_label;

			/* Unmap the Tooltips */
			TooltipsUnmap(NULL);

			/* Set the new selected label and set it as
			 * highlighted
			 */
			p->sel_label = label_num;
			p->sel_label_state = PAGE_LABEL_STATE_HIGHLIGHTED;

			/* Schedual the hint message to be shown */
			if(label->op_code != XSW_ACTION_NONE)
			{
			    if(label->hint_mesg != NULL)
			    {
				TooltipsSetMessage(
				    NULL, w, label->hint_mesg
				);
	                        TooltipsSchedualMap(
				    NULL, (time_t)-1, w
				);
			    }
			}

			/* Redraw the labels */
			PageDrawLabel(p, w, image, prev_sel_label, False);
			PageDrawLabel(p, w, image, p->sel_label, True);

			/* Play the highlighted sound */
			if((option.sounds > XSW_SOUNDS_NONE) &&
			   (label->op_code != XSW_ACTION_NONE)
			)
			    SoundPlay(
				&sound,
				SW_SNDREF_MENU_HIGHLIGHT,
				1.0f, 1.0f,
				0, 
				0
			    );

			if(label->op_code != XSW_ACTION_NONE)
			    events_handled++;
		    }
		}
	        else
	        {
		    /* Pointer is not over any label */

		    /* Was a label previously selected? */
		    if(PageIsLabelAllocated(p, p->sel_label))
		    {
			const int prev_sel_label = p->sel_label;

			/* Set the label as not selected and unarmed */
			p->sel_label = -1;
			p->sel_label_state = PAGE_LABEL_STATE_UNARMED;

			TooltipsUnmap(NULL);

			/* Redraw this label */
			PageDrawLabel(p, w, image, prev_sel_label, True);

		        events_handled++;
		    }

		}
	    }
#if 0
	    else
	    {
		/* MotionNotify occured on another window, check if
		 * there is a selected label
		 */
		if(PageIsLabelAllocated(p, p->sel_label))
		{
		    /* Unselect it and unmap its Tooltips */
		    p->sel_label = -1;
		    TooltipsUnmap(NULL);
		}
	    }
#endif
	    break;

	  case EnterNotify:
	    if(event_w == w)
	    {
		/* Do nothing */
	    }
	    break;

	  case LeaveNotify:
	    if(event_w == w)
	    {
		/* Was there a selected label? */
		if(PageIsLabelAllocated(p, p->sel_label))
		{
		    /* Unselect the label and unmap the tooltips */
		    p->sel_label = -1;
		    TooltipsUnmap(NULL);
		    DRAW_PAGE_COMPLETE(p, True);
		    events_handled++;
		}
	    }
	    break;
	}

#undef DRAW_PAGE_COMPLETE

	return(events_handled);
}

void PageMap(
	page_struct *p,
	win_t w, shared_image_t *image
)
{
	if((p == NULL) || (w == 0) || (image == NULL))
	    return;

	if(!p->map_state)
	{
	    OSWUnsetWindowCursor(w);

	    /* Change selected watched events */
	    OSWSetWindowInput(
		w,
		ButtonPressMask | ButtonReleaseMask |
		PointerMotionMask | ExposureMask |
		EnterWindowMask | LeaveWindowMask |
		VisibilityChangeMask
	    );
	}

	/* Map by drawing it */
	p->map_state = False;
	PageDraw(p, w, image, PAGE_DRAW_AMOUNT_COMPLETE, True);
}

void PageUnmap(
	page_struct *p,
	win_t w, shared_image_t *image
)
{
	if((p == NULL) || (w == 0) || (image == NULL))
	    return;

	/* Unmap as needed */
	if(p->map_state)
	{
	    /* Destroy the page background image */
	    OSWDestroyImage(&p->bg_image);

	    /* Set unmapped map state */
	    p->map_state = False;

	    /* Unmap the Tooltips */
	    TooltipsUnmap(NULL);
	}
}

/*
 *	Deletes all resources for the page.
 */
void PageDestroy(
	page_struct *p,
	win_t w, shared_image_t *image
)
{
	int i, n, m;
	xsw_imglabel_struct *imglabel;
	page_label_struct *plabel;

	if(p == NULL)
	    return;

	/* Delete the hint message for this page */
	TooltipsDeleteMessage(NULL, w);

	/* Free background image filename */
	free(p->bg_filename);
	p->bg_filename = NULL;

	/* Destroy background image */
	OSWDestroyImage(&p->bg_image);

	/* Delete all labels */
	for(i = 0; i < p->total_labels; i++)
	{
	    plabel = p->label[i];
	    if(plabel == NULL)
		continue;

	    m = (int)sizeof(plabel->imglabel) /
		(int)sizeof(xsw_imglabel_struct);
	    for(n = 0; n < m; n++)
	    {
		imglabel = &plabel->imglabel[n];
	        ImgLabelReset(imglabel);
		/* Do not delete the imglabel itself */
	    }

	    free(plabel->hint_mesg); 

	    free(plabel);
	}
	free(p->label);
	p->label = NULL;
	p->total_labels = 0;

	/* Reset values */
	p->map_state = False;
}

/*
 *	Performs main menu action.
 */
void PageDoAction(
	page_struct *p,
	win_t w, shared_image_t *image,
	const int action,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	XSWAction(
	    action, NULL, xsw_bridge_win,
	    cur_ms, cur_sec, tc
	);
}
