#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "../include/disk.h"
#include "../include/swsoundcodes.h"

#include "blitting.h"
#include "sound.h"
#include "xsw.h"
#include "splash.h"
#include "imagepaths.h"


/* Default icon data */
#include "icons/xsw.h"


static void SplashDoFadeIn(
	splash_win_struct *sw, win_t w, image_t *src_img
);

splash_win_struct *SplashNew(void);
static void SplashDrawProgressBar(
	splash_win_struct *sw, const float progress_coeff
);
void SplashProgress(
	splash_win_struct *sw,
	const float progress_coeff, const char *msg
);
void SplashDelete(splash_win_struct *sw);


/*
 *	Splash Window Size:
 *
 *	Used only if the background image is missing.
 */
#define SW_DEF_WIDTH			320
#define SW_DEF_HEIGHT			240

/*
 *	Progress bar width coefficient:
 */
#define SW_LS_BAR_WIDTHC		0.9

/*
 *	Progress bar y position coefficient:
 */
#define SW_LS_BAR_YC			0.78


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)
#define IS_NUM_ODD(n)	((long)(n) & 1)


/*
 *	Does the opening fadein affect.
 */
static void SplashDoFadeIn(
	splash_win_struct *sw, win_t w, image_t *src_img
)
{
	float gamma;
	shared_image_t *tar_img;
	time_t cur_ms, last_ms;

	if((sw == NULL) || (w == 0) || (src_img == NULL))
	    return;

	/* Create tempory target image */
	if(OSWCreateSharedImage(&tar_img, src_img->width, src_img->height))
	    return;

	/* Map window */
	OSWMapRaised(w);
	sw->map_state = True;
	sw->has_focus = True;
	sw->visibility_state = VisibilityUnobscured;

#define BLIT_FADE_AND_PUT(_gamma_)	{	\
 BlitBufFade(					\
  tar_img->bytes_per_pixel, tar_img->depth,	\
  tar_img->data,				\
  0, 0,						\
  tar_img->width, tar_img->height,		\
      tar_img->bytes_per_line,			\
  tar_img->width, tar_img->height,		\
  (_gamma_)					\
 );						\
 OSWPutSharedImageToDrawable(tar_img, w);	\
}

#define UPDATE_GAMMA	{			\
 usleep(0);					\
 last_ms = cur_ms;				\
 cur_ms = OSWTimeMS();				\
 gamma += (float)MAX((cur_ms - last_ms), 0l) / 1000.0f; \
 if(gamma > 1.0f)				\
  gamma = 1.0f;					\
 OSWFlush();					\
}

	/* Begin doing fade in */
	cur_ms = last_ms = OSWTimeMS();

	/* From gamma 0.0 to 1.0 */
	gamma = 0.0f;
	while(gamma < 1.0f)
	{
	    WidgetCopyImageToShared(tar_img, src_img);
	    BLIT_FADE_AND_PUT(gamma);
	    UPDATE_GAMMA;
	}

	/* One last time at 1.0 gamma */
	WidgetCopyImageToShared(tar_img, src_img);
	BLIT_FADE_AND_PUT(1.0f);
	UPDATE_GAMMA;

#undef BLIT_FADE_AND_PUT
#undef UPDATE_GAMMA

	/* Destroy shared target image */
	OSWSyncSharedImage(tar_img, w);
	OSWDestroySharedImage(&tar_img);

	/* Map state needs to be set False so it gets mapped when drawn
	 * again
	 */
	sw->map_state = False;
}


/*
 *	Creates a new splash window.
 */
splash_win_struct *SplashNew(void)
{
	const char *s;
	win_t w, parent;
	pixmap_t pixmap;
	image_t *image;
	osw_gui_struct *gui = &osw_gui[0];
	splash_win_struct *sw = SPLASH_WIN(calloc(
	    1, sizeof(splash_win_struct)
	));
	if(sw == NULL)
	    return(NULL);

	/* Load the background and progress bar images */
	s = PrefixPaths(dname.images, XSW_IMG_PATH_SPLASH_BG);
	sw->bkg_img = WidgetLoadImageFromTgaFile(s);

	/* Set values */
	sw->map_state = False;
	sw->has_focus = False;
	sw->visibility_state = VisibilityFullyObscured;
	sw->x = 0;
	sw->y = 0;
	/* Width and height from splash background image */
	if(sw->bkg_img == NULL)
	{
	    sw->width = SW_DEF_WIDTH;
	    sw->height = SW_DEF_HEIGHT;
	}
	else
	{
	    sw->width = sw->bkg_img->width;
	    sw->height = sw->bkg_img->height;
	}
	if(sw->width < 2)
	    sw->width = 2;
	if(sw->height < 2)
	    sw->height = 2;
	sw->cursor = OSWCursorOpenID(XC_watch);

	/* Create toplevel */
	parent = gui->root_win;
	OSWCreateWindow(
	    &w, parent,
	    0, 0,
	    sw->width, sw->height
	);
	sw->toplevel = w;
	OSWSetWindowInput(w, OSW_EVENTMASK_TOPLEVEL);
	OSWSetWindowCursor(w, sw->cursor);
	WidgetCenterWindow(w, WIDGET_WINDOW_RELATIVE_ROOT);

	OSWGetWindowGeometry(
	    w, &sw->x, &sw->y, &sw->width, &sw->height, NULL
	);

	/* Load icon */
	image = WidgetLoadImageFromTgaData(xsw_icon_tga);
	pixmap = OSWCreatePixmapFromImage(image);
	OSWDestroyImage(&image);

	/* WM properties */
	OSWSetWindowWMProperties(
	    w,
	    PROG_NAME_FULL,			/* Title */
	    PROG_NAME,			/* Icon title */
	    pixmap,			/* Icon */
	    False,			/* Do not let WM set coordinates */
	    sw->x, sw->y,
	    sw->width, sw->height,	/* Min size */
	    sw->width, sw->height,	/* Max size */
	    OSW_WINDOW_FRAME_NONE,
	    0, NULL
	);
	pixmap = 0;		/* Do not destroy the icon pixmap */


	/* Play splash sound */
	if(option.sounds > XSW_SOUNDS_NONE)
	    SoundPlay(
		&sound, SW_SNDREF_XSW_LOGO,
		1.0f, 1.0f,
		0,
		0
	    );

	/* Do fade in */
	SplashDoFadeIn(sw, w, sw->bkg_img);

	return(sw);
}

/*
 *	Draws the splash window's progress bar.
 */
static void SplashDrawProgressBar(
	splash_win_struct *sw, const float progress_coeff
)
{
	const int nticks = 12;
	int x, y, cx, cx_end;
	unsigned int width, height, bwidth, wwidth, wheight;
	char *s;
	win_t w;
	image_t *pb_l_img, *pb_r_img, *pb_t_img, *bkg_img;
	image_t *tmp_img;

	if(sw == NULL)
	    return;

	w = sw->toplevel;

	/* Load background image as needed */
	if(sw->bkg_img == NULL)
	{
	    s = STRDUP(PrefixPaths(
		dname.images, XSW_IMG_PATH_SPLASH_BG
	    ));
	    sw->bkg_img = WidgetLoadImageFromTgaFile(s);
	    free(s);
	}

	/* Load progress bar images as needed */
	if(sw->pb_l_img == NULL)
	{
	    s = STRDUP(PrefixPaths(
		dname.images, XSW_IMG_PATH_SPLASH_PBAR_LEFT
	    ));
	    sw->pb_l_img = WidgetLoadImageFromTgaFile(s);
	    free(s);
	}
	if(sw->pb_r_img == NULL)
	{
	    s = STRDUP(PrefixPaths(
		dname.images, XSW_IMG_PATH_SPLASH_PBAR_RIGHT
	    ));
	    sw->pb_r_img = WidgetLoadImageFromTgaFile(s);
	    free(s);
	}
	if(sw->pb_t_img == NULL)
	{
	    s = STRDUP(PrefixPaths(
		dname.images, XSW_IMG_PATH_SPLASH_PBAR_TICK
	    ));
	    sw->pb_t_img = WidgetLoadImageFromTgaFile(s);
	    free(s);

	    /* Resize ticks image */
	    if(sw->pb_t_img != NULL)
	    {
		image_t *src_img = sw->pb_t_img;

	        OSWGetWindowGeometry(w, NULL, NULL, &wwidth, &wheight, NULL);
	        width = (unsigned int)((int)wwidth * SW_LS_BAR_WIDTHC);
		width -= (int)((sw->pb_l_img != NULL) ?
		    sw->pb_l_img->width : 0
		);
		width -= (int)((sw->pb_r_img != NULL) ?
		    sw->pb_r_img->width : 0
		);

		width = MAX((int)width / nticks, 2);
		if(IS_NUM_ODD(width) && (width > 2))
		    width--;

		if(!OSWCreateImage(
		    &tmp_img,
		    width,
		    src_img->height
		))
		{
		    
		    WidgetResizeImageBuffer(
			tmp_img->bits_per_pixel,
			(u_int8_t *)tmp_img->data,
			(const u_int8_t *)src_img->data,
			tmp_img->width, tmp_img->height,
			tmp_img->bytes_per_line,
			src_img->width, src_img->height,
			src_img->bytes_per_line
		    );
		    OSWDestroyImage(&sw->pb_t_img);
		    sw->pb_t_img = tmp_img;
		}
	    }
	}

	bkg_img = sw->bkg_img;
	pb_l_img = sw->pb_l_img;
	pb_r_img = sw->pb_r_img;
	pb_t_img = sw->pb_t_img;

        OSWGetWindowGeometry(w, NULL, NULL, &wwidth, &wheight, NULL);
	width = (unsigned int)((int)wwidth * SW_LS_BAR_WIDTHC);
	bwidth = MAX(
	    (int)width - (int)((pb_l_img != NULL) ? pb_l_img->width : 0) -
		(int)((pb_r_img != NULL) ? pb_r_img->width : 0),
	    0
	);

	x = ((int)wwidth - (int)width) / 2;
	y = (int)((int)wheight * SW_LS_BAR_YC);

	height = (int)((pb_l_img != NULL) ? pb_l_img->height : 0);
	height = MAX(
	    (int)((pb_r_img != NULL) ? pb_r_img->height : 0),
	    (int)height
	);
	height = MAX(
	    (int)((pb_t_img != NULL) ? pb_t_img->height : 0),
	    (int)height
	);

	/* Recreate buffer as needed */
	if(sw->toplevel_buf == None)
	    OSWCreatePixmap(
		&sw->toplevel_buf, wwidth, wheight
	    );

	/* Clear portion of background */
	OSWPutImageToDrawableSect(
	    bkg_img, sw->toplevel_buf,
	    x, y,
	    x, y,
	    width, height
	);

	/* Draw ticks */
	if(pb_t_img != NULL)
	{
	    for(cx = x + (int)((pb_l_img == NULL) ? 0 : pb_l_img->width),
		cx_end = (int)(cx + ((int)bwidth * progress_coeff));
		cx < cx_end;
		cx += MAX((int)pb_t_img->width, 2)
	    )
	    {
		WidgetPutImageNormal(
		    sw->toplevel_buf, pb_t_img,
		    cx - 7, y,
		    True
		);
	    }
	}

	/* Draw the ends of the progress bar */
	if(pb_l_img != NULL)
	    WidgetPutImageNormal(
		sw->toplevel_buf, pb_l_img,
		x, y,
		True
	    );
	if(pb_r_img != NULL)
	    WidgetPutImageNormal(
		sw->toplevel_buf, pb_r_img,
		x + (int)width - (int)pb_r_img->width, y,
		True
	    );

	OSWCopyDrawablesCoord(
	    sw->toplevel,
	    sw->toplevel_buf,
	    x, y,
	    width, height,
	    x, y
	);
}


/*
 *	Updates the splash window's progress bar and message.
 */
void SplashProgress(
        splash_win_struct *sw,
        const float progress_coeff, const char *msg       
)
{
	int	lbearing, rbearing, swidth, ascent, descent,
		font_height;
	const char *s;
	win_t w;
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;
	font_t *font;

	if(sw == NULL)
	    return;

	/* Draw the progress bar */
	SplashDrawProgressBar(
	    sw,
	    CLIP(progress_coeff, 0.0f, 1.0f)
	);

	w = sw->toplevel;

	/* Draw labels */
	OSWSetForeground(gui->black_pix);
	font = wglobal->std_font;
	OSWSetFont(font);

	/* Draw version label on lower right corner */
	s = "Version: " PROG_VERSION;
	OSWGetTextBounds(
	    font, s, -1,
	    &lbearing, &rbearing, &swidth, &ascent, &descent
	);
	font_height = ascent + descent;
	OSWDrawString(
	    w,
	    (int)sw->width - swidth - 5 - lbearing,
	    (int)sw->height - font_height - 5 + ascent,
	    s
	);
}


/*
 *	Deletes the splash window.
 */
void SplashDelete(splash_win_struct *sw)
{
	if(sw == NULL)
	    return;

	if(IDC())
	{
	    OSWDestroyImage(&sw->pb_l_img);
	    OSWDestroyImage(&sw->pb_r_img);
	    OSWDestroyImage(&sw->pb_t_img);
	    OSWDestroyImage(&sw->bkg_img);

	    OSWDestroyPixmap(&sw->toplevel_buf);
	    OSWDestroyWindow(&sw->toplevel);

	    OSWDestroyCursor(&sw->cursor);
	}

	free(sw);
}
