#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
extern "C" {
#include <fnmatch.h>
}

#if defined(__SOLARIS__)
# include <sys/mnttab.h>
# include <sys/vfstab.h>
#elif defined(__FreeBSD__)
/* #  include <mntent.h> */
#elif defined(__linux__)
# include <mntent.h>
#endif

#include <unistd.h>

#include "../include/cfgfmt.h"
#include "../include/string.h"
#include "../include/strexp.h"
#include "../include/disk.h"
#include "../include/widget.h"

#include "../include/oswkeyvals.h"


static const char *FBrowserGetChild(const char *path);
static const char *FBrowserGetSelectionName(fb_struct *fb);

static bool_t FBrowserHasMask(const char *path);
static char *FBrowserGetPathMask(const char *path);
static char *FBrowserGetJustPath(const char *path);

static fb_object_struct *FBrowserGetSelObject(fb_struct *fb);

static void FBrowserDoOK(fb_struct *fb, bool_t from_ok_button);

static void FBrowserResizeCB(void *widget, void *data);
static void FBrowserDrawCB(void *widget, void *data);
static void FBrowserSBValueChangedCB(
	scroll_bar_struct *sb, float x, float y, void *data
);
static void FBrowserLocationPUListSelectCB(
	popup_list_struct *list, int sel_item, void *data
);
static void FBrowserParentPBCB(push_button_struct *btn, void *data);
static void FBrowserHomePBCB(push_button_struct *btn, void *data);
static void FBrowserOKPBCB(push_button_struct *btn, void *data);
static void FBrowserCancelPBCB(push_button_struct *btn, void *data);
static void FBrowserRefreshPBCB(push_button_struct *btn, void *data);
static void FBrowserPromptActivateCB(
	prompt_struct *prompt, const char *s, void *data
);
static void FBrowserMenuActivateCB(
	menu_struct *menu, int id, void *data
);
static void FBrowserCVPromptActivateCB(
	prompt_struct *prompt, const char *s, void *data
);

static void FBrowserDefaultOKCB(
	fb_struct *fb, const char *path, void *data
);
static void FBrowserDefaultCancelCB(fb_struct *fb, void *data);

static void FBrowserChangeDir(fb_struct *fb, const char *path);
static void FBrowserRefreshLocations(fb_struct *fb);
static void FBrowserRepositionListObjects(fb_struct *fb);
static void FBrowserRefreshList(fb_struct *fb);

bool_t FBrowserGetResponse(
	fb_struct *fb,
	const char *title,
	const char *ok_btn_label,
	const char *cancel_btn_label,
	const char *path,
	char ***paths_rtn, int *npaths_rtn,
	int map_rel, win_t toplevel,
	void (*main_push_func)(void *),
	void *main_push_data,
	void (*main_pop_func)(void *),
	void *main_pop_data
);
void FBrowserSetOpMesg(
	fb_struct *fb,
	const char *title,
	const char *ok_btn_label,
	const char *cancel_btn_label
);
void FBrowserSetTransientFor(fb_struct *fb, win_t toplevel);
void FBrowserSetCallbacks(
	fb_struct *fb,
	void (*ok_cb)(fb_struct *, const char *, void *),
	void *ok_data,
	void (*cancel_cb)(fb_struct *, void *),
	void *cancel_data
);
void FBrowserClearCallbacks(fb_struct *fb);

int FBrowserInit(fb_struct *fb, int style);
void FBrowserResize(fb_struct *fb);
void FBrowserQueueResize(fb_struct *fb);
void FBrowserDraw(fb_struct *fb, int amount);
void FBrowserQueueDraw(fb_struct *fb, int amount);
int FBrowserManage(fb_struct *fb, event_t *event);
void FBrowserMapCVPrompt(fb_struct *fb, int mode);
void FBrowserUnmapCVPrompt(fb_struct *fb);
void FBrowserMap(fb_struct *fb);
void FBrowserMapPath(fb_struct *fb, const char *path);
void FBrowserMapSearchMask(fb_struct *fb, const char *mask);
void FBrowserUnmap(fb_struct *fb);
void FBrowserDestroy(fb_struct *fb);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


#define FB_DEF_WIDTH			480
#define FB_DEF_HEIGHT			300

#define FB_MARGIN			10

#define FB_ICON_BTN_WIDTH		26
#define FB_ICON_BTN_HEIGHT		26

#define FB_BTN_WIDTH			70
#define FB_BTN_HEIGHT			28

#define FB_LIST_BORDER			5
#define FB_LIST_ROW_HEIGHT		20


/* Directory window width in percent */
#define FB_DIR_LIST_WIDTH_PERCENT	35

/* X colum shift increment per directory level (in pixels) for
 * the file browser's list windows
 */
#define FB_LIST_COLUM_LEVEL_INC         26


#define FB_PROMPT_HEIGHT		30

/* Number of items visible at once on the locations popup list */
#define FB_LOCATIONS_LIST_NITEMS_VISIBLE	7

/*
 *	Operations:
 */
#define FB_OP_NONE			0
#define FB_OP_SELECT			20
#define FB_OP_GOTO_PARENT		30
#define FB_OP_GOTO_HOME			31
#define FB_OP_REFRESH                   50

/*
 *	Change value prompt modes:
 */
#define FB_CVP_MODE_NONE	0
#define FB_CVP_MODE_RENAME	1
#define FB_CVP_MODE_COPY	2
#define FB_CVP_MODE_MOVE	3
#define FB_CVP_MODE_MODE	4	/* Change mode */
#define FB_CVP_MODE_RUN		5	/* run program to view selected file */


/*
 *	Returns a statically allocated string containing only the child
 *	of the path statement path.
 */
static const char *FBrowserGetChild(const char *path)
{
	static char rtn_str[PATH_MAX];
	strncpy(rtn_str, path, sizeof(rtn_str));
	rtn_str[sizeof(rtn_str) - 1] = '\0';
	StripAbsolutePath(rtn_str);
	return(rtn_str);
}

/*
 *	Returns a statically allocated string containing the name
 *	of the item selected by the file browser fb.
 *
 *	Can return NULL if there is no object selected.
 */
static const char *FBrowserGetSelectionName(fb_struct *fb)
{
	int i;
	fb_object_struct *obj;
	static char rtn_str[PATH_MAX + NAME_MAX];

	if(fb == NULL)
	    return(NULL);

	if(fb->style == FB_STYLE_SINGLE_LIST)
	{
	    i = fb->sel_object;
	    if(i < 0)
		return(NULL);

	    if(i < fb->dir_list_items)
	    {
		obj = fb->dir_list[i];
		if(obj != NULL)
		{
		    strncpy(
			rtn_str, obj->name,
		 	sizeof(rtn_str)
		    );
		    rtn_str[sizeof(rtn_str) - 1] = '\0';
		    return(rtn_str);
		}
	    }
	    else
	    {
		i -= fb->dir_list_items;
		if((i >= 0) && (i < fb->file_list_items))
		{
		    obj = fb->file_list[i];
		    if(obj != NULL)
		    {
			strncpy(
			    rtn_str, obj->name,
			    sizeof(rtn_str)
			);
			rtn_str[sizeof(rtn_str) - 1] = '\0';
			return(rtn_str);
		    }
		}
	    }
	}
	else	/* FB_STYLE_SPLIT_LIST */
	{
	    i = fb->sel_dir;
	    if((i > -1) &&
	       (i < fb->dir_list_items)
	    )
	    {
		obj = fb->dir_list[i];
	        if(obj != NULL)
	        {
		    strncpy(
		        rtn_str, obj->name, 
		        sizeof(rtn_str)
		    );
		    rtn_str[sizeof(rtn_str) - 1] = '\0';
		    return(rtn_str);
	        }
	    }

	    i = fb->sel_file;
	    if((i > -1) &&
	       (i < fb->file_list_items)
	    )
	    {
		obj = fb->file_list[i];
		if(obj != NULL)
		{
		    strncpy(  
			rtn_str, obj->name,
			sizeof(rtn_str)
		    );
		    rtn_str[sizeof(rtn_str) - 1] = '\0';
		    return(rtn_str);
		}
	    }
	}

	return(NULL);
}


/*
 *	Checks if the specified path has a mask specified in it.
 */
static bool_t FBrowserHasMask(const char *path)
{
	const char *s;

	if(path == NULL)
	    return(False);

	s = strrchr(path, '/');
	if(s == NULL)
	    s = path;
	else
	    s = s + 1;

	while(*s != '\0')
	{
	    /* Following wild card characters suggest selection mask */
	    if((*s == '*') || (*s == '?'))
		return(True);
	    s++;
	}

	return(False);
}


/*
 *	Returns a copy of the the mask at the end of specified path or
 *	"*" if none was specified.
 */
static char *FBrowserGetPathMask(const char *path)
{
	const char *s, *mask;

	if(path == NULL)
	    return(strdup("*"));

	/* Seek mask to the last segment of the path */
	mask = strrchr(path, DIR_DELIMINATOR);
	if(mask == NULL)
	    return(strdup("*"));

	mask++;	/* Seek past the deliminator */

	/* Now mask is at the last segment of the path, check if it
	 * has any wildcard characters that would indicate that it is
	 * a mask
	 */
	for(s = mask; *s != '\0'; s++)
	{
	    if((*s == '*') || (*s == '?'))
		return(strdup(mask));
	}

	return(strdup("*"));
}

/*
 *	Returns a copy of the path-only portion (without the mask)
 *	of the specified path.
 */
static char *FBrowserGetJustPath(const char *path)
{
	int len;
	const char *path_end;
	char *rtn_path;

	if(path == NULL)
	    return(strdup("/"));

	/* Does not have a mask? */
	if(!FBrowserHasMask(path))
	    return(strdup(path));

	/* Find the last deliminator character */
	path_end = strrchr(path, DIR_DELIMINATOR);
	if(path_end == NULL)
	    return(strdup("/"));

	/* Calculate the length of the path up to the last deliminator
	 * character (not including the last deliminator character) so
	 * that we exclude the last deliminator character and everything
	 * past it
	 */
	len = path_end - path;
	if(len <= 0)
	    return(strdup("/"));

	/* Allocate the return path */
	rtn_path = (char *)malloc((len + 1) * sizeof(char));
	if(rtn_path == NULL)
	    return(strdup("/"));

	/* Copy the path up to the last deliminator character to the
	 * return path
	 */
	memcpy(rtn_path, path, len);
	rtn_path[len] = '\0';

	return(rtn_path);
}


/*
 *	Returns the selected object or NULL if there no objects are
 *	selected.
 */
static fb_object_struct *FBrowserGetSelObject(fb_struct *fb) 
{
	int i;

	if(fb == NULL)
	    return(NULL);

	if(fb->style == FB_STYLE_SINGLE_LIST)
	{
	    i = fb->sel_object;
	    if((i >= 0) && (i < (fb->dir_list_items + fb->file_list_items)))
	    {
		if(i < fb->dir_list_items)
		    return(fb->dir_list[i]);

		i -= fb->dir_list_items;
		if((i >= 0) && (i < fb->file_list_items))
		    return(fb->file_list[i]);
	    }
	}
	else	/* FB_STYLE_SPLIT_LIST */
	{
	    i = fb->sel_dir;
	    if((i >= 0) && (i < fb->dir_list_items))
		return(fb->dir_list[i]);

	    i = fb->sel_file;
	    if((i >= 0) && (i < fb->file_list_items))
		return(fb->file_list[i]);
	}

	return(NULL);
}


/*
 *	OK procedure.
 *
 *	Gets the selected object (or the current location if no object
 *	is selected) and either reports selection of the object(s)
 *	(if they are files) or changes to the directory if the selected
 *	object or location is a directory.
 *
 *	The from_ok_btn specifies that the OK button was pressed
 *	in order to call this function. This modifies the behavior when
 *	selecting directories. If a directory is selected and
 *	from_button_press is True then the directory will be reported
 *	as selected instead of chanting to that directory.
 */
static void FBrowserDoOK(fb_struct *fb, bool_t from_ok_btn)
{
	const char *name;
	char *path_mask, *parent;
	prompt_struct *prompt;

	if(fb == NULL)
	    return;

	prompt = fb->prompt;

	/* Get the selected item's name */
	name = FBrowserGetSelectionName(fb); 
	if(name != NULL)
	{
	    struct stat stat_buf;

	    /* Prefix the location prompt's value with the item's
	     * name to form the full path
	     */
	    char *parent = FBrowserGetJustPath(PromptGetS(prompt));

	    /* Parent does not exist? */
	    if((parent != NULL) ? stat(parent, &stat_buf) : True)
	    {
		/* Get the grand parent as the parent */
		char *grand_parent = STRDUP(GetParentDir(parent));
		free(parent);
		parent = grand_parent;
	    }
	    else
	    {
		/* Is the parent not a directory? */
#ifdef S_ISDIR
		if(!S_ISDIR(stat_buf.st_mode))
#else
		if(True)
#endif
		{
		    /* Get the grand parent as the parent */
		    char *grand_parent = STRDUP(GetParentDir(parent));
		    free(parent);
		    parent = grand_parent;
		}
	    }
	    if(parent == NULL)
		return;

	    path_mask = STRDUP(PrefixPaths(parent, name));
	    free(parent);
	}
	else
	{
	    /* No item selected, so just use the location prompt's
	     * value by itself
	     */
	    path_mask = STRDUP(PromptGetS(prompt));
	}
	if(path_mask == NULL)
	    return;

	/* Get the path */
	parent = FBrowserGetJustPath(path_mask);

	/* Path leads to a directory? */
	if(ISPATHDIR(parent) && !from_ok_btn)
	{
	    FBrowserChangeDir(fb, parent);
	}
	else if(parent != NULL)
	{
	    /* Selected object is something that should be returned
	     * to the calling function
	     */
	    struct stat stat_buf;

	    if(stat(parent, &stat_buf))
	    {
		/* Object does not exist, should we accept that? */
		if(!(fb->options & FB_FLAG_MUST_EXIST))
		{
		    /* Does not have to exist, so report it */

		    /* Unmap file browser */
		    if(fb->options & FB_FLAG_CLOSE_ON_OK)
			FBrowserUnmap(fb);

		    /* Call OK callback */
		    if(fb->ok_cb != NULL)
			fb->ok_cb(
			    fb,
			    parent,
			    fb->ok_data
			);

		    if(!WidgetIsRegistered(fb))
		    {
			free(parent);
			free(path_mask);
			return;
		    }
		}
	    }
	    else
	    {
		/* Object exists */

		/* Unmap file browser */
		if(fb->options & FB_FLAG_CLOSE_ON_OK)
		    FBrowserUnmap(fb);

		/* Call OK callback */
		if(fb->ok_cb != NULL)
		    fb->ok_cb(
			fb,
			parent,
			fb->ok_data
		    );

		if(!WidgetIsRegistered(fb))
		{
		    free(parent);
		    free(path_mask);
		    return;
		}
	    }
	}

	free(parent);
	free(path_mask);
}


/*
 *	Resize callback.
 */
static void FBrowserResizeCB(void *widget, void *data)
{
	fb_struct *fb = (fb_struct *)widget;
	if(fb == NULL)
	    return;   

	FBrowserResize(fb);
}

/*
 *	Draw callback.
 */
static void FBrowserDrawCB(void *widget, void *data)
{
	int draw_amount = (int)data;
	fb_struct *fb = (fb_struct *)widget;
	if(fb == NULL)
	    return;

	if(fb->map_state)
	    FBrowserDraw(fb, draw_amount);
}

/*
 *	Scroll bar value changed callback.
 */
static void FBrowserSBValueChangedCB(
	scroll_bar_struct *sb, float x, float y, void *data
)
{
	fb_struct *fb = FB(data);
	if(fb == NULL)
	    return;

	if(fb->map_state)
	    FBrowserQueueDraw(fb, FB_DRAW_ALLLISTS);
}

/*
 *	Location popup list select callback.
 */
static void FBrowserLocationPUListSelectCB(
	popup_list_struct *list, int sel_item, void *data
)
{
	char *path;
	popup_list_item_struct *item;
	fb_struct *fb = FB(data);
	if((list == NULL) || (fb == NULL))
	    return;

	item = PUListGetItem(list, sel_item);
	if(item == NULL)
	    return;

	/* Change location */
	path = STRDUP(item->text);
	FBrowserUnmapCVPrompt(fb);
	FBrowserChangeDir(fb, path);

	free(path);
}

/*
 *	Parent callback.
 */
static void FBrowserParentPBCB(push_button_struct *btn, void *data)
{
	fb_struct *fb = FB(data);
	if(fb == NULL)
	    return;

	FBrowserChangeDir(fb, "..");
}

/*
 *	Home callback.
 */
static void FBrowserHomePBCB(push_button_struct *btn, void *data)
{
	char *path;
	fb_struct *fb = FB(data);
	if(fb == NULL)
	    return;

	path = STRDUP(getenv("HOME"));
	FBrowserChangeDir(fb, path);
	free(path);
}

/*
 *	OK button callback.
 */
static void FBrowserOKPBCB(push_button_struct *btn, void *data)
{
	fb_struct *fb = FB(data);
	if(fb == NULL)
	    return;

	FBrowserDoOK(fb, True);
}

/*
 *	Cancel button callback.
 */
static void FBrowserCancelPBCB(push_button_struct *btn, void *data)
{
	fb_struct *fb = FB(data);
	if(fb == NULL)
	    return;

	/* Call cancel callback */
	if(fb->cancel_cb != NULL)
	    fb->cancel_cb(
		fb,
		fb->cancel_data
	    );

	if(!WidgetIsRegistered(fb))
	    return;

	/* Unmap the file browser */
	if(fb->options & FB_FLAG_CLOSE_ON_CANCEL)
	    FBrowserUnmap(fb);
}

/*
 *	Refresh button callback.
 */
static void FBrowserRefreshPBCB(push_button_struct *btn, void *data)
{
	fb_struct *fb = FB(data);
	if(fb == NULL)
	    return;

	/* Reget listings */
	FBrowserRefreshLocations(fb);
	FBrowserRefreshList(fb);

	/* Redraw listings */
	if(fb->map_state)
	{
	    FBrowserQueueDraw(fb, FB_DRAW_ALLLISTS);
	    FBrowserQueueDraw(fb, FB_DRAW_SCROLLBARS);
	}
}

/*
 *	Location prompt activate callback.
 */
static void FBrowserPromptActivateCB(
	prompt_struct *prompt, const char *s, void *data
)
{
	fb_struct *fb = FB(data);
	if(fb == NULL)
	    return;

	FBrowserDoOK(fb, False);
}


/*
 *	Menu activate callback.
 */
static void FBrowserMenuActivateCB(
	menu_struct *menu, int id, void *data
)
{
	const int op = id;
	fb_struct *fb = FB(data);
	if(fb == NULL)
	    return;

	switch(op)
	{
	  case FB_OP_SELECT:
	    FBrowserDoOK(fb, False);
	    break;
	  case FB_OP_GOTO_PARENT:
	    FBrowserParentPBCB(fb->parent_btn, fb);
	    break;
	  case FB_OP_GOTO_HOME:
	    FBrowserHomePBCB(fb->home_btn, fb);
	    break;
	  case FB_OP_REFRESH:
	    FBrowserRefreshPBCB(fb->refresh_btn, fb);
	    break;
	}
}

/*
 *      Apply changes from cv prompt.
 */
static void FBrowserCVPromptActivateCB(
	prompt_struct *prompt, const char *s, void *data
)
{
	char	tar_path[PATH_MAX + NAME_MAX],
		src_path[PATH_MAX + NAME_MAX],
		new_name[PATH_MAX + NAME_MAX],
		old_name[PATH_MAX + NAME_MAX];
	struct stat stat_buf;
	widget_global_struct *wglobal = &widget_global;
	fb_struct *fb = FB(data);
	if((s == NULL) || (fb == NULL))
	    return;

	s = PromptGetS(fb->prompt);
	if(!ISPATHABSOLUTE(s))
	    return;

	s = fb->cv_prompt_target;
	if(s == NULL)
	    return;
	if(!ISPATHABSOLUTE(s))
	    return;

	switch(fb->cv_prompt_mode)
	{
	  case FB_CVP_MODE_RENAME:
	    /* Target name must not be absolute */
	    if(ISPATHABSOLUTE(PromptGetS(fb->cv_prompt)))
		return;

	    /* Record old name */
	    s = FBrowserGetChild(fb->cv_prompt_target);
	    strncpy(old_name, s, sizeof(old_name));
	    old_name[sizeof(old_name) - 1] = '\0';

	    /* Get source paths */
	    strncpy(src_path, fb->cv_prompt_target, sizeof(src_path));
	    src_path[sizeof(src_path) - 1] = '\0';

	    /* Get target path */
	    s = GetParentDir(fb->cv_prompt_target);
	    strncpy(
		tar_path,
		(s != NULL) ? s : "/",
		sizeof(tar_path)
	    );
	    tar_path[sizeof(tar_path) - 1] = '\0';

	    s = PrefixPaths(
		tar_path, PromptGetS(fb->cv_prompt)
	    );
	    strncpy(tar_path, (s != NULL) ? s : "", sizeof(tar_path));
	    tar_path[sizeof(tar_path) - 1] = '\0';


	    /* Target already exists? */
	    if(!stat(tar_path, &stat_buf))
		return;

	    if(!rename(src_path, tar_path))
	    {
		/* Update name on lists */
		int i, icon_label_spacing, swidth;
		unsigned int icon_width, icon_height;
		font_t *font = fb->font;
		pixmap_t pixmap = wglobal->icon_folder_closed_20x20_pixmap;
		fb_object_struct *obj;

		if(pixmap != None)
		{
		    OSWGetPixmapGeometry(
			pixmap, &icon_width, &icon_height, NULL
		    );
		    icon_label_spacing = 2;
		}
		else
		{
		    icon_width = 0;
		    icon_height = 0;
		    icon_label_spacing = 0;
		}

	        strncpy(
		    new_name, PromptGetS(fb->cv_prompt),
		    sizeof(new_name)
		);
	        new_name[sizeof(new_name) - 1] = '\0';

	        for(i = 0; i < fb->dir_list_items; i++)
	        {
		    obj = fb->dir_list[i];
		    if(obj == NULL)
			continue;

		    if(obj->name == NULL)
			continue;

		    if(!strcmp(obj->name, old_name))
		    {
		        free(obj->name);
			obj->name = STRDUP(new_name);
			OSWGetTextBounds(
			    font, new_name, -1,
			    NULL, NULL, &swidth, NULL, NULL
			);
			obj->width = icon_width + icon_label_spacing +
			    swidth;
		    }
	        }
		for(i = 0; i < fb->file_list_items; i++)
		{
		    obj = fb->file_list[i];
		    if(obj == NULL)
			continue;

		    if(obj->name == NULL)
			continue;

		    if(!strcmp(obj->name, old_name))
		    {
			free(obj->name);           
			obj->name = STRDUP(new_name);
			OSWGetTextBounds(
			    font, new_name, -1,
			    NULL, NULL, &swidth, NULL, NULL
			);
			obj->width = icon_width + icon_label_spacing +
			    swidth;
		    }
		}
	    }
	    break;
	}

	FBrowserUnmapCVPrompt(fb);

	if(fb->map_state)
	{
	    FBrowserQueueDraw(fb, FB_DRAW_ALLLISTS);
	    FBrowserQueueDraw(fb, FB_DRAW_SCROLLBARS);
	}
}

/*
 *	Default OK callback.
 */
static void FBrowserDefaultOKCB(
	fb_struct *fb, const char *path, void *data
)
{
	if(fb == NULL)
	    return;

	fb->response_paths = strlistappend(
	    fb->response_paths,
	    &fb->nresponse_paths,
	    path
	);

	if(fb->main_pop_func != NULL)
	    fb->main_pop_func(fb->main_pop_data);
}

/*
 *	Default cancel callback.
 */
static void FBrowserDefaultCancelCB(fb_struct *fb, void *data)
{
	if(fb == NULL)
	    return;

	strlistfree(fb->response_paths, fb->nresponse_paths);
	fb->response_paths = NULL;
	fb->nresponse_paths = 0;

	if(fb->main_pop_func != NULL)
	    fb->main_pop_func(fb->main_pop_data);
}



/*
 *	Changes the current location based on the specified path,
 *	regets the listings, and redraws.
 */
static void FBrowserChangeDir(fb_struct *fb, const char *path)
{
	char *parent, *mask, *new_path;
	prompt_struct *prompt;

	if((fb == NULL) || STRISEMPTY(path))
	    return;

	prompt = fb->prompt;

	/* Get path and search string */
	parent = FBrowserGetJustPath(PromptGetS(prompt));
	mask = FBrowserGetPathMask(PromptGetS(prompt));
	if(parent == NULL)
	{
	    free(parent);
	    free(mask);
	    return;
	}

	if(ISPATHABSOLUTE(path))
	{
	    new_path = STRDUP(path);
	}
	else if(!strcmp(path, "."))
	{
	    new_path = STRDUP(parent);
	}
	else if(!strcmp(path, ".."))
	{
	    new_path = STRDUP(GetParentDir(parent));
	}
	else if(*path == '~')
	{
	    new_path = STRDUP(PathSubHome(path));
	}
	else
	{
	    new_path = STRDUP(PrefixPaths(parent, path));
	}
	if(new_path != NULL)
	{
	    /* Update the location prompt */
	    char *path_mask = STRDUP(PrefixPaths(new_path, mask));
	    PromptSetS(prompt, path_mask);
	    PromptMarkBuffer(prompt, PROMPT_POS_END);
	    free(path_mask);
	}

	free(mask);
	free(parent);
	free(new_path);

	/* Reget listings */
	FBrowserRefreshLocations(fb);
	FBrowserRefreshList(fb);

	/* Redraw listings */
	if(fb->map_state)
	{
	    FBrowserQueueDraw(fb, FB_DRAW_ALLLISTS);
	    FBrowserQueueDraw(fb, FB_DRAW_SCROLLBARS);
	}
}


/*
 *	Deletes all items in the location popup list and gets a new
 *	listing based on the current location.
 */
static void FBrowserRefreshLocations(fb_struct *fb)
{
	const char *s;
	char *path;
#if defined(__linux__) || defined(__SOLARIS__)
	FILE *fp;
#endif
	prompt_struct *prompt;
	popup_list_struct *pulist;
	widget_global_struct *wglobal = &widget_global;

	if(fb == NULL)
	    return;

	prompt = fb->prompt;
	pulist = fb->location_pulist;

	/* Delete the existing listing */
	PUListClear(pulist);

	/* Add the current path and its parent paths */
	s = PromptGetS(prompt);
	path = FBrowserGetJustPath(s);
	if(path != NULL)
	{
	    int n = 0;
	    char *s2;

	    do
	    {
		if(n == 0)
		    PUListAddItem(
			pulist,
			0,
			path,
			wglobal->icon_folder_opened_20x20_pixmap,
			wglobal->icon_folder_opened_20x20_mask
		    );
		else
		    PUListAddItem(
			pulist,
			0,
			path,
			wglobal->icon_folder_closed_20x20_pixmap,
			wglobal->icon_folder_closed_20x20_mask
		    );

		s2 = strrchr(path, '/');
		if(s2 != NULL)
		    *s2 = '\0';

		n++;

	    } while((s2 != NULL) && (*path != '\0'));

	    free(path);
	}

	/* Add the user's home directory */
	path = STRDUP(getenv("HOME"));
	if(path != NULL)
	{
	    PUListAddItem(
		pulist,
		0,
		path,
		wglobal->icon_folder_closed_20x20_pixmap,
		wglobal->icon_folder_closed_20x20_mask
	    );
	    free(path);
	}

#if defined(__SOLARIS__)
	/* Add the paths from the filesystem configuration file */
	fp = fopen("/etc/vfstab", "rb");
	if(fp != NULL)
	{
	    struct vfstab *vfs_ptr = (struct vfstab *)malloc(
		sizeof(struct vfstab)
	    );
	    int mtback = getvfsent(fp, vfs_ptr);
	    for(; mtback != 0; mtback = getmntent(fp, vfs_ptr))
	    {
		if(!ISPATHABSOLUTE(vfs_ptr->vfs_mountp))
		    continue;

		PUListAddItem(
		    pulist,
		    0,
		    vfs_ptr->vfs_mountp,
		    wglobal->dir_closed_pixmap,
		    wglobal->dir_closed_mask
		);
	    }

	    fclose(fp);
	}
#elif defined(__linux__)
	fp = setmntent("/etc/fstab", "rb");
	if(fp != NULL)
	{
	    struct mntent *me = getmntent(fp);
	    for(; me != NULL; me = getmntent(fp))
	    {
		if(!ISPATHABSOLUTE(me->mnt_dir))
		    continue;

		PUListAddItem(
		    pulist,
		    0,
		    me->mnt_dir,
		    wglobal->icon_folder_closed_20x20_pixmap,
		    wglobal->icon_folder_closed_20x20_mask
		);
	    }
	    endmntent(fp);
	}
#endif

	/* Select the first item */
	PUListSetSelItem(pulist, 0);

	/* Redraw the locations list */
	if(fb->map_state)
	    WidgetQueueDraw(pulist);
}

/*
 *	Updates the positions of the objects in the list based
 *	on the current size of the list.
 *
 *	Does not redraw.
 */
static void FBrowserRepositionListObjects(fb_struct *fb)
{
	widget_global_struct *wglobal = &widget_global;

	if(fb == NULL)
	    return;

	if(fb->style == FB_STYLE_SINGLE_LIST)
	{
	    int	i, x, y, column_num, row_num,
		swidth,
		longest_obj_width, icon_label_spacing, rows_visible;
	    unsigned int        width, height,
				icon_width, icon_height;
	    pixmap_t icon_pixmap;
	    font_t *font = fb->font;
	    win_t w = fb->list_win;
	    fb_object_struct *obj;

	    /* Get the geometry of the list window */
	    OSWGetWindowGeometry(
		w, NULL, NULL, &width, &height, NULL
	    );

	    /* Calculate the number of rows completely visible
	     * (account for the scroll bar regardless of if it is
	     * mapped or not)
	     */
	    fb->list_rows_visible = rows_visible = MAX(
		(((int)height - SCROLLBAR_XBAR_HEIGHT) / FB_LIST_ROW_HEIGHT),
		1
	    );

	    /* Get size of icon (use the directory icon) */
	    icon_pixmap = wglobal->icon_folder_closed_20x20_pixmap;
	    if(icon_pixmap != None)
	    {
		OSWGetPixmapGeometry(
		    icon_pixmap, &icon_width, &icon_height, NULL
		);
		icon_label_spacing = 2;
	    }
	    else
	    {
		icon_width = 0;
		icon_height = 0;
		icon_label_spacing = 0;
	    }

	    x = 0;
	    y = 0;
	    column_num = 0;
	    row_num = 0;
	    longest_obj_width = 0;

	    /* Directories */
	    for(i = 0; i < fb->dir_list_items; i++)
	    {
		obj = fb->dir_list[i];
		if(obj == NULL)
		    continue;

		OSWGetTextBounds(
		    font, obj->name, -1,
		    NULL, NULL, &swidth, NULL, NULL
		);

		obj->x = x;
		obj->y = y;
		obj->width = icon_width + icon_label_spacing + swidth;
		obj->height = FB_LIST_ROW_HEIGHT;

		if((int)obj->width > longest_obj_width)
		    longest_obj_width = (int)obj->width;

		row_num++;
		y += (int)obj->height;
		if(row_num >= rows_visible)
		{
		    x += longest_obj_width + FB_LIST_BORDER;
		    y = 0;
		    column_num++;
		    row_num = 0;
		    longest_obj_width = 0;
		}
	    }

	    /* Files and others */
	    for(i = 0; i < fb->file_list_items; i++)
	    {
		obj = fb->file_list[i];
		if(obj == NULL)
		    continue;

		OSWGetTextBounds(
		    font, obj->name, -1,
		    NULL, NULL, &swidth, NULL, NULL
		);

		obj->x = x;
		obj->y = y;
		obj->width = icon_width + icon_label_spacing + swidth;
		obj->height = FB_LIST_ROW_HEIGHT;

		if((int)obj->width > longest_obj_width)
		    longest_obj_width = (int)obj->width;

		row_num++;
		y += (int)obj->height;
		if(row_num >= rows_visible)
		{
		    x += longest_obj_width + FB_LIST_BORDER;
		    y = 0;
		    column_num++;
		    row_num = 0;
		    longest_obj_width = 0;
		}
	    }

	    /* Recalculate the list_max_width */
	    fb->list_max_width = x + longest_obj_width + (2 * FB_LIST_BORDER);

	    /* Resize the list's scroll bar */
	    SBarQueueResize(
		fb->list_win_sb,
		width, height,
		fb->list_max_width, height
	    );
	}
}

/*
 *	Deletes the current listing and gets a new listing from the
 *	current location.
 */
static void FBrowserRefreshList(fb_struct *fb)
{
	bool_t show_hidden = True;
	int i, n;
	const char *name;
	char *parent, *mask, **dent, *full_path;
	struct stat stat_buf;
	prompt_struct *prompt;
	fb_object_struct *obj;

	if(fb == NULL)
	    return;

	prompt = fb->prompt;

	/* Get the parent and mask */
	parent = FBrowserGetJustPath(PromptGetS(prompt));
	mask = FBrowserGetPathMask(PromptGetS(prompt));

	/* Location must refer to an existing directory */
	if(!ISPATHDIR(parent))
	{
	    free(parent);
	    free(mask);
	    return;
	}

	/* Delete the current listing */
	for(i = 0; i < fb->dir_list_items; i++)
	{
	    obj = fb->dir_list[i];
	    if(obj == NULL)
		continue;
 
	    free(obj->name);
	    free(obj);
	}
	free(fb->dir_list);
	fb->dir_list = NULL;
	fb->dir_list_items = 0;
	fb->sel_dir = -1;
	
	for(i = 0; i < fb->file_list_items; i++)
	{
	    obj = fb->file_list[i];
	    if(obj == NULL)
		continue;

	    free(obj->name);
	    free(obj);
	}
	free(fb->file_list);
	fb->file_list = NULL;
	fb->file_list_items = 0;
	fb->sel_file = -1;
	fb->sel_object = -1;


#define IS_OBJ_HIDDEN(_name_)			\
 ((*(_name_) == '.') ? ((_name_)[1] != '\0') : False)
	    
	/* Get new listing */
	dent = GetDirEntNames(parent);
	if(dent == NULL)
	{
	    free(parent);
	    free(mask);
	    return;
	}

	/* Sort dents */
	i = 0;
	while(dent[i] != NULL)
	    i++;
	StringQSort(dent, i);

	/* Add to list */
	i = 0;
	while(dent[i] != NULL)
	{
	    name = dent[i];
	    if(name == NULL)
	    {
		i++;
		continue;
	    }

	    /* Get the full path to this object */
	    full_path = STRDUP(PrefixPaths(parent, name));
	    if(full_path == NULL)
	    {
		free(dent[i]);
		dent[i] = NULL;
		i++;
		continue;
	    }

	    /* Get local stats */
	    if(lstat(full_path, &stat_buf))
		memset(&stat_buf, 0x00, sizeof(struct stat));

	    /* Skip special directory notations */
	    if(!strcmp(name, "."))
	    {
		free(dent[i]);
		dent[i] = NULL;
		free(full_path);
		i++;
		continue;
	    }

	    /* Accept parent directory notation? */
	    if(fb->style == FB_STYLE_SINGLE_LIST)
	    { 
		/* In FB_STYLE_SINGLE_LIST style, always skip */
		if(!strcmp(name, ".."))
		{
		    free(dent[i]);
		    dent[i] = NULL;
		    free(full_path);
		    i++;
		    continue;
		}
	    }
	    else
	    {
		/* Skip parent if toplevel */
		if(!strcmp(parent, "/") &&
		   !strcmp(name, "..")
		)
		{
		    free(dent[i]);
		    dent[i] = NULL;
		    free(full_path);
		    i++;
		    continue;
		}
	    }

	    /* Is the object hidden? */
	    if(!show_hidden)
	    {
		if(IS_OBJ_HIDDEN(name))
		{
		    free(dent[i]);
		    dent[i] = NULL;
		    free(full_path);
		    i++;
		    continue;
		}
	    }

	    /* Is this a directory? (follow links) */
	    if(ISPATHDIR(full_path))
	    {
		/* Add to directory list */
		n = fb->dir_list_items;
		fb->dir_list_items = n + 1;
		fb->dir_list = (fb_object_struct **)realloc(
		    fb->dir_list,
		    fb->dir_list_items * sizeof(fb_object_struct *)
		);
		if(fb->dir_list == NULL)
		{
		    fb->dir_list_items = 0;
		    free(dent[i]);
		    dent[i] = NULL;
		    free(full_path);
		    i++;
		    continue;
		}

		/* Create new object */
		fb->dir_list[n] = obj = (fb_object_struct *)calloc(
		    1, sizeof(fb_object_struct)
		);
		if(obj == NULL)
		{
		    free(dent[i]);
		    dent[i] = NULL;
		    free(full_path);
		    i++;
		    continue;
		}

		/* Set new object values */
		obj->name = STRDUP(name);
		obj->mode = stat_buf.st_mode;
		obj->uid = stat_buf.st_uid;
		obj->gid = stat_buf.st_gid;
		obj->size = stat_buf.st_size;
		obj->atime = stat_buf.st_atime;
		obj->ctime = stat_buf.st_ctime;
		obj->mtime = stat_buf.st_mtime;
	    }
	    else
	    {
		/* Name of file matched in search string? */
		if(fnmatch(mask, name, 0))
		{
		    /* Need to free this entry name */
		    free(dent[i]);
		    dent[i] = NULL;
		    free(full_path);
		    i++;
		    continue;
		}

		/* Add to files list */
		n = fb->file_list_items;
		fb->file_list_items = n + 1;
		fb->file_list = (fb_object_struct **)realloc(
		    fb->file_list,
		    fb->file_list_items * sizeof(fb_object_struct *)
		);
		if(fb->file_list == NULL)  
		{
		    fb->file_list_items = 0;
		    free(dent[i]);
		    dent[i] = NULL;
		    free(full_path);
		    i++;
		    continue;
		}

		/* Create new object */
		fb->file_list[n] = obj = (fb_object_struct *)calloc(
		    1, sizeof(fb_object_struct)
		);
		if(obj == NULL)
		{
		    free(dent[i]);
		    dent[i] = NULL;
		    free(full_path);
		    i++;
		    continue;
		}

		/* Set new object values */
		obj->name = STRDUP(name);
		obj->mode = stat_buf.st_mode;
		obj->uid = stat_buf.st_uid;  
		obj->gid = stat_buf.st_gid;
		obj->size = stat_buf.st_size;  
		obj->atime = stat_buf.st_atime;
		obj->ctime = stat_buf.st_ctime;
		obj->mtime = stat_buf.st_mtime;
	    }

	    free(full_path);
	    free(dent[i]);
	    dent[i] = NULL;
	    i++;
	}

	/* Delete dent pointer array (each entry has been either
	 * transfered or deleted
	 */
	free(dent);

#undef IS_OBJ_HIDDEN

	/* Set the positions for the new objects */
	FBrowserRepositionListObjects(fb);

	/* Reset the scroll bar positions */
	SBarSetPositionX(fb->dir_win_sb, 0.0f);
	SBarSetPositionY(fb->dir_win_sb, 0.0f);
	SBarSetPositionX(fb->file_win_sb, 0.0f);
	SBarSetPositionY(fb->file_win_sb, 0.0f);
	SBarSetPositionX(fb->list_win_sb, 0.0f);
	SBarSetPositionY(fb->list_win_sb, 0.0f);

	free(parent);
	free(mask);
}


/*
 *	Maps the file browser and blocks until user response.
 *
 *	Returns True if the user selected one or more paths or False if
 *	the user canceled.
 */
bool_t FBrowserGetResponse(
	fb_struct *fb,
	const char *title,
	const char *ok_btn_label,
	const char *cancel_btn_label,
	const char *path,            
	char ***paths_rtn, int *npaths_rtn,
	int map_rel, win_t toplevel,
	void (*main_push_func)(void *),
	void *main_push_data,
	void (*main_pop_func)(void *),
	void *main_pop_data           
)
{
	bool_t response;

	if(paths_rtn != NULL)
	    *paths_rtn = NULL;
	if(npaths_rtn != NULL)
	    *npaths_rtn = 0;

	if(fb == NULL)
	    return(False);

	/* Not initialized or already mapped? */
	if((fb->toplevel == None) || fb->map_state)
	    return(False);

	/* Do not query if blocking input */
	if(OSWBlockInputLevel() > 0)
	    return(False);

	/* Set values */
	FBrowserSetCallbacks(
	    fb,
	    FBrowserDefaultOKCB, NULL,
	    FBrowserDefaultCancelCB, NULL
	);
	OSWSetTransientFor(fb->toplevel, toplevel);
	fb->main_push_func = main_push_func;
	fb->main_push_data = main_push_data;
	fb->main_pop_func = main_pop_func;
	fb->main_pop_data = main_pop_data;
	fb->response_paths = NULL;
	fb->nresponse_paths = 0;
	FBrowserSetOpMesg(fb, title, ok_btn_label, cancel_btn_label);

	/* Set the file browser's initial position */
	switch(map_rel)
	{
	  case WIDGET_WINDOW_RELATIVE_PARENT:
	    WidgetCenterWindowToWindow(fb->toplevel, toplevel);
	    break;
	  case WIDGET_WINDOW_RELATIVE_ROOT:
	  case WIDGET_WINDOW_RELATIVE_POINTER:
	    WidgetCenterWindow(fb->toplevel, map_rel);
	    break;
	}

	/* Map the file browser */
	FBrowserMapPath(fb, path);

	/* Push main loop level */
	if(fb->main_push_func != NULL)
	    fb->main_push_func(fb->main_push_data);

	/* Get response? */
	response = (fb->response_paths != NULL) ? True : False;

	/* Set response paths */
	if((paths_rtn != NULL) && (npaths_rtn != NULL))
	{
	    *paths_rtn = fb->response_paths;
	    *npaths_rtn = fb->nresponse_paths;
	    fb->response_paths = NULL;
	    fb->nresponse_paths = 0;
	}
	else
	{
	    strlistfree(fb->response_paths, fb->nresponse_paths);
	    fb->response_paths = NULL;
	    fb->nresponse_paths = 0;
	}

	return(response);
}


/*
 *	Set operations messages on file browser fb.
 *
 *	title is the message that will appear on the file browser's
 *	title bar.
 *
 *	ok_btn_name is the label to put on the ok button.
 */
void FBrowserSetOpMesg(
	fb_struct *fb,
	const char *title,
	const char *ok_btn_label,
	const char *cancel_btn_label
)
{
	if(fb == NULL)
	    return;

	if(fb->toplevel == None)
	    return;

	if(title != NULL)
	    OSWSetWindowTitle(fb->toplevel, title);

	if(ok_btn_label != NULL)
	    PBtnSetLabel(
		fb->ok_btn,
		FB_BTN_WIDTH, FB_BTN_HEIGHT,
		ok_btn_label,
		PBTN_LABEL_ALIGN_CENTER,
		None, None
	    );
	if(cancel_btn_label != NULL)
	    PBtnSetLabel(
		fb->cancel_btn,
		FB_BTN_WIDTH, FB_BTN_HEIGHT,
		cancel_btn_label,
		PBTN_LABEL_ALIGN_CENTER,
		None, None
	    );
}

/*
 *	Sets the file browser's transient for.
 */
void FBrowserSetTransientFor(fb_struct *fb, win_t toplevel)
{
	if(fb == NULL)
	    return;

	if(fb->toplevel == None)
	    return;

	OSWSetTransientFor(fb->toplevel, toplevel);
}

/*
 *	Sets the file browser's callbacks.
 */
void FBrowserSetCallbacks(
	fb_struct *fb,
	void (*ok_cb)(fb_struct *, const char *, void *),
	void *ok_data,
	void (*cancel_cb)(fb_struct *, void *),
	void *cancel_data
)
{
	if(fb == NULL)
	    return;

	fb->ok_cb = ok_cb;
	fb->ok_data = ok_data;
	fb->cancel_cb = cancel_cb;
	fb->cancel_data = cancel_data;
}

/*
 *	Clears the file browser's callbacks.
 */
void FBrowserClearCallbacks(fb_struct *fb)
{
	if(fb == NULL)
	    return;

	fb->ok_cb = NULL;
	fb->ok_data = NULL;
	fb->cancel_cb = NULL;
	fb->cancel_data = NULL;
}


/*
 *	Initializes the file browser.
 */
int FBrowserInit(fb_struct *fb, int style)
{
	int x, y;
	unsigned int width, height;
	keyval_t accel_keys[4];
	key_state_t accel_mods[4];
	label_struct *label;
	push_button_struct *btn;
	prompt_struct *prompt;
	scroll_bar_struct *sb;
	osw_gui_struct *gui = &osw_gui[0];
	win_t w, parent = gui->root_win;
	widget_global_struct *wglobal = &widget_global;

	if(fb == NULL)
	    return(-2);

	fb->type = WIDGET_TYPE_FILE_BROWSER;
	fb->map_state = False;
	fb->has_focus = False;
	fb->disabled = False;
	fb->visibility_state = VisibilityFullyObscured;
	fb->x = 0;
	fb->y = 0;
	fb->width = FB_DEF_WIDTH;
	fb->height = FB_DEF_HEIGHT;
	fb->font = wglobal->std_font;
	fb->next = NULL;
	fb->prev = NULL;
	fb->data = NULL;

	fb->options = 0;
	fb->style = style;

	fb->ok_btn = NULL;
	fb->cancel_btn = NULL;

	fb->location_label = NULL;
	fb->location_pulist = NULL;
	fb->parent_btn = NULL;
	fb->home_btn = NULL;
	fb->refresh_btn = NULL;

	fb->dir_win = None;
	fb->dir_win_buf = None;
	fb->dir_win_sb = NULL;

	fb->file_win = None;
	fb->file_win_buf = None;
	fb->file_win_sb = NULL;

	fb->list_win = None;
	fb->list_win_buf = None;
	fb->list_win_sb = NULL;
	fb->list_max_width = 0;
	fb->list_rows_visible = 0;

	fb->prompt = NULL;

	fb->menu = NULL;

	fb->cv_prompt = NULL;
	fb->cv_prompt_target = NULL;
	fb->cv_prompt_mode = FB_CVP_MODE_NONE;

	fb->dir_list = NULL;
	fb->dir_list_items = 0;
	fb->file_list = NULL;
	fb->file_list_items = 0;
	fb->sel_dir = -1;
	fb->sel_file = -1;
	fb->sel_object = -1;

	fb->button_pressed = 0;
	fb->button_last_x = 0;
	fb->button_last_y = 0;
	fb->button_pressed_time = 0l;

	fb->ok_cb = NULL;
	fb->ok_data = NULL;
	fb->cancel_cb = NULL;
	fb->cancel_data = NULL;

	fb->main_push_func = NULL;
	fb->main_push_data = NULL;
	fb->main_pop_func = NULL;
	fb->main_pop_data = NULL;

	fb->response_paths = NULL;
	fb->nresponse_paths = 0;  


	/* Create toplevel */
	x = fb->x;
	y = fb->y;
	width = fb->width;
	height = fb->height;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	fb->toplevel = w;
	OSWSetWindowInput(w, OSW_EVENTMASK_TOPLEVEL);

	/* WM properties */
	OSWSetWindowWMProperties(
	    w,
	    "File Browser",		/* Title */
	    "File Browser",		/* Icon title */
	    wglobal->std_icon_pm,	/* Icon */
	    False,			/* Do not let WM set coordinates */
	    x, y,
	    width, height,
	    width, height,
	    OSW_WINDOW_FRAME_DIALOG,
	    0, NULL
	);
	parent = w;


	/* Create lists by style */
	if(fb->style == FB_STYLE_SINGLE_LIST)
	{
	    int n;
	    popup_list_struct *pulist;
	    menu_struct *menu;

	    /* Locations label and popup list */
	    fb->location_label = label = LABEL(malloc(sizeof(label_struct)));
	    x = FB_MARGIN;
	    y = FB_MARGIN;
	    width = 65;
	    height = 0;
	    LabelInit(
		label, parent,
		x, y, width, height,
		"Location:"
	    );
	    LabelSetJustify(label, LABEL_JUSTIFY_RIGHT);

	    fb->location_pulist = pulist = POPUP_LIST(malloc(sizeof(popup_list_struct)));
	    x += (int)width + 2;
	    width = MAX(
		(int)fb->width - x - (3 * 28) - (3 * 10) - (1 * 5),
		40
	    );
	    height = 0;
	    PUListInit(
		pulist, parent,
		x, y, width, height,
		FB_LOCATIONS_LIST_NITEMS_VISIBLE,
		PULIST_POPUP_BELOW
	    );
	    pulist->select_cb = FBrowserLocationPUListSelectCB;
	    pulist->select_data = fb;

	    /* Parent button */
	    fb->parent_btn = btn = PUSH_BUTTON(malloc(sizeof(push_button_struct)));
	    x = (int)fb->width - (3 * 28) - (2 * 10) - (1 * 5);
	    y = 10;
	    width = FB_ICON_BTN_WIDTH;
	    height = FB_ICON_BTN_HEIGHT;
	    PBtnInit(
		btn, parent,
		x, y, width, height,
		NULL, PBTN_LABEL_ALIGN_CENTER,
		wglobal->parent_dir_pixmap,
		wglobal->parent_dir_mask,
		FBrowserParentPBCB, fb
	    );
	    accel_keys[0] = OSW_KEY_p;
	    accel_mods[0] = OSW_KEY_STATE_CTRL;
	    PBtnSetAccelKeys(btn, accel_keys, accel_mods, 1);
	    PBtnSetTooltip(btn, "Parent");

	    /* Home Button */
	    fb->home_btn = btn = PUSH_BUTTON(malloc(sizeof(push_button_struct)));
	    x = (int)fb->width - (2 * 28) - (2 * 10) - (0 * 5);
	    PBtnInit(
		btn, parent,
		x, y, width, height,
		NULL, PBTN_LABEL_ALIGN_CENTER,
		wglobal->home_dir_pixmap,
		wglobal->home_dir_mask,
		FBrowserHomePBCB, fb
	    );
	    accel_keys[0] = OSW_KEY_h;
	    accel_mods[0] = OSW_KEY_STATE_CTRL;
	    PBtnSetAccelKeys(btn, accel_keys, accel_mods, 1);
	    PBtnSetTooltip(btn, "Home");

	    /* Refresh Button */
	    fb->refresh_btn = btn = PUSH_BUTTON(malloc(sizeof(push_button_struct)));
	    x = (int)fb->width - (1 * 28) - (1 * 10) - (0 * 5);
	    PBtnInit(
		btn, parent,
		x, y, width, height,
		NULL, PBTN_LABEL_ALIGN_CENTER,
		wglobal->refresh_pixmap,
		wglobal->refresh_mask,
		FBrowserRefreshPBCB, fb
	    );
	    accel_keys[0] = OSW_KEY_r;
	    accel_mods[0] = OSW_KEY_STATE_CTRL;
	    accel_keys[1] = OSW_KEY_F5;
	    accel_mods[1] = 0;
	    PBtnSetAccelKeys(btn, accel_keys, accel_mods, 2);
	    PBtnSetTooltip(btn, "Refresh");


	    /* List Window */
	    x = 10;
	    y = 45;
	    width = (unsigned int)MAX(
		((int)fb->width - (2 * 10)),
		0
	    );
	    height = MAX(
		((int)fb->height - 10 - FB_BTN_HEIGHT - 5 -
		FB_BTN_HEIGHT - 10 - 2 - 10 - y),
		0
	    );
	    OSWCreateWindow(&w, parent, x, y, width, height);
	    fb->list_win = w;
	    OSWSetWindowInput(
		w,
		ExposureMask |
		ButtonPressMask | ButtonReleaseMask |
		PointerMotionMask
	    );

	    /* List scroll Bars */
	    fb->list_win_sb = sb = SCROLL_BAR(malloc(sizeof(scroll_bar_struct)));
	    SBarInit(
		sb, w,
		width, height,
		0, 0
	    );
	    sb->value_changed_cb = FBrowserSBValueChangedCB;
	    sb->value_changed_data = fb;

	    /* Right-click menu */
	    fb->menu = menu = MENU(malloc(sizeof(menu_struct)));
	    MenuInit(
		menu, gui->root_win,
		FBrowserMenuActivateCB, fb
	    );
#define APPEND_MENU_ITEM(_t_,_s_,_id_)	{	\
 n = MenuItemAppend(				\
  menu, (_t_), (_s_), None, None, (_id_)	\
 );						\
}
#define SET_MENU_ITEM_ACCEL_KEY(_k_,_m_) {	\
 MenuSetItemAccelKey(				\
  menu, n, (_k_), (_m_)				\
 );						\
}
	    APPEND_MENU_ITEM(
		MENU_ITEM_LABEL,
		"Select",
		FB_OP_SELECT
	    );
	    APPEND_MENU_ITEM(
		MENU_ITEM_SEPARATOR,
		NULL,
		FB_OP_NONE
	    );
	    APPEND_MENU_ITEM(
		MENU_ITEM_LABEL,
		"Go To Parent",
		FB_OP_GOTO_PARENT
	    );
	    APPEND_MENU_ITEM(
		MENU_ITEM_LABEL,
		"Go To Home",
		FB_OP_GOTO_HOME
	    );
	    APPEND_MENU_ITEM(
		MENU_ITEM_SEPARATOR,
		NULL,
		FB_OP_SELECT
	    );
	    APPEND_MENU_ITEM(
		MENU_ITEM_LABEL,
		"Refresh",
		FB_OP_REFRESH
	    );
#undef APPEND_MENU_ITEM
#undef SET_MENU_ITEM_ACCEL_KEY
	}
	else
	{
	    /* Split window style */

	    /* Directory list window */
	    x = 10;
	    y = 25;
	    width = MAX(
		(int)(fb->width * FB_DIR_LIST_WIDTH_PERCENT / 100),
		2
	    );   
	    height = MAX(
		(int)fb->height - FB_PROMPT_HEIGHT - FB_BTN_HEIGHT - 75,
		2
	    );
	    OSWCreateWindow(&w, parent, x, y, width, height);
	    fb->dir_win = w;
	    OSWSetWindowInput(
		w,
		ExposureMask |
		ButtonPressMask | ButtonReleaseMask |
		PointerMotionMask
	    );
	    /* Directory list window scroll bars */
	    fb->dir_win_sb = sb = SCROLL_BAR(malloc(sizeof(scroll_bar_struct)));
	    SBarInit(
		sb, w,
		width, height,
		0, 0
	    );
	    sb->value_changed_cb = FBrowserSBValueChangedCB;
	    sb->value_changed_data = fb;

	    /* File List Window */
	    x = ((int)fb->width * FB_DIR_LIST_WIDTH_PERCENT / 100) + 20;
	    y = 25;
	    width = MAX(
		(int)((int)fb->width *
		    (100 - FB_DIR_LIST_WIDTH_PERCENT) / 100) - 30,
		2
	    );
	    height = MAX(
		(int)fb->height - FB_PROMPT_HEIGHT - FB_BTN_HEIGHT - 75,
		2
	    );
	    OSWCreateWindow(&w, parent, x, y, width, height);
	    fb->file_win = w;
	    OSWSetWindowInput(
		w,
		ExposureMask |
		ButtonPressMask | ButtonReleaseMask |
		PointerMotionMask
	    );
	    /* File list window scroll Bars */
	    fb->file_win_sb = sb = SCROLL_BAR(malloc(sizeof(scroll_bar_struct)));
	    SBarInit(
		sb, w,
		width, height,
		0, 0
	    );
	    sb->value_changed_cb = FBrowserSBValueChangedCB;
	    sb->value_changed_data = fb;
	}

	/* Location prompt */
	fb->prompt = prompt = PROMPT(malloc(sizeof(prompt_struct)));
	width = (unsigned int)MAX(
	    ((int)fb->width - (3 * 10) - FB_BTN_WIDTH),
	    0
	);
	height = FB_PROMPT_HEIGHT;
	x = 10;
	y = (int)fb->height - 10 - FB_BTN_HEIGHT - 5 - FB_BTN_HEIGHT;
	PromptInit(
	    prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Location:",
	    PATH_MAX + NAME_MAX, 0
	);
	prompt->options |= PROMPT_FLAG_KEEP_VALUE;
	prompt->activate_cb = FBrowserPromptActivateCB;
	prompt->activate_data = fb;

	/* OK button */
	fb->ok_btn = btn = PUSH_BUTTON(malloc(sizeof(push_button_struct)));
	width = FB_BTN_WIDTH;
	height = FB_BTN_HEIGHT;
	x = (int)fb->width - 10 - (int)width;
	y = (int)fb->height - 10 - (int)height - 5 - (int)height;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "OK",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    FBrowserOKPBCB, fb
	);
	accel_keys[0] = OSW_KEY_o;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	accel_keys[1] = OSW_KEY_s;
	accel_mods[1] = OSW_KEY_STATE_CTRL; 
	accel_keys[2] = OSW_KEY_Return;
	accel_mods[2] = 0;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 3);

	/* Cancel button */
	fb->cancel_btn = btn = PUSH_BUTTON(malloc(sizeof(push_button_struct)));;
	y += 5 + (int)height;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Cancel",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    FBrowserCancelPBCB, fb
	);
	accel_keys[0] = OSW_KEY_c;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	accel_keys[1] = OSW_KEY_w;
	accel_mods[1] = OSW_KEY_STATE_CTRL;
	accel_keys[2] = OSW_KEY_Escape;
	accel_mods[2] = 0;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 3);


	/* Add this widget to the regeristry */
	WidgetRegister(fb, fb->type);

	return(0);
}

/*
 *	Resizes file browser widget.
 */
void FBrowserResize(fb_struct *fb)
{
	int x, y;
	unsigned int width, height;
	win_t w;
	label_struct *label;
	scroll_bar_struct *sb;
	push_button_struct *btn;
	prompt_struct *prompt;
	popup_list_struct *pulist;

	if(fb == NULL)
	    return;

	w = fb->toplevel;
	if(w == None)
	    return;

	OSWGetWindowGeometry(w, &x, &y, &width, &height, NULL);
	fb->x = x;
	fb->y = y;
	fb->width = width;
	fb->height = height;

	if(fb->style == FB_STYLE_SINGLE_LIST)
	{
	    /* Locations label & popup list */
	    label = fb->location_label;
	    pulist = fb->location_pulist;
	    x = 10;
	    y = 10 + (((int)pulist->height - (int)label->height) / 2);
	    width = 65;
	    height = label->height;
	    OSWMoveResizeWindow(label->toplevel, x, y, width, height);
	    WidgetQueueResize(label);

	    x += (int)width + 2;
	    y = 10;
	    width = MAX(
		(int)fb->width - x - (3 * 28) - (3 * 10) - (1 * 5),
		40
	    );
	    height = pulist->height;
	    OSWMoveResizeWindow(pulist->toplevel, x, y, width, height);
	    PUListQueueResize(
		pulist,
		FB_LOCATIONS_LIST_NITEMS_VISIBLE
	    );

	    /* Parent button */
	    btn = fb->parent_btn;
	    x = (int)fb->width - (3 * 28) - (2 * 10) - (1 * 5);
	    y = 10;
	    OSWMoveWindow(btn->toplevel, x, y);

	    /* Home button */
	    btn = fb->home_btn;
	    x = (int)fb->width - (2 * 28) - (2 * 10) - (0 * 5);
	    OSWMoveWindow(btn->toplevel, x, y);

	    /* Refresh button */
	    btn = fb->refresh_btn;
	    x = (int)fb->width - (1 * 28) - (1 * 10) - (0 * 5);
	    OSWMoveWindow(btn->toplevel, x, y);

	    /* List window */
	    w = fb->list_win;
	    x = 10;
	    y = 45;
	    width = (unsigned int)MAX(
		((int)fb->width - (2 * 10)),
		0
	    );
	    height = MAX(  
		((int)fb->height - 10 - FB_BTN_HEIGHT - 5 -
		FB_BTN_HEIGHT - 10 - 2 - 10 - y),
		0
	    );
	    OSWMoveResizeWindow(w, x, y, width, height);
	    OSWDestroyPixmap(&fb->list_win_buf);

	    /* Reposition objects */
	    FBrowserRepositionListObjects(fb);

	    /* List scroll bar */
	    sb = fb->list_win_sb;
	    SBarResize(
		sb,
		width, height,
		fb->list_max_width, height
	    );
	    SBarSetPositionY(sb, 0.0f);
	}
	else	/* FB_STYLE_SPLIT_LIST */
	{
	    int x, y;
	    unsigned int width, height;

	    /* Directory list window */
	    w = fb->dir_win;
	    x = 10;
	    y = 25;
	    width = MAX(
		(int)(fb->width * FB_DIR_LIST_WIDTH_PERCENT / 100),
		2
	    );
	    height = MAX(
		(int)fb->height - FB_PROMPT_HEIGHT - FB_BTN_HEIGHT - 75,
		2
	    );
	    OSWMoveResizeWindow(w, x, y, width, height);
	    OSWDestroyPixmap(&fb->dir_win_buf);

	    sb = fb->dir_win_sb;
	    SBarResize(
		sb,
		width, height,
		width,
		(FB_LIST_ROW_HEIGHT * fb->dir_list_items) +
			FB_LIST_ROW_HEIGHT
	    );
	    SBarSetPositionX(sb, 0.0f);
	    SBarSetPositionY(sb, 0.0f);

	    /* File list window */
	    w = fb->file_win;
	    x = ((int)fb->width * FB_DIR_LIST_WIDTH_PERCENT / 100) + 20;
	    y = 25;
	    width = MAX(
		(int)((int)fb->width *
		    (100 - FB_DIR_LIST_WIDTH_PERCENT) / 100) - 30,
		2
	    );
	    height = MAX(
		(int)fb->height - FB_PROMPT_HEIGHT - FB_BTN_HEIGHT - 75,
		2
	    );
	    OSWMoveResizeWindow(w, x, y, width, height);
	    OSWDestroyPixmap(&fb->file_win_buf);

	    sb = fb->file_win_sb;
	    SBarResize(
		sb,
		width, height,
		width,
		(FB_LIST_ROW_HEIGHT * fb->file_list_items) +
		    FB_LIST_ROW_HEIGHT
	    );
	    SBarSetPositionX(sb, 0.0f);
	    SBarSetPositionY(sb, 0.0f);
	}

	/* Location prompt */
	prompt = fb->prompt;
	width = (unsigned int)MAX(
	    ((int)fb->width - (3 * 10) - FB_BTN_WIDTH),
	    0
	);   
	height = FB_PROMPT_HEIGHT;
	x = 10;
	y = (int)fb->height - 10 - FB_BTN_HEIGHT - 5 - FB_BTN_HEIGHT;
	OSWMoveResizeWindow(
	    prompt->toplevel, x, y, width, height
	);
	WidgetQueueResize(prompt);

	/* OK button */
	btn = fb->ok_btn;
	x = (int)fb->width - 10 - (int)btn->width;
	y = (int)fb->height - 10 - (int)btn->height - 5 -
	    (int)btn->height;
	OSWMoveWindow(btn->toplevel, x, y);

	/* Cancel button */
	btn = fb->cancel_btn;
	y += 5 + (int)btn->height;
	OSWMoveWindow(btn->toplevel, x, y);
}

/*
 *	Queues the file browser to resize.
 */
void FBrowserQueueResize(fb_struct *fb)
{
	if(fb == NULL)
	    return;

	if(fb->toplevel == None)
	    return;

	WidgetQueueResizeAction(FBrowserResizeCB, fb, NULL);
}

/*
 *	Draws the file browser.
 */
void FBrowserDraw(fb_struct *fb, int amount)
{
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if(fb == NULL)
	    return;

	if(fb->toplevel == None)
	    return;

	/* Toplevel */
	if(amount == FB_DRAW_COMPLETE)
	{
	    int x, y;
	    unsigned int width, height;
	    font_t *font = fb->font;
	    win_t w = fb->toplevel;

	    OSWGetWindowGeometry(
		w, NULL, NULL, &width, &height, NULL
	    );

	    /* Draw horizontal separator and labels */
	    if(wglobal->force_mono)
	    {
		OSWSetForeground(gui->black_pix);
		OSWDrawSolidRectangle(w, 0, 0, width, height);

		/* Horizontal separator */
		y = (int)fb->height - 10 - FB_BTN_HEIGHT - 5 -
		    FB_BTN_HEIGHT - 10 - 1;
		OSWSetForeground(gui->white_pix);
		OSWDrawLine(
		    w,
		    0, y,
		    (int)fb->width, y
		);
	    }
	    else
	    {
		WidgetPutImageTile(
		    w, wglobal->std_bkg_img,
		    width, height
		);

		/* Horizontal separator */
		y = (int)fb->height - 10 - FB_BTN_HEIGHT - 5 -
		    FB_BTN_HEIGHT - 10 - 2;
		OSWSetForeground(wglobal->surface_shadow_pix);
		OSWDrawLine(
		    w,
		    0, y, (int)fb->width, y
		);  
		y += 1;
		OSWSetForeground(wglobal->surface_highlight_pix);
		OSWDrawLine(
		    w,
		    0, y, (int)fb->width, y
		);
 	    }

	    /* Single List? */
	    if(fb->style == FB_STYLE_SINGLE_LIST)
	    {

	    }
	    /* Split List */
	    else
	    {
		const char *s;

		OSWSetFont(font);
		OSWSetForeground(
		    wglobal->force_mono ?
			gui->white_pix : wglobal->normal_text_pix
		);

		/* Directories list label */
		s = "Directories";
		x = 14;
		y = 18;
		OSWDrawString(
		    w, x, y, s
		);

		/* Files list label */
		s = "Files";
		x = ((int)fb->width *
		    FB_DIR_LIST_WIDTH_PERCENT / 100) + 23;
		y = 18;
		OSWDrawString(w, x, y, s);
	    }
	}

	/* Draw the Locations Popup List */
	if(amount == FB_DRAW_COMPLETE)
	{
	    popup_list_struct *pulist = fb->location_pulist;
	    PUListDraw(pulist, PULIST_DRAW_AMOUNT_COMPLETE);
	}

	/* Draw the List (FB_STYLE_SINGLE_LIST only) */
	if(((amount == FB_DRAW_COMPLETE) ||
	    (amount == FB_DRAW_DIRLIST) ||
	    (amount == FB_DRAW_FILELIST) ||
	    (amount == FB_DRAW_ALLLISTS)) &&
	   (fb->style == FB_STYLE_SINGLE_LIST)
	)
	{
	    int	i, x, y,
		lbearing, rbearing, ascent, descent, font_height,
		icon_label_spacing;
	    unsigned int width, height, icon_width, icon_height;
	    font_t *font = fb->font;
	    win_t w = fb->list_win;
	    pixmap_t pixmap, icon_pixmap;
	    bitmap_t icon_mask;
	    scroll_bar_struct *sb = fb->list_win_sb;
	    fb_object_struct *obj;

	    OSWGetWindowGeometry(
		w, NULL, NULL, &width, &height, NULL
	    );

	    if(fb->list_win_buf == None)
	    {
		if(OSWCreatePixmap(
		    &fb->list_win_buf, width, height
		))
		    return;
	    }
	    pixmap = fb->list_win_buf;

	    /* Get size of icon */
	    icon_pixmap = wglobal->icon_folder_closed_20x20_pixmap;
	    if(icon_pixmap != None)
	    {
		OSWGetPixmapGeometry(
		    icon_pixmap, &icon_width, &icon_height, NULL
		);
		icon_label_spacing = 2;
	    }
	    else
	    {
		icon_width = 0;
		icon_height = 0;
		icon_label_spacing = 0;
	    }

	    /* Clear buffer */
	    OSWSetForeground(wglobal->surface_editable_pix);
	    OSWDrawSolidRectangle(
		pixmap, 0, 0, width, height
	    );

	    OSWSetFont(font);

	    /* Get starting dir item number */
	    for(i = 0; i < fb->dir_list_items; i++)
	    {
		obj = fb->dir_list[i];
	        if(obj == NULL)
		    continue;

		if((obj->x + (int)obj->width +
		    FB_LIST_BORDER) >= (int)SBarGetPositionX(sb)
		)
		    break;
	    }
	    if(i < fb->dir_list_items)
	    {
	        for(; i < fb->dir_list_items; i++)
		{
		    obj = fb->dir_list[i];
		    if(obj == NULL)
			continue;

		    if(obj->name == NULL)
			continue;

		    if((obj->x + FB_LIST_BORDER -
			(int)SBarGetPositionX(sb)) > (int)width
		    )
			break;

		    x = FB_LIST_BORDER + obj->x -
			(int)SBarGetPositionX(sb);
		    y = FB_LIST_BORDER + obj->y;

		    /* Is this directory selected? */
		    if(fb->sel_object == i)
		    {
			int	x2 = x,
				y2 = y;

			/* Draw selection background */
			if(wglobal->force_mono)
			    OSWSetForeground(gui->white_pix);
			else
			    OSWSetForeground(wglobal->surface_selected_pix);
			OSWDrawSolidRectangle(
			    pixmap, x2, y2, obj->width, obj->height
			);

			if(wglobal->force_mono)
			    OSWSetForeground(gui->black_pix);              
			else
		            OSWSetForeground(wglobal->selected_text_pix);

#ifdef S_ISLNK
			if(S_ISLNK(obj->mode))
#else
			if(False)
#endif
			{
			    icon_pixmap = wglobal->icon_link_20x20_pixmap;
			    icon_mask = wglobal->icon_link_20x20_mask;
			}
			else
			{
			    icon_pixmap = wglobal->icon_folder_closed_20x20_pixmap;
			    icon_mask = wglobal->icon_folder_closed_20x20_mask;
			}
		    }
		    else
		    {
			if(wglobal->force_mono)
			    OSWSetForeground(gui->white_pix);
			else
			    OSWSetForeground(wglobal->editable_text_pix);

#ifdef S_ISLNK
			if(S_ISLNK(obj->mode))
#else
			if(False)
#endif
			{
			    icon_pixmap = wglobal->icon_link_20x20_pixmap;
			    icon_mask = wglobal->icon_link_20x20_mask;
			}
			else
			{
			    icon_pixmap = wglobal->icon_folder_closed_20x20_pixmap;
			    icon_mask = wglobal->icon_folder_closed_20x20_mask;
			}
		    }
		    /* Draw the pixmap */
		    if(icon_pixmap != None)
		    {
			int	x2 = x,
				y2 = y + (((int)obj->height - icon_height) / 2);
			OSWSetClipMask(icon_mask);
			OSWSetClipOrigin(x2, y2);
			OSWCopyDrawablesCoord(
			    pixmap, icon_pixmap,
			    x2, y2,
			    icon_width, icon_height,
			    0, 0
			);
			OSWSetClipMask(None);
		    }
		    /* Draw the label */
		    if(obj->name != NULL)
		    {
			int x2, y2;
			const char *text = obj->name;
			OSWGetTextBounds(
			    font, text, -1,
			    &lbearing, &rbearing, NULL,
			    &ascent, &descent
			);
			font_height = ascent + descent;
			x2 = x + (int)icon_width + icon_label_spacing;
			y2 = y + (((int)obj->height - (int)font_height) / 2);
			OSWDrawString(
			    pixmap,
			    x2 - lbearing,
			    y2 + ascent,
			    text
			);
		    }
		}
	    }

	    /* Get starting file item number */
	    for(i = 0; i < fb->file_list_items; i++)
	    {
		obj = fb->file_list[i];
		if(obj == NULL)
		    continue;

		if((obj->x + (int)obj->width +
		    FB_LIST_BORDER) >= (int)SBarGetPositionX(sb)
		)
		    break;
	    }
	    if(i < fb->file_list_items)  
	    {
		for(; i < fb->file_list_items; i++)
		{
		    obj = fb->file_list[i];
		    if(obj == NULL)
			continue;

		    if(obj->name == NULL)
			continue;

		    if((obj->x + FB_LIST_BORDER - 
			(int)SBarGetPositionX(sb)) > (int)width
		    )
			break;

		    x = FB_LIST_BORDER + obj->x -
			(int)SBarGetPositionX(sb);
		    y = FB_LIST_BORDER + obj->y;

		    /* Is this file selected? */
		    if(fb->sel_object == (i + fb->dir_list_items))
		    {
			int	x2 = x,
				y2 = y;

			/* Draw selection background */
			if(wglobal->force_mono) 
			    OSWSetForeground(gui->white_pix);
			else
			    OSWSetForeground(wglobal->surface_selected_pix);

			OSWDrawSolidRectangle(
			    pixmap, x2, y2, obj->width, obj->height
			);

			if(wglobal->force_mono)
			    OSWSetForeground(gui->black_pix);
			else
			    OSWSetForeground(wglobal->selected_text_pix);

#ifdef S_ISLNK
			if(S_ISLNK(obj->mode))
#else
			if(False)
#endif
			{
			    icon_pixmap = wglobal->icon_link_20x20_pixmap;
			    icon_mask = wglobal->icon_link_20x20_mask;
			}
#if defined(S_IXUSR) && defined(S_IXGRP) && defined(S_IXOTH)
			else if((S_IXUSR & obj->mode) ||
				(S_IXGRP & obj->mode) ||
				(S_IXOTH & obj->mode)
			)
#else
			else if(False)
#endif
			{
			    icon_pixmap = wglobal->icon_executable_20x20_pixmap;
			    icon_mask = wglobal->icon_executable_20x20_mask;
			}
			else
			{
			    icon_pixmap = wglobal->icon_file_20x20_pixmap;
			    icon_mask = wglobal->icon_file_20x20_mask;    
			}
		    }
		    else
		    {
			if(wglobal->force_mono)
			    OSWSetForeground(gui->white_pix);
			else
			    OSWSetForeground(wglobal->editable_text_pix);

#ifdef S_ISLNK
			if(S_ISLNK(obj->mode))
#else
			if(False)
#endif
			{
			    icon_pixmap = wglobal->icon_link_20x20_pixmap;
			    icon_mask = wglobal->icon_link_20x20_mask;
			}
#ifdef S_ISFIFO
			else if(S_ISFIFO(obj->mode))
#else
			else if(False)
#endif
			{
			    icon_pixmap = wglobal->icon_pipe_20x20_pixmap;
			    icon_mask = wglobal->icon_pipe_20x20_mask;
			}
#ifdef S_ISBLK
			else if(S_ISBLK(obj->mode))
#else
			else if(False)
#endif
			{
			    icon_pixmap = wglobal->icon_device_block_20x20_pixmap;
			    icon_mask = wglobal->icon_device_block_20x20_mask;
			}
#ifdef S_ISCHR
			else if(S_ISCHR(obj->mode))
#else
			else if(False)
#endif
			{
			    icon_pixmap = wglobal->icon_device_character_20x20_pixmap;
			    icon_mask = wglobal->icon_device_character_20x20_mask;
			}
#if defined(S_IXUSR) && defined(S_IXGRP) && defined(S_IXOTH)
			else if((S_IXUSR & obj->mode) ||
				(S_IXGRP & obj->mode) ||
				(S_IXOTH & obj->mode)
			)
#else
			else if(False)
#endif
			{
			    icon_pixmap = wglobal->icon_executable_20x20_pixmap;
			    icon_mask = wglobal->icon_executable_20x20_mask;
			}
			else
			{
			    icon_pixmap = wglobal->icon_file_20x20_pixmap;
			    icon_mask = wglobal->icon_file_20x20_mask;
			}
 		    }
		    /* Draw the pixmap */
		    if(icon_pixmap != None)
		    {
			int	x2 = x,
				y2 = y + (((int)obj->height - icon_height) / 2);
			OSWSetClipMask(icon_mask);
			OSWSetClipOrigin(x2, y2);
			OSWCopyDrawablesCoord(
			    pixmap, icon_pixmap,
			    x2, y2,
			    icon_width, icon_height,
			    0, 0
			);
			OSWSetClipMask(None);
		    }
		    /* Draw the label */
		    if(obj->name != NULL)
		    {
			int x2, y2;
			const char *text = obj->name;
			OSWGetTextBounds(
			    font, text, -1,
			    &lbearing, &rbearing,
			    NULL,
			    &ascent, &descent
			);
			font_height = ascent + descent;
			x2 = x + (int)icon_width + icon_label_spacing;
			y2 = y + (((int)obj->height - (int)font_height) / 2);
			OSWDrawString(
			    pixmap,
			    x2 - lbearing,
			    y2 + ascent,
			    text
			);
		    }
		}
	    }

	    WidgetFrameButtonPixmap(
		pixmap, True, width, height,
		wglobal->surface_highlight_pix,
		wglobal->surface_shadow_pix
	    );

	    OSWPutDrawableToWindow(w, pixmap);
	}


	/* Draw directory list */
	if(((amount == FB_DRAW_COMPLETE) ||
	    (amount == FB_DRAW_DIRLIST) ||   
	    (amount == FB_DRAW_ALLLISTS)) &&
	   (fb->style == FB_STYLE_SPLIT_LIST)
	)
	{
	    int lbearing, rbearing, ascent, descent, font_height,
		rows_visible;
	    unsigned int width, height, icon_width, icon_height;
	    font_t *font = fb->font;
	    win_t w = fb->dir_win;
	    pixmap_t pixmap, icon_pixmap;
	    bitmap_t icon_mask;
	    fb_object_struct *obj;

	    OSWGetWindowGeometry(
		w, NULL, NULL, &width, &height, NULL
	    );

	    /* Recreate buffers */ 
	    if(fb->dir_win_buf == None)
	    {
		if(OSWCreatePixmap(
		    &fb->dir_win_buf, width, height
		))
		    return;
	    }

	    pixmap = fb->dir_win_buf;

	    /* Clear buffer */
	    OSWSetForeground(wglobal->surface_editable_pix);
	    OSWDrawSolidRectangle(
		pixmap, 0, 0, width, height
	    );

	    if((fb->dir_list != NULL) && (fb->dir_list_items > 0))
	    {
		int	i, y, x_level,
			dlist_y_pos = MAX(FB_LIST_ROW_HEIGHT - 4, 0);

	        /* Calculate starting index */
	        i = (int)SBarGetPositionY(fb->dir_win_sb) / FB_LIST_ROW_HEIGHT;
	        if(i < 0)
		    i = 0;

	        /* Calculate total possible rows z visable on window */
	        rows_visible = ((int)height / FB_LIST_ROW_HEIGHT) + 1;

		OSWSetFont(font);

	        y = 0;
	        x_level = 0; /* Always 0 for now, since no dir tree support */
	        while(i < fb->dir_list_items)
	        {
		    if(y >= rows_visible)
		        break;

		    obj = fb->dir_list[i];
		    if(obj == NULL)
		    {
		        i++;
			continue;
		    }

		    if(obj->name == NULL)
		    {
			i++;
			continue;
		    }

		    if(fb->sel_dir == i)
		    {
		        /* Selected directory entry */
			OSWSetForeground(wglobal->surface_selected_pix);
			OSWDrawSolidRectangle(
			    pixmap,
			    (x_level * FB_LIST_COLUM_LEVEL_INC) +
				FB_LIST_COLUM_LEVEL_INC +
				FB_LIST_BORDER + 6,
			    dlist_y_pos - 12,
			    MAX(
			        (((int)strlen(obj->name) * 7) + 3),
			        7
			    ),
			    MAX(FB_LIST_ROW_HEIGHT - 2, 1)
			);

#ifdef S_ISLNK
		        if(S_ISLNK(obj->mode))
#else
			if(False)
#endif
			{
			    icon_pixmap = wglobal->icon_link_20x20_pixmap;
			    icon_mask = wglobal->icon_link_20x20_mask;
			}
		        else
			{
			    icon_pixmap = wglobal->icon_folder_closed_20x20_pixmap;
			    icon_mask = wglobal->icon_folder_closed_20x20_mask;
			}
			OSWSetForeground(wglobal->selected_text_pix);
		    }
		    else
		    {
#ifdef S_ISLNK
			if(S_ISLNK(obj->mode))
#else
			if(False)
#endif
			{
			    icon_pixmap = wglobal->icon_link_20x20_pixmap;
			    icon_mask = wglobal->icon_link_20x20_mask;
			}
			else 
			{
			    icon_pixmap = wglobal->icon_folder_closed_20x20_pixmap;
			    icon_mask = wglobal->icon_folder_closed_20x20_mask;
			}

		        /* Normal directory entry */
			OSWSetForeground(wglobal->editable_text_pix);
		    }
		    /* Draw the pixmap */
		    if(icon_pixmap != None)
		    {
			int	x2 = (x_level * FB_LIST_COLUM_LEVEL_INC) +
			    FB_LIST_BORDER,
				y2 = dlist_y_pos - 13;
			OSWGetPixmapGeometry(
			    icon_pixmap, &icon_width, &icon_height, NULL
			);
			OSWSetClipMask(icon_mask);
			OSWSetClipOrigin(x2, y2);
			OSWCopyDrawablesCoord( 
			    pixmap, icon_pixmap,
			    x2, y2,
			    icon_width, icon_height,
			    0, 0
			);
			OSWSetClipMask(None);
		    }

		    OSWGetTextBounds(
			font, obj->name, -1,
			&lbearing, &rbearing, NULL,
			&ascent, &descent
		    );
		    font_height = ascent + descent;
		    OSWDrawString(
			pixmap,
			(x_level * FB_LIST_COLUM_LEVEL_INC) +
			    FB_LIST_COLUM_LEVEL_INC +
			    FB_LIST_BORDER -
			    lbearing,
		        dlist_y_pos + ascent,
			obj->name
		    );

		    dlist_y_pos += FB_LIST_ROW_HEIGHT;
		    i++;
		    y++;
	        }
	    }
	    WidgetFrameButtonPixmap(
		pixmap, True, width, height,
	        wglobal->surface_highlight_pix,
	        wglobal->surface_shadow_pix
	    );
	    OSWPutDrawableToWindow(w, pixmap);
	}

	/* Draw file list */
	if(((amount == FB_DRAW_COMPLETE) ||
	    (amount == FB_DRAW_FILELIST) ||
	    (amount == FB_DRAW_ALLLISTS)) &&
	   (fb->style == FB_STYLE_SPLIT_LIST)
	)
	{
	    int i, y,
		lbearing, rbearing, ascent, descent, font_height,
		rows_visible,
		flist_x_filesize_pos;
	    unsigned int width, height, icon_width, icon_height;
	    font_t *font = fb->font;
	    win_t w = fb->file_win;
	    pixmap_t pixmap, icon_pixmap;
	    bitmap_t icon_mask;
	    fb_object_struct *obj;

	    OSWGetWindowGeometry(
		w, NULL, NULL, &width, &height, NULL
	    );

 	    /* Recreate buffers */
	    if(fb->file_win_buf == None)
	    {
		if(OSWCreatePixmap(
		    &fb->file_win_buf, width, height
		))
		    return;
	    }
	    pixmap = fb->file_win_buf;
 
	    /* Calculate flist_x_filesize_pos */
	    flist_x_filesize_pos = 0;
	    for(i = 0; i < fb->file_list_items; i++)
	    {
		obj = fb->file_list[i];
	        if(obj == NULL)
		    continue;

	        if(obj->name == NULL)
		    continue;

	        y = (strlen(obj->name) * 8) + 50;
	        if(y > flist_x_filesize_pos)
		    flist_x_filesize_pos = y;
	    }
 
	    /* Clear buffer */
	    OSWSetForeground(wglobal->surface_editable_pix);
	    OSWDrawSolidRectangle(
		pixmap, 0, 0, width, height
	    );

	    if((fb->file_list != NULL) && (fb->file_list_items > 0))
	    {
		int flist_y_pos = MAX(FB_LIST_ROW_HEIGHT - 4, 0);
		char text[256];

	        /* Calculate starting row index */
		i = (int)SBarGetPositionY(fb->file_win_sb) / FB_LIST_ROW_HEIGHT;
		if(i < 0)
		    i = 0;  

		/* Calculate total possible rows y visable on window */
		rows_visible = ((int)height / FB_LIST_ROW_HEIGHT) + 1;

		OSWSetFont(font);

	        y = 0;
		while(i < fb->file_list_items)
		{
		    if(y >= rows_visible)
			break;

		    obj = fb->file_list[i];
		    if(obj == NULL)
		    {
			i++;
			continue;
	            }
		    if(obj->name == NULL)
		    {
			i++;
			continue;
		    }

		    if(fb->sel_file == i)
		    {
		        /* Selected file entry */
		        OSWSetForeground(wglobal->surface_selected_pix);

		        OSWDrawSolidRectangle(
			    pixmap,
			    FB_LIST_COLUM_LEVEL_INC +
				FB_LIST_BORDER + 6,
			    flist_y_pos - 12,
			    MAX(
				(((int)strlen(obj->name) * 7) + 3),
				7
			    ),
			    MAX(FB_LIST_ROW_HEIGHT - 2, 1)
		        );

			/* Get the icon based on the object type */
#ifdef S_ISLNK
			if(S_ISLNK(obj->mode))
#else
			if(False)
#endif
			{
			    icon_pixmap = wglobal->icon_link_20x20_pixmap;
			    icon_mask = wglobal->icon_link_20x20_mask;
			}
#if defined(S_IXUSR) && defined(S_IXGRP) && defined(S_IXOTH)
			else if((S_IXUSR & obj->mode) ||
				(S_IXGRP & obj->mode) ||
				(S_IXOTH & obj->mode)
			)
#else
			else if(False)
#endif
			{
			    icon_pixmap = wglobal->icon_executable_20x20_pixmap;
			    icon_mask = wglobal->icon_executable_20x20_mask;
			}
			else
			{
			    icon_pixmap = wglobal->icon_file_20x20_pixmap;
			    icon_mask = wglobal->icon_file_20x20_mask;
			}

		        OSWSetForeground(wglobal->selected_text_pix);
		    }
		    else
		    {
			/* Get the icon based on the object type */
#ifdef S_ISLNK
			if(S_ISLNK(obj->mode))
#else
			if(False)
#endif
			{
			    icon_pixmap = wglobal->icon_link_20x20_pixmap;
			    icon_mask = wglobal->icon_link_20x20_mask;
			}
#if defined(S_IXUSR) && defined(S_IXGRP) && defined(S_IXOTH)
			else if((S_IXUSR & obj->mode) ||
				(S_IXGRP & obj->mode) ||
				(S_IXOTH & obj->mode)
			)
#else
			else if(False)
#endif
			{
			    icon_pixmap = wglobal->icon_executable_20x20_pixmap;
			    icon_mask = wglobal->icon_executable_20x20_mask;
			}
			else
			{
			    icon_pixmap = wglobal->icon_file_20x20_pixmap;
			    icon_mask = wglobal->icon_file_20x20_mask;
			}

		        /* Normal file entry */
		        OSWSetForeground(wglobal->editable_text_pix);
		    }
		    /* Draw the pixmap */                          
		    if(icon_pixmap != None)
		    {
			int     x2 = FB_LIST_BORDER,
				y2 = flist_y_pos - 13;
			OSWGetPixmapGeometry(
			    icon_pixmap, &icon_width, &icon_height, NULL
			);
			OSWSetClipMask(icon_mask);
			OSWSetClipOrigin(x2, y2);
			OSWCopyDrawablesCoord(
			    pixmap, icon_pixmap,
			    x2, y2,
			    icon_width, icon_height,
			    0, 0
			);
			OSWSetClipMask(None);
		    }

		    /* Draw name */
		    OSWDrawString(
			pixmap,
			FB_LIST_COLUM_LEVEL_INC + FB_LIST_BORDER + 6,
			flist_y_pos + 4,
			obj->name
		    );

		    /* Draw filesize and permissions */
		    sprintf(text, "%ld", (long)obj->size);
		    StringTailSpaces(text, 14);
#if defined(S_IXUSR) && defined(S_IXGRP) && defined(S_IXOTH)
		    text[14] = (obj->mode & S_IRUSR) ? 'r' : '-';
		    text[15] = (obj->mode & S_IWUSR) ? 'w' : '-';
		    text[16] = (obj->mode & S_IXUSR) ? 'x' : '-';
		    text[17] = (obj->mode & S_IRGRP) ? 'r' : '-';
		    text[18] = (obj->mode & S_IWGRP) ? 'w' : '-';
		    text[19] = (obj->mode & S_IXGRP) ? 'x' : '-';
		    text[20] = (obj->mode & S_IROTH) ? 'r' : '-';
		    text[21] = (obj->mode & S_IWOTH) ? 'w' : '-';
		    text[22] = (obj->mode & S_IXOTH) ? 'x' : '-';
#endif
		    text[23] = '\0';

		    OSWSetForeground(wglobal->editable_text_pix);
		    OSWGetTextBounds(
			font, text, -1,
			&lbearing, &rbearing, NULL,
			&ascent, &descent
		    );
		    font_height = ascent + descent;
		    OSWDrawString(
			pixmap,
		        flist_x_filesize_pos - lbearing,
		        flist_y_pos + ascent,
		        text
		    );

		    flist_y_pos += FB_LIST_ROW_HEIGHT;
		    i++;
	            y++;
		}
	    }

	    WidgetFrameButtonPixmap(
		pixmap, True, width, height,
		wglobal->surface_highlight_pix,
		wglobal->surface_shadow_pix
	    );
	    OSWPutDrawableToWindow(w, pixmap);
	}

	/* Draw scroll bars */
	if((amount == FB_DRAW_COMPLETE) ||
	   (amount == FB_DRAW_SCROLLBARS)
	)
	{
	    if(fb->style == FB_STYLE_SINGLE_LIST)
	    {
		SBarDraw(fb->list_win_sb);
	    }
	    else
	    {
		SBarDraw(fb->dir_win_sb);
		SBarDraw(fb->file_win_sb);
	    }
	}
}

/*
 *	Queues the file browser to draw.
 */
void FBrowserQueueDraw(fb_struct *fb, int amount)
{
	if(fb == NULL)
	    return;

	if(fb->toplevel == None)
	    return;

	WidgetQueueDrawAction(FBrowserDrawCB, fb, (void *)amount);
}

/*
 *	Manages the file browser event.
 */
int FBrowserManage(fb_struct *fb, event_t *event)
{
	int events_handled = 0;
	time_t t;
	keycode_t keycode;
	key_state_t key_state;
	keyval_t keyval;
	win_t w;
	event_key_t *key;
	event_button_t *button;
	event_motion_t *motion;
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if((event == NULL) || (fb == NULL))
	    return(events_handled);

	if(fb->toplevel == None)
	    return(events_handled);

	w = event->xany.window;

	/* Change value prompt */
	if(fb->map_state && fb->has_focus &&
	   (fb->cv_prompt_mode != FB_CVP_MODE_NONE)
	)
	{
	    events_handled += WidgetManage(fb->cv_prompt, event);
	    if(events_handled > 0)
		return(events_handled);
	}

	switch(event->type)	
	{
	  case ConfigureNotify:
	    if(w == fb->toplevel)
	    {
		FBrowserQueueResize(fb);
		if(fb->map_state)
		    FBrowserQueueDraw(fb, FB_DRAW_COMPLETE);
		events_handled++;
	    }
	    break;

	  case Expose:
	    if(w == fb->toplevel)
	    {
		FBrowserQueueDraw(fb, FB_DRAW_COMPLETE);
		events_handled++;
	    }
	    else if(w == fb->dir_win)
	    {
		FBrowserQueueDraw(fb, FB_DRAW_DIRLIST);
		events_handled++;
	    }
	    else if(w == fb->file_win)
	    {
		FBrowserQueueDraw(fb, FB_DRAW_FILELIST);
		events_handled++;
	    }
	    else if(w == fb->list_win)
	    {
		FBrowserQueueDraw(fb, FB_DRAW_ALLLISTS);
		events_handled++;
	    }
	    break;

	  case KeyPress:
	    if(!fb->has_focus)
		return(events_handled);

	    if(WidgetHasFocus(fb->location_pulist) ||
	       WidgetHasFocus(fb->prompt) ||
	       WidgetHasFocus(fb->list_win_sb) ||
	       WidgetHasFocus(fb->dir_win_sb) ||
	       WidgetHasFocus(fb->file_win_sb)
	    )
		break;

	    key = &event->xkey;
	    keycode = key->keycode;
	    key_state = key->state;
	    keyval = OSWGetKeyValFromKeyCode(keycode, key_state);

#define FB_SCROLL_LIST_AS_NEEDED(_obj_)		{		\
 unsigned int width, height;					\
 if(OSWGetWindowGeometry(w, NULL, NULL, &width, &height, NULL)) { \
  const int sb_x = (int)SBarGetPositionX(sb);			\
  const int x2 = (_obj_)->x - sb_x + (int)(_obj_)->width;	\
  width = (unsigned int)MAX(((int)width - (2 * FB_LIST_BORDER)), 0); \
  if(x2 > (int)width)						\
   SBarSetPositionX(						\
    sb,								\
    (float)((_obj_)->x - (int)width + (int)(_obj_)->width)	\
   );								\
  else if((_obj_)->x < sb_x)                                    \
   SBarSetPositionX(sb, (float)(_obj_)->x);                     \
 }                                                              \
}
	    /* Left */
	    if((keyval == OSW_KEY_Left) ||
	       (keyval == OSW_KEY_Page_Up) ||
	       (keyval == OSW_KEY_KP_Left) ||
	       (keyval == OSW_KEY_KP_Page_Up)
	    )
	    {
		if(fb->style == FB_STYLE_SINGLE_LIST)
		{
		    const int prev_sel_object = fb->sel_object;
		    scroll_bar_struct *sb = fb->list_win_sb;

		    fb->sel_object = prev_sel_object -
			MAX(fb->list_rows_visible, 1);
		    if(fb->sel_object < 0)
			fb->sel_object = 0;

		    /* Check if the selected object is not visible and
		     * scroll as needed
		     */
		    w = fb->list_win;
		    if(fb->sel_object >= fb->dir_list_items)
		    {
			const int i = fb->sel_object - fb->dir_list_items;
			if((i >= 0) && (i < fb->file_list_items))
			{
			    fb_object_struct *obj = fb->file_list[i];
			    if(obj != NULL)
			    {
				FB_SCROLL_LIST_AS_NEEDED(obj);
			    }
			}
		    }
		    else
		    {
			const int i = fb->sel_object;
			if((i >= 0) && (i < fb->dir_list_items))
			{
			    fb_object_struct *obj = fb->dir_list[i];
			    if(obj != NULL)
			    {
				FB_SCROLL_LIST_AS_NEEDED(obj);
			    }
			}
		    }
		    FBrowserQueueDraw(fb, FB_DRAW_ALLLISTS);
		}
		events_handled++;
	    }
	    /* Right */
	    else if((keyval == OSW_KEY_Right) ||
		    (keyval == OSW_KEY_Page_Down) ||
		    (keyval == OSW_KEY_KP_Right) ||
		    (keyval == OSW_KEY_KP_Page_Down)
	    )
	    {
		if(fb->style == FB_STYLE_SINGLE_LIST)
		{
		    const int	prev_sel_object = fb->sel_object,
				total = fb->dir_list_items + fb->file_list_items;
		    scroll_bar_struct *sb = fb->list_win_sb;

		    w = fb->list_win;

		    fb->sel_object = prev_sel_object +
			MAX(fb->list_rows_visible, 1);
		    if(fb->sel_object >= total)
			fb->sel_object = total - 1;

		    /* Check if the selected object is not visible and
		     * scroll as needed
		     */
		    if(fb->sel_object >= fb->dir_list_items)
		    {
			const int i = fb->sel_object - fb->dir_list_items;
			if((i >= 0) && (i < fb->file_list_items))
			{
			    fb_object_struct *obj = fb->file_list[i];
			    if(obj != NULL)
			    {
				FB_SCROLL_LIST_AS_NEEDED(obj);   
			    }
			}
		    }
		    else
		    {
			const int i = fb->sel_object;
			if((i >= 0) && (i < fb->dir_list_items)) 
			{
			    fb_object_struct *obj = fb->dir_list[i];
			    if(obj != NULL)
			    {
				FB_SCROLL_LIST_AS_NEEDED(obj);
			    }
			}    
		    }    
		    FBrowserQueueDraw(fb, FB_DRAW_ALLLISTS);
		}
		events_handled++;
 	    }
	    /* Up */
	    else if((keyval == OSW_KEY_Up) ||
		    (keyval == OSW_KEY_KP_Up)
	    )
	    {
		if(fb->style == FB_STYLE_SINGLE_LIST)
		{
		    const int   prev_sel_object = fb->sel_object;
		    scroll_bar_struct *sb = fb->list_win_sb;

		    w = fb->list_win;

		    fb->sel_object = prev_sel_object - 1;
		    if(fb->sel_object < 0)
			fb->sel_object = 0;

		    /* Check if the selected object is not visible and
		     * scroll as needed
		     */
		    if(fb->sel_object >= fb->dir_list_items)
		    {
			const int i = fb->sel_object - fb->dir_list_items;
			if((i >= 0) && (i < fb->file_list_items))
			{
			    fb_object_struct *obj = fb->file_list[i];
			    if(obj != NULL)
			    {
				FB_SCROLL_LIST_AS_NEEDED(obj);
			    }
			}    
		    }
		    else
		    {   
			const int i = fb->sel_object;
			if((i >= 0) && (i < fb->dir_list_items))
			{
			    fb_object_struct *obj = fb->dir_list[i];
			    if(obj != NULL)
			    {
				FB_SCROLL_LIST_AS_NEEDED(obj);
			    }
			}
		    }    
		    FBrowserQueueDraw(fb, FB_DRAW_ALLLISTS);
		}
		events_handled++;
	    }
	    /* Down */
	    else if((keyval == OSW_KEY_Down) ||
		    (keyval == OSW_KEY_KP_Down)
	    )
	    {
		if(fb->style == FB_STYLE_SINGLE_LIST)
		{
		    const int   prev_sel_object = fb->sel_object,
				total = fb->dir_list_items + fb->file_list_items;
		    scroll_bar_struct *sb = fb->list_win_sb;

		    w = fb->list_win;

		    fb->sel_object = prev_sel_object + 1;
		    if(fb->sel_object >= total)
			fb->sel_object = total - 1;

		    /* Check if the selected object is not visible and
		     * scroll as needed
		     */
		    if(fb->sel_object >= fb->dir_list_items)
		    {
			const int i = fb->sel_object - fb->dir_list_items;
			if((i >= 0) && (i < fb->file_list_items))
			{
			    fb_object_struct *obj = fb->file_list[i];
			    if(obj != NULL)
			    {
				FB_SCROLL_LIST_AS_NEEDED(obj);
			    }
			}
		    }
		    else
		    {
			const int i = fb->sel_object;
			if((i >= 0) && (i < fb->dir_list_items))
			{
			    fb_object_struct *obj = fb->dir_list[i];
			    if(obj != NULL)
			    {
				FB_SCROLL_LIST_AS_NEEDED(obj);
			    }
			}
		    }
		    FBrowserQueueDraw(fb, FB_DRAW_ALLLISTS);
		}
		events_handled++;
	    }
	    /* Home */
	    else if((keyval == OSW_KEY_Home) ||
		    (keyval == OSW_KEY_KP_Home)
	    )
	    {
		if(fb->style == FB_STYLE_SINGLE_LIST)
		{
		    scroll_bar_struct *sb = fb->list_win_sb;

		    w = fb->list_win;

		    fb->sel_object = 0;

		    /* Check if the selected object is not visible and
		     * scroll as needed
		     */
		    if(fb->sel_object >= fb->dir_list_items)
		    {
			const int i = fb->sel_object - fb->dir_list_items;
			if((i >= 0) && (i < fb->file_list_items))
			{
			    fb_object_struct *obj = fb->file_list[i];
			    if(obj != NULL)
			    {
				FB_SCROLL_LIST_AS_NEEDED(obj);
			    }
			}
		    }
		    else
		    {
			const int i = fb->sel_object;
			if((i >= 0) && (i < fb->dir_list_items))
			{
			    fb_object_struct *obj = fb->dir_list[i];
			    if(obj != NULL)
			    {
				FB_SCROLL_LIST_AS_NEEDED(obj);
			    }
			}
		    }
		    FBrowserQueueDraw(fb, FB_DRAW_ALLLISTS);
		}
		events_handled++;
	    }
	    /* End */
	    else if((keyval == OSW_KEY_End) ||
		    (keyval == OSW_KEY_KP_End)
	    )
	    {
		if(fb->style == FB_STYLE_SINGLE_LIST)
		{
		    const int total = fb->dir_list_items + fb->file_list_items;
		    scroll_bar_struct *sb = fb->list_win_sb;

		    w = fb->list_win;

		    fb->sel_object = total - 1;

		    /* Check if the selected object is not visible and
		     * scroll as needed
		     */
		    if(fb->sel_object >= fb->dir_list_items)
		    {
			const int i = fb->sel_object - fb->dir_list_items;
			if((i >= 0) && (i < fb->file_list_items))
			{
			    fb_object_struct *obj = fb->file_list[i];
			    if(obj != NULL)
			    {
				FB_SCROLL_LIST_AS_NEEDED(obj);
			    }
			}
		    }
		    else
		    {
			const int i = fb->sel_object;
			if((i >= 0) && (i < fb->dir_list_items))
			{
			    fb_object_struct *obj = fb->dir_list[i];
			    if(obj != NULL)
			    {
				FB_SCROLL_LIST_AS_NEEDED(obj);
			    }
			}
		    }
		    FBrowserQueueDraw(fb, FB_DRAW_ALLLISTS);
		}
		events_handled++;
	    }
	    /* Backspace */
	    else if(keyval == OSW_KEY_BackSpace)
	    {
		if(fb->style == FB_STYLE_SINGLE_LIST)
		{
		    FBrowserChangeDir(fb, "..");
		}
		events_handled++;
	    }
	    /* Enter */
	    else if((keyval == OSW_KEY_Return) ||
	            (keyval == OSW_KEY_KP_Enter) ||
		    (keyval == OSW_KEY_ISO_Enter)
	    )
	    {
		/* In change values mode? */
		if(fb->cv_prompt_mode != FB_CVP_MODE_NONE)
		{
		    /* Do apply change values routine */
		    prompt_struct *prompt = fb->cv_prompt;
		    char *s = STRDUP(PromptGetS(prompt));
		    FBrowserCVPromptActivateCB(prompt, s, fb);
		    free(s);

		    events_handled++;
		    return(events_handled);
		}
		else
		{
		    FBrowserDoOK(fb, False);
		}

		events_handled++;
	    }
#undef FB_SCROLL_LIST_AS_NEEDED
	    break;

	  case KeyRelease:
	    if(!fb->has_focus)
		return(events_handled);

	    if(WidgetHasFocus(fb->location_pulist) ||
	       WidgetHasFocus(fb->prompt) ||
	       WidgetHasFocus(fb->list_win_sb) ||
	       WidgetHasFocus(fb->dir_win_sb) ||
	       WidgetHasFocus(fb->file_win_sb)
	    )
		break;

	    key = &event->xkey;
	    keycode = key->keycode;
	    key_state = key->state;
	    keyval = OSWGetKeyValFromKeyCode(keycode, key_state);

	    break;

	  case ButtonPress:
	    button = &event->xbutton;

            /* Grab focus as needed */
            if(!fb->has_focus)
            {
                if(OSWCheckWindowAncestory(fb->toplevel, w))
                    OSWGrabFocus(fb->toplevel);
            }

	    /* Update focus */
	    if((w == fb->toplevel) ||
	       (w == fb->dir_win) ||
	       (w == fb->file_win) ||
	       (w == fb->list_win)
	    )
	    {
		/* Clear widget focus */
		WidgetSetFocus(fb->dir_win_sb, False);
		WidgetSetFocus(fb->file_win_sb, False);
		WidgetSetFocus(fb->list_win_sb, False);
		WidgetSetFocus(fb->prompt, False);
		WidgetSetFocus(fb->location_pulist, False);
	    }

	    /* List window */
	    if(w == fb->list_win)
	    {
		const int	button_num = button->button,
				prev_sel_object = fb->sel_object;
		scroll_bar_struct *sb = fb->list_win_sb;
		int	i,
			x = (int)button->x + (int)SBarGetPositionX(sb) -
			    FB_LIST_BORDER,
			y = (int)button->y + (int)SBarGetPositionY(sb) -
			    FB_LIST_BORDER;
		time_t dt = (time_t)button->time - fb->button_pressed_time;
		fb_object_struct *obj;

		fb->button_pressed = button_num;
		fb->button_last_x = (int)button->x;
		fb->button_last_y = (int)button->y;
		fb->button_pressed_time = (time_t)button->time;

		/* Drag? */
		if(button_num == Button2)
		{
		    OSWGrabPointerTime(
			w,
			True,
			ButtonReleaseMask | PointerMotionMask,
			GrabModeAsync, GrabModeAsync,
			None,
			None,
			(time_t)button->time
		    );
		    events_handled++;
		    break;
		}
		/* Scroll button pressed? */
		else if(button_num == Button4)
		{
		    unsigned int width, height;
		    float inc;
		    OSWGetWindowGeometry(w, NULL, NULL, &width, &height, NULL);
		    inc = MAX(
			(0.25f * width),
			sb->x_step_inc
		    );
		    SBarSetPositionX(
			sb,
			SBarGetPositionX(sb) - inc
		    );
		    events_handled++;
		    break;
		}
		else if(button_num == Button5)
		{
		    unsigned int width, height;
		    float inc;
		    OSWGetWindowGeometry(w, NULL, NULL, &width, &height, NULL);
		    inc = MAX(
			(0.25f * width),
			sb->x_step_inc
		    );
		    SBarSetPositionX(
			sb,
			SBarGetPositionX(sb) + inc
		    );
		    events_handled++;
		    break;
		}

#define FB_SCROLL_LIST_AS_NEEDED(_obj_)		{		\
 unsigned int width, height;					\
 if(OSWGetWindowGeometry(w, NULL, NULL, &width, &height, NULL)) { \
  const int sb_x = (int)SBarGetPositionX(sb);			\
  const int x2 = (_obj_)->x - sb_x + (int)(_obj_)->width;	\
  width = (unsigned int)MAX(((int)width - (2 * FB_LIST_BORDER)), 0); \
  if(x2 > (int)width)						\
   SBarSetPositionX(						\
    sb,								\
    (float)((_obj_)->x - (int)width + (int)(_obj_)->width)	\
   );								\
  else if((_obj_)->x < sb_x)                                    \
   SBarSetPositionX(sb, (float)(_obj_)->x);                     \
 }                                                              \
}
		/* Begin selecting object by finding which object the
		 * button press occured on
		 */
		fb->sel_object = -1;
		for(i = 0; i < fb->dir_list_items; i++)
		{
		    obj = fb->dir_list[i];
		    if(obj == NULL)
			continue;

		    if((x >= obj->x) &&
		       (y >= obj->y) &&
		       (x < (obj->x + (int)obj->width)) &&
		       (y < (obj->y + (int)obj->height))
		    )
		    {
			fb->sel_object = i;
			FB_SCROLL_LIST_AS_NEEDED(obj);
		        break;
		    }
		}
		/* Iterate through files list */
		if(fb->sel_object == -1)
		{
		    for(i = 0; i < fb->file_list_items; i++)
		    {
			obj = fb->file_list[i];
			if(obj == NULL)
			    continue;

			if((x >= obj->x) &&
			   (y >= obj->y) &&
			   (x < (obj->x + (int)obj->width)) &&
			   (y < (obj->y + (int)obj->height))
			)
			{
			    /* Offset selected object index */
			    const int n = i + fb->dir_list_items;
			    fb->sel_object = n;
			    FB_SCROLL_LIST_AS_NEEDED(obj);
			    break;
			}
		    }
		}
#undef FB_SCROLL_LIST_AS_NEEDED
		if(fb->sel_object >= (fb->file_list_items + fb->dir_list_items))
		    fb->sel_object = -1;

		/* Map the right-click menu? */
		if(button_num == Button3)
		{
		    int x, y;
		    FBrowserQueueDraw(fb, FB_DRAW_ALLLISTS);
		    OSWGetPointerCoords(
			gui->root_win,
			&x, &y,
			NULL, NULL
		    );
		    MenuMapCoordinates(fb->menu, x, y);

		    /* Need to reset the button pressed */
		    fb->button_pressed = 0;

		    events_handled++;
		    break;
		}    

		/* Double clicked? */
		if((dt >= 0l) && (dt < wglobal->double_click_int))
		{
		    if((fb->sel_object > -1) &&
		       (fb->sel_object == prev_sel_object) &&
		       (button->button == Button1)
		    )
		    {
			FBrowserDoOK(fb, False);
			events_handled++;  
			return(events_handled);
		    }
		}
		/* Slow double clicked? */
		else if((dt >= 0l) && (dt < wglobal->rename_click_int))
		{
		    if((fb->sel_object > -1) &&
		       (fb->sel_object == prev_sel_object) &&
		       (button->button == Button1)   
		    )
		    {
			FBrowserMapCVPrompt(fb, FB_CVP_MODE_RENAME);
		    }
		}
		FBrowserQueueDraw(fb, FB_DRAW_ALLLISTS);
		events_handled++;
	    }
	    /* Directory list window */
	    else if(w == fb->dir_win)
	    {
		int y = (int)button->y;
		time_t dt = (time_t)button->time - fb->button_pressed_time;
		scroll_bar_struct *sb = fb->dir_win_sb;

		/* Select dir */
		fb->sel_dir = ((int)SBarGetPositionY(sb) / FB_LIST_ROW_HEIGHT) +
		    ((y - 4) / FB_LIST_ROW_HEIGHT);
		if(fb->sel_dir < 0)
		    fb->sel_dir = -1;
		else if(fb->sel_dir >= fb->dir_list_items)
		    fb->sel_dir = -1;

		/* Unselect file when the directory list is clicked on */
		fb->sel_file = -1;

		/* Map change value prompt? */
		if(button->button == Button3)
		{
		    if(button->state & OSW_KEY_STATE_SHIFT)
			FBrowserMapCVPrompt(fb, FB_CVP_MODE_MODE);
		    else
		        FBrowserMapCVPrompt(fb, FB_CVP_MODE_RENAME);
		}

		/* Update last clicked time */
		fb->button_pressed_time = (time_t)button->time;

		/* Double click */
		if((dt > 0l) && (dt < wglobal->double_click_int))
		{
		    if((fb->sel_dir > -1) &&
		       (button->button == Button1)
		    )
		    {
			FBrowserDoOK(fb, False);
			events_handled++;
			return(events_handled);
		    }
		}

		FBrowserQueueDraw(fb, FB_DRAW_ALLLISTS);
		events_handled++;
	    }
	    /* File list window */
	    else if(w == fb->file_win)
	    {
		int y = (int)button->y;
		time_t dt = (time_t)button->time - fb->button_pressed_time;
		scroll_bar_struct *sb = fb->file_win_sb;

		/* Select file */
		fb->sel_file = ((int)SBarGetPositionY(sb) / FB_LIST_ROW_HEIGHT) +
		    ((y - 4) / FB_LIST_ROW_HEIGHT);
		if(fb->sel_file < 0)
		    fb->sel_file = -1;
		else if(fb->sel_file >= fb->file_list_items) 
		    fb->sel_file = -1;

		/* Unselect directory when the file list is clicked on */
		fb->sel_dir = -1;

		/* Map change values prompt? */
		if(button->button == Button3)
		{
		    if(button->state & OSW_KEY_STATE_SHIFT)
			FBrowserMapCVPrompt(fb, FB_CVP_MODE_MODE);
		    else
			FBrowserMapCVPrompt(fb, FB_CVP_MODE_RENAME);
		}

		/* Update last clicked time */
		fb->button_pressed_time = (time_t)button->time;
				
		/* Double click */
		if((dt > 0l) && (dt < wglobal->double_click_int))      
		{
		    if((fb->sel_file > -1) &&
		       (button->button == Button1)
		    )
		    {
			FBrowserDoOK(fb, False);
			events_handled++;
			return(events_handled);
		    }
		}

		FBrowserQueueDraw(fb, FB_DRAW_ALLLISTS);
		events_handled++;
	    }
	    break;

	  case ButtonRelease:
	    button = &event->xbutton;

	    /* Was in drag? */
	    if(fb->button_pressed > 0)
	    {
		if(OSWIsPointerGrabbed())
		    OSWUngrabPointerTime((time_t)button->time);
	    }

	    /* Reset drag information */
	    fb->button_pressed = 0;
	    fb->button_last_x = 0;
	    fb->button_last_y = 0;

	    break;

	  case MotionNotify:
	    motion = &event->xmotion;

	    /* Drag on list window? */
	    if((fb->button_pressed == Button2) &&
	       (w == fb->list_win)
	    )
	    {
		int	x = (int)motion->x,
			y = (int)motion->y,
			dx = x - fb->button_last_x;
		scroll_bar_struct *sb = fb->list_win_sb;

		SBarSetPositionX(
		    sb,
		    SBarGetPositionX(sb) - dx
		);

		fb->button_last_x = x;
		fb->button_last_y = y;

		events_handled++;
	    }
	    break;

	  case FocusIn:
	    if(w == fb->toplevel)
	    {
		fb->has_focus = True;
		events_handled++;
	    }
	    break;

	  case FocusOut:
	    if(w == fb->toplevel)
	    {
		fb->has_focus = False;
		events_handled++;
	    }
	    break;

	  case MapNotify:
	    if(w == fb->toplevel)
	    {
		if(!fb->map_state)
		{
		    if(fb->style == FB_STYLE_SINGLE_LIST)
		    {
			OSWMapWindow(fb->list_win);
			WidgetMap(fb->location_label);
			WidgetMap(fb->location_pulist);
			WidgetMap(fb->parent_btn);
			WidgetMap(fb->home_btn);
			WidgetMap(fb->refresh_btn);
			SBarDraw(fb->list_win_sb);
		    }
		    else
		    {
			scroll_bar_struct *sb;

			OSWMapWindow(fb->dir_win);
			OSWMapWindow(fb->file_win);

			sb = fb->dir_win_sb;
			SBarSetPositionX(sb, 0.0f);
			SBarSetPositionY(sb, 0.0f);
			SBarDraw(sb);

			sb = fb->file_win_sb;
			SBarSetPositionX(sb, 0.0f);
			SBarSetPositionY(sb, 0.0f);
			SBarDraw(sb);
		    }

		    WidgetMap(fb->ok_btn);
		    WidgetMap(fb->cancel_btn);
		    WidgetMap(fb->prompt);

		    fb->map_state = True;
		}
		events_handled++;
	    }
	    break;

	  case UnmapNotify:
	    if(w == fb->toplevel)
	    {
		if(fb->map_state)
		{
		    WidgetUnmap(fb->menu);
		    FBrowserUnmapCVPrompt(fb);
		    if(fb->style == FB_STYLE_SINGLE_LIST)
		    {
			WidgetUnmap(fb->location_label);
			WidgetUnmap(fb->location_pulist);
			WidgetUnmap(fb->parent_btn);
			WidgetUnmap(fb->home_btn);
			WidgetUnmap(fb->refresh_btn);
		    }
		    else
		    {

		    }
		    WidgetUnmap(fb->prompt);
		    WidgetUnmap(fb->ok_btn);
		    WidgetUnmap(fb->cancel_btn);

		    OSWDestroyPixmap(&fb->dir_win_buf);
		    OSWDestroyPixmap(&fb->file_win_buf);
		    OSWDestroyPixmap(&fb->list_win_buf);

		    fb->map_state = False;
		}
		events_handled++;
	    }
	    break;

	  case VisibilityNotify:
	    if(w == fb->toplevel)
	    {
		fb->visibility_state = event->xvisibility.state;
		events_handled++;
	    }
	    break;

	  case ClientMessage:
	    if(OSWEventIsTakeFocus(fb->toplevel, event, &t))
	    {
		OSWSetInputFocus(fb->toplevel, RevertToParent, t);
		events_handled++;
		return(events_handled);
	    }
	    else if(OSWEventIsDeleteWindow(fb->toplevel, event))
	    {
		/* Call cancel callback */
		if(fb->cancel_cb != NULL)
		    fb->cancel_cb(
			fb,
			fb->cancel_data
		    );

		events_handled++;

		if(!WidgetIsRegistered(fb))
		    return(events_handled);

		/* Unmap the file browser */
		if(fb->options & FB_FLAG_CLOSE_ON_CANCEL)
		{
		    FBrowserUnmap(fb);
		    return(events_handled);
		}
	    }
	    break;
	}

	/* Style specific widgets */
	if(fb->style == FB_STYLE_SINGLE_LIST)
	{
	    /* Locations label & popup list */
	    if(events_handled == 0)
		events_handled += WidgetManage(fb->location_label, event);
	    if(events_handled == 0)
		events_handled += WidgetManage(fb->location_pulist, event);

	    /* Parent button */
	    if(events_handled == 0)
		events_handled += WidgetManage(fb->parent_btn, event);

	    /* Home button */
	    if(events_handled == 0)
		events_handled += WidgetManage(fb->home_btn, event);

	    /* Refresh button */
	    if(events_handled == 0)
		events_handled += WidgetManage(fb->refresh_btn, event);


	    /* List window scroll bars */
	    if(events_handled == 0)
		events_handled += WidgetManage(fb->list_win_sb, event);
	}
	else
	{
	    /* Dir list and file list window scroll bars */
	    if(events_handled == 0)
	    {
		events_handled += WidgetManage(fb->dir_win_sb, event);
		events_handled += WidgetManage(fb->file_win_sb, event);
	    }
	}

	if(events_handled == 0)
	    events_handled += WidgetManage(fb->menu, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(fb->prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(fb->ok_btn, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(fb->cancel_btn, event);

	return(events_handled);
}


/*
 *	Maps and sets up the change values prompt on
 *	file browser fb.
 */
void FBrowserMapCVPrompt(fb_struct *fb, int mode)
{
	int x, y;
	unsigned int width, height;
	const char *s;
	prompt_struct *prompt;
	fb_object_struct *obj;

	if(fb == NULL)
	    return;

	if((fb->options & FB_FLAG_WRITE_PROTECT) ||
	   (fb->cv_prompt_mode != FB_CVP_MODE_NONE)
	)
	    return;

	/* Recreate the change value prompt */
	WidgetDestroy(fb->cv_prompt);
	free(fb->cv_prompt);
	fb->cv_prompt = prompt = PROMPT(malloc(sizeof(prompt_struct)));
	PromptInit(
	    prompt, fb->toplevel,
	    10,
	    ((int)fb->height / 2) - (FB_PROMPT_HEIGHT / 2),
	    MAX((int)fb->width - 20, 100),
	    FB_PROMPT_HEIGHT,
	    PROMPT_STYLE_NOBORDER,
	    "New value:",
	    PATH_MAX + NAME_MAX, 0
	);
	prompt->activate_cb = FBrowserCVPromptActivateCB;
	prompt->activate_data = fb;

	/* Change focus */
	fb->has_focus = True;
	WidgetSetFocus(fb->list_win_sb, False);
	WidgetSetFocus(fb->dir_win_sb, False);
	WidgetSetFocus(fb->file_win_sb, False);
	WidgetSetFocus(fb->prompt, False);
	WidgetSetFocus(fb->location_pulist, False);

	/* Reset the change value prompt target */
	free(fb->cv_prompt_target);
	fb->cv_prompt_target = NULL;

	/* Set the cv prompt mode */
	fb->cv_prompt_mode = mode;

	switch(mode)
	{
	  case FB_CVP_MODE_RENAME:
	    if(fb->style == FB_STYLE_SINGLE_LIST)
	    {
		/* Get selected object's name */
		s = FBrowserGetSelectionName(fb);
		if(s != NULL)
		{
		    /* Get coordinates */
		    obj = FBrowserGetSelObject(fb);
		    if(obj != NULL)
		    {
			scroll_bar_struct *sb = fb->list_win_sb;

		        x = obj->x - (int)SBarGetPositionX(sb) +
			    FB_LIST_BORDER + 26;
		        y = obj->y + FB_LIST_BORDER;
/*
			width = MAX(obj->width - 22, 4);
 */
			width = 120;
			height = obj->height;
		    }
		    else
		    {
			x = 0;
			y = 0;
			width = 100;
			height = 30;
		    }
		    OSWMoveResizeWindow(prompt->toplevel, x, y, width, height);

		    /* Set the change value prompt target */
		    fb->cv_prompt_target = STRDUP(PrefixPaths(
			PromptGetS(fb->prompt), s
		    ));

		    /* Set the initial value */
		    PromptSetS(prompt, s);
		    PromptUnmarkBuffer(prompt, PROMPT_POS_END);
		}
	    }
	    else
	    {


	    }
	    break;
	}

	/* Map the change value prompt and set focus */
	WidgetMap(prompt);
	WidgetSetFocus(prompt, True);
}


/*
 *	Unmaps change values prompt on file browser.
 */
void FBrowserUnmapCVPrompt(fb_struct *fb)
{
	if(fb == NULL)
	    return;

	/* Free recorded target path */
	free(fb->cv_prompt_target);
	fb->cv_prompt_target = NULL;

	/* Reset mode */
	fb->cv_prompt_mode = FB_CVP_MODE_NONE;

	/* Destroy the change value prompt */
	WidgetDestroy(fb->cv_prompt);
	free(fb->cv_prompt);
	fb->cv_prompt = NULL;
}


/*
 *	Maps file browser.
 */
void FBrowserMap(fb_struct *fb)
{
	if(fb == NULL)
	    return;

	/* Map */
	OSWMapRaised(fb->toplevel);

	/* Unset focus */
	WidgetSetFocus(fb->dir_win_sb, False);
	WidgetSetFocus(fb->file_win_sb, False);
	WidgetSetFocus(fb->list_win_sb, False);
	WidgetSetFocus(fb->prompt, False);
	WidgetSetFocus(fb->location_pulist, False);

	/* Set modal and grab focus */
	OSWSetModal(fb->toplevel);
	OSWGrabFocus(fb->toplevel);
}


/*
 *	Maps the file browser and sets its location to the one
 *	specified by path.
 */
void FBrowserMapPath(fb_struct *fb, const char *path)
{
	char *full_path;
	prompt_struct *prompt;

	if(fb == NULL)
	    return;

	if(fb->toplevel == None)
	    return;

	prompt = fb->prompt;

	if(path == NULL)
	{
	    const char *s = PromptGetS(prompt);
	    if(STRISEMPTY(s))
	    {
		char cwd[PATH_MAX];
		if(getcwd(cwd, sizeof(cwd)) != NULL)
		{
		    cwd[sizeof(cwd) - 1] = '\0';
		    full_path = STRDUP(PrefixPaths(cwd, path));
		    PromptSetS(prompt, full_path);
		    PromptMarkBuffer(prompt, PROMPT_POS_END);
		    free(full_path);
		}
		else
		{
		    full_path = STRDUP(PrefixPaths("/", path));
		    PromptSetS(prompt, full_path);
		    PromptMarkBuffer(prompt, PROMPT_POS_END);
		    free(full_path);
		}
	    }
	    FBrowserRefreshLocations(fb);
	    FBrowserRefreshList(fb);
	    FBrowserMap(fb);
	    return;
	}

	full_path = STRDUP(PathSubHome(path));
	if(full_path == NULL)
	    full_path = strdup(path);
	if(ISPATHABSOLUTE(full_path))
	{
	    PromptSetS(prompt, full_path);
	    PromptMarkBuffer(prompt, PROMPT_POS_END);
	}
	else
	{
	    char cwd[PATH_MAX];
	    if(getcwd(cwd, sizeof(cwd)) != NULL)
	    {
		cwd[sizeof(cwd) - 1] = '\0';
		free(full_path);
		full_path = STRDUP(PrefixPaths(cwd, path));
		PromptSetS(prompt, full_path);
		PromptMarkBuffer(prompt, PROMPT_POS_END);
	    }
	    else
	    {
		free(full_path);
		full_path = STRDUP(PrefixPaths("/", path));
		PromptSetS(prompt, full_path);
		PromptMarkBuffer(prompt, PROMPT_POS_END);
	    }
	}
	FBrowserRefreshLocations(fb);
	FBrowserRefreshList(fb);

	/* Map */
	FBrowserMap(fb);

	free(full_path);
}

/*
 *	Set file browser's search string pattern prepended to
 *	whatever path it currently has, reget listing and map.
 */
void FBrowserMapSearchMask(fb_struct *fb, const char *mask)
{
	const char *s;
	char *parent, *path_mask;
	prompt_struct *prompt;

	if(fb == NULL)
	    return;

	if(fb->toplevel == None)
	    return;

	if(mask == NULL)
	{
	    FBrowserRefreshLocations(fb);
	    FBrowserRefreshList(fb);
	    FBrowserMap(fb);
	    return;
	}

	prompt = fb->prompt;

	/* No location set? */
	s = PromptGetS(prompt);
	if(STRISEMPTY(s))
	{
	    char cwd[PATH_MAX];
	    if(getcwd(cwd, sizeof(cwd)) != NULL)
		cwd[sizeof(cwd) - 1] = '\0';
	    else
		*cwd = '\0';
	    PromptSetS(prompt, cwd);
	}

	/* Get the parent of the current location */
	parent = FBrowserGetJustPath(PromptGetS(prompt));

	/* Parent is not a directory? */
	if(!ISPATHDIR(parent))
	{
	    /* Get the grand parent as the parent */
	    char *grand_parent = STRDUP(GetParentDir(parent));
	    free(parent);
	    parent = grand_parent;
	}

	/* Set new location */
	path_mask = STRDUP(PrefixPaths(parent, mask));
	PromptSetS(prompt, path_mask);
	PromptMarkBuffer(prompt, PROMPT_POS_END);
	free(path_mask);

	/* Reget listings */
	FBrowserRefreshLocations(fb);
	FBrowserRefreshList(fb);

	FBrowserMap(fb);

	free(parent);
}

/*
 *	Unmaps the file browser.
 */
void FBrowserUnmap(fb_struct *fb)
{
	if(fb == NULL)
	    return;

	OSWUnmapWindow(fb->toplevel);
}

/*
 *	Destroys file browser.
 */
void FBrowserDestroy(fb_struct *fb)
{
	int i;
	fb_object_struct *obj;

	if(fb == NULL)
	    return;

	/* Delete the change values prompt's target object path */
	free(fb->cv_prompt_target);
	fb->cv_prompt_target = NULL;

	WidgetDestroy(fb->menu);
	free(fb->menu);
	fb->menu = NULL;
	WidgetDestroy(fb->cv_prompt);
	free(fb->cv_prompt);
	fb->cv_prompt = NULL;

	WidgetDestroy(fb->location_label);
	free(fb->location_label);
	fb->location_label = NULL;
	WidgetDestroy(fb->location_pulist);
	free(fb->location_pulist);
	fb->location_pulist = NULL;
	WidgetDestroy(fb->dir_win_sb);
	free(fb->dir_win_sb);
	fb->dir_win_sb = NULL;
	WidgetDestroy(fb->file_win_sb);
	free(fb->file_win_sb);
	fb->file_win_sb = NULL;
	WidgetDestroy(fb->list_win_sb);
	free(fb->list_win_sb);
	fb->list_win_sb = NULL;
	WidgetDestroy(fb->prompt);
	free(fb->prompt);
	fb->prompt = NULL;

	WidgetDestroy(fb->refresh_btn);
	free(fb->refresh_btn);
	fb->refresh_btn = NULL;
	WidgetDestroy(fb->home_btn);
	free(fb->home_btn);
	fb->home_btn = NULL;
	WidgetDestroy(fb->parent_btn);
	free(fb->parent_btn);
	fb->parent_btn = NULL;
	WidgetDestroy(fb->ok_btn);
	free(fb->ok_btn);
	fb->ok_btn = NULL;
	WidgetDestroy(fb->cancel_btn);
	free(fb->cancel_btn);
	fb->cancel_btn = NULL;

	OSWDestroyWindow(&fb->list_win);
	OSWDestroyPixmap(&fb->list_win_buf);

	OSWDestroyWindow(&fb->dir_win);
	OSWDestroyPixmap(&fb->dir_win_buf);

	OSWDestroyWindow(&fb->file_win);
	OSWDestroyPixmap(&fb->file_win_buf);

	OSWDestroyWindow(&fb->toplevel);

	/* Delete the directory listing */
	for(i = 0; i < fb->dir_list_items; i++)
	{
	    obj = fb->dir_list[i];
	    if(obj == NULL)
		continue;

	    free(obj->name);
	    free(obj);
	}
	free(fb->dir_list);
	fb->dir_list = NULL;
	fb->dir_list_items = 0;
	fb->sel_dir = -1;

	/* Free the file listing */
	for(i = 0; i < fb->file_list_items; i++)
	{
	    obj = fb->file_list[i];
	    if(obj == NULL)
		continue;

	    free(obj->name);
	    free(obj);
	}
	free(fb->file_list);
	fb->file_list = NULL;
	fb->file_list_items = 0;
	fb->sel_file = -1;
	fb->sel_object = -1;

	fb->list_max_width = 0;
	fb->list_rows_visible = 0;

	fb->button_pressed = 0;
	fb->button_last_x = 0;
	fb->button_last_y = 0;
	fb->button_pressed_time = 0l;

	fb->type = WIDGET_TYPE_UNKNOWN;
	fb->map_state = False;
	fb->has_focus = False;
	fb->disabled = False;
	fb->x = 0;
	fb->y = 0;
	fb->width = 0;
	fb->height = 0;
	fb->visibility_state = VisibilityFullyObscured;
	fb->font = NULL;
	fb->next = NULL;
	fb->prev = NULL;

	fb->options = 0;

	fb->ok_cb = NULL;
	fb->ok_data = NULL;
	fb->cancel_cb = NULL;
	fb->cancel_data = NULL;

	fb->main_push_func = NULL;
	fb->main_push_data = NULL;
	fb->main_pop_func = NULL;
	fb->main_pop_data = NULL;

	strlistfree(fb->response_paths, fb->nresponse_paths);
	fb->response_paths = NULL;
	fb->nresponse_paths = 0;  
}
