#include <stdlib.h>
#include <string.h>

#include "../include/string.h"

#include "../include/widget.h"

#include "../include/oswkeyvals.h"


static bool_t DialogGetMessageSize(
	dialog_struct *d, font_t *font, const char *msg,  
	unsigned int *width_rtn, unsigned int *height_rtn,
	int *longest_line_rtn,			/* In characters */
	int *nlines_rtn
);

static void DialogResizeCB(void *widget, void *data);
static void DialogDrawCB(void *widget, void *data);
static void DialogOKPBCB(push_button_struct *btn, void *data);

int DialogInit(dialog_struct *d);
void DialogSetMapRelative(dialog_struct *d, int map_rel);
void DialogSetTransientFor(dialog_struct *d, win_t toplevel);
void DialogSetTitle(dialog_struct *d, const char *title);
void DialogSetIcon(dialog_struct *d, image_t *icon);
void DialogSetMessage(dialog_struct *d, const char *msg);
void DialogResize(dialog_struct *d);
void DialogQueueResize(dialog_struct *d);
void DialogDraw(dialog_struct *d, const char *msg);
void DialogQueueDraw(dialog_struct *d);
int DialogManage(dialog_struct *d, event_t *event);
void DialogMap(dialog_struct *d);
void DialogUnmap(dialog_struct *d);
void DialogDestroy(dialog_struct *d);

void printdw(dialog_struct *d, const char *msg, win_t toplevel);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Sizes (in pixels):
 */
#define DLG_DEF_WIDTH			350
#define DLG_DEF_HEIGHT			80

#define DLG_BTN_WIDTH			70
#define DLG_BTN_HEIGHT			28

#define DLG_MARGIN			10

#define DLG_DEF_TITLE			"Dialog"
#define DLG_DEF_ICON_TITLE		"Dialog"


/*
 *	Gets the size of the specified message in pixels by measuring
 *	each line and returning the maximum extent size.
 *
 *	The longest_line_rtn is in characters (not pixels).
 *
 *	The nlines_rtn contains the number of lines in the message.
 */
static bool_t DialogGetMessageSize(
	dialog_struct *d, font_t *font, const char *msg,
	unsigned int *width_rtn, unsigned int *height_rtn,
	int *longest_line_rtn,			/* In characters */
	int *nlines_rtn
)
{
	int len, swidth, ascent, descent, longest_line, nlines;
	unsigned int width, height;
	const char *s, *s_end;

	if(width_rtn != NULL)
	    *width_rtn = 0;
	if(height_rtn != NULL)
	    *height_rtn = 0;
	if(longest_line_rtn != NULL)
	    *longest_line_rtn = 0;
	if(nlines_rtn != NULL)
	    *nlines_rtn = 0;

	if((d == NULL) || (font == NULL))
	    return(False);

	/* Calculate the message size */
	width = 0;
	height = 0;
	longest_line = 0;
	nlines = 0;
	s = msg;
	while(s != NULL)
	{
	    nlines++;
	    s_end = strchr(s, '\n');
	    len = (s_end != NULL) ? (int)(s_end - s) : STRLEN(s);
	    if(OSWGetTextBounds(
		font, s, len,
		NULL, NULL,
		&swidth,
		&ascent, &descent
	    ))
	    {
		if((unsigned int)swidth > width)
		{
		    width = (unsigned int)swidth;
		    longest_line = len;
		}
		height += (unsigned int)(ascent + descent);
	    }
	    s = (s_end != NULL) ? (s_end + 1) : NULL;
	}

	if(width_rtn != NULL)
	    *width_rtn = width;
	if(height_rtn != NULL)
	    *height_rtn = height;
	if(longest_line_rtn != NULL)
	    *longest_line_rtn = longest_line;
	if(nlines_rtn != NULL)
	    *nlines_rtn = nlines;

	return(True);
}


/*
 *	Resize callback.
 */
static void DialogResizeCB(void *widget, void *data)
{
	dialog_struct *d = (dialog_struct *)widget;
	if(d == NULL)
	    return;

	DialogResize(d);
}

/*
 *	Draw callback.
 */
static void DialogDrawCB(void *widget, void *data)
{
	dialog_struct *d = (dialog_struct *)widget;
	if(d == NULL)
	    return;

	if(d->map_state)
	    DialogDraw(d);
}

/*
 *	OK button callback.
 */
static void DialogOKPBCB(push_button_struct *btn, void *data)
{
	dialog_struct *dw = (dialog_struct *)data;
	if(dw == NULL)
	    return;

	DialogUnmap(dw);
}


/*
 *	Initializes the dialog.
 */
int DialogInit(dialog_struct *d)
{
	widget_global_struct *wglobal = &widget_global;

	if(d == NULL)
	    return(-2);

	/* Only reset values, the dialog's window is actually
	 * created when it is mapped
	 */

	d->type = WIDGET_TYPE_DIALOG;
	d->map_state = False;
	d->has_focus = False;
	d->disabled = False;
	d->visibility_state = VisibilityFullyObscured;
	d->x = 0;
	d->y = 0;
	d->width = 0;
	d->height = 0;
	d->font = wglobal->std_font;
	d->next = NULL;
	d->prev = NULL;
	d->data = NULL;

	d->toplevel = None;
	d->toplevel_buf = None;

	d->map_rel = WIDGET_WINDOW_RELATIVE_POINTER;
	d->transient_for_w = None;

	d->title = NULL;
	d->icon = NULL;
	d->message = NULL;

	d->ok_btn = NULL;

	/* Add widget to regeristry */
	WidgetRegister(d, d->type);

	return(0);
}

/*
 *	Sets the dialog's map relative.
 */
void DialogSetMapRelative(dialog_struct *d, int map_rel)
{
	win_t w;

	if(d == NULL)
	    return;

	d->map_rel = map_rel;

	w = d->toplevel;
	if(w != None)
	{
	    switch(d->map_rel)
	    {
	      case WIDGET_WINDOW_RELATIVE_PARENT:
		WidgetCenterWindowToWindow(w, d->transient_for_w);
		break;
	      case WIDGET_WINDOW_RELATIVE_ROOT:
	      case WIDGET_WINDOW_RELATIVE_POINTER:
		WidgetCenterWindow(w, d->map_rel);
		break;
	    }
	}
}

/*
 *	Sets the dialog's transient for hint to the toplevel window
 *      specified by toplevel.   
 */
void DialogSetTransientFor(dialog_struct *d, win_t toplevel) 
{
	win_t w;

	if(d == NULL)
	    return;

	d->transient_for_w = toplevel;
	w = d->toplevel;
	if(w != None)
	    OSWSetTransientFor(w, d->transient_for_w);
}

/*
 *	Sets the dialog's title.
 */
void DialogSetTitle(dialog_struct *d, const char *title)
{
	if(d == NULL)
	    return;

	if(d->title != title)
	{
	    free(d->title);
	    d->title = STRDUP(title);
	}

	if(d->toplevel != None)
	    OSWSetWindowTitle(d->toplevel, title);
}

/*
 *	Sets the dialog's icon.
 */
void DialogSetIcon(dialog_struct *d, image_t *icon)
{
	if(d == NULL)
	    return;

	if(d->icon == icon)
	    return;

	d->icon = icon;

	if(d->toplevel != None)
	{
	    if(d->map_state)
		DialogQueueDraw(d);
	}
}

/*
 *	Sets the dialog's message.
 */
void DialogSetMessage(dialog_struct *d, const char *msg)
{
	if(d == NULL)
	    return;

	if(d->message != msg)
	{
	    free(d->message);
	    d->message = STRDUP(msg);
	}

	if(d->toplevel != None)
	{
	    if(d->map_state)
		DialogQueueDraw(d);
	}
}

/*
 *	Resizes the dialog.
 */
void DialogResize(dialog_struct *d)
{
	int x, y;
	unsigned int width, height;
	win_t w;
	push_button_struct *btn;

	if(d == NULL)
	    return;

	w = d->toplevel;
	if(w == None)
	    return;

	OSWGetWindowGeometry(w, &x, &y, &width, &height, NULL);
	d->x = x;
	d->y = y;
	d->width = width;
	d->height = height;

	OSWDestroyPixmap(&d->toplevel_buf);

	/* Button */
	btn = d->ok_btn;
	if(btn != NULL)
	{
	    x = ((int)d->width - (int)btn->width) / 2;
	    y = (int)d->height - (int)btn->height - DLG_MARGIN;
	    OSWMoveWindow(btn->toplevel, x, y);
	    WidgetQueueResize(btn);
	}
}

/*
 *	Queues the dialog to resize.
 */
void DialogQueueResize(dialog_struct *d)
{
	if(d == NULL)
	    return;

	if(d->toplevel == None)
	    return;

	WidgetQueueResizeAction(DialogResizeCB, d, NULL);
}

/*
 *	Draws the dialog.
 */
void DialogDraw(dialog_struct *d)
{
	unsigned int	icon_width = 0, icon_height = 0,
			mesg_width = 0, mesg_height = 0;
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if(d == NULL)
	    return;

	/* Calculate the icon's size */
	if(d->icon != NULL)
	{
	    image_t *icon = d->icon;
	    icon_width = icon->width;
	    icon_height = icon->height;
	}

	/* Calculate the message size */
	DialogGetMessageSize(
	    d, d->font, d->message,
	    &mesg_width, &mesg_height, NULL, NULL
	);

	/* Need to create the dialog's windows? */
	if(d->toplevel == None)
	{
	    const unsigned int	max_width = gui->display_width,
				max_height = gui->display_height;
	    unsigned int	width = DLG_MARGIN +
				    icon_width +
				    ((icon_width > 0) ? DLG_MARGIN : 0) +
				    mesg_width + DLG_MARGIN,
				height = DLG_MARGIN +
				    MAX(icon_height, mesg_height) +
				    DLG_MARGIN + 2 + DLG_MARGIN +
				    DLG_BTN_HEIGHT +
				    DLG_MARGIN;
	    int x = 0, y = 0;
	    keyval_t accel_keys[5];
	    win_t w, parent = gui->root_win;
	    push_button_struct *btn;

	    if(width > max_width)
		width = max_width;
	    if(height > max_height)
		height = max_height;

	    /* Destroy previously allocated resources */
	    WidgetDestroy(d->ok_btn);
	    free(d->ok_btn);
	    d->ok_btn = NULL;
	    OSWDestroyPixmap(&d->toplevel_buf);

	    /* Create toplevel */
	    OSWCreateWindow(
		&w, parent,
		x, y,
		width, height
	    );
	    d->toplevel = w;
	    OSWSetWindowInput(w, OSW_EVENTMASK_TOPLEVEL);
	    OSWSetWindowWMProperties(
		w,
		(d->title != NULL) ? d->title : DLG_DEF_TITLE,
		DLG_DEF_ICON_TITLE,	/* Icon title */
		wglobal->std_icon_pm,	/* Icon */
		False,			/* Do not let WM set coordinates */
		x, y,
		DLG_BTN_WIDTH, DLG_BTN_HEIGHT,
		max_width, max_height,
		OSW_WINDOW_FRAME_DIALOG,
		0, NULL
	    );
	    OSWSetTransientFor(w, d->transient_for_w);
	    switch(d->map_rel)
	    {
	      case WIDGET_WINDOW_RELATIVE_PARENT:
		WidgetCenterWindowToWindow(w, d->transient_for_w);
		break;
	      case WIDGET_WINDOW_RELATIVE_ROOT:
	      case WIDGET_WINDOW_RELATIVE_POINTER:
		WidgetCenterWindow(w, d->map_rel);
		break;
	    }
	    OSWGetWindowGeometry(
		w, &d->x, &d->y, &d->width, &d->height, NULL
	    );
	    parent = w;

	    /* Create the OK button */
	    d->ok_btn = btn = PUSH_BUTTON(malloc(sizeof(push_button_struct)));
	    PBtnInit(
		btn, parent,
		((int)d->width - DLG_BTN_WIDTH) / 2,
		(int)d->height - DLG_BTN_HEIGHT - DLG_MARGIN,
		DLG_BTN_WIDTH,
		DLG_BTN_HEIGHT,
		"OK",
		PBTN_LABEL_ALIGN_CENTER,
		None, None,
		DialogOKPBCB, d
	    );
	    accel_keys[0] = OSW_KEY_space;
	    accel_keys[1] = OSW_KEY_Escape;
	    accel_keys[2] = OSW_KEY_Return;
	    accel_keys[3] = OSW_KEY_o;
	    PBtnSetAccelKeys(btn, accel_keys, NULL, 4);
	}

	/* Recreate pixmap buffer as needed */
	if(d->toplevel_buf == None)
	    OSWCreatePixmap(&d->toplevel_buf, d->width, d->height);

	/* Draw the dialog */
	if(d->toplevel != None)
	{
	    int x, y;
	    unsigned int width, height;
	    font_t *font = d->font;
	    win_t w = d->toplevel;
	    pixmap_t pixmap = d->toplevel_buf;
	    image_t *icon = d->icon;

	    OSWGetWindowGeometry(
		w, NULL, NULL, &width, &height, NULL
	    );

	    /* Draw background */
	    if(wglobal->force_mono)
	        OSWClearPixmap(
		    pixmap, width, height,
		    gui->black_pix
	        );
	    else
	        WidgetPutImageTile(
	            pixmap,
	            wglobal->std_bkg_img,
	            width, height
		);

	    /* Draw message */
	    if((d->message != NULL) && (font != NULL))
	    {
		int len, lbearing, rbearing, swidth, ascent, descent;
		const char *s = d->message, *s_end;

	        /* Set text color */
	        if(wglobal->force_mono)
		    OSWSetForeground(gui->white_pix);
	        else
		    OSWSetForeground(wglobal->normal_text_pix);
		OSWSetFont(font);

	        /* Begin drawing each line */
		x = DLG_MARGIN + (int)icon_width +
		    ((icon_width > 0) ? DLG_MARGIN : 0);
		if(icon_height > mesg_height)
		    y = DLG_MARGIN +
			((int)(icon_height - mesg_height) / 2);
		else
		    y = DLG_MARGIN;

	        while(s != NULL)
	        {
		    s_end = strchr(s, '\n');
		    len = (s_end != NULL) ? (int)(s_end - s) : strlen(s);
		    if(OSWGetTextBounds(
			font, s, len,  
			&lbearing, &rbearing,
			&swidth,
			&ascent, &descent
		    ))
		    {
			OSWDrawStringLimited(
			    pixmap,
			    x - lbearing, y + ascent,
			    s, len
			);
			y += ascent + descent;
		    }
		    s = (s_end != NULL) ? (s_end + 1) : NULL;
		}
	    }

	    /* Draw horizontal rule */
	    y = (int)d->height - DLG_BTN_HEIGHT - (2 * DLG_MARGIN);
	    if(wglobal->force_mono)
	    {
		OSWSetForeground(gui->black_pix);
		OSWDrawLine(
		    pixmap,
		    0, y, (int)width, y
		);
	    }
	    else
	    {
		OSWSetForeground(wglobal->surface_shadow_pix);
		OSWDrawLine(
		    pixmap,
		    0, y, (int)width, y
		);
		OSWSetForeground(wglobal->surface_highlight_pix);
		OSWDrawLine(
		    pixmap,
		    0, y - 1, (int)width, y - 1
		);
	    }

	    /* Draw icon */
	    if((icon != NULL) && !wglobal->force_mono)
	    {
		x = DLG_MARGIN;
		y = DLG_MARGIN + ((mesg_height > icon_height) ?
		    ((int)(mesg_height - icon_height) / 2) : 0);
		WidgetPutImageRaised(
		    pixmap, icon,
		    x, y,
		    6			/* Altitude */
		);
	    }

	    /* Put the buffer to the window */
	    OSWPutDrawableToWindow(w, pixmap);
	}

	/* Map as needed */
	if(!d->map_state)
	{
	    OSWMapRaised(d->toplevel);
	    d->map_state = True;
	    WidgetMap(d->ok_btn);
	    OSWPutDrawableToWindow(d->toplevel, d->toplevel_buf);
	}
}

/*
 *	Queues the dialog to draw.
 */
void DialogQueueDraw(dialog_struct *d)
{
	if(d == NULL)
	    return;

	WidgetQueueDrawAction(DialogDrawCB, d, NULL);
}

/*
 *	Manages the dialog event.
 */
int DialogManage(dialog_struct *dw, event_t *event)
{
	int events_handled = 0;
	time_t t;
	win_t w;
	event_visibility_t *visibility;

	if((dw == NULL) || (event == NULL))
	    return(events_handled);

	if(dw->toplevel == None)
	    return(events_handled);

	w = event->xany.window;

	switch(event->type)
	{
	  case ConfigureNotify:
	    if(w == dw->toplevel)
	    {
		DialogQueueResize(dw);
		if(dw->map_state)
		    DialogQueueDraw(dw);
		events_handled++;
	    }
	    break;

	  case Expose:
	    if(w == dw->toplevel)
	    {
		DialogQueueDraw(dw);
		events_handled++;
	    }
	    break;

	  case KeyPress:
	    if(!dw->has_focus)
		return(events_handled);

	    break;

	  case KeyRelease:
	    if(!dw->has_focus)     
		return(events_handled);

	    break;

	  case ButtonPress:
            /* Grab focus as needed */
            if(!dw->has_focus)
            {
                if(OSWCheckWindowAncestory(dw->toplevel, w))
                    OSWGrabFocus(dw->toplevel);
            }
	    break;

	  case ButtonRelease:
	    break;

	  case FocusIn:
	    if(w == dw->toplevel)
	    {
		dw->has_focus = True;
		events_handled++;
	    }
	    break;

	  case FocusOut:
	    if(w == dw->toplevel)
	    {
		dw->has_focus = False;
		events_handled++;
	    }
	    break;

	  case MapNotify:
	    if(w == dw->toplevel)
	    {
		dw->map_state = True;
		events_handled++;
	    }
	    break;

	  case UnmapNotify:
	    if(w == dw->toplevel)
	    {
		dw->map_state = False;
		events_handled++;
	    }
	    break;

	  case VisibilityNotify:
	    visibility = &event->xvisibility;
	    if(w == dw->toplevel)
	    {
		dw->visibility_state = visibility->state;
		events_handled++;
	    }
	    break;

	  case ClientMessage:
	    if(OSWEventIsTakeFocus(dw->toplevel, event, &t))
	    {
		OSWSetInputFocus(dw->toplevel, RevertToParent, t);
		events_handled++;
	    }
	    else if(OSWEventIsDeleteWindow(dw->toplevel, event))
	    {
		DialogUnmap(dw);
		events_handled++;
		return(events_handled);
	    }
	    break;
	}

	/* Manage widgets */
	if(events_handled == 0)
	    events_handled += WidgetManage(dw->ok_btn, event);

	return(events_handled);
}

/*
 *	Maps the dialog.
 */
void DialogMap(dialog_struct *d)
{
	win_t w;

	if(d == NULL)
	    return;

	/* Dialogs may not be used while blocking input since user
	 * will not be able to respond
	 */
	if(OSWBlockInputLevel() > 0)        
	    return;

	/* Map */
	d->map_state = False;
	DialogDraw(d);

	/* Set modal and grab focus */
	w = d->toplevel;
	OSWSetModal(w);
	OSWGrabFocus(w);
}

/*
 *	Unmaps the dialog.
 */
void DialogUnmap(dialog_struct *d)
{
	if(d == NULL)
	    return;

	OSWUnmapWindow(d->toplevel);

	WidgetDestroy(d->ok_btn);
	free(d->ok_btn);
	d->ok_btn = NULL;

	OSWDestroyWindow(&d->toplevel);
	OSWDestroyPixmap(&d->toplevel_buf);

	d->map_state = False;
	d->has_focus = False;
	d->visibility_state = VisibilityFullyObscured;
	d->x = 0;
	d->y = 0;
	d->width = 0;
	d->height = 0;
}

/*
 *	Deletes the dialog.
 */
void DialogDestroy(dialog_struct *d)
{
	if(d == NULL)
	    return;

	d->map_rel = 0;
	d->transient_for_w = None;

	free(d->title);
	d->title = NULL;

	d->icon = NULL;		/* Shared */

	free(d->message);
	d->message = NULL;

	WidgetDestroy(d->ok_btn);
	free(d->ok_btn);
	d->ok_btn = NULL;
	OSWDestroyWindow(&d->toplevel);
	OSWDestroyPixmap(&d->toplevel_buf);

	d->type = WIDGET_TYPE_UNKNOWN;
	d->map_state = False;
	d->has_focus = False;
	d->visibility_state = VisibilityFullyObscured;
	d->x = 0;
	d->y = 0;
	d->width = 0;
	d->height = 0;
	d->disabled = False;
	d->font = NULL;
	d->next = NULL;
	d->prev = NULL;
}


/*
 *	Macro to set message, map, and draw dialog window.
 */
void printdw(dialog_struct *d, const char *msg, win_t toplevel)
{
	if(d == NULL)
	    return;

	/* Dialogs may not be used while blocking input since user
	 * will not be able to respond
	 */
	if(OSWBlockInputLevel() > 0)        
	    return;

	d->map_state = False;
	d->transient_for_w = toplevel;
	DialogSetMessage(d, msg);
	DialogDraw(d);
}
