#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/string.h"
#include "../include/graphics.h"

#include "../include/widget.h"

#include "../include/oswkeyvals.h"


int ListGetItemNumByData(list_struct *list, void *data);
int ListIsItemAllocated(list_struct *list, int n);

static void ListCVPromptActivateCB(
	prompt_struct *prompt, const char *s, void *data
);
static void ListCVPromptCancelCB(
	prompt_struct *prompt, void *data
);
static void ListWinResizeCB(void *widget, void *data);
static void ListWinDrawCB(void *widget, void *data);
static void ListSBValueChangedCB(
	scroll_bar_struct *sb, float x, float y, void *data
);
static void ListDragScrollCB(void *widget, void *data);

int ListGetItemFromCoordinates(list_struct *list, int x, int y);
int ListGetRowFromCoordinates(list_struct *list, int x, int y);
bool_t ListGetCoordinatesFromItem(
	list_struct *list, int i,
	int *x_rtn, int *y_rtn,
	unsigned int *width_rtn, unsigned int *height_rtn
);
visibility_t ListGetItemVisibility(list_struct *list, int item_num);
void ListMoveTo(list_struct *list, int item_num, float coeff);

int ListInsert(
	list_struct *list,
	int type,
	const char *text,
	pixmap_t pixmap, bitmap_t mask,
    	int pos,
	void *data
);
int ListAppend(
	list_struct *list,
	int type,
	const char *text,
	pixmap_t pixmap, bitmap_t mask,
	void *data
);
void ListRemove(list_struct *list, int i);
void ListClear(list_struct *list);

void *ListGetItemData(list_struct *list, int i);
void ListSetItemData(
	list_struct *list, int i,
	void *data, void (*destroy_cb)(void *)
);

void ListSelect(list_struct *list, int i);
void ListUnselect(list_struct *list, int i);

void ListWinDNDEnterNotify(list_struct *list, time_t t);
void ListWinDNDPositionNotify(
	list_struct *list, win_t tar_w,
	int x, int y,
	int root_x, int root_y,
	atom_t action,
	time_t t
);
void ListWinDNDLeaveNotify(list_struct *list, time_t t);

int ListWinInit(
	list_struct *list, win_t parent,
	int x, int y,
	unsigned int width, unsigned int height
);
void ListWinResize(list_struct *list);
void ListWinQueueResize(list_struct *list);
void ListWinDraw(list_struct *list);
void ListWinQueueDraw(list_struct *list);
int ListWinManage(list_struct *list, event_t *event);
static void ListWinCVPromptApply(list_struct *list, const char *s);
static void ListWinMapCVPrompt(list_struct *list, int mode);
static void ListWinUnmapCVPrompt(list_struct *list);
void ListWinMap(list_struct *list);
void ListWinUnmap(list_struct *list);
void ListWinDestroy(list_struct *list);

static win_t ListWinCreateDNDIcon(
	list_struct *list, int item_num,
	int x, int y
);
static void ListWinDragBegin(
	list_struct *list, int item_num,
	int x, int y,
	time_t t
);
static void ListWinDragStop(
	list_struct *list, bool_t successful,
	time_t t
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Sizes (in pixels):
 */
#define LW_DEF_ROW_HEIGHT		26
#define LW_PROMPT_HEIGHT		24

/*
 *	Default name to give to an item who's name is NULL.
 */
#define LW_DEF_ITEM_NAME		"Untitled"

/*
 *	Drag start tolorance (in pixels).
 */
#define LW_DRAG_TOLORANCE		6


/*
 *	Operations:
 */
#define LW_OP_CODE_NONE			0
#define LW_OP_CODE_SCROLL_UP		1	/* One unit */
#define LW_OP_CODE_SCROLL_DOWN		2	/* One unit */


/*
 *	Checks if the list is currently in a DND drag and that it is
 *	the DND drag source.
 */
#define LIST_IS_DND_SOURCE(_list_)	\
 (((_list_) != NULL) ? ((_list_)->drag_item_num > -1) : False)


/*
 *	Returns the list item that has the specified data.
 */
int ListGetItemNumByData(list_struct *list, void *data)
{
	int i;
	list_item_struct *item;

	if((list == NULL) || (data == NULL))
	    return(-1);

	for(i = 0; i < list->total_items; i++)
	{
	    item = list->item[i];
	    if(item == NULL)
		continue;

	    if(item->data == data)
		return(i);
	}

	return(-1);
}

/*
 *	Checks if the item is allocated on the list.
 */
int ListIsItemAllocated(
	list_struct *list, int n
)
{
	if(list == NULL)
	    return(0);
	else if((list->item == NULL) ||
		(n < 0) ||
		(n >= list->total_items)
	)
	    return(0);
	else if(list->item[n] == NULL)
	    return(0);
	else
	    return(1);
}


/*
 *	Change values prompt activate callback.
 */
static void ListCVPromptActivateCB(
	prompt_struct *prompt, const char *s, void *data
)
{
	list_struct *list = (list_struct *)data;
	if(list == NULL)
	    return;

	if(list->cv_prompt_mode == LIST_CV_PROMPT_MODE_NONE)
	    return;

	ListWinCVPromptApply(list, s);
	ListWinUnmapCVPrompt(list);
	if(list->map_state)
	    ListWinQueueDraw(list);
}

/*
 *	Change values prompt cancel callback.
 */
static void ListCVPromptCancelCB(
	prompt_struct *prompt, void *data
)
{
	list_struct *list = (list_struct *)data;
	if(list == NULL)
	    return;

	if(list->cv_prompt_mode == LIST_CV_PROMPT_MODE_NONE)
	    return;

	ListWinUnmapCVPrompt(list);
}

/*
 *	Resize callback.
 */
static void ListWinResizeCB(void *widget, void *data)
{
	list_struct *list = (list_struct *)widget;
	if(list == NULL)
	    return;

	ListWinResize(list); 
}

/*
 *	Draw callback.
 */
static void ListWinDrawCB(void *widget, void *data)
{
	list_struct *list = (list_struct *)widget;
	if(list == NULL)
	    return;

	if(list->map_state)
	    ListWinDraw(list);
}

/*
 *	Scroll bar value changed callback.
 *
 *	Redraws the list when the scroll bar's value has changed and
 *	if the list is mapped.
 */
static void ListSBValueChangedCB(
	scroll_bar_struct *sb, float x, float y, void *data
)
{
	list_struct *list = (list_struct *)data;
	if(list == NULL)
	    return;

	if(list->map_state)
	    ListWinQueueDraw(list);
}

/*
 *	Drag scroll callback.
 *
 *	Scrolls the list based on the current Y position of the
 *	pointer off the list.
 */
static void ListDragScrollCB(void *widget, void *data)
{
	int x, y;
	scroll_bar_struct *sb;
	widget_global_struct *wglobal = &widget_global;
	list_struct *list = (list_struct *)widget;
	if(list == NULL)
	    return;

	if(!list->map_state)
	    return;

	sb = list->sb;
	OSWGetPointerCoords(
	    list->toplevel, NULL, NULL, &x, &y
	);
	if(y < 0)
	{
	    int step = -y,
		y_pos = (int)SBarGetPositionY(sb);
	    SBarSetPositionY(
		sb,
		(int)(y_pos - step)
	    );
	}
	else if(y >= (int)sb->height)
	{
	    int	step = y - (int)sb->height + 1,
		y_pos = (int)SBarGetPositionY(sb);
	    SBarSetPositionY(
		sb,
		(int)(y_pos + step)
	    );
	}

	WidgetTimeoutAdd(
	    ListDragScrollCB,
	    widget, data,
	    wglobal->sbar_scroll_repeat_int
	);
}


/*
 *	Calculates the item number from the specified coordinates.
 *
 *	Both the x and y coordinates (not just the y coordinate) must
 *	fall within the list and on an item or else -1 is returned.
 */
int ListGetItemFromCoordinates(list_struct *list, int x, int y)
{
	int i, item_x, item_y;
	unsigned int item_width, item_height;
	list_item_struct *item;

	if(list == NULL)
	    return(-1);

	/* The specified coordinates must fall within the list */
	if((x < 0) || (x >= (int)list->width) ||
	   (y < 0) || (y >= (int)list->height)
	)
	    return(-1);

	/* Calculate the item number from the specified coordinates */
	i = ListGetRowFromCoordinates(list, x, y);
	if((i < 0) || (i >= list->total_items))
	    return(-1);

	item = list->item[i];
	if(item == NULL)
	    return(-1);

	/* Calculate the coordinates of the item */
	if(ListGetCoordinatesFromItem(
	    list, i, &item_x, &item_y, &item_width, &item_height
	))
	{
	    /* Is the specified coordinates outside of the item? */
	    if((x < item_x) || (y < item_y) ||
	       (x >= (item_x + (int)item_width)) ||
	       (y >= (item_y + (int)item_height))
	    )
		return(-1);
	    else
		return(i);
	}
	{
	    return(-1);
	}
}

/*
 *	Calculates the row number from the specified coordinates.
 *
 *	Does not take into consideration if the coordinates fall within
 *	the list and on an item.
 *
 *	The returned value may not refer to a valid item.
 */
int ListGetRowFromCoordinates(list_struct *list, int x, int y)
{
	int row_height;
	scroll_bar_struct *sb;

	if(list == NULL)
	    return(-1);
		       
	row_height = MAX((int)list->row_height, 1);
	sb = list->sb;

	return(
	    ((int)SBarGetPositionY(sb) + y - (int)list->top_margin) /
	    row_height
	);
}

/*
 *	Calculates the coordinates of the specified item.
 *
 *	The returned window coordinates will be relative to the list.
 */
bool_t ListGetCoordinatesFromItem(
	list_struct *list, int i,
	int *x_rtn, int *y_rtn,
	unsigned int *width_rtn, unsigned int *height_rtn
)
{
	int row_height;
	unsigned int width, height;
	scroll_bar_struct *sb;
	list_item_struct *item;

	if(x_rtn != NULL)
	    *x_rtn = 0;
	if(y_rtn != NULL)
	    *y_rtn = 0;
	if(width_rtn != NULL)
	    *width_rtn = 0;
	if(height_rtn != NULL)
	    *height_rtn = 0;

	if(list == NULL)
	    return(False);

	if((i < 0) || (i >= list->total_items))
	    return(False);

	item = list->item[i];
	if(item == NULL)
	    return(False);

	row_height = MAX((int)list->row_height, 1);
	sb = list->sb;

	width = 0;
	height = 0;
	if(item->pixmap != None)
	{
	    unsigned int icon_width, icon_height;
	    pixmap_t pixmap = item->pixmap;
	    OSWGetPixmapGeometry(
		pixmap, &icon_width, &icon_height, NULL
	    );
	    width += icon_width + 2;
	    height += icon_height;
	}
	if(item->text != NULL)
	{
	    int swidth, ascent, descent, font_height;
	    const char *s = item->text;
	    font_t *font = list->font;

	    OSWGetTextBounds(
		font, s, -1,
		NULL, NULL, &swidth, &ascent, &descent
	    );
	    width += swidth;
	    font_height = ascent + descent;
	    if(font_height > (int)height)
		height = (unsigned int)font_height;
	}

	if(x_rtn != NULL)
	    *x_rtn = (int)list->left_margin;
	if(y_rtn != NULL)
	    *y_rtn = (int)list->top_margin + (i * row_height) -
		(int)SBarGetPositionY(sb) +
		((row_height - (int)height) / 2);

	if(width_rtn != NULL)
	    *width_rtn = width;
	if(height_rtn != NULL)
	    *height_rtn = height;

	return(True);
}

/*
 *	Returns the item's visibility.
 */
visibility_t ListGetItemVisibility(list_struct *list, int item_num)
{
	int y, height, row_height;
	scroll_bar_struct *sb;

	if(list == NULL)
	    return(VisibilityFullyObscured);

	if((item_num < 0) || (item_num >= list->total_items))
	    return(VisibilityFullyObscured);

	row_height = MAX((int)list->row_height, 1);
	sb = list->sb;
	height = (int)sb->height - (2 * list->top_margin);

	y = (item_num * row_height) - (int)SBarGetPositionY(sb);
	if(y < 0)
	    return(((y + row_height) > 0) ?
		VisibilityPartiallyObscured : VisibilityFullyObscured
	    );
	else if((y + row_height) > height)
	    return((y < height) ?
		VisibilityPartiallyObscured : VisibilityFullyObscured
	    );
	else
	    return(VisibilityUnobscured);
}

/*
 *	Scrolls to the specified item with the specified coefficient
 *	applied as the offset.
 */
void ListMoveTo(list_struct *list, int item_num, float coeff)
{
	int y, last_y, height, row_height;
	scroll_bar_struct *sb;

	if(list == NULL)
	    return;
		   
	if((item_num < 0) || (item_num >= list->total_items))
	    return;

	row_height = MAX((int)list->row_height, 1);

	sb = list->sb;
	height = (int)sb->height - (2 * list->top_margin);
	last_y = (int)SBarGetPositionY(sb);

	/* Move to */
	y = (item_num * row_height) -
	    (int)(coeff * (height - row_height));
	if(y != last_y)
	    SBarSetPositionY(sb, (float)y);
}


/*
 *	Inserts a new item.
 */
int ListInsert(
	list_struct *list,
	int type,
	const char *text,
	pixmap_t pixmap, bitmap_t mask,
	int pos,
	void *data
)
{
	int i, n;
	list_item_struct *item;

	if(list == NULL)
	    return(-1);

	if(text == NULL)
	    text = LW_DEF_ITEM_NAME;

	/* Allocate a new item */
	item = (list_item_struct *)calloc(
	    1, sizeof(list_item_struct)
	);
	if(item == NULL)
	    return(-3);

	if(list->total_items < 0)
	    list->total_items = 0;
	list->total_items++;
	list->item = (list_item_struct **)realloc(
	    list->item,
	    list->total_items * sizeof(list_item_struct *)
	);
	if(list->item == NULL)
	{
	    free(item);
	    list->total_items = 0;
	    return(-3); 
	}

	/* Append? */
	if((pos < 0) || (pos >= list->total_items))
	{
	    /* Append */
	    n = list->total_items - 1;
	}
	else
	{
	    /* Insert */
	    n = pos;

	    /* Shift entry pointers */
	    for(i = list->total_items - 1; i > n; i--)
		list->item[i] = list->item[i - 1];
	}

	/* Set new entry pointer to point to the just allocated struct */
	list->item[n] = item;

	/* Set new values */
	item->type = type;
	item->text = STRDUP(text);
	item->pixmap = pixmap;
	item->mask = mask;
	item->data = data;
	item->destroy_cb = NULL;

	SBarQueueResize(
	    list->sb,
	    list->width, list->height,
	    list->width,
	    (list->total_items * list->row_height) +
		(2 * list->top_margin)
	);
	WidgetQueueDraw(list->sb);
	ListWinQueueDraw(list);

	return(n);
}

/*
 *	Appends a new item.
 */
int ListAppend(
	list_struct *list,
	int type,
	const char *text,
	pixmap_t pixmap, bitmap_t mask,
	void *data
)
{
	return(ListInsert(
	    list, type, text, pixmap, mask, -1, data
	));
}

/*
 *	Deletes the specified item from the list.
 */
void ListRemove(list_struct *list, int i)
{
	int j;
	list_item_struct *item;

	if(list == NULL)
	    return;

	if((i < 0) || (i >= list->total_items))
	    return;

	/* If this is the selected item then report it being unselected
	 * first
	 */
	if(list->sel_item == i)
	{
	    list->sel_item = -1;
	    if(list->unselect_cb != NULL)
		list->unselect_cb(
		    list,
		    i,
		    list->unselect_data
		);
	}

	/* Delete the specified item */
	item = list->item[i];
	if(item != NULL)
	{
	    if(item->destroy_cb != NULL)
		item->destroy_cb(item->data);

	    free(item->text);
	    free(item);
	    list->item[i] = item = NULL;
	}

	/* Shift entries */
	for(j = i; j < (list->total_items - 1); j++)
	    list->item[j] = list->item[j + 1];

	/* Reallocate entry pointers */
	list->total_items--;
	if(list->total_items > 0)
	{
	    list->item = (list_item_struct **)realloc(
		list->item,
		list->total_items * sizeof(list_item_struct *)
	    );
	    if(list->item == NULL)
	    {
		list->total_items = 0;
		return;
	    }
	}
	else
	{
	    free(list->item);
	    list->item = NULL;
	    list->total_items = 0;
	}

	/* Update the seleted item index */
	if(list->sel_item > -1)
	{
	    if(i < list->sel_item)
		list->sel_item--;
	}

	SBarQueueResize(
	    list->sb,
	    list->width, list->height,
	    list->width,
	    (list->total_items * list->row_height) +
		(2 * list->top_margin)
	);
	WidgetQueueDraw(list->sb);
	ListWinQueueDraw(list);
}

/*
 *	Deletes all items from the list.
 */
void ListClear(list_struct *list)
{
	int i;
	list_item_struct *item;

	if(list == NULL)
	    return;

	/* If there is an item selected then report it being
	 * unselected first
	 */
	if((list->sel_item >= 0) && (list->sel_item < list->total_items))
	{
	    if(list->unselect_cb != NULL)
		list->unselect_cb(
		    list,
		    list->sel_item,
		    list->unselect_data
		);
	}

	/* Delete all items */
	for(i = 0; i < list->total_items; i++)
	{
	    item = list->item[i];
	    if(item == NULL)
		continue;

	    if(item->destroy_cb != NULL)
		item->destroy_cb(item->data);

	    free(item->text);
	    free(item);
	    list->item[i] = NULL;
	}
	free(list->item);
	list->item = NULL;
	list->total_items = 0;
	list->sel_item = -1;

	SBarSetPositionX(list->sb, 0.0f);
	SBarSetPositionY(list->sb, 0.0f);
	SBarQueueResize(
	    list->sb,
	    list->width, list->height,
	    list->width,
	    (list->total_items * list->row_height) +
		(2 * list->top_margin)
	);
	WidgetQueueDraw(list->sb);
	ListWinQueueDraw(list);
}


/*
 *	Returns the data for the item.
 */
void *ListGetItemData(list_struct *list, int i)
{
	list_item_struct *item;

	if(list == NULL)
	    return(NULL);

	if((i < 0) || (i >= list->total_items))
	    return(NULL);

	item = list->item[i];
	return((item != NULL) ? item->data : NULL);
}

/*
 *	Sets the item data.
 */
void ListSetItemData(
	list_struct *list, int i,
	void *data, void (*destroy_cb)(void *)
)
{
	list_item_struct *item;

	if(list == NULL)
	    return;

	if((i < 0) || (i >= list->total_items))
	    return;

	item = list->item[i];
	if(item == NULL)
	    return;

	/* Call the destroy callback for the previous data as needed */
	if(item->destroy_cb != NULL)
	    item->destroy_cb(item->data);

	item->data = data;
	item->destroy_cb = destroy_cb;
}


/*
 *	Selects the specified item.
 */
void ListSelect(list_struct *list, int i)
{
	if(list == NULL)
	    return;

	if((i < 0) || (i >= list->total_items))
	    return;

	if(i != list->sel_item)
	{
	    if(list->unselect_cb != NULL)
		list->unselect_cb(
		    list,
		    list->sel_item,
		    list->unselect_data
		);

	    list->sel_item = i;
	    ListWinQueueDraw(list);

	    if(list->select_cb != NULL)
		list->select_cb(
		    list,
		    list->sel_item,
		    list->select_data
		);
	}
}

/*
 *	Unselects the specified item.
 */
void ListUnselect(list_struct *list, int i)
{
	if(list == NULL)
	    return;

	if(list->sel_item > -1)
	{
	    if(list->unselect_cb != NULL)
		list->unselect_cb(
		    list,
		    list->sel_item,
		    list->unselect_data
		);

	    list->sel_item = -1;
	    ListWinQueueDraw(list);
	}
}


/*
 *	DND Enter Notify.
 *
 *	Initializes DND values on the list.
 */
void ListWinDNDEnterNotify(list_struct *list, time_t t)
{
	if(list == NULL)
	    return;

	/* Reset DND values */
	list->dnd_action = None;
	list->dnd_drag_line = -1;

	if(list->map_state)
	    ListWinQueueDraw(list);
}

/*
 *	DND Position Notify.
 *
 *	Calculates and draws the drag position on the list.
 */
void ListWinDNDPositionNotify(
	list_struct *list, win_t tar_w,
	int x, int y,
	int root_x, int root_y,
	atom_t action,
	time_t t
)
{
	bool_t need_redraw = False;
	int row_height, prev_drag_line;
	scroll_bar_struct *sb;
	osw_gui_struct *gui = &osw_gui[0];

	if(list == NULL)
	    return;

	row_height = MAX((int)list->row_height, 1);
	sb = list->sb;

	/* Calculate the drag line based on the specified cursor
	 * position
	 */
	prev_drag_line = list->dnd_drag_line;
	if((tar_w == list->toplevel) &&
	   (x >= 0) && (y >= 0) &&
	   (x < (int)list->width) && (y < (int)list->height)
	)
	{
	    list->dnd_drag_line = ((int)SBarGetPositionY(sb) +
		(int)y - (int)list->top_margin + (row_height / 2)) /
		row_height;
	    if(list->dnd_drag_line < 0)
		list->dnd_drag_line = 0;
	}
	else if(tar_w != None)
	{
	    list->dnd_drag_line = -1;
	}
	if(prev_drag_line != list->dnd_drag_line)
	    need_redraw = True;

	/* Are we the DND source? */
	if(LIST_IS_DND_SOURCE(list))
	{
	    int	x_offset = 8, y_offset = 8;
	    win_t dnd_icon = list->dnd_icon;

/* Need to offset the DND icon to the lower right of the cursor so
 * that the pointer position is never reported to be on the DND
 * icon itself
 */
	    /* Move the DND icon to the position of the cursor */
	    OSWMoveWindow(dnd_icon, root_x + x_offset, root_y + y_offset);

	    /* Update the cursor based on the specified action */
	    if(action != list->dnd_action)
	    {
		cursor_t cur;
		if(action == osw_atom.dnd_action_move)
		    cur = gui->drag_move_cursor;
	        else if(action == osw_atom.dnd_action_copy)
		    cur = gui->drag_copy_cursor;
	        else if(action == osw_atom.dnd_action_link)
		    cur = gui->drag_link_cursor;
		else
		    cur = gui->drag_unaccepted_cursor;
		OSWDNDSetCursor(cur, t);
	    }
	}

	/* Update the current action */ 
	list->dnd_action = action;

	if(list->map_state && need_redraw)
	    ListWinQueueDraw(list);
}

/*
 *	DND Leave Notify.
 *
 *	Clears any drag draws.
 */
void ListWinDNDLeaveNotify(list_struct *list, time_t t)
{
	osw_gui_struct *gui = &osw_gui[0];

	if(list == NULL)
	    return;

	/* Are we the DND source? */
	if(LIST_IS_DND_SOURCE(list))
	{
	    /* Update the cursor as needed */
	    if(list->dnd_action != None)
	    {
		cursor_t cur = gui->drag_unaccepted_cursor;
		OSWDNDSetCursor(cur, t);
	    }
	}

	/* Reset DND values */
	list->dnd_action = None;
	list->dnd_drag_line = -1;

	if(list->map_state)
	    ListWinQueueDraw(list);
}


/*
 *	Initialize the list window.
 */
int ListWinInit(
	list_struct *list, win_t parent,
	int x, int y,
	unsigned int width, unsigned int height
)
{
	win_t w;
	scroll_bar_struct *sb;
	widget_global_struct *wglobal = &widget_global;

	if((list == NULL) || (parent == None))
	    return(-2);

	if(width < 2)
	    width = 2;
	if(height < 2)
	    height = 2;

	list->type = WIDGET_TYPE_LIST;
	list->map_state = False;
	list->has_focus = False;
	list->disabled = False;
	list->visibility_state = VisibilityFullyObscured;
	list->x = x;
	list->y = y;
	list->width = width;
	list->height = height;
	list->font = wglobal->std_font;
	list->next = NULL;
	list->prev = NULL;
	list->data = NULL;

	list->options = 0;

	list->row_height = LW_DEF_ROW_HEIGHT;

	list->item = NULL;
	list->total_items = 0;
	list->sel_item = -1;

	list->left_margin = 5;
	list->top_margin = 5;

	list->button_pressed = 0;
	list->button_last_x = 0;
	list->button_last_y = 0;
	list->button_pressed_time = 0l;

	list->drag_item_num = -1;
	list->dnd_icon = None;
	list->dnd_action = None;
	list->dnd_drag_line = -1;

	list->activate_cb = NULL;
	list->activate_data = NULL;
	list->select_cb = NULL;
	list->select_data = NULL;
	list->unselect_cb = NULL;
	list->unselect_data = NULL;

	/* Create toplevel */
	x = list->x;
	y = list->y;
	width = list->width;
	height = list->height;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	list->toplevel = w;
	list->toplevel_buf = 0;
	OSWSetWindowInput(
	    w,
	    StructureNotifyMask | ExposureMask |
	    KeyPressMask | KeyReleaseMask |
	    ButtonPressMask | ButtonReleaseMask | PointerMotionMask |
	    VisibilityChangeMask
	);

	parent = w;

	/* Scroll Bar */
	list->sb = sb = SCROLL_BAR(malloc(sizeof(scroll_bar_struct)));
	SBarInit(
	    sb, w,
	    width, height,
	    width,
	    (list->total_items * list->row_height) +
		(2 * list->top_margin)
	);
	sb->value_changed_cb = ListSBValueChangedCB;
	sb->value_changed_data = list;

	list->cv_prompt = NULL;

	/* Add this widget to the widget regeristry */
	WidgetRegister(list, WIDGET_TYPE_LIST);

	return(0);
}       

/*
 *	Resizes the list window.
 */
void ListWinResize(list_struct *list)
{
	int x, y;
	unsigned int width, height;
	win_t w;

	if(list == NULL)
	    return;

	w = list->toplevel;
	OSWGetWindowGeometry(w, &x, &y, &width, &height, NULL);
	list->x = x;
	list->y = y;
	list->width = width;
	list->height = height;

	OSWDestroyPixmap(&list->toplevel_buf);

	/* Resize scroll bars */
	SBarResize(
	    list->sb,
	    list->width, list->height,
	    list->width,
	    (list->total_items * list->row_height) +
		(2 * list->top_margin)
	);
	WidgetQueueDraw(list->sb);
}

/*
 *	Queues the list window to resize.
 */
void ListWinQueueResize(list_struct *list)
{
	if(list == NULL)
	    return;

	WidgetQueueResizeAction(ListWinResizeCB, list, NULL); 
}

/*
 *	Draws the list window.
 */
void ListWinDraw(list_struct *list)
{
	win_t w;
	pixmap_t pixmap;
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if(list == NULL)
	    return;

	w = list->toplevel;
	if(w == None)
	    return;

	/* Recreate buffers as needed */
	if(list->toplevel_buf == None)
	    OSWCreatePixmap(
		&list->toplevel_buf, list->width, list->height
	    );
	pixmap = list->toplevel_buf;

	/* Begin drawing */
	if(pixmap != None)
	{
	    const int	row_height = (int)MAX(list->row_height, 1),
			total_items = list->total_items;
	    const unsigned int	width = list->width,
				height = list->height;
	    int	i, x, y,
		lbearing, rbearing, swidth, ascent, descent,
		font_height, icon_label_spacing,
		items_drawn, items_visible;
	    const char *name;
	    font_t *font = list->font;
	    pixel_t sel_bg_pix, sel_fg_pix, text_fg_pix;
	    pixmap_t icon_pixmap;
	    bitmap_t icon_mask;
	    list_item_struct	**items_list = list->item,
				*item;
	    unsigned int icon_width, icon_height;
	    scroll_bar_struct *sb = list->sb;

	    /* Draw the background */
	    OSWSetForeground(
		wglobal->force_mono ?
		    gui->black_pix :
		    wglobal->surface_editable_pix
	    );
	    OSWDrawSolidRectangle(pixmap, 0, 0, width, height);

	    /* Begin drawing each item, get starting index, position,
	     * and pointers in list
	     */
	    if((total_items > 0) && (items_list != NULL))
	    {
		/* Get colors */
		if(wglobal->force_mono)
		{
		    sel_bg_pix = gui->white_pix;
		    sel_fg_pix = gui->black_pix;
		    text_fg_pix = gui->white_pix;
		}
		else
		{
		    sel_bg_pix = wglobal->surface_selected_pix;
		    sel_fg_pix = wglobal->selected_text_pix;
		    text_fg_pix = wglobal->editable_text_pix;
		}

	        /* Calculate the first item to draw */
	        i = CLIP(
		    ((int)SBarGetPositionY(sb) / row_height),
		    0, total_items - 1
		);

		/* Calculate the starting coordinates */
		x = (int)list->left_margin;
		y = (int)list->top_margin -
		    ((int)SBarGetPositionY(sb) % row_height);

	        /* Calculate number of visible entries plus 2 (to
		 * account for partially visible items at the top and
		 * bottom)
		 */
		items_visible = ((int)height / row_height) + 2;

		/* Begin drawing each entry */
		items_drawn = 0;
		while((i < total_items) &&
		      (items_drawn < items_visible)
		)
		{
		    item = items_list[i];
		    if(item == NULL)
		    {
			i++;
			continue;
		    }

		    /* Get the item's name */
		    name = item->text;
		    if(name == NULL)
			name = "";
		    OSWSetFont(font);
		    OSWGetTextBounds(
			font, name, -1,
			&lbearing, &rbearing, &swidth,
			&ascent, &descent
		    );
		    font_height = ascent + descent;

		    /* Get the item's icon */
		    icon_pixmap = item->pixmap;
		    icon_mask = item->mask;
		    if(icon_pixmap != None)
		    {
			OSWGetPixmapGeometry(
			    icon_pixmap,
			    &icon_width, &icon_height, NULL
			);
			icon_label_spacing = 2;
		    }
		    else
		    {
			icon_width = 0;
			icon_height = 0;
			icon_label_spacing = 0;
		    }

		    /* Draw drag line at this item? */
		    if(list->dnd_drag_line == i)
		    {
			/* Draw drag line above the item */
			int	x2 = x,
				y2 = y;
			OSWSetForeground(sel_bg_pix);
			OSWDrawLine(
			    pixmap,
			    x2,
			    y2,
			    x2 + MAX(
				((int)icon_width + icon_label_spacing +
				    swidth + 2), 100
			    ),
			    y2
			);
		    }
		    /* Draw line past last item? */
		    else if((list->dnd_drag_line >= total_items) &&
			    ((i + 1) == total_items)
		    )
		    {
			/* Draw drag line below this item */
			int	x2 = x,
				y2 = y + row_height;
			OSWSetForeground(sel_bg_pix);
			OSWDrawLine(
			    pixmap,
			    x2,
			    y2,
			    x2 + MAX(
				((int)icon_width + icon_label_spacing +
				    swidth + 2), 100
			    ),
			    y2
			);
		    }

		    /* Is entry selected and in drag? */
		    if(list->drag_item_num == i)
		    {
			/* Pretend we drew this entry */
			i++;
			y += row_height;
			items_drawn++;
			continue;
		    }

		    /* Selected? */
		    if(list->sel_item == i)
		    {
			/* Draw selected rectangle around this item */
			int	w2 = icon_width + icon_label_spacing +
				    swidth + 2,
				h2 = MIN(
				    (MAX((int)icon_height, font_height) + 2),
				    row_height
				),
				x2 = x - 1,
				y2 = y + ((row_height - h2) / 2);
			OSWSetForeground(sel_bg_pix);
			OSWDrawSolidRectangle(
			    pixmap,
			    x2, y2, w2, h2
			);
			OSWSetForeground(sel_fg_pix);
		    }
		    else
		    {
			OSWSetForeground(text_fg_pix);
		    }

		    /* Draw icon? */
		    if(icon_pixmap != None)
		    {
			int	x2 = x,
				y2 = y + ((row_height - (int)icon_height) / 2);
			OSWSetClipMask(icon_mask);
			OSWSetClipOrigin(x2, y2);
			OSWCopyDrawablesCoord(
			    pixmap, icon_pixmap,
			    x2, y2,
			    icon_width, icon_height,
			    0, 0
			);
			OSWSetClipMask(None);
		    }

		    OSWDrawString( 
			pixmap,
			x + (int)icon_width + icon_label_spacing -
			    lbearing,
			y + ((row_height - font_height) / 2) +
			    ascent,
			name
		    );

		    i++;
		    y += row_height;
		    items_drawn++;
		}
	    }

	    /* Draw the frame */
	    if(wglobal->force_mono)
		WidgetFrameButtonPixmap(
		    pixmap, True, width, height,
		    gui->white_pix,
		    gui->white_pix
		);
	    else
		WidgetFrameButtonPixmap(
		    pixmap, True, width, height,
		    wglobal->surface_highlight_pix,
		    wglobal->surface_shadow_pix
		);

	    if(list->dnd_drag_line > -1)
	    {
		OSWSetForeground(
		    wglobal->force_mono ?
			gui->white_pix : wglobal->surface_selected_pix
		);
		OSWDrawRectangle(pixmap, 0, 0, width - 1, height - 1);
	    }		    

	    OSWPutDrawableToWindow(w, pixmap);
	}
}

/*
 *	Queues the list window to draw.
 */
void ListWinQueueDraw(list_struct *list)
{
	if(list == NULL)
	    return;

	WidgetQueueDrawAction(ListWinDrawCB, list, NULL);
}

/*
 *	Manages the list window event.
 */
int ListWinManage(list_struct *list, event_t *event)
{
	int events_handled = 0;
	win_t w;
	keycode_t keycode;
	key_state_t key_state;
	keyval_t keyval;
	event_key_t *key;
	event_button_t *button;
	event_motion_t *motion;
	event_crossing_t *crossing;
	event_visibility_t *visibility;
	widget_global_struct *wglobal = &widget_global;

	if((list == NULL) || (event == NULL))
	    return(events_handled);

	w = event->xany.window;

	/* If the change values prompt is mapped then it must be able
	 * to managed the event first
	 */
	if(list->map_state && list->has_focus &&
	   (list->cv_prompt_mode != LIST_CV_PROMPT_MODE_NONE)
	)
	{
	    events_handled += WidgetManage(list->cv_prompt, event);
	    if(events_handled > 0)
		return(events_handled);
	}

	switch(event->type)
	{
	  case ConfigureNotify:
	    if(w == list->toplevel)
	    {
		if(list->map_state)
		    ListWinQueueDraw(list);
		events_handled++;
	    }
	    break;

	  case Expose:
	    if(w == list->toplevel)
	    {
		ListWinQueueDraw(list);
		events_handled++;
	    }
	    break;

	  case KeyPress:
	    key = &event->xkey;
	    if(!list->has_focus)
		return(events_handled);

	    /* Do not handle keys if the change value prompt is mapped */
	    if(list->cv_prompt_mode != LIST_CV_PROMPT_MODE_NONE)
		break;

	    keycode = key->keycode;
	    key_state = key->state;
	    keyval = OSWGetKeyValFromKeyCode(keycode, key_state);

	    /* Dragging? */
	    if(LIST_IS_DND_SOURCE(list))
	    {
		atom_t action = OSWDNDQueryAction(
		    list->toplevel,
		    osw_atom.dnd_action_move,
		    OSW_DND_ACTION_MOVE
		);
		ListWinDNDPositionNotify(
		    list, None,
		    (int)key->x, (int)key->y,
		    (int)key->x_root, (int)key->y_root,
		    action,
		    (time_t)key->time
		);
		break;
	    }

	    /* Up */
	    if((keyval == OSW_KEY_Up) ||
	       (keyval == OSW_KEY_KP_Up)
	    )
	    {
		const int prev_sel_item = list->sel_item;

		/* Select the previous item */
		list->sel_item = MAX(
		    (prev_sel_item - 1),
		    0
		);

		/* Scroll to the selected item if it is not fully
		 * visible
		 */
		if(ListGetItemVisibility(
		    list, list->sel_item) != VisibilityUnobscured
		)
		    ListMoveTo(list, list->sel_item, 0.0f);

		/* Report selection if it has changed */
		if(list->sel_item != prev_sel_item)
		{
		    if((prev_sel_item > -1) &&
		       (list->unselect_cb != NULL)
		    )
			list->unselect_cb(
			    list,
			    prev_sel_item,
			    list->unselect_data
			);
		    if(list->select_cb != NULL)
			list->select_cb(
			    list,
			    list->sel_item,
			    list->select_data
			);
		}
		ListWinQueueDraw(list);
		events_handled++;
	    }
	    /* Down */
	    else if((keyval == OSW_KEY_Down) ||
		    (keyval == OSW_KEY_KP_Down)
	    )
	    {
		const int prev_sel_item = list->sel_item;

		/* Select the next item */
		list->sel_item = MIN(
		    (prev_sel_item + 1),
		    MAX((list->total_items - 1), 0)
		);

		/* Scroll to the selected item if it is not fully
		 * visible
		 */
		if(ListGetItemVisibility(
		    list, list->sel_item) != VisibilityUnobscured
		)
		    ListMoveTo(list, list->sel_item, 1.0f);

		/* Report selection if it has changed */
		if(list->sel_item != prev_sel_item)
		{
		    if((prev_sel_item > -1) && 
		       (list->unselect_cb != NULL) 
		    )
			list->unselect_cb(
			    list,
			    prev_sel_item,
			    list->unselect_data
			);
		    if(list->select_cb != NULL)
			list->select_cb(
			    list,
			    list->sel_item,
			    list->select_data
			);
		}
		ListWinQueueDraw(list);
		events_handled++;
	    }
	    /* Page Up */
	    else if((keyval == OSW_KEY_Page_Up) ||
		    (keyval == OSW_KEY_KP_Page_Up)
	    )
	    {
		const int	prev_sel_item = list->sel_item,
				row_height = MAX((int)list->row_height, 1),
				height = (int)list->height -
				    (2 * list->top_margin),
				di = MAX(((height / row_height) - 1), 1);

		/* Select the previous page item */
		list->sel_item = MAX( 
		    (prev_sel_item - di),
		    0
		);

		/* Scroll to the selected item if it is not fully
		 * visible
		 */
		if(ListGetItemVisibility(
		    list, list->sel_item) != VisibilityUnobscured
		)
		    ListMoveTo(list, list->sel_item, 0.0f);

		/* Report selection if it has changed */
		if(list->sel_item != prev_sel_item)
		{
		    if((prev_sel_item > -1) &&
		       (list->unselect_cb != NULL)
		    )
			list->unselect_cb(
			    list,
			    prev_sel_item,
			    list->unselect_data
			);
		    if(list->select_cb != NULL)
			list->select_cb(
			    list,
			    list->sel_item,
			    list->select_data
			);
		}
		ListWinQueueDraw(list);
		events_handled++;
	    }
	    /* Page Down */
	    else if((keyval == OSW_KEY_Page_Down) ||
		    (keyval == OSW_KEY_KP_Page_Down)
	    )
	    {
		const int	prev_sel_item = list->sel_item,
				row_height = MAX((int)list->row_height, 1),
				height = (int)list->height -
				    (2 * list->top_margin),
				di = MAX(((height / row_height) - 1), 1);

		/* Select the next page item */
		list->sel_item = MIN(
		    (prev_sel_item + di),
		    MAX((list->total_items - 1), 0)
		);

		/* Scroll to the selected item if it is not fully
		 * visible
		 */
		if(ListGetItemVisibility(
		    list, list->sel_item) != VisibilityUnobscured
		)
		    ListMoveTo(list, list->sel_item, 1.0f);

		/* Report selection if it has changed */
		if(list->sel_item != prev_sel_item)
		{
		    if((prev_sel_item > -1) &&
		       (list->unselect_cb != NULL)
		    )
			list->unselect_cb(
			    list,
			    prev_sel_item,
			    list->unselect_data
			);
		    if(list->select_cb != NULL)
			list->select_cb(
			    list,
			    list->sel_item,
			    list->select_data
			);
		}
		ListWinQueueDraw(list);
		events_handled++;
	    }
	    /* Home */
	    else if((keyval == OSW_KEY_Home) ||
		    (keyval == OSW_KEY_KP_Home)
	    )
	    {
		const int	prev_sel_item = list->sel_item;

		/* Select the first item */
		list->sel_item = 0;

		/* Scroll to the selected item if it is not fully
		 * visible
		 */
		if(ListGetItemVisibility(
		    list, list->sel_item) != VisibilityUnobscured
		)
		    ListMoveTo(list, list->sel_item, 0.0f);

		/* Report selection if it has changed */
		if(list->sel_item != prev_sel_item)
		{
		    if((prev_sel_item > -1) &&
		       (list->unselect_cb != NULL)
		    )
			list->unselect_cb(
			    list,
			    prev_sel_item,
			    list->unselect_data
			);
		    if(list->select_cb != NULL)
			list->select_cb(
			    list,
			    list->sel_item,
			    list->select_data
			);
		}
		ListWinQueueDraw(list);
		events_handled++;
	    }
	    /* End */
	    else if((keyval == OSW_KEY_End) ||
		    (keyval == OSW_KEY_KP_End)
	    )
	    {
		const int       prev_sel_item = list->sel_item;

		/* Select the last item */
		list->sel_item = MAX((list->total_items - 1), 0);

		/* Scroll to the selected item if it is not fully
		 * visible
		 */
		if(ListGetItemVisibility(
		    list, list->sel_item) != VisibilityUnobscured
		)
		    ListMoveTo(list, list->sel_item, 1.0f);

		/* Report selection if it has changed */
		if(list->sel_item != prev_sel_item)
		{
		    if((prev_sel_item > -1) &&
		       (list->unselect_cb != NULL)
		    )
			list->unselect_cb(
			    list,
			    prev_sel_item,
			    list->unselect_data
			);
		    if(list->select_cb != NULL)   
			list->select_cb(
			    list,
			    list->sel_item,
			    list->select_data
			);
		}
		ListWinQueueDraw(list);
		events_handled++;
	    }
	    /* Rename */
	    else if(keyval == OSW_KEY_F2)
	    {
		if(list->sel_item > -1)
		{
		     ListWinMapCVPrompt(list, LIST_CV_PROMPT_MODE_RENAME);
		     events_handled++;
		}
	    }
	    /* Enter */
	    else if((keyval == OSW_KEY_Return) ||
		    (keyval == OSW_KEY_KP_Enter) ||
		    (keyval == OSW_KEY_ISO_Enter)
	    )
	    {
		/* Call activate callback */
		if(list->activate_cb != NULL)
		{
		    list->activate_cb(
			list,
			list->activate_data	/* Data */
		    );
		    if(!WidgetIsRegistered(list))
		    {
			events_handled++;
			return(events_handled);
		    }
		    events_handled++;
		}
	    }
	    break;

	  case KeyRelease:
	    key = &event->xkey;
	    if(!list->has_focus)
		return(events_handled);

	    /* Do not handle keys if the change value prompt is mapped */
	    if(list->cv_prompt_mode != LIST_CV_PROMPT_MODE_NONE)
		break;

	    /* Dragging? */
	    if(LIST_IS_DND_SOURCE(list))
	    {
		atom_t action = OSWDNDQueryAction(
		    list->toplevel,
		    osw_atom.dnd_action_move,
		    OSW_DND_ACTION_MOVE
		);
		ListWinDNDPositionNotify(
		    list, None,
		    (int)key->x, (int)key->y,
		    (int)key->x_root, (int)key->y_root,
		    action,
		    (time_t)key->time
		);
		break;
	    }

	    break;

	  case ButtonPress:
	    button = &event->xbutton;
	    if(w == list->toplevel)
	    {
		const int	button_num = (int)button->button,
				x = (int)button->x,
				y = (int)button->y,
				prev_sel_item = list->sel_item;
		time_t dt = (time_t)button->time - list->button_pressed_time;
		scroll_bar_struct *sb = list->sb;

		list->has_focus = True;

		/* Record button state */
		list->button_pressed = button_num;
		list->button_last_x = x;
		list->button_last_y = y;

		/* Update last click time */
		list->button_pressed_time = (time_t)button->time;

		/* Unmap the change values prompt as needed */
		if(list->cv_prompt_mode != LIST_CV_PROMPT_MODE_NONE)
		    ListWinUnmapCVPrompt(list);

		switch(button_num)
		{
		  case Button1:
		  case Button3:
		    /* Calculate the selected item number */
		    list->sel_item = ListGetItemFromCoordinates(
			list, x, y
		    );
		    break;
		  case Button4:
		    SBarSetPositionY(
			sb,
			SBarGetPositionY(sb) - sb->y_step_inc
		    );
		    break;
		  case Button5:
		    SBarSetPositionY(
			sb,
			SBarGetPositionY(sb) + sb->y_step_inc
		    );
		    break;
		}

		/* Report selection if it has changed */
		if(list->sel_item != prev_sel_item)
		{
		    if((prev_sel_item > -1) &&
		       (list->unselect_cb != NULL)
		    )
			list->unselect_cb(
			    list,
			    prev_sel_item,
			    list->unselect_data
			);
		    if((list->sel_item > -1) &&
		       (list->select_cb != NULL)
		    )
			list->select_cb(
			    list,
			    list->sel_item,
			    list->select_data
		        );
		}

		/* Scroll to the selected item if it is not fully
		 * visible
		 */
		if(list->sel_item > -1)
		{
		    if(ListGetItemVisibility(
			list, list->sel_item) != VisibilityUnobscured
		    )
			ListMoveTo(list, list->sel_item, 0.5f);
		}

		/* Double clicked? */
		if((dt > 0l) && (dt < wglobal->double_click_int))
		{
		    /* Activate? */
		    if((button_num == Button1) &&
		       (list->sel_item > -1) &&
		       (list->sel_item == prev_sel_item)
		    )
		    {
			/* Call activate callback */
			if(list->activate_cb != NULL)
			    list->activate_cb(
				list,
				list->activate_data	/* Data */
			    );
			if(!WidgetIsRegistered(list))
			{
			    events_handled++;
			    return(events_handled);
			}
		    }
		}
		/* Slow double click? */
		else if((dt > 0l) && (dt < wglobal->rename_click_int))
		{
		    /* Rename? */
		    if((button_num == Button1) &&
		       (list->sel_item > -1) &&
		       (list->sel_item == prev_sel_item)
		    )
		    {
			ListWinMapCVPrompt(list, LIST_CV_PROMPT_MODE_RENAME);
		    }
		}
		ListWinQueueDraw(list);
		events_handled++;
	    }
	    else
	    {
		/* Unfocus the list as needed but do not report that    
		 * the event was handled
		 */
		if(list->has_focus)
		{
		    list->has_focus = False;
		    if(list->map_state)
			ListWinQueueDraw(list);
		}
	    }
	    break;

	  case ButtonRelease:
	    button = &event->xbutton;
	    if(w == list->toplevel)
	    {
		bool_t drop_successful = False;
		int	x = (int)button->x,
			y = (int)button->y;

		list->button_pressed = 0;

		/* Currently dragging an item and released on the
		 * list?
		 */
		if((list->drag_item_num >= 0) &&
		   (list->drag_item_num < list->total_items) &&
		   (x >= 0) && (y >= 0) &&
		   (x < (int)list->width) && (y < (int)list->height)
		)
		{
		    /* Calculate which position ButtonRelease was on */
		    int	src_item_num = list->drag_item_num,
			tar_row = list->dnd_drag_line,
			tar_item_num = CLIP(
			    tar_row,
			    0, (list->total_items - 1)
			);
		    list_item_struct *src_item = list->item[src_item_num];
		    if((src_item != NULL) &&
		       (tar_item_num >= 0) &&
		       (tar_item_num < list->total_items)
		    )
		    {
			if(list->dnd_action == osw_atom.dnd_action_move)
			{
			    /* Shift positions */
			    if(src_item_num > tar_item_num)
			    {
			        int i;
			        for(i = src_item_num; i > tar_item_num; i--)
				    list->item[i] = list->item[i - 1];
			    }
			    else if(src_item_num < tar_item_num)
			    {
			        int i;
			        if((tar_item_num > 0) &&
			           (tar_row < list->total_items)
			        )
				    tar_item_num--;
			        for(i = src_item_num; i < tar_item_num; i++)
				    list->item[i] = list->item[i + 1];
			    }

			    /* Set item in its new position */
			    list->item[tar_item_num] = src_item;
			}
			else if(list->dnd_action == osw_atom.dnd_action_copy)
			{
/* TODO */
			}

			if(tar_item_num != list->sel_item)
			{
			    /* Report the previous item being unselected */
			    if(list->unselect_cb != NULL)
				list->unselect_cb(
				    list,
				    list->sel_item,
				    list->unselect_data
			        );

			    list->sel_item = tar_item_num;

			    /* Report the new item being selected */
			    if(list->select_cb != NULL)
				list->select_cb(
				    list,
				    list->sel_item,
				    list->select_data
				);
			}

			drop_successful = True;
		    }
		}

		/* End drag as needed */
		if(LIST_IS_DND_SOURCE(list))
		{
		    time_t t = (time_t)button->time;
		    ListWinDNDLeaveNotify(list, t);
		    ListWinDragStop(
			list, drop_successful, t
		    );
		}
		ListWinQueueDraw(list);
		events_handled++;
	    }
	    break;

	  case MotionNotify:
	    motion = &event->xmotion;
	    if(w == list->toplevel)
	    {
		int	button_num = list->button_pressed,
			x = (int)motion->x,
			y = (int)motion->y,
			root_x = (int)motion->x_root,
			root_y = (int)motion->y_root,
			sel_item = list->sel_item;

		/* Not previously in a DND drag and there is an item
		 * selected so that we may have to start a DND drag?
		 */
		if(!LIST_IS_DND_SOURCE(list) &&
		   (sel_item > -1) &&
		   (button_num == Button1) &&
		   (list->options & CLIST_FLAG_ALLOW_DRAG)
		)
		{
		    /* Check drag tolorance */
		    if((x < (list->button_last_x - (LW_DRAG_TOLORANCE / 2))) ||
		       (x >= (list->button_last_x + (LW_DRAG_TOLORANCE / 2))) ||
		       (y < (list->button_last_y - (LW_DRAG_TOLORANCE / 2))) ||
		       (y >= (list->button_last_y + (LW_DRAG_TOLORANCE / 2)))
		    )
		    {
			/* Start a DND drag with this list as the DND
			 * drag source
			 */
			time_t t = (time_t)motion->time;
			atom_t action = OSWDNDQueryAction(
			    w,
			    osw_atom.dnd_action_move,
			    OSW_DND_ACTION_MOVE
			);

			/* Begin drag */
			ListWinDragBegin(
			    list, sel_item,
			    x, y, t
			);
			ListWinDNDEnterNotify(list, t);
			ListWinDNDPositionNotify(
			    list, w,
			    x, y, root_x, root_y,
			    action,
			    t
			);
		    }
		    events_handled++;
		    break;
		}
	    }

	    /* Dragging? */
	    if(LIST_IS_DND_SOURCE(list))
	    {
		int	x = (int)motion->x,
			y = (int)motion->y,
			root_x = (int)motion->x_root,
			root_y = (int)motion->y_root;
		ListWinDNDPositionNotify(
		    list, w,
		    x, y, root_x, root_y,
		    list->dnd_action,
		    (time_t)motion->time
		);
	    }
	    break;

	  case EnterNotify:
	    crossing = &event->xcrossing;
	    if(w == list->toplevel)
	    {
		WidgetTimeoutRemoveByFuncWidget(ListDragScrollCB, list);
		if(LIST_IS_DND_SOURCE(list))
		{
		    const time_t t = (time_t)crossing->time;
		    atom_t action = OSWDNDQueryAction(
			w,
			osw_atom.dnd_action_move,
			OSW_DND_ACTION_MOVE
		    );
		    ListWinDNDEnterNotify(list, t);
		    ListWinDNDPositionNotify(
			list, w,
			(int)crossing->x, (int)crossing->y,
			(int)crossing->x_root, (int)crossing->y_root,
			action,
			t
		    );
		}
	    }
	    break;

	  case LeaveNotify:
	    crossing = &event->xcrossing;
	    if(w == list->toplevel)
	    {
		WidgetTimeoutRemoveByFuncWidget(ListDragScrollCB, list);
		if(LIST_IS_DND_SOURCE(list))
		{
		    const time_t t = (time_t)crossing->time;
		    WidgetIdleAdd(
			ListDragScrollCB,
			list, NULL
		    );
		    ListWinDNDLeaveNotify(list, t);
		}
	    }
	    break;

	  case FocusIn:
	    if(w == list->toplevel)
	    {
		list->has_focus = True;
		events_handled++;
	    }
	    break;

	  case FocusOut:
	    if(w == list->toplevel)
	    {
		list->has_focus = False;
		events_handled++;
	    }
	    break;

	  case VisibilityNotify:
	    visibility = &event->xvisibility;
	    if(w == list->toplevel)
	    {
		list->visibility_state = visibility->state;
		events_handled++;
	    }
	    break;
	}

	if(events_handled == 0)
	    events_handled += WidgetManage(list->sb, event);

	return(events_handled);
}

/*
 *	Applies changes on the list's change values prompt.
 */
static void ListWinCVPromptApply(list_struct *list, const char *s)
{
	int i;
	list_item_struct *item;

	if((list == NULL) || (s == NULL))
	    return;

	if(list->cv_prompt_mode == LIST_CV_PROMPT_MODE_NONE)
	    return;

	i = list->cv_prompt_entry_pos;
	if(!ListIsItemAllocated(list, i))
	    return;

	item = list->item[i];
	free(item->text);
	item->text = STRDUP(s);
}

/*
 *	Maps the list's change values prompt.
 */
static void ListWinMapCVPrompt(list_struct *list, int mode)
{
	int i, x, y, row_height, icon_label_spacing;
	unsigned int	width, height,
			icon_width, icon_height;
	list_item_struct *item;
	scroll_bar_struct *sb;
	prompt_struct *prompt;

	if(list == NULL)
	    return;

	row_height = (int)MAX(list->row_height, 1);
	sb = list->sb;

	/* Already changing values in another mode? */
	if(list->cv_prompt_mode != LIST_CV_PROMPT_MODE_NONE)
	    return;

	/* Get selected item */
	i = list->sel_item;
	if(!ListIsItemAllocated(list, i))
	    return;

	item = list->item[i];

	/* Recreate the change value prompt */
	WidgetDestroy(list->cv_prompt);
	free(list->cv_prompt);
	list->cv_prompt = prompt = PROMPT(malloc(sizeof(prompt_struct)));
	x = 0;
	y = 0;
	width = 0;
	height = 0;
	PromptInit(
	    prompt, list->toplevel,
	    x, y, width, height,
	    PROMPT_STYLE_NOBORDER,
	    NULL,
	    1024, 0
	);
	prompt->style = PROMPT_STYLE_NOBORDER;
	prompt->activate_cb = ListCVPromptActivateCB;
	prompt->activate_data = list;
	prompt->cancel_cb = ListCVPromptCancelCB;
	prompt->cancel_data = list;

	/* Make sure that the list is in focus */
	WidgetSetFocus(list, True);

	/* Record selected entry for change values prompt */
	list->cv_prompt_entry_pos = list->sel_item;

	switch(mode)
	{
	  case LIST_CV_PROMPT_MODE_RENAME:
	    /* Get the size of the icon */
	    if(item->pixmap != None)
	    {
		pixmap_t pixmap = item->pixmap;
		OSWGetPixmapGeometry(
		    pixmap, &icon_width, &icon_height, NULL
		);
		icon_label_spacing = 2;
	    }
	    else
	    {
		icon_width = 0;
		icon_height = 0;
		icon_label_spacing = 0;
	    }

	    x = list->left_margin + (int)icon_width +
		icon_label_spacing;
	    y = list->top_margin + (row_height * i) -
		(int)SBarGetPositionY(sb) +
		((row_height - LW_PROMPT_HEIGHT) / 2);
	    width = MAX((int)list->width - x - SCROLLBAR_YBAR_WIDTH, 100);
	    height = LW_PROMPT_HEIGHT;
	    OSWMoveResizeWindow(prompt->toplevel, x, y, width, height);
	    WidgetQueueResize(prompt);

	    PromptSetS(prompt, item->text);
	    PromptMarkBuffer(prompt, PROMPT_POS_END);

	    break;
	}

	list->cv_prompt_mode = mode;
	WidgetMap(prompt);
	WidgetSetFocus(prompt, True);
}

/*
 *	Unmaps the list's change values prompt.
 */
static void ListWinUnmapCVPrompt(list_struct *list)
{
	if(list == NULL)
	    return;

	WidgetDestroy(list->cv_prompt);
	free(list->cv_prompt);
	list->cv_prompt = NULL;

	list->cv_prompt_mode = LIST_CV_PROMPT_MODE_NONE;
	list->cv_prompt_entry_pos = -1;
}

/*
 *	Maps the list.
 */
void ListWinMap(list_struct *list)
{
	if(list == NULL)
	    return;

	list->button_pressed = 0;
	list->button_pressed_time = 0l;
	list->drag_item_num = -1;

	if(!list->map_state)
	{
	    OSWMapWindow(list->toplevel);
	    list->map_state = True;
	    WidgetQueueDraw(list->sb);
	}
}

/*
 *	Unmaps the list.
 */
void ListWinUnmap(list_struct *list)
{
	if(list == NULL)
	    return;

	/* Is this list the DND drag source? */
	if(LIST_IS_DND_SOURCE(list))
	{
	    /* Stop the drag */
	    time_t t = (time_t)CurrentTime;
	    ListWinDNDLeaveNotify(list, t);
	    ListWinDragStop(list, False, t);
	}

	OSWUnmapWindow(list->toplevel);
	list->map_state = False;

	list->button_pressed = 0;
	list->button_pressed_time = 0l;
	list->drag_item_num = -1;

	/* Destroy buffers */
	OSWDestroyPixmap(&list->toplevel_buf);
}

/*
 *	Destroys the list.
 */
void ListWinDestroy(list_struct *list)
{  
	if(list == NULL)
	    return;

	/* Is this list the DND drag source? */
	if(LIST_IS_DND_SOURCE(list))
	{
	    /* Stop the drag */
	    time_t t = (time_t)CurrentTime;
	    ListWinDNDLeaveNotify(list, t);
	    ListWinDragStop(list, False, t);
	}

	/* Delete all items */
	ListClear(list);

	WidgetDestroy(list->cv_prompt);
	free(list->cv_prompt);
	list->cv_prompt = NULL;
	WidgetDestroy(list->sb);
	free(list->sb);
	list->sb = NULL;
	OSWDestroyWindow(&list->toplevel);
	OSWDestroyPixmap(&list->toplevel_buf);

	list->type = WIDGET_TYPE_UNKNOWN;
	list->map_state = False;
	list->has_focus = False;
	list->disabled = False;
	list->visibility_state = VisibilityFullyObscured;
	list->x = 0;
	list->y = 0;
	list->width = 0;
	list->height = 0;
	list->font = NULL;
	list->next = NULL;
	list->prev = NULL;
	list->data = NULL;

	list->options = 0;

	list->row_height = 0;

	list->left_margin = 0;
	list->top_margin = 0;
			 
	list->button_pressed = 0;
	list->button_last_x = 0;
	list->button_last_y = 0;
	list->button_pressed_time = 0l;

	list->drag_item_num = -1;
	OSWDestroyWindow(&list->dnd_icon);

	list->activate_cb = NULL;
	list->activate_data = NULL;
	list->select_cb = NULL;
	list->select_data = NULL;
	list->unselect_cb = NULL;
	list->unselect_data = NULL;
}

/*
 *	Creates a DND icon window based on the specified list item's
 *	text and pixmap.
 */
static win_t ListWinCreateDNDIcon(
	list_struct *list, int item_num,
	int x, int y
)
{
#if defined(HAVE_X)

	int	lbearing, rbearing, swidth, ascent, descent,
		font_height;
	unsigned int	width, height,
			icon_width, icon_height,
			icon_spacing;
	const char *s;
	display_t *dpy;
	gc_t gc;
	gc_val_t gcv;
	font_t *font;
	bitmap_t mask, icon_mask;
	pixmap_t pixmap, icon_pixmap;
	win_t w;
	XSetWindowAttributes swattr;
	list_item_struct *item;
	osw_gui_struct *gui = &osw_gui[0];

	if(list == NULL)
	    return(None);

	if(!ListIsItemAllocated(list, item_num))
	    return(None);

	dpy = gui->display;
	item = list->item[item_num];
	font = list->font;
	width = 0;
	height = 0;
	icon_pixmap = item->pixmap;
	icon_mask = item->mask;
	s = item->text;

	/* Calculate the size of the icon to create */
	if(icon_pixmap != None)
	{
	    OSWGetPixmapGeometry(
		icon_pixmap, &icon_width, &icon_height, NULL
	    );
	    icon_spacing = STRISEMPTY(s) ? 0 : 2;
	}
	else
	{
	    icon_width = 0;
	    icon_height = 0;
	    icon_spacing = 0;
	}

	/* Calculate the size of the text to create */
	if(s != NULL)
	{
	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, &rbearing, &swidth, &ascent, &descent
	    );
	}
	else
	{
	    lbearing = rbearing = swidth = ascent = descent = 0;
	}
	font_height = ascent + descent;

	/* Calculate the size of the drag icon window */
	width = MAX(icon_width + icon_spacing + swidth + 2, 2);
	height = MAX(MAX((int)icon_height, font_height + 2), 2);

	/* Create the mask */
	if(OSWCreateBitmap(&mask, width, height))
	    return(None);

	/* Create the gc for drawing the mask */
	gcv.function = GXset;
	gcv.plane_mask = 1;			/* 1 bit */
	gcv.font = font->xfont->fid;
	gcv.foreground = gui->white_pix;
	gcv.background = gui->black_pix;
	gc = XCreateGC(
	    dpy, mask,
	    GCFunction | GCPlaneMask | GCFont |
		GCForeground | GCBackground,
	    &gcv
	);
	if(gc == None)
	{
	    OSWDestroyBitmap(&mask);
	    return(None);
	}

	/* Create the pixmap */
	if(OSWCreatePixmap(&pixmap, width, height))
	{
	    XFreeGC(dpy, gc);
	    OSWDestroyBitmap(&mask);
	    return(None);
	}

	/* Clear the mask */
	XSetFunction(dpy, gc, GXclear);
	XFillRectangle(dpy, mask, gc, 0, 0, width, height);

	/* Draw the icon? */
	if(icon_pixmap != None)
	{
	    int	x = 0,
		y = 0;
	    OSWSetClipMask(icon_mask);
	    OSWSetClipOrigin(x, y);
	    OSWCopyDrawablesCoord(
		pixmap, icon_pixmap,
		x, y,
		icon_width, icon_height,
		0, 0
	    );
	    OSWSetClipMask(None);

	    XSetClipMask(dpy, gc, icon_mask);
	    XSetClipOrigin(dpy, gc, x, y);
	    XSetFunction(dpy, gc, GXset);
	    XFillRectangle(
		dpy, mask, gc,
		x, y, icon_width, icon_height
	    );
	    XSetClipMask(dpy, gc, None);
	}

	/* Draw the text? */
	if(s != NULL)
	{
	    int	x = (int)icon_width + (int)icon_spacing,
		y = ((int)height - (int)font_height) / 2;

	    XSetFunction(dpy, gc, GXset);
#define DRAW_STRING(_x_, _y_)		\
 XDrawString(				\
  dpy, mask, gc,			\
  (_x_) - lbearing, (_y_) + ascent,	\
  s, strlen(s)				\
 )
	    DRAW_STRING(x + 1, y + 1);
	    DRAW_STRING(x + 1, y);
	    DRAW_STRING(x + 2, y + 1);
	    DRAW_STRING(x + 1, y + 2);
	    DRAW_STRING(x, y + 1);
#undef DRAW_STRING

	    /* Draw the string on the label */
	    OSWSetForeground(gui->black_pix);
	    OSWSetFont(font);
#define DRAW_STRING(_x_, _y_)		\
 OSWDrawString(				\
  pixmap,				\
  (_x_) - lbearing, (_y_) + ascent,	\
  s					\
 )
	    DRAW_STRING(x + 1, y);
	    DRAW_STRING(x + 2, y + 1);
	    DRAW_STRING(x + 1, y + 2);
	    DRAW_STRING(x, y + 1);
	    OSWSetForeground(gui->white_pix);
	    DRAW_STRING(x + 1, y + 1);
#undef DRAW_STRING
	}

	XFreeGC(dpy, gc);

	if(OSWCreateWindow(
	    &w, gui->root_win,
	    x, y, width, height
	))
	{
	    OSWDestroyBitmap(&mask);
	    OSWDestroyPixmap(&pixmap);
	    return(None);
	}
	swattr.background_pixmap = pixmap;
	swattr.save_under = True;
	swattr.override_redirect = True;
	XChangeWindowAttributes(
	    dpy, w,
	    CWBackPixmap | CWSaveUnder | CWOverrideRedirect,
	    &swattr
	);
	XShapeCombineMask(
	    dpy, (XID)w,
	    ShapeBounding,
	    0, 0,
	    mask,
	    ShapeSet
	);

	OSWDestroyBitmap(&mask);
	OSWDestroyBitmap(&pixmap);

	return(w);
#else
	return(NULL);
#endif
}


/*
 *	Starts a DND drag and with the specified list window as the
 *	DND source.
 */
void ListWinDragBegin(
	list_struct *list, int item_num,
	int x, int y,
	time_t t
)
{
	win_t w;
	cursor_t cur;
	osw_gui_struct *gui = &osw_gui[0];

	if(list == NULL)
	    return;

	if((item_num < 0) || (item_num >= list->total_items))
	    return;

	/* Unmap change values prompt (as needed) */
	if(list->cv_prompt_mode != LIST_CV_PROMPT_MODE_NONE)
	    ListWinUnmapCVPrompt(list);

	/* Set up drag values */
	list->dnd_action = None;
	list->drag_item_num = item_num;

	/* Recreate the drag icon */
	OSWDestroyWindow(&list->dnd_icon);
	list->dnd_icon = w = ListWinCreateDNDIcon(
	    list, item_num, x, y
	);
	if(w != None)
	    OSWMapRaised(w);

	/* Grab pointer */
	cur = gui->drag_unaccepted_cursor;
	OSWGrabPointerTime(
	    list->toplevel,
	    True,
	    ButtonReleaseMask | PointerMotionMask |
	    EnterWindowMask | LeaveWindowMask,
	    GrabModeAsync, GrabModeAsync,
	    None,
	    cur,
	    t
	);

	if(list->map_state)
	    ListWinQueueDraw(list);
}

/*
 *	Stops a DND drag with the specified list being the DND source.
 *
 *	This should be called after a drop or cancellation when the
 *	specified list is the DND source.
 */
void ListWinDragStop(
	list_struct *list, bool_t successful,
	time_t t
)
{
	osw_gui_struct *gui = &osw_gui[0];

	if(list == NULL)
	    return;

	/* Drop not successful? */
	if(!successful)
	{
	    /* Move the DND icon back to the DND source position */
	    if(LIST_IS_DND_SOURCE(list) &&
	       (list->dnd_icon != None)
	    )
	    {
		int	i = list->drag_item_num,
			x, y, orig_x, orig_y;
		win_t w = list->dnd_icon;
		scroll_bar_struct *sb = list->sb;

		/* Calculate the DND source position */
		OSWTranslateCoordinatesToParent(
		    list->toplevel, gui->root_win,
		    0, 0,
		    &orig_x, &orig_y
		);
		orig_x += (int)list->left_margin;
		orig_y += (i * (int)list->row_height) +
		    (int)list->top_margin - (int)SBarGetPositionY(sb);

		/* Get the DND icon's current position */
		OSWGetWindowGeometry(w, &x, &y, NULL, NULL, NULL);

		/* Move the DND icon back to the DND source position */
		OSWDNDIconAnimateReturn(w, x, y, orig_x, orig_y);
	    }
	}

	OSWUngrabPointerTime(t);

	WidgetTimeoutRemoveByFuncWidget(ListDragScrollCB, list);

	/* Destroy the DND icon window */
	OSWDestroyWindow(&list->dnd_icon);

	list->dnd_action = None;
	list->drag_item_num = -1;

	if(list->map_state)
	    ListWinQueueDraw(list);
}
