#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "../include/widget.h"


static void MenuResizeCB(void *widget, void *data);
static void MenuDrawCB(void *widget, void *data);

int MenuIsItemAllocated(menu_struct *menu, const int item_num);
int MenuGetItemNumberByID(menu_struct *menu, const int id);
bool_t MenuGetItemState(menu_struct *menu, const int item_num);
void MenuSetItemState(menu_struct *menu, const int item_num, bool_t state);
void MenuSetItemAccelKey(
	menu_struct *menu, const int item_num,
	const keyval_t accel_key, const key_state_t accel_mods
);

int MenuItemInsert(
	menu_struct *menu, int item_num,
	int type,
	const char *text,
	pixmap_t pixmap, bitmap_t mask,
	int id
);
int MenuItemAppend(
	menu_struct *menu,
	int type,
	const char *text, 
	pixmap_t pixmap, bitmap_t mask,
	int id
);
void MenuDeleteAllItems(menu_struct *menu);

static bool_t MenuGetMinSize(
	menu_struct *menu,
	unsigned int *width_rtn, unsigned int *height_rtn
);

int MenuInit(
	menu_struct *menu, win_t parent,
	void (*activate_cb)(menu_struct *, int, void *),
	void *activate_data
);
void MenuResize(menu_struct *menu);
void MenuQueueResize(menu_struct *menu);
void MenuDraw(menu_struct *menu);
void MenuQueueDraw(menu_struct *menu);
int MenuManage(menu_struct *menu, event_t *event);
void MenuMap(menu_struct *menu);
void MenuMapCoordinates(menu_struct *menu, int x, int y);
void MenuUnmap(menu_struct *menu);
void MenuDestroy(menu_struct *menu);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Sizes (in pixels):
 */
#define MENU_ROW_HEIGHT			25

#define MENU_HR_HEIGHT			6

#define MENU_FRAME_MARGIN		2
#define MENU_ITEM_MARGIN_X		5

#define MENU_ACCELERATOR_MARGIN		20


/*
 *	Resize callback.
 */
static void MenuResizeCB(void *widget, void *data)
{
	menu_struct *menu = (menu_struct *)widget;
	if(menu == NULL)                          
	    return;

	MenuResize(menu);
}

/*
 *	Draw callback.
 */
static void MenuDrawCB(void *widget, void *data)
{
	menu_struct *menu = (menu_struct *)widget;
	if(menu == NULL)
	    return;

	if(menu->map_state)
	    MenuDraw(menu);
}


/*
 *	Checks if item n is allocated on menu.
 */
int MenuIsItemAllocated(menu_struct *menu, const int item_num)
{
	if(menu == NULL)
	    return(0);
	else if((item_num < 0) || (item_num >= menu->total_items))
	    return(0);
	else if(menu->item[item_num] == NULL)
	    return(0);
	else
	    return(1);
}

/*
 *	Returns the item number on the menu that matches the
 *	given id code. Can return -1 for no match or error.
 */
int MenuGetItemNumberByID(menu_struct *menu, const int id)
{
	int i;

	if(menu == NULL)
	    return(-1);

	for(i = 0; i < menu->total_items; i++)
	{
	    if(menu->item[i] == NULL)
		continue;

	    if(menu->item[i]->id == id)
		return(i);
	}

	return(-1);
}

/*
 *	Returns the state of the menu item.
 *
 *	The menu item should be of type MENU_ITEM_CHECK (but this is
 *	not checked).
 */
bool_t MenuGetItemState(menu_struct *menu, const int item_num)
{
	menu_item_struct *item;

	if(!MenuIsItemAllocated(menu, item_num))
	    return(False);

	item = menu->item[item_num];
	return(item->flags & MENU_ITEM_FLAG_TOGGLED);
}

/*
 *	Sets the state of menu item.
 *
 *	The menu item should be of type MENU_ITEM_CHECK (but this is    
 *	not checked).
 */
void MenuSetItemState(
	menu_struct *menu, const int item_num, bool_t state
)
{
	menu_item_struct *item;

	if(!MenuIsItemAllocated(menu, item_num))
	    return;

	item = menu->item[item_num];
	if(state)
	    item->flags |= MENU_ITEM_FLAG_TOGGLED;
	else
	    item->flags &= ~MENU_ITEM_FLAG_TOGGLED;
}

/*
 *	Sets the accelerator key for the specified menu item.
 */
void MenuSetItemAccelKey(
	menu_struct *menu, const int item_num,
	const keyval_t accel_key, const key_state_t accel_mods     
)
{
	menu_item_struct *item;

	if(!MenuIsItemAllocated(menu, item_num))
	    return;

	item = menu->item[item_num];
	if((item->accel_key != accel_key) ||
	   (item->accel_mods != accel_mods)
	)
	{
	    item->accel_key = accel_key;
	    item->accel_mods = accel_mods;

	    free(item->accel_text);
	    item->accel_text = NULL;

	    if(accel_key != OSW_KEY_None)
	    {
		char	*name = STRDUP(OSWGetKeyValName(accel_key)),
			*s = (char *)malloc(40 + 40 + STRLEN(name) + 80);

		*s = '\0';

		if(accel_mods & OSW_KEY_STATE_CTRL)
		    strcat(s, "Ctrl+");
		if(accel_mods & OSW_KEY_STATE_SHIFT)
		    strcat(s, "Shift+");
		if(accel_mods & OSW_KEY_STATE_ALT)
		    strcat(s, "Alt+");

		if(name != NULL)
		{
		    *name = toupper(*name);
		    strcat(s, name);
		}

		free(name);

		item->accel_text = s;
	    }
	}
}

/*
 *	Inserts a new item to the menu.
 */
int MenuItemInsert(
	menu_struct *menu, int item_num,
	int type,
	const char *text, 
	pixmap_t pixmap, bitmap_t mask,
	int id
)
{
	int n;
	menu_item_struct *item;

	if(menu->total_items < 0)
	    menu->total_items = 0;

	/* Append? */
	if((item_num < 0) || (item_num >= menu->total_items))
	{
	    n = menu->total_items;
	    menu->total_items = n + 1;
	    menu->item = (menu_item_struct **)realloc(
	        menu->item,
	        menu->total_items * sizeof(menu_item_struct *)
	    );
	    if(menu->item == NULL)
	    {
	        menu->total_items = 0;
	        return(-3);
	    }
	}
	else
	{
	    int i;
	    n = item_num;
	    menu->total_items++;
	    menu->item = (menu_item_struct **)realloc(
		menu->item,
		menu->total_items * sizeof(menu_item_struct *)
	    );
	    if(menu->item == NULL)
	    {
		menu->total_items = 0;
		return(-3);
	    }
	    for(i = menu->total_items - 1; i > n; i--)
		menu->item[i] = menu->item[i - 1];
	}

	/* Create a new item */
	menu->item[n] = item = (menu_item_struct *)calloc(
	    1, sizeof(menu_item_struct)
	);
	if(item == NULL)
	    return(-3);

	/* Set values */
	item = menu->item[n];
	item->type = type;
	item->flags = MENU_ITEM_FLAG_SENSITIVE;
	item->text = STRDUP(text);
	item->accel_text = NULL;
	item->pixmap = pixmap;
	item->mask = mask;
	item->accel_key = 0;
	item->accel_mods = 0;
	item->id = id;

	return(n);
}

/*
 *	Appends a new item to the menu.
 */
int MenuItemAppend(
	menu_struct *menu,
	int type,
	const char *text,
	pixmap_t pixmap, bitmap_t mask,
	int id
)
{
	return(MenuItemInsert(
	    menu, -1,
	    type, text, pixmap, mask, id
	));
}

/*
 *	Deletes all items in menu.
 */
void MenuDeleteAllItems(menu_struct *menu)
{
	int i;
	menu_item_struct *item;

	for(i = 0; i < menu->total_items; i++)
	{
	    item = menu->item[i];
	    if(item == NULL)
		continue;

	    free(item->text);
	    free(item->accel_text);
	    free(item);
	}

	free(menu->item);
	menu->item = NULL;
	menu->total_items = 0;
	menu->selected_item = -1;
}


/*
 *	Calculates the minimum size of the menu.
 */
static bool_t MenuGetMinSize(
	menu_struct *menu,
	unsigned int *width_rtn, unsigned int *height_rtn
)
{
	const int icon_spacing = 2;
	int i, row_height, item_width, swidth;
	unsigned int width = 0, height = 0;
	const char *text;
	font_t *font = menu->font;
	pixmap_t pixmap;
	image_t *img;
	menu_item_struct *item;
	widget_global_struct *wglobal = &widget_global;

	if(width_rtn != NULL)
	    *width_rtn = width;
	if(height_rtn != NULL)
	    *height_rtn = height;

	if(menu == NULL)
	    return(False);

	row_height = MAX(menu->row_height, 1);

	/* Calculate the new width and height by adding up the 
	 * geometries of all the items
	 */
	for(i = 0; i < menu->total_items; i++)
	{
	    item = menu->item[i];
	    if(item == NULL)
		continue;

	    switch(item->type)
	    {
	      case MENU_ITEM_CHECK:
		pixmap = item->pixmap;
		img = (item->flags & MENU_ITEM_FLAG_TOGGLED) ?
		    wglobal->check_button_armed_img :
		    wglobal->check_button_unarmed_img;
		text = item->text;
		height += row_height;
		break;
	      case MENU_ITEM_SEPARATOR:
		pixmap = None;
		img = NULL;
		text = NULL;
		height += MENU_HR_HEIGHT;
		break;
	      default:
		pixmap = item->pixmap;
		img = NULL;
		text = item->text;
		height += row_height;
		break;
	    }
	    if(text != NULL)
	    {
		OSWGetTextBounds(
		    font, text, -1,
		    NULL, NULL, &swidth, NULL, NULL
		);
		item_width = ((img != NULL) ?
		    ((int)img->width + icon_spacing) : 0) + swidth;
	    }
	    else
	    {
		item_width = (img != NULL) ?
		    (int)img->width : 0;
	    }
	    if(pixmap != None)
	    {
		unsigned int pwidth, pheight;
		OSWGetPixmapGeometry(pixmap, &pwidth, &pheight, NULL);
		item_width += pwidth + MENU_ITEM_MARGIN_X;
	    }
	    item_width += (2 * MENU_FRAME_MARGIN) + (2 * MENU_ITEM_MARGIN_X);

	    /* Add accelerator key label width */
	    if(item->accel_text != NULL)
	    {
		OSWGetTextBounds(
		    font, item->accel_text, -1,
		    NULL, NULL, &swidth, NULL, NULL
		);
		item_width += swidth + MENU_ACCELERATOR_MARGIN;
	    }

	    if(item_width > (int)width)
		width = item_width;
	}
	height += (2 * MENU_FRAME_MARGIN);

	if(width_rtn != NULL)
	    *width_rtn = width;
	if(height_rtn != NULL)
	    *height_rtn = height;

	return(True);
}


/*
 *	Initializes the menu.
 */
int MenuInit(
	menu_struct *menu, win_t parent,
        void (*activate_cb)(menu_struct *, int, void *),
        void *activate_data
)
{
	int	x = 0,
		y = 0;
	unsigned int	width = 2,
			height = 2;
	win_t w;
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if(menu == NULL)
	    return(-2);

	if(parent == None)
	    parent = gui->root_win;
	if(parent == None)
	    return(-2);

	menu->type = WIDGET_TYPE_MENU;
	menu->map_state = False;
	menu->has_focus = False;
	menu->disabled = False;
	menu->visibility_state = VisibilityFullyObscured;
	menu->x = x;
	menu->y = y;
	menu->width = width;
	menu->height = height;
	menu->font = wglobal->menu_font;
	menu->prev = NULL;
	menu->next = NULL;
	menu->data = NULL;

	menu->bg_img = NULL;
	menu->item_img = NULL;
	menu->item_highlighted_img = NULL;

	menu->row_height = MENU_ROW_HEIGHT;

	menu->item = NULL;
	menu->total_items = 0;
	menu->selected_item = -1;

	menu->pointer_map_x = 0;
	menu->pointer_map_y = 0;

	menu->activate_cb = activate_cb;
	menu->activate_data = activate_data;

	/* Create toplevel */
	x = menu->x;
	y = menu->y;
	width = menu->width;
	height = menu->height;
	OSWCreateWindow(  
	    &w, parent,
	    x, y, width, height
	);
	menu->toplevel = w;
	menu->toplevel_buf = None;
	OSWSetWindowWMProperties(
	    w,
	    "", "",
	    None,
	    False,			/* Do not let WM set coordinates */
	    x, y,
	    2, 2,
	    gui->display_width, gui->display_height,
	    OSW_WINDOW_FRAME_NONE,
	    0, NULL
	);  
	OSWSetWindowInput(
	    w,
	    StructureNotifyMask | ExposureMask |
	    ButtonPressMask | ButtonReleaseMask |
	    PointerMotionMask |
	    VisibilityChangeMask
	);

	/* Add this widget to the regeristry */
	WidgetRegister(menu, menu->type);

	return(0);
}

/*
 *	Resizes the menu.
 */
void MenuResize(menu_struct *menu)
{
	int x, y, row_height;
	unsigned int width, height;
	win_t w;
	widget_global_struct *wglobal = &widget_global;

	if(menu == NULL)
	    return;

	w = menu->toplevel;
	if(w == None)
	    return;

	OSWGetWindowGeometry(w, &x, &y, &width, &height, NULL);
	menu->x = x;
	menu->y = y;
	menu->width = width;
	menu->height = height;

	row_height = MAX(menu->row_height, 1);

	/* Resize background images */
	if(wglobal->menu_item_img != NULL)
	{
	    image_t *tar_img, *src_img = wglobal->menu_item_img;

	    width = (unsigned int)MAX(
		(int)menu->width - (2 * MENU_FRAME_MARGIN),
		1
	    );
	    height = (unsigned int)MAX(
		(int)menu->height - (2 * MENU_FRAME_MARGIN),
		1
	    );

	    OSWDestroyImage(&menu->bg_img);
	    OSWCreateImage(&tar_img, width, height);
	    WidgetResizeImageBuffer(
		tar_img->bits_per_pixel,
		(u_int8_t *)tar_img->data,
		(const u_int8_t *)src_img->data,
		tar_img->width, tar_img->height,
		tar_img->bytes_per_line,                           
		src_img->width, src_img->height,
		src_img->bytes_per_line
	    );
	    menu->bg_img = tar_img;
	}

	/* Resize menu item images */
	if(wglobal->menu_item_img != NULL)
	{
	    image_t *tar_img, *src_img = wglobal->menu_item_img;

	    width = (unsigned int)MAX(
		(int)menu->width - (2 * MENU_FRAME_MARGIN),
		1
	    );
	    height = (unsigned int)row_height;

	    OSWDestroyImage(&menu->item_img);
	    OSWCreateImage(&tar_img, width, height);
	    WidgetResizeImageBuffer(       
		tar_img->bits_per_pixel,
		(u_int8_t *)tar_img->data,
		(const u_int8_t *)src_img->data,
		tar_img->width, tar_img->height,
		tar_img->bytes_per_line,                           
		src_img->width, src_img->height,
		src_img->bytes_per_line
	    );
	    menu->item_img = tar_img;
	}
	if(wglobal->menu_item_highlighted_img != NULL)
	{
	    image_t *tar_img, *src_img = wglobal->menu_item_highlighted_img;

	    width = (unsigned int)MAX(
		(int)menu->width - (2 * MENU_FRAME_MARGIN),
		1
	    );
	    height = (unsigned int)row_height;

	    OSWDestroyImage(&menu->item_highlighted_img);
	    OSWCreateImage(&tar_img, width, height);
	    WidgetResizeImageBuffer(
		tar_img->bits_per_pixel,
		(u_int8_t *)tar_img->data,
		(const u_int8_t *)src_img->data,
		tar_img->width, tar_img->height,
		tar_img->bytes_per_line,
		src_img->width, src_img->height,
		src_img->bytes_per_line
	    );
	    menu->item_highlighted_img = tar_img;
	}
}

/*
 *	Queues the menu to resize.
 */
void MenuQueueResize(menu_struct *menu)
{
	if(menu == NULL)
	    return;

	WidgetQueueResizeAction(MenuResizeCB, menu, NULL);
}

/*
 *	Draws the menu.
 */
void MenuDraw(menu_struct *menu)
{
	unsigned int width = 0, height = 0;
	image_t *img;
	pixmap_t pixmap;
	win_t w;
	menu_item_struct *item;
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if(menu == NULL)
	    return;

	w = menu->toplevel;
	if(w == None)
	    return;

	/* Get the size of the menu */
	width = menu->width;
	height = menu->height;

	/* Create toplevel buffer */
	if(menu->toplevel_buf == None)
	{
	    if(OSWCreatePixmap(&menu->toplevel_buf, width, height))
	        return;
	}
	pixmap = menu->toplevel_buf;

	/* Draw the background */
	if(wglobal->force_mono)
	{
	    OSWClearPixmap(
		pixmap, width, height,
		gui->black_pix
	    );
	}
#if 0
	else
	{
	    if(menu->bg_img != NULL)
	    {
		const int	x = MENU_FRAME_MARGIN,
				y = MENU_FRAME_MARGIN;
		OSWPutImageToDrawablePos(menu->bg_img, pixmap, x, y);
	    }
	    else
	    {
		WidgetPutImageTile(
		    pixmap, wglobal->std_bkg_img,
		    width, height
		);
	    }
	}
#endif

	/* Draw menu items */
	if(menu->total_items > 0)
	{
	    const int row_height = MAX(menu->row_height, 1);
	    bool_t is_selected;
	    int	i,
		x = MENU_FRAME_MARGIN,
		y = MENU_FRAME_MARGIN,
		icon_text_spacing;
	    unsigned int icon_width, icon_height;
	    pixmap_t icon_pixmap;
	    bitmap_t icon_mask;

	    for(i = 0; i < menu->total_items; i++)
	    {
		item = menu->item[i];
		if(item == NULL)
		    continue;

		is_selected = (i == menu->selected_item) ? True : False;

		/* Toggle? */
		if(item->type == MENU_ITEM_CHECK)
		{
		    if(item->flags & MENU_ITEM_FLAG_TOGGLED)
		    {
			img = wglobal->check_button_armed_img;
		    }
		    else
		    {
			img = (i == menu->selected_item) ?
			    wglobal->check_button_highlighted_img :
			    wglobal->check_button_unarmed_img;
		    }
		}
		else
		{
		    img = NULL;
		}

		/* Get the item's icon */
		icon_pixmap = item->pixmap;
		icon_mask = item->mask;

	        /* Draw the item's background */
	        if(is_selected)
	        {
		    const int	x2 = x,
				y2 = y;
		    unsigned int	w2 = (unsigned int)MAX(
			((int)width - (2 * MENU_FRAME_MARGIN)),
			1
		    ),
					h2 = (unsigned int)MAX(
			row_height, 1
		    );
		    if(menu->item_highlighted_img != NULL)
		    {
			OSWPutImageToDrawablePos(
			    menu->item_highlighted_img, pixmap, x2, y2
			);
		    }
		    else
		    {
			if(wglobal->force_mono)
			    OSWSetForeground(gui->white_pix);
			else
			    OSWSetForeground(wglobal->surface_selected_pix);
			OSWDrawSolidRectangle(pixmap, x2, y2, w2, h2);
		    }
	        }
		else
		{
		    const int   x2 = x,
				y2 = y;
		    unsigned int        w2 = (unsigned int)MAX(
			((int)width - (2 * MENU_FRAME_MARGIN)),
			1
		    ),
					h2 = (unsigned int)MAX(
			row_height, 1
		    );
		    if(menu->item_img != NULL)
		    {
			OSWPutImageToDrawablePos(
			    menu->item_img, pixmap, x2, y2
			);
		    }
		    else
		    {   
			if(wglobal->force_mono)
			    OSWSetForeground(gui->black_pix);
			else
			    OSWSetForeground(gui->black_pix);
			OSWDrawSolidRectangle(pixmap, x2, y2, w2, h2);
		    }
		}

		/* Horizontal Rule? */
		if(item->type == MENU_ITEM_SEPARATOR)
		{
		    const int	x2 = x + MENU_ITEM_MARGIN_X,
				y2 = y + (MENU_HR_HEIGHT / 2),
				x3 = (int)width - MENU_FRAME_MARGIN -
				    MENU_ITEM_MARGIN_X;

		    if(!wglobal->force_mono)
		    {
		        OSWSetForeground(wglobal->surface_shadow_pix);
		        OSWDrawLine(
			    pixmap,
			    x2, y2 - 1, x3, y2 - 1
		        );
		    }

		    if(wglobal->force_mono)
		        OSWSetForeground(gui->white_pix);
		    else
		        OSWSetForeground(wglobal->surface_highlight_pix);
		    OSWDrawLine(
		        pixmap,
			x2, y2, x3, y2
		    );
		    y += MENU_HR_HEIGHT;
		    continue;
	        }

		/* Get the icon size */
		if(img != NULL)
		{
		    icon_width = img->width;
		    icon_height = img->height;
		    icon_text_spacing = STRISEMPTY(item->text) ?
			0 : MENU_ITEM_MARGIN_X;
		}
		else if(icon_pixmap != None)
		{
		    OSWGetPixmapGeometry(
			icon_pixmap, &icon_width, &icon_height, NULL
		    );
		    icon_text_spacing = STRISEMPTY(item->text) ?
			0 : MENU_ITEM_MARGIN_X;
		}
		else
		{
		    icon_width = 0;
		    icon_height = 0;
		    icon_text_spacing = 0;
		}

		/* Draw icon? */
		if(img != NULL)
		{
		    const int	x2 = x + MENU_ITEM_MARGIN_X,
				y2 = y + ((row_height - (int)icon_height) / 2);
		    WidgetPutImageNormal(pixmap, img, x2, y2, True);
		}
		else if(icon_pixmap != None)
		{
		    const int	x2 = x + MENU_ITEM_MARGIN_X,
				y2 = y + ((row_height - (int)icon_height) / 2);
		    OSWSetClipMask(icon_mask);
		    OSWSetClipOrigin(x2, y2);
		    OSWCopyDrawablesCoord(
			pixmap, icon_pixmap,
			x2, y2,
			icon_width, icon_height,
			0, 0
		    );   
		    OSWSetClipMask(None);
		}

		/* Draw text */
		if(wglobal->force_mono)
		    OSWSetForeground(
			is_selected ?
			    gui->black_pix : gui->white_pix
		    );
		else
		    OSWSetForeground(
			(item->flags & MENU_ITEM_FLAG_SENSITIVE) ?
			    wglobal->normal_text_pix :
			    wglobal->disabled_text_pix
		    );
	        if(!STRISEMPTY(item->text))
	        {
		    int	x2, y2, lbearing, rbearing, ascent, descent,
			font_height;
		    const char *name = item->text;
		    font_t *font = menu->font;

		    OSWSetFont(font);
		    OSWGetTextBounds(
		        font, name, -1,
		        &lbearing, &rbearing, NULL,
		        &ascent, &descent
		    );
		    font_height = ascent + descent;
		    x2 = x + MENU_ITEM_MARGIN_X + (int)icon_width +
			icon_text_spacing;
		    y2 = y + ((row_height - font_height) / 2);
		    OSWDrawString(
		        pixmap,
			x2 - lbearing,
		        y2 + ascent,
			name
		    );
		}

	        /* Draw accelerator text? */
	        if(item->accel_text != NULL)
	        {
		    int	x2, y2,
			lbearing, rbearing, swidth, ascent, descent,
			font_height;
		    const char *text = item->accel_text;
		    font_t *font = menu->font;

		    OSWSetFont(font);
		    OSWGetTextBounds(
			font, text, -1,
			&lbearing, &rbearing, &swidth,
			&ascent, &descent
		    );
		    font_height = ascent + descent;
		    x2 = (int)width - swidth - MENU_FRAME_MARGIN -
			MENU_ITEM_MARGIN_X;
		    y2 = y + ((row_height - font_height) / 2);
		    OSWDrawString(
			pixmap,
			x2 - lbearing,
			y2 + ascent,
			text
		    );
		}

	        y += row_height;
	    }
	}

	/* Draw frame */
	if(wglobal->force_mono)
	    WidgetFrameButtonPixmap(
		pixmap, False,
		width, height,
		gui->white_pix,
		gui->white_pix
	    );
	else
	    WidgetFrameButtonPixmap(
		pixmap, False,
		width, height,
		wglobal->surface_highlight_pix,
		wglobal->surface_shadow_pix
	    );

	OSWPutDrawableToWindow(w, pixmap);
}

/*
 *	Queues the menu to draw.
 */
void MenuQueueDraw(menu_struct *menu)
{
	if(menu == NULL)
	    return;

	WidgetQueueDrawAction(MenuDrawCB, menu, NULL);
}

/*
 *	Manages the menu event.
 */
int MenuManage(menu_struct *menu, event_t *event)
{
	int events_handled = 0;
	win_t w;
	event_key_t *key;
	event_button_t *button;
	event_motion_t *motion;

	if((menu == NULL) || (event == NULL))
	    return(events_handled);

	w = event->xany.window;

	switch(event->type)
	{
	  case ConfigureNotify:
	    if(w == menu->toplevel)
	    {
		MenuQueueResize(menu);
		if(menu->map_state)
		    MenuQueueDraw(menu);
		events_handled++;
	    }
	    break;

	  case Expose:
	    if(w == menu->toplevel)
	    {
		MenuQueueDraw(menu);
		events_handled++;
	    }
	    break;

	  case KeyPress:
	    key = &event->xkey;

	    /* Do not handle key events if not mapped */
	    if(!menu->map_state)
		return(events_handled);

	    if(menu->activate_cb != NULL)
	    {
		int i;
		keycode_t keycode = key->keycode;
		key_state_t key_state = key->state;
		keyval_t keyval = OSWGetKeyValFromKeyCode(keycode, key_state);
		menu_item_struct *item;

	        for(i = 0; i < menu->total_items; i++)
		{
		    item = menu->item[i];
		    if(item == NULL)
			continue;

		    if(item->accel_key == OSW_KEY_None)
			continue;

		    if(item->accel_key != keyval)
			continue;

		    /* Check modifier keys */
		    if(item->accel_mods & OSW_KEY_STATE_ALT)
		    {
			if(!(key_state & OSW_KEY_STATE_ALT))
			    continue;
		    }
		    else
		    {
			if(key_state & OSW_KEY_STATE_ALT)
			    continue;
		    }

		    if(item->accel_mods & OSW_KEY_STATE_CTRL)
		    {
			if(!(key_state & OSW_KEY_STATE_CTRL))
			    continue;
		    }
		    else
		    {
			if(key_state & OSW_KEY_STATE_CTRL)
			    continue; 
		    }

		    if(item->accel_mods & OSW_KEY_STATE_SHIFT)
		    {
			if(!(key_state & OSW_KEY_STATE_SHIFT))
			    continue;
		    }
		    else
		    {
			if(key_state & OSW_KEY_STATE_SHIFT)
			    continue; 
		    }

		    /* Toggle state as needed */
		    if(item->type == MENU_ITEM_CHECK)
		    {
			if(item->flags & MENU_ITEM_FLAG_TOGGLED)
			    item->flags &= ~MENU_ITEM_FLAG_TOGGLED;
			else
			    item->flags |= MENU_ITEM_FLAG_TOGGLED;
		    }

		    /* Call callback function */
		    menu->activate_cb(
			menu,
			item->id,
			menu->activate_data
		    );

		    if(!WidgetIsRegistered(menu))
		    {
			events_handled++;
			return(events_handled);
		    }

		    if(menu->map_state)
			MenuQueueDraw(menu);

		    events_handled++;
		    break;
		}
	    }
	    break;

	  case KeyRelease:
	    key = &event->xkey;

	    /* Do not handle key events if not mapped */
	    if(!menu->map_state)
		return(events_handled);

	    break;

	  case ButtonPress:
	    button = &event->xbutton;  
	    if(menu->map_state)
	    {
		events_handled++;
	    }
	    break;

	  case ButtonRelease:
	    button = &event->xbutton;
	    if(menu->map_state)
	    {
		int	x = (int)button->x,
			y = (int)button->y;
		win_t menu_w = menu->toplevel;

		/* If the button was pressed on another window then
		 * translate the coordinates to be relative to this
		 * menu
		 */
		if(w != menu_w)
		{
		    const int	tx = x,
				ty = y;
		    OSWTranslateCoordinatesToChild(
			w, menu_w,
			tx, ty,
			&x, &y
		    );
		}

		/* Was the button released at a position different
		 * from its original map position?
		 */
		if((button->x_root != menu->pointer_map_x) ||
		   (button->y_root != menu->pointer_map_y)
		)
		{
		    /* Unmap the menu */
		    MenuUnmap(menu);

		    /* Did the button press fall on the menu? */
		    if((x >= 0) && (x < (int)menu->width) &&
		       (y >= 0) && (y < (int)menu->height)
		    )
		    {
			const int i = menu->selected_item;
			if((i >= 0) && (i < menu->total_items))
			{
			    menu_item_struct *item = menu->item[i];
			    if((item != NULL) ?
				(item->flags & MENU_ITEM_FLAG_SENSITIVE) : False
			    )
			    {
				/* Check item? */
				if(item->type == MENU_ITEM_CHECK)
				{
				    /* Toggle the check item */
				    if(item->flags & MENU_ITEM_FLAG_TOGGLED)
					item->flags &= ~MENU_ITEM_FLAG_TOGGLED;
				    else
					item->flags |= MENU_ITEM_FLAG_TOGGLED;
				}

				/* Call activate callback */
				if(menu->activate_cb != NULL)
				    menu->activate_cb(
					menu,         
					item->id,
					menu->activate_data
				    );
				if(!WidgetIsRegistered(menu))
				{
				    events_handled++;
				    return(events_handled);
				}
			    }
			}
		    }
		}
		events_handled++;
	    }
	    break;

	  case MotionNotify:
	    motion = &event->xmotion;
	    if(menu->map_state)
	    {
		int     x = (int)motion->x,
			y = (int)motion->y;
		win_t menu_w = menu->toplevel;

		/* If the pointer moved on a different window then
		 * translate the coordinates to be relative to this
		 * menu
		 */
		if(w != menu_w)
		{
		    const int   tx = x,
				ty = y;                          
		    OSWTranslateCoordinatesToChild(
			w, menu_w,
			tx, ty,
			&x, &y
		    );
		}

		/* Did the motion occure over the menu? */
		if((x >= 0) && (x < (int)menu->width) &&
		   (y >= 0) && (y < (int)menu->height)
		)
		{
		    /* Find the item that the pointer moved over */
		    const int prev_sel_item = menu->selected_item;
		    int	i,
			item_y = MENU_FRAME_MARGIN,
			sel_item = -1,
			row_height = menu->row_height;
		    menu_item_struct *item;

		    for(i = 0; i < menu->total_items; i++)
		    {
			item = menu->item[i];
			if(item == NULL)
			    continue;

			/* Did we find the selected item? */
			if(sel_item > -1)
			    break;

			/* Check based on type */
			switch(item->type)
			{
			  case MENU_ITEM_SEPARATOR:
			    item_y += MENU_HR_HEIGHT;
			    break;

			  default:
			    if((y >= item_y) &&
			       (y < (item_y + row_height))
			    )
			    {
				if(item->flags & MENU_ITEM_FLAG_SENSITIVE)
				    sel_item = i;
			    }
			    item_y += row_height;
			    break;
			}
		    }
		    /* Pointer moved to a new selected item? */
		    if((sel_item != prev_sel_item) &&
		       (sel_item >= 0) && (sel_item < menu->total_items)
		    )
		    {
			menu->selected_item = sel_item;
			MenuQueueDraw(menu);
			events_handled++;
		    }
		    else if((sel_item < 0) && (prev_sel_item > -1))
		    {
			menu->selected_item = -1;
			MenuQueueDraw(menu);
		    }
		}
		else
		{
		    menu->selected_item = -1;
		    MenuQueueDraw(menu);
		}
		events_handled++;
	    }
	    break;

	  case MapNotify:
	    if(w == menu->toplevel)
	    {
		if(!menu->map_state)
		{
		    menu->map_state = True;
		}
		events_handled++;
	    }
	    break;

	  case UnmapNotify:
	    if(w == menu->toplevel)
	    {
		if(menu->map_state)
		{
		    OSWDestroyPixmap(&menu->toplevel_buf);

		    OSWDestroyImage(&menu->bg_img);
		    OSWDestroyImage(&menu->item_img);
		    OSWDestroyImage(&menu->item_highlighted_img);

		    menu->map_state = False;
		}
		events_handled++;
	    }
	    break;

	  case VisibilityNotify:
	    if(w == menu->toplevel)
	    {
		menu->visibility_state = event->xvisibility.state;
		events_handled++;
	    }
	    break;
	}

	return(events_handled);
}

/*
 *	Maps the menu.
 */
void MenuMap(menu_struct *menu)
{
	unsigned int width, height;
	win_t w;

	if(menu == NULL)
	    return;

	w = menu->toplevel;
	if(w == None)
	    return;

	/* Reset the selected item */
	menu->selected_item = -1;

	/* Record the original pointer position at map time */
	OSWGetPointerCoords(
	    w,
	    &menu->pointer_map_x, &menu->pointer_map_y,
	    NULL, NULL
	);

	/* Get the minimum size of the menu */
	MenuGetMinSize(
	    menu,
	    &width, &height
	);

	/* Resize the menu */
	OSWResizeWindow(
	    w,
	    width, height
	);
	MenuQueueResize(menu);

	/* Map the menu */
	OSWMapRaised(w);

	/* Ungrab the pointer in case the calling function was
	 * grabbing the pointer
	 */
	OSWUngrabPointer();
}

/*
 *	Maps the menu at the specified position.
 */
void MenuMapCoordinates(menu_struct *menu, int x, int y)
{
	unsigned int width, height;
	win_t w, parent;

	if(menu == NULL)
	    return;

	w = menu->toplevel;
	if(w == None)
	    return;

	/* Reset the selected item */
	menu->selected_item = -1;

	/* Record the original pointer position at map time */
	OSWGetPointerCoords(
	    w,
	    &menu->pointer_map_x, &menu->pointer_map_y,
	    NULL, NULL
	);

	MenuGetMinSize(menu, &width, &height);

	/* Keep the menu within its parent window */
	parent = OSWGetWindowParent(w);
	if(parent != None)
	{
	    unsigned int pwidth, pheight;

	    OSWGetWindowGeometry(
		parent, NULL, NULL, &pwidth, &pheight, NULL
	    );

	    if((x + (int)width) >= (int)pwidth)
		x = (int)pwidth - (int)width;
	    if((y + (int)height) >= (int)pheight)
		y = (int)pheight - (int)height;
	    if(x < 0)
		x = 0;
	    if(y < 0)
		y = 0;
	}

	/* Move and resize the menu */
	OSWMoveResizeWindow(
	    w,
	    x, y,
	    width, height
	);
	MenuQueueResize(menu);

	/* Map the menu */
	OSWMapRaised(w);

	/* Ungrab the pointer in case the calling function was
	 * grabbing the pointer
	 */
	OSWUngrabPointer();
}

/*
 *	Unmaps the menu.
 */
void MenuUnmap(menu_struct *menu)
{
	if(menu == NULL)
	    return;

	OSWUnmapWindow(menu->toplevel);
}

/*
 *	Destroys the menu.
 */
void MenuDestroy(menu_struct *menu)
{
	if(menu == NULL)
	    return;

	/* Delete all the menu items */
	MenuDeleteAllItems(menu);

	OSWDestroyWindow(&menu->toplevel);
	OSWDestroyPixmap(&menu->toplevel_buf);

	OSWDestroyImage(&menu->bg_img);
	OSWDestroyImage(&menu->item_img);
	OSWDestroyImage(&menu->item_highlighted_img);

	menu->type = WIDGET_TYPE_UNKNOWN;
	menu->map_state = False;
	menu->has_focus = False;
	menu->disabled = False;
	menu->visibility_state = VisibilityFullyObscured;
	menu->x = 0;
	menu->y = 0;
	menu->width = 0;
	menu->height = 0;
	menu->font = NULL;
	menu->prev = NULL;
	menu->next = NULL;

	menu->row_height = 0;

	menu->item = NULL;
	menu->total_items = 0;
	menu->selected_item = 0;

	menu->pointer_map_x = 0;
	menu->pointer_map_y = 0;

	menu->activate_cb = NULL;
	menu->activate_data = NULL;
}
