#include <stdlib.h>
#include <string.h>
#include <cmath>

#include "../include/widget.h"

#include "../include/oswkeyvals.h"


popup_list_item_struct *PUListGetItem(popup_list_struct *list, int i);
int PUListGetSelItem(popup_list_struct *list);
const char *PUListGetSelItemName(popup_list_struct *list);
void PUListSetSelItem(popup_list_struct *list, int n);
int PUListGetItemByData(popup_list_struct *list, void *data);

/* Item Add/Delete */
int PUListAddItem(
	popup_list_struct *list,
	unsigned int flags,
	const char *text,
	pixmap_t pixmap,
	bitmap_t mask
);
void PUListClear(popup_list_struct *list);

/* Item Data */
void PUListItemSetData(
	popup_list_struct *list, int i,
	void *data, void (*destroy_cb)(void *)
);
void *PUListItemGetData(popup_list_struct *list, int i);

static void PUListResizeCB(void *widget, void *data);
static void PUListDrawCB(void *widget, void *data);
static void PUListScrollTimeoutCB(void *widget, void *data);

static void PUListMapBtnPressedCB(push_button_struct *btn, void *data);

static void PUListAddScrollTimeout(
	popup_list_struct *list, const time_t interval_ms
);
static void PUListScrollRepeatStop(popup_list_struct *list);

int PUListInit(
	popup_list_struct *list, win_t parent,
	int x, int y,
	unsigned int width, unsigned int height,
	int vis_items,			/* Items visible on list */
	int direction			/* One of PULIST_POPUP_* */
);
void PUListResize(popup_list_struct *list, int list_items_visible);
void PUListQueueResize(popup_list_struct *list, int list_items_visible);
void PUListDraw(popup_list_struct *list, int amount);
void PUListQueueDraw(popup_list_struct *list, int amount);
int PUListManage(popup_list_struct *list, event_t *event);
void PUListMap(popup_list_struct *list);
void PUListUnmap(popup_list_struct *list);
void PUListMapList(popup_list_struct *list);
void PUListUnmapList(popup_list_struct *list);
void PUListDestroy(popup_list_struct *list);


#define PULIST_DEF_WIDTH		90
#define PULIST_DEF_HEIGHT		26

#define PULIST_DEF_ITEMS_VISIBLE	7

#define PULIST_ITEM_HEIGHT		20


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Returns the pointer to the item index specified by i.
 */
popup_list_item_struct *PUListGetItem(popup_list_struct *list, int i)
{
	if(list == NULL)
	    return(NULL);
	else if((i < 0) || (i >= list->total_items))
	    return(NULL);
	else
	    return(list->item[i]);
}

/*
 *	Returns the selected item.
 */
int PUListGetSelItem(popup_list_struct *list)
{
	return((list != NULL) ? list->sel_item : 0);
}

/*
 *	Returns the selected item's name.
 */
const char *PUListGetSelItemName(popup_list_struct *list)
{
	int i;
	popup_list_item_struct *item;

	if(list == NULL)
	    return(NULL);

	i = list->sel_item;
	item = ((i >= 0) && (i < list->total_items)) ?
	    list->item[i] : NULL;
	return((item != NULL) ? item->text : NULL);
}

/*
 *	Sets the selected item.
 */
void PUListSetSelItem(popup_list_struct *list, int n)
{
	if(list == NULL)
	    return;

	if(list->sel_item != n)
	{
	    list->sel_item = n;

	    /* Popup list not mapped? */
	    if(!list->popup_map_state)
	    {
		const int y_max = (int)list->list_max_height -
		    (list->popup_list_vis_items * PULIST_ITEM_HEIGHT);

		/* Scroll to the selected item */
		list->list_y_pos = n * PULIST_ITEM_HEIGHT;
		if(list->list_y_pos > y_max)
		    list->list_y_pos = y_max;
		if(list->list_y_pos < 0)
		    list->list_y_pos = 0;

		/* Record the previous selected item */
		list->prev_sel_item = list->sel_item;
	    }

	    if(list->map_state)
		PUListQueueDraw(list, PULIST_DRAW_AMOUNT_LABEL);
	    if(list->popup_map_state)
		PUListQueueDraw(list, PULIST_DRAW_AMOUNT_PULIST);
	}
}

/*
 *	Returns the first item number who's data matches the specified
 *	data.
 */
int PUListGetItemByData(popup_list_struct *list, void *data)
{
	int i;
	popup_list_item_struct *item;

	if(list == NULL)
	    return(-1);

	for(i = 0; i < list->total_items; i++)
	{
	    item = list->item[i];
	    if(item == NULL)
		continue;

	    if(item->data == data)
		return(i);
	}

	return(-1);
}


/*
 *	Allocates a new item on list.
 *
 *	Returns list item number or -1 on error.
 */
int PUListAddItem(
	popup_list_struct *list,
	unsigned int flags,
	const char *text,  
	pixmap_t pixmap, 
	bitmap_t mask   
)
{
	int i;
	unsigned int width, height;
	popup_list_item_struct *item;

	if(list == NULL)
	    return(-2);

	i = MAX(list->total_items, 0);
	list->total_items = i + 1;
	list->item = (popup_list_item_struct **)realloc(
	    list->item,
	    list->total_items * sizeof(popup_list_item_struct *)
	);
	if(list->item == NULL)
	{
	    list->total_items = 0;
	    return(-3);
	}

	list->item[i] = item = (popup_list_item_struct *)calloc(
	    1, sizeof(popup_list_item_struct)
	);
	if(item == NULL)
	    return(-3);

	OSWGetWindowGeometry(
	    list->popup_list, NULL, NULL, &width, &height, NULL
	);

	item->flags = flags;
	item->text = STRDUP(text);
	item->pixmap = pixmap;
	item->mask = mask;
	item->data = NULL;
	item->destroy_cb = NULL;

	list->list_max_width = width;
	list->list_max_height = list->total_items * PULIST_ITEM_HEIGHT;

	return(i);
}

/*
 *	Delete all items on the popup list.
 */
void PUListClear(popup_list_struct *list) 
{
	int i;
	popup_list_item_struct *item;

	if(list == NULL)
	    return;

	for(i = 0; i < list->total_items; i++)
	{
	    item = list->item[i];
	    if(item == NULL)
		continue;

	    if(item->destroy_cb != NULL)
		item->destroy_cb(item->data);

	    free(item->text);
	    free(item);
	}
	free(list->item);
	list->item = NULL;
	list->total_items = 0;
	list->sel_item = -1;
	list->prev_sel_item = -1;
}

/*
 *	Sets the item's data.
 */
void PUListItemSetData(
	popup_list_struct *list, int i, 
	void *data, void (*destroy_cb)(void *)
)
{
	popup_list_item_struct *item = PUListGetItem(list, i);
	if(item == NULL)
	    return;

	if(item->destroy_cb != NULL)
	    item->destroy_cb(item->data);

	item->data = data;
	item->destroy_cb = destroy_cb;
}

/*
 *	Returns the item's data.
 */
void *PUListItemGetData(popup_list_struct *list, int i)
{
	popup_list_item_struct *item = PUListGetItem(list, i);
	return((item != NULL) ? item->data : NULL);
}


/*
 *	Resize callback.
 */
static void PUListResizeCB(void *widget, void *data)
{
	popup_list_struct *list = (popup_list_struct *)widget;
	if(list == NULL)
	    return;     

	PUListResize(list, list->popup_list_vis_items);
}

/*
 *	Draw callback.
 */
static void PUListDrawCB(void *widget, void *data)
{
	int amount = (int)data;
	popup_list_struct *list = (popup_list_struct *)widget;
	if(list == NULL)
	    return;

	if(list->map_state)
	    PUListDraw(list, amount);
}

/*
 *	Scroll timeout callback.
 */
static void PUListScrollTimeoutCB(void *widget, void *data)
{
	int x, y, root_x, root_y, y_max, prev_sel_item;
	unsigned int width, height;
	int prev_y_pos;
	win_t w;
	widget_global_struct *wglobal = &widget_global;
	popup_list_struct *list = (popup_list_struct *)widget;
	if(list == NULL)
	    return;

	w = list->popup_list;
	prev_y_pos = list->list_y_pos;
	prev_sel_item = list->sel_item;

	OSWGetWindowGeometry(
	    w, NULL, NULL, &width, &height, NULL
	);
	OSWGetPointerCoords(
	    w, &root_x, &root_y, &x, &y
	);

	/* Check the pointer's position */
	if(y < 0)
	{
	    list->sel_item = 0;
	    list->list_y_pos -= PULIST_ITEM_HEIGHT;
	}
	else if(y >= (int)height)
	{
	    list->sel_item = list->total_items - 1;
	    list->list_y_pos += PULIST_ITEM_HEIGHT;
	}
	else
	{
	    PUListAddScrollTimeout(
		list, wglobal->pulist_repeat_int
	    );
	    return;
	}

	/* Clip scroll position */
	y_max = (int)list->list_max_height -
	    (list->popup_list_vis_items * PULIST_ITEM_HEIGHT);
	if(list->list_y_pos > y_max)
	    list->list_y_pos = y_max;
	if(list->list_y_pos < 0)
	    list->list_y_pos = 0;

	if((prev_y_pos != list->list_y_pos) ||
	   (list->sel_item != prev_sel_item)
	)
	{
	    if(list->map_state)
		PUListQueueDraw(list, PULIST_DRAW_AMOUNT_LABEL);
	    if(list->popup_map_state)
		PUListQueueDraw(list, PULIST_DRAW_AMOUNT_PULIST);
	}

	PUListAddScrollTimeout(
	    list, wglobal->pulist_repeat_int
	);
}

/*
 *	Map button pressed callback.
 */
static void PUListMapBtnPressedCB(push_button_struct *btn, void *data)
{
	popup_list_struct *list = POPUP_LIST(data);
	if(list == NULL)
	    return;

	if(list->disabled)
	    return;

	if(list->popup_map_state)
	    PUListUnmapList(list);
	else
	    PUListMapList(list);
}


/*
 *	Scheduals a scroll timeout.
 */
static void PUListAddScrollTimeout(
	popup_list_struct *list, const time_t interval_ms
)
{
	if(list == NULL)
	    return;

	WidgetTimeoutAdd(
	    PUListScrollTimeoutCB,
	    list, NULL,
	    interval_ms
	);
}

/*
 *	Removes all timeouts associated with the specified popup
 *	list.
 */
static void PUListScrollRepeatStop(popup_list_struct *list)
{
	if(list == NULL)
	    return;

	WidgetTimeoutRemoveByFuncWidget(
	    PUListScrollTimeoutCB, list
	);
}


/*
 *	Initializes the popup list.
 */
int PUListInit(
	popup_list_struct *list, win_t parent,
	int x, int y,
	unsigned int width, unsigned int height,
	int nitems_visible,		/* Items visible on list */
	int direction			/* One of PULIST_POPUP_* */
)
{
	unsigned int list_height, list_width;
	win_t w;
	push_button_struct *btn;
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if((list == NULL) || (parent == None))
	    return(-2);

	if(width == 0)
	    width = PULIST_DEF_WIDTH;
	if(height == 0)
	    height = PULIST_DEF_HEIGHT;
	if(nitems_visible <= 0)
	    nitems_visible = PULIST_DEF_ITEMS_VISIBLE;

	list->type = WIDGET_TYPE_POPUP_LIST;
	list->map_state = False;
	list->has_focus = False;
	list->disabled = False;
	list->visibility_state = VisibilityFullyObscured;
	list->x = x;
	list->y = y;
	list->width = width;
	list->height = height;
	list->font = wglobal->std_font;
	list->next = NULL;
	list->prev = NULL;
	list->data = NULL;

	list->popup_map_state = False;
	list->pointer_map_x = 0;
	list->pointer_map_y = 0;
	list->popup_list_vis_items = MAX(nitems_visible, 1);

	list->list_max_width = 0;
	list->list_max_height = 0;
	list->list_y_pos = 0;

	list->item = NULL;
	list->total_items = 0;
	list->sel_item = 0;
	list->prev_sel_item = 0;

	list->direction = direction;

	list->select_cb = NULL;
	list->select_data = NULL;

	/* Label toplevel */
	x = list->x;
	y = list->y;
	width = list->width;
	height = list->height;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	list->toplevel = w;
	OSWSetWindowInput(
	    w,
	    StructureNotifyMask | ExposureMask |
	    ButtonPressMask | ButtonReleaseMask
	);

	/* Map button */
	list->map_btn = btn = PUSH_BUTTON(malloc(sizeof(push_button_struct)));
	width = PULIST_MAP_BTN_WIDTH;
	height = (unsigned int)MAX(
	    ((int)list->height - (2 * 2)),
	    1
	);
	x = (int)list->width - (int)width - 2;
	y = ((int)list->height - (int)height) / 2;
	PBtnInit(
	    btn, w,
	    x, y, width, height,
	    NULL, PBTN_LABEL_ALIGN_CENTER,
	    wglobal->pulist_map_pixmap,
	    wglobal->pulist_map_mask,
	    NULL, list
	);
	btn->pressed_cb = PUListMapBtnPressedCB;
	btn->pressed_data = list;

	/* Create the popup toplevel */
	parent = gui->root_win;
	list_width = (unsigned int)MAX(((int)list->width - 4), 0);
	list_height = (unsigned int)(
	    PULIST_ITEM_HEIGHT * list->popup_list_vis_items
	);
	x = 0;
	y = 0;
	width = list_width + 4;
	height = list_height + 4;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	list->popup_toplevel = w;
	list->popup_toplevel_buf = None;
	OSWSetWindowWMProperties(               
	    w,
	    "", "",
	    None,
	    False,			/* Do not let WM set coordinates */
	    x, y,
	    1, 1,
	    gui->display_width, gui->display_height,
	    OSW_WINDOW_FRAME_NONE,
	    0, NULL
	);
	OSWSetWindowInput(
	    w,
	    ExposureMask
	);
	parent = w;

	/* Create popup list */
	x = 2;
	y = 2;
	width = list_width;
	height = list_height;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	list->popup_list = w;
	list->popup_list_buf = None;
	OSWSetWindowInput(
	    w,
	    ExposureMask |
	    ButtonPressMask | ButtonReleaseMask |
	    PointerMotionMask |
	    EnterWindowMask | LeaveWindowMask
	);

	/* Add this to widget the regeristry */
	WidgetRegister(list, list->type);

	return(0);
}

/*
 *	Resizes the popup list.
 */
void PUListResize(popup_list_struct *list, int list_items_visible)
{
	int x, y;
	unsigned int width, height, list_width, list_height;
	win_t w, parent;
	push_button_struct *btn;
/*	widget_global_struct *wglobal = &widget_global; */

	if(list == NULL)
	    return;

	/* Set the number of items visible on the popup list */
	if(list_items_visible > 0)
	    list->popup_list_vis_items = MAX(list_items_visible, 1);

	/* Destroy buffers */
	OSWDestroyPixmap(&list->popup_list_buf);
	OSWDestroyPixmap(&list->popup_toplevel_buf);

	/* Get new sizes */
	w = list->toplevel;
	OSWGetWindowGeometry(w, &x, &y, &width, &height, NULL);
	list->x = x;
	list->y = y;
	list->width = width;
	list->height = height;
	parent = w;

	/* Map Button */
	btn = list->map_btn;
	width = PULIST_MAP_BTN_WIDTH;
	height = (unsigned int)MAX(
	    ((int)list->height - (2 * 2)),
	    1  
	);   
	x = (int)list->width - (int)width - 2;   
	y = ((int)list->height - (int)height) / 2;
	OSWMoveResizeWindow(btn->toplevel, x, y, width, height);
	WidgetQueueResize(btn);


	/* Resize the popup list's toplevel */
	w = list->popup_toplevel;
	list_width = (unsigned int)MAX(((int)list->width - 4), 0);
	list_height = (unsigned int)(
	    PULIST_ITEM_HEIGHT * list->popup_list_vis_items
	);
	width = list_width + 4;
	height = list_height + 4;
	OSWResizeWindow(w, width, height);

	/* Resize the popup list */
	w = list->popup_list;
	x = 2;
	y = 2;
	width = list_width;
	height = list_height;
	OSWMoveResizeWindow(w, x, y, width, height);
}

/*
 *	Queues the popup list to resize.
 */
void PUListQueueResize(popup_list_struct *list, int list_items_visible)
{
	if(list == NULL)
	    return;

	/* Set the number of items visible on the popup list */
	if(list_items_visible > 0)
	    list->popup_list_vis_items = MAX(list_items_visible, 1);

	WidgetQueueResizeAction(PUListResizeCB, list, NULL);
}

/*
 *	Draws the popup list.
 */
void PUListDraw(popup_list_struct *list, int amount)
{
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if(list == NULL)
	    return;

	/* Recreate buffers as needed */
	if(list->popup_toplevel_buf == None)
	{
	    unsigned int width, height;
	    OSWGetWindowGeometry(
		list->popup_toplevel, NULL, NULL, &width, &height, NULL
	    );
	    OSWCreatePixmap(
		&list->popup_toplevel_buf, width, height
	    );
	}
	if(list->popup_list_buf == None)
	{
	    unsigned int width, height;
	    OSWGetWindowGeometry(
		list->popup_list, NULL, NULL, &width, &height, NULL
	    );
	    OSWCreatePixmap(
		&list->popup_list_buf, width, height
	    );
	}

	/* Draw label? */
	if((amount == PULIST_DRAW_AMOUNT_COMPLETE) ||
	   (amount == PULIST_DRAW_AMOUNT_LABEL)
	)
	{
	    const int x_margin = 4;
	    const unsigned int	width = list->width,
				height = list->height;
	    int i;
	    font_t *font = list->font;
	    win_t w = list->toplevel;
	    popup_list_item_struct *item;

	    /* Draw the background */
	    OSWSetForeground(
		wglobal->force_mono ?
		    gui->black_pix :
		    wglobal->surface_editable_pix
	    );
	    OSWDrawSolidRectangle(w, 0, 0, width, height);

	    /* Always use the previous selected item when drawing the
	     * label
	     */
	    i = list->prev_sel_item;
	    item = PUListGetItem(list, i);

	    /* Draw the selected item */
	    if(item != NULL)
	    {
		int	x2, y2, icon_spacing = 0,
			lbearing, rbearing, swidth,
			ascent, descent, font_height;
		unsigned int icon_width = 0, icon_height = 0;
		const char *text = item->text;
		pixmap_t icon_pixmap = item->pixmap;
		bitmap_t icon_mask = item->mask;

		/* Get icon geometry */
		if(icon_pixmap != None)
		{
		    OSWGetPixmapGeometry(
			icon_pixmap, &icon_width, &icon_height, NULL
		    );
		    icon_spacing = STRISEMPTY(text) ? 0 : 2;
		}

		/* Get text geometry */
		OSWGetTextBounds(
		    font, text, -1,
		    &lbearing, &rbearing, &swidth,
		    &ascent, &descent
		);
		font_height = ascent + descent;

		/* Draw the pixmap */
		if(icon_pixmap != None)
		{
		    x2 = (2 * x_margin);
		    y2 = ((int)height - (int)icon_height) / 2;
		    OSWSetClipMask(icon_mask);
		    OSWSetClipOrigin(x2, y2);
		    OSWCopyDrawablesCoord(
			w, icon_pixmap,
			x2, y2,
			icon_width, icon_height,
			0, 0
		    );
		    OSWSetClipMask(None);
		}

		if(item->flags & PULIST_ITEM_FLAG_DISABLED)
		    OSWSetForeground(
			wglobal->force_mono ?
			    gui->white_pix :
			    wglobal->disabled_text_pix
		    );
		else
		    OSWSetForeground(
			wglobal->force_mono ?
			    gui->white_pix :
			    wglobal->editable_text_pix
			);

		/* Draw the text */
		if(text != NULL)
		{
		    x2 = (2 * x_margin) + (int)icon_width + icon_spacing;
		    y2 = ((int)height - font_height) / 2;
		    OSWSetFont(font);
		    OSWDrawString(
			w,
			x2 - lbearing,
			y2 + ascent,
			text
		    );
		}
	    }

	    if(wglobal->force_mono)
		WidgetFrameButtonPixmap(
		    w, True, width, height,
		    gui->white_pix,
		    gui->white_pix
		);
	    else
	        WidgetFrameButtonPixmap(
		    w, True, width, height,
		    wglobal->surface_highlight_pix,
		    wglobal->surface_shadow_pix
	        );
	}

	/* Draw popup list? */
	if((amount == PULIST_DRAW_AMOUNT_COMPLETE) ||
	   (amount == PULIST_DRAW_AMOUNT_PULIST)
	)
	{
	    /* Is popup list mapped? */
	    if(list->popup_map_state)
	    {
		const int	x_margin = 4,
				y_scroll_pos = list->list_y_pos,
				nitems_visible = MAX(list->popup_list_vis_items, 1),
				total_items = list->total_items;
		int	i, x, y, x2, y2, icon_spacing,
			lbearing, rbearing, swidth,
			ascent, descent, font_height,
			items_drawn = 0;
		unsigned int width, height, icon_width, icon_height;
		const char *text;
		font_t *font = list->font;
		pixmap_t icon_pixmap;
		bitmap_t icon_mask;
		popup_list_item_struct *item;
		win_t w = list->popup_toplevel;
		pixmap_t pixmap;

		OSWGetWindowGeometry(
		    w, NULL, NULL, &width, &height, NULL
		);

		if(list->popup_toplevel_buf == None)
		    OSWCreatePixmap(
			&list->popup_toplevel_buf, width, height
		    );
		pixmap = list->popup_toplevel_buf;
 
		OSWClearPixmap(
		    pixmap, width, height,
		    wglobal->force_mono ?
			gui->black_pix :
			wglobal->surface_editable_pix
		);

		/* Draw frame */
		if(wglobal->force_mono)
		    WidgetFrameButtonPixmap(
			pixmap, False, width, height,
			gui->white_pix,
			gui->white_pix
		    );
		else
		    WidgetFrameButtonPixmap(
			pixmap, False, width, height,
			wglobal->surface_highlight_pix,
			wglobal->surface_shadow_pix
		    );

		OSWPutDrawableToWindow(w, pixmap);


		/* Popup list */
		w = list->popup_list;

		OSWGetWindowGeometry(
		    w, NULL, NULL, &width, &height, NULL
		);
		if(list->popup_list_buf == None)
		    OSWCreatePixmap(&list->popup_list_buf, width, height);
		pixmap = list->popup_list_buf;

		OSWClearPixmap(
		    pixmap, width, height,
		    wglobal->force_mono ?
			gui->black_pix :
			wglobal->surface_editable_pix
		);

		/* Calculate starting item pos */
		x = (2 * x_margin);
		y = 0 - (y_scroll_pos % PULIST_ITEM_HEIGHT);
		items_drawn = 0;

		for(i = MAX(y_scroll_pos / PULIST_ITEM_HEIGHT, 0);
		    i < total_items;
		    i++
		)
	        {
		    /* All visible items drawn? */
		    if(items_drawn > nitems_visible)
			break;

		    item = list->item[i];
		    if(item == NULL)
			continue;

		    /* Get text and icon */
		    text = item->text;
		    icon_pixmap = item->pixmap;
		    icon_mask = item->mask;

		    /* Get icon geometry */
		    if(icon_pixmap != None)
		    {
			OSWGetPixmapGeometry(
			    icon_pixmap, &icon_width, &icon_height, NULL
			);
			icon_spacing = STRISEMPTY(text) ? 0 : 2;
		    }
		    else
		    {
			icon_width = icon_height = 0;
			icon_spacing = 0;
		    }

		    /* Get text geometry */
		    OSWGetTextBounds(
			font, text, -1,
			&lbearing, &rbearing, &swidth,
			&ascent, &descent
		    );
		    font_height = ascent + descent;

		    /* Is this the selected item? */
		    if(i == list->sel_item)
		    {
			/* Draw selected background */
			OSWSetForeground(
			    wglobal->force_mono ?
				gui->white_pix :
				wglobal->surface_selected_pix
			);
			OSWDrawSolidRectangle(
			    pixmap,
			    0, y,
			    width, PULIST_ITEM_HEIGHT
			);

			if(wglobal->force_mono)
			    OSWSetForeground(gui->black_pix);
			else
			    OSWSetForeground(wglobal->selected_text_pix);
		    }
		    /* Previous selected item? */
		    else if(i == list->prev_sel_item)
		    {
#ifdef HAVE_X
			/* Draw selection outline */
			Display *dpy = gui->display;
			GC gc = gui->gc;
			unsigned long vm = GCCapStyle | GCLineStyle |
					   GCDashOffset | GCLineWidth |
					   GCJoinStyle;
			XGCValues gcv, prev_gcv;

			XGetGCValues(dpy, gc, vm, &prev_gcv);
			gcv.cap_style = CapNotLast;
			gcv.join_style = JoinBevel;   
			gcv.line_width = 1;
			gcv.line_style = LineOnOffDash;
			gcv.dash_offset = 0;
			gcv.dashes = 1;
			XChangeGC(dpy, gc, vm, &gcv);
#endif
			OSWSetForeground(
			    wglobal->force_mono ?
				gui->white_pix :
				wglobal->editable_text_pix
			);
			OSWDrawLine(
			    pixmap,
			    0, y,
			    (int)width - 1, y
			);
			OSWDrawLine(
			    pixmap,
			    (int)width - 1, y,
			    (int)width - 1, y + PULIST_ITEM_HEIGHT - 1
			);
			OSWDrawLine(
			    pixmap,
			    (int)width - 1, y + PULIST_ITEM_HEIGHT - 1,
			    0, y + PULIST_ITEM_HEIGHT - 1
			);
			OSWDrawLine(
			    pixmap,
			    0, y + PULIST_ITEM_HEIGHT - 1,
			    0, y
			);
#ifdef HAVE_X
			XChangeGC(dpy, gc, vm, &prev_gcv);
#endif
		    }
		    else
		    {
			OSWSetForeground(
			    wglobal->force_mono ?
				gui->white_pix :
				wglobal->editable_text_pix
			);
		    }
		    /* Is this item disabled? */
		    if(item->flags & PULIST_ITEM_FLAG_DISABLED)
			OSWSetForeground(
			    wglobal->force_mono ?
				gui->white_pix :
				wglobal->disabled_text_pix
			);

		    /* Draw the pixmap */
		    if(icon_pixmap != None)
		    {
			const int h = PULIST_ITEM_HEIGHT;
			x2 = x;
			y2 = y + ((h - (int)icon_height) / 2);
			OSWSetClipMask(icon_mask);
			OSWSetClipOrigin(x2, y2);
			OSWCopyDrawablesCoord(
			    pixmap, icon_pixmap,
			    x2, y2,
			    icon_width, icon_height,
			    0, 0
			);
			OSWSetClipMask(None);
		    }

		    /* Draw the text */
		    if(text != NULL)
		    {
			const int h = PULIST_ITEM_HEIGHT;
			x2 = x + (int)icon_width + icon_spacing;
			y2 = y + (h - font_height) / 2;
			OSWSetFont(font);
			OSWDrawString(
			    pixmap,
			    x2 - lbearing,
			    y2 + ascent,
			    text
			);
		    }

		    y += PULIST_ITEM_HEIGHT;
		    items_drawn++;
		}

		/* Draw scroll position? */
		if((total_items > nitems_visible) &&
		   (list->list_max_height > 0)
		)
		{
		    int	wb = SCROLLBAR_YBAR_WIDTH,
			hb = (int)height * nitems_visible / total_items,
			xb = (int)width - wb,
			yb = (int)height * y_scroll_pos /
			    (int)list->list_max_height;

		    /* Make sure that the height of the handle is
		     * atleast 3 pixels
		     */
		    if(hb < 3)
		    {
			yb -= (3 - hb) / 2;
			hb = 3;
		    }
		    if((yb + hb) > (int)height)
			yb = (int)height - hb;
		    if(yb < 0)
			yb = 0;
		    OSWSetForeground(
			wglobal->force_mono ?
			    gui->black_pix :
			    wglobal->scroll_bar_bg_pix
		    );
		    OSWDrawSolidRectangle(
			pixmap,
			xb, 0,
			(unsigned int)wb, height
		    );
		    OSWSetForeground(
			wglobal->force_mono ?
			    gui->white_pix :
			    wglobal->scroll_bar_handle_pix
		    );
		    OSWDrawSolidRectangle(
			pixmap,      
			xb, yb,
			(unsigned int)wb, (unsigned int)hb
		    );
		    OSWSetForeground(
			wglobal->force_mono ?
			    gui->white_pix :
			    wglobal->scroll_bar_frame_pix
		    );
		    OSWDrawRectangle(
			pixmap,
			xb, 0,
			(unsigned int)(wb - 1),
			(unsigned int)((int)height - 1)
		    );
		    OSWDrawRectangle(
			pixmap,
			xb, yb,
			(unsigned int)(wb - 1),
			(unsigned int)(hb - 1)
		    );
		}

		OSWPutDrawableToWindow(w, pixmap);
	    }
	}
}

/*
 *	Queues the popup list to draw.
 */
void PUListQueueDraw(popup_list_struct *list, int amount)
{
	if(list == NULL)
	    return;

	WidgetQueueDrawAction(PUListDrawCB, list, (void *)amount);
}

/*
 *	Manage the popup list event.
 */
int PUListManage(popup_list_struct *list, event_t *event)
{
	int events_handled = 0;
	keycode_t keycode;
	key_state_t key_state;
	keyval_t keyval;
	win_t w;
	event_key_t *key;
	event_button_t *button;
	event_motion_t *motion;

	if((list == NULL) || (event == NULL))
	    return(events_handled);

	w = event->xany.window;

	switch(event->type)
	{
	  case ConfigureNotify:
	    if(w == list->toplevel)
	    {
		if(list->map_state)
		    PUListQueueDraw(list, PULIST_DRAW_AMOUNT_LABEL);
		events_handled++;
	    }
	    break;

	  case Expose:
	    if(w == list->toplevel)
	    {
		PUListQueueDraw(list, PULIST_DRAW_AMOUNT_LABEL);
		events_handled++;
	    }
	    else if(w == list->popup_toplevel)
	    {
		PUListQueueDraw(list, PULIST_DRAW_AMOUNT_PULIST);
		events_handled++;
	    }
	    else if(w == list->popup_list)
	    {
		PUListQueueDraw(list, PULIST_DRAW_AMOUNT_PULIST);
		events_handled++;
	    }
	    break;

	  case KeyPress:
	    key = &event->xkey;
	    if(!list->has_focus || list->disabled)
		return(events_handled);

	    keycode = key->keycode;
	    key_state = key->state;
	    keyval = OSWGetKeyValFromKeyCode(keycode, key_state);

	    /* Up */
	    if((keyval == OSW_KEY_Up) ||
	       (keyval == OSW_KEY_KP_Up)
	    )
	    {
		const int prev_sel_item = list->sel_item;
		list->sel_item--;
		if(list->sel_item < 0)
		    list->sel_item = 0;
		if(prev_sel_item != list->sel_item)
		{
		    list->prev_sel_item = list->sel_item;

		    if(list->map_state)
			PUListQueueDraw(list, PULIST_DRAW_AMOUNT_LABEL);
		    if(list->popup_map_state)
			PUListQueueDraw(list, PULIST_DRAW_AMOUNT_PULIST);

		    /* Call select callback */
		    if(list->select_cb != NULL)
		    {
			list->select_cb(
			    list,
			    list->sel_item,
			    list->select_data
			);
			if(!WidgetIsRegistered(list))
			{
			    events_handled++;
			    return(events_handled);
			}
		    }
		}
		events_handled++;
	    }
	    /* Down */
	    else if((keyval == OSW_KEY_Down) ||
		    (keyval == OSW_KEY_KP_Down)
	    )
	    {
		const int prev_sel_item = list->sel_item;
		list->sel_item++;
		if(list->sel_item >= list->total_items)
		    list->sel_item = list->total_items - 1;
		if(list->sel_item < 0)
		    list->sel_item = 0;
		if(prev_sel_item != list->sel_item)
		{
		    list->prev_sel_item = list->sel_item;

		    if(list->map_state)
			PUListQueueDraw(list, PULIST_DRAW_AMOUNT_LABEL);
		    if(list->popup_map_state)
			PUListQueueDraw(list, PULIST_DRAW_AMOUNT_PULIST);

		    /* Call select callback */
		    if(list->select_cb != NULL)
		    {
			list->select_cb(
			    list,
			    list->sel_item,
			    list->select_data
			);
			if(!WidgetIsRegistered(list))
			{
			    events_handled++;
			    return(events_handled);
			}
		    }
		}
 		events_handled++;
	    }
	    /* Space or Enter */
	    else if((keyval == OSW_KEY_space) ||
		    (keyval == OSW_KEY_Return) ||
		    (keyval == OSW_KEY_KP_Enter) ||
		    (keyval == OSW_KEY_ISO_Enter)
	    )
	    {
		if(list->popup_map_state)
		    PUListUnmapList(list);
		else
		    PUListMapList(list);
		events_handled++;
	    }
	    break;

	  case KeyRelease:
	    key = &event->xkey;
	    if(!list->has_focus || list->disabled)
		return(events_handled);

	    break;

	  case ButtonPress:
	    button = &event->xbutton;
	    if(w == list->toplevel)
	    {
		if(!list->disabled)
		{
		    /* Set into focus regardless of which button was
		     * pressed
		     */
		    list->has_focus = True;

		    switch(button->button)
		    {
		      case Button1:	/* Map/unmap the popup list */
			if(list->popup_map_state)
			    PUListUnmapList(list);
			else
			    PUListMapList(list);
			events_handled++;
			break;
		    }
		}
	    }
	    break;

	  case ButtonRelease:
	    button = &event->xbutton;
	    if(w == list->popup_list)
	    {
		if(!list->disabled && (button->button == Button1))
		{
		    /* Button released at a different position from
		     * where it was when the list was mapped?
		     */
		    if((button->x_root != list->pointer_map_x) ||
		       (button->y_root != list->pointer_map_y)
		    )
		    {
			const int	x = button->x,
					y = button->y;
			unsigned int width, height;

			OSWGetWindowGeometry(w, NULL, NULL, &width, &height, NULL);

			/* Released on the list? */
			if((x >= 0) && (y >= 0) &&
			   (x < (int)width) && (y < (int)height)
			)
			{
			    /* Select the item if the item is not
			     * disabled, otherwise, restore the previous
			     * selected item
			     */
			    popup_list_item_struct *item = PUListGetItem(
				list, list->sel_item
			    );
			    if((item != NULL) ?
			       (item->flags & PULIST_ITEM_FLAG_DISABLED) : False
			    )
				list->sel_item = list->prev_sel_item;
			    else
				list->prev_sel_item = list->sel_item;

			    /* Unmap the popup list */
			    PUListUnmapList(list);

			    if(list->map_state)
				PUListQueueDraw(list, PULIST_DRAW_AMOUNT_LABEL);

			    /* Call the select callback */
			    if(list->select_cb != NULL)
			    {
				list->select_cb(
				    list,
				    list->sel_item,
				    list->select_data
				);
				if(!WidgetIsRegistered(list))
				{
				    events_handled++;
				    return(events_handled);
				}
			    }
			}
			else
			{
			    /* Released outside of the popup list,
			     * just unmap the popup list
			     */
			    PUListUnmapList(list);
			}

			events_handled++;
		    }

		    /* Reset the pointer map position after the first 
		     * button release   
		     */
		    list->pointer_map_x = 0;
		    list->pointer_map_y = 0;
		}
	    }
	    break;

	  case MotionNotify:
	    motion = &event->xmotion;
	    if(w == list->popup_list)
	    {
		const int	x = motion->x,
				y = motion->y;
		unsigned int width, height;

		OSWGetWindowGeometry(w, NULL, NULL, &width, &height, NULL);

		if((x >= 0) && (x < (int)width) &&
		   (y >= 0) && (y < (int)height)
		)
		{
		    const int prev_sel_item = list->sel_item;

		    list->sel_item = CLIP(
			((list->list_y_pos + y) / PULIST_ITEM_HEIGHT),
			0, list->total_items - 1
		    );

		    if(prev_sel_item != list->sel_item)
		    {
			if(list->map_state)
			    PUListQueueDraw(list, PULIST_DRAW_AMOUNT_LABEL);
			if(list->popup_map_state)
			    PUListQueueDraw(list, PULIST_DRAW_AMOUNT_PULIST);
		    }

		    events_handled++;
		}
	    }
	    break;

	  case EnterNotify:
	    if(w == list->popup_list)
	    {
		if(list->popup_map_state)
		{
		    /* Stop any scroll repeats */
		    PUListScrollRepeatStop(list);
		}
		events_handled++;
	    }
	    break;

	  case LeaveNotify:
	    if(w == list->popup_list)
	    {
		if(list->popup_map_state)
		{
		    /* Schedual scroll repeats */
		    PUListScrollRepeatStop(list);
		    PUListAddScrollTimeout(list, 0l);
		}
		events_handled++;
	    }
	    else if(list->popup_map_state)
	    {
		/* Since the popup list grabs the pointer,
		 * receiving this event means that other windows
		 * will have received LeaveNotify events
		 */
		push_button_struct *btn = list->map_btn;

		/* If the button received a LeaveNotify we need to make
		 * sure that it stays armed by intercepting this event
		 */
		if(w == btn->toplevel)
		{
		    PBtnSetState(btn, PBTN_ARMED);
		    events_handled++;
		}
	    }
	    break;
	}

	/* Map button */
	if(events_handled == 0)
	    events_handled += WidgetManage(list->map_btn, event);

	return(events_handled);
}

/*
 *	Maps the popup list label.
 */
void PUListMap(popup_list_struct *list)
{
	if(list == NULL)
	    return;

	if(!list->map_state)
	{
	    OSWMapRaised(list->toplevel);
	    list->map_state = True;
	    WidgetMap(list->map_btn);
	}
}

/*
 *	Unmaps the popup list label.
 */
void PUListUnmap(popup_list_struct *list)
{
	if(list == NULL)
	    return;

	PUListUnmapList(list);

	OSWUnmapWindow(list->toplevel);
	list->map_state = False;

	WidgetUnmap(list->map_btn);
}


/*
 *	Maps the popup list.
 */
void PUListMapList(popup_list_struct *list)
{
	int x, y, root_xw, root_yw;
	win_t w, parent, root;
	unsigned int width, height;
	osw_gui_struct *gui = &osw_gui[0];

	if(list == NULL)
	    return;

	/* Skip if already mapped */
	if(list->popup_map_state)
	    return;

	/* Record the position of pointer at time of mapping */
	w = list->popup_toplevel;
	OSWGetPointerCoords(
	    list->popup_toplevel,
	    &list->pointer_map_x, &list->pointer_map_y,
	    NULL, NULL
	);

	/* Calculate the position of the popup list to be mapped */
	parent = list->toplevel;
	root = gui->root_win;
	root_xw = 0;
	root_yw = 0;
	while((parent != root) && (parent != None))
	{
	    OSWGetWindowGeometry(parent, &x, &y, NULL, NULL, NULL);
	    root_xw += x;
	    root_yw += y;
	    parent = OSWGetWindowParent(parent);
	}
	w = list->popup_toplevel;
	OSWGetWindowGeometry(w, NULL, NULL, &width, &height, NULL);
	switch(list->direction)
	{
	  case PULIST_POPUP_UP:
	    root_yw = root_yw - (int)height + (int)list->height;
	    break;
	  case PULIST_POPUP_DOWN:
	    break;
	  case PULIST_POPUP_CENTER:
	    root_yw = root_yw - (((int)height - (int)list->height) / 2);
	    break;
	  default:      /* PULIST_POPUP_BELOW */
	    root_yw = root_yw + (int)list->height;
	    break;
	}
	if((root_xw + (int)width) > (int)gui->display_width)
	    root_xw = (int)gui->display_width - (int)width;
	if(root_xw < 0)
	    root_xw = 0;
	if((root_yw + (int)height) > (int)gui->display_height)
	    root_yw = (int)gui->display_height - (int)height;
	if(root_yw < 0)
	    root_yw = 0;
	OSWMoveWindow(w, root_xw, root_yw);

	/* Map list */
	list->popup_map_state = True;
	list->prev_sel_item = list->sel_item;
	OSWMapRaised(list->popup_toplevel);
	OSWMapWindow(list->popup_list);
	PUListQueueDraw(list, PULIST_DRAW_AMOUNT_PULIST);

	/* Ungrab pointer from window of button press that mapped us */
	if(OSWIsPointerGrabbed())
	    OSWUngrabPointer();
	OSWGrabPointer(
	    list->popup_list,
	    True,
	    ButtonReleaseMask | PointerMotionMask |
	    EnterWindowMask | LeaveWindowMask,
	    GrabModeAsync, GrabModeAsync,
	    None,
	    None
	);

	/* Set the map button to armed */
	PBtnSetState(list->map_btn, PBTN_ARMED);
}

/*
 *	Unmaps the popup list.
 */
void PUListUnmapList(popup_list_struct *list)
{
	if(list == NULL)
	    return;

	if(OSWIsPointerGrabbed())
	    OSWUngrabPointer();

	/* Stop any scroll repeats */
	PUListScrollRepeatStop(list);

	/* Unmap the list */
	OSWUnmapWindow(list->popup_toplevel);
	list->popup_map_state = False;

	/* Set the map button to unarmed */
	PBtnSetState(list->map_btn, PBTN_UNARMED);

	/* Destroy buffers */
	OSWDestroyPixmap(&list->popup_list_buf);
	OSWDestroyPixmap(&list->popup_toplevel_buf);
}

/*
 *	Destroy pulist.
 */
void PUListDestroy(popup_list_struct *list)
{
	if(list == NULL)
	    return;

	if(OSWIsPointerGrabbed())
	    OSWUngrabPointer();

	/* Stop any scroll timeouts */
	PUListScrollRepeatStop(list);

	/* Delete all items in list */
	PUListClear(list);

	OSWDestroyWindow(&list->popup_list);
	OSWDestroyPixmap(&list->popup_list_buf);

	OSWDestroyWindow(&list->popup_toplevel);
	OSWDestroyPixmap(&list->popup_toplevel_buf);

	WidgetDestroy(list->map_btn);
	free(list->map_btn);
	list->map_btn = NULL;

	OSWDestroyWindow(&list->toplevel);

	list->type = WIDGET_TYPE_UNKNOWN;
	list->map_state = False;
	list->has_focus = False;
	list->disabled = False;
	list->visibility_state = VisibilityFullyObscured;
	list->x = 0;
	list->y = 0;
	list->width = 0;
	list->height = 0;
	list->font = NULL;
	list->next = NULL;
	list->prev = NULL;

	list->popup_map_state = False;
	list->popup_list_vis_items = 0;

	list->list_max_width = 0;
	list->list_max_height = 0;
	list->list_y_pos = 0;

	list->direction = PULIST_POPUP_CENTER;

	list->select_cb = NULL;
	list->select_data = NULL;
}
