#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "../include/disk.h"

#include "../include/widget.h"

#include "../include/oswkeyvals.h"


static int ViewerCountLines(char *buf, int len);
int ViewerUpdateMaxWidth(file_viewer_struct *fv);

int ViewerLoadFile( 
	file_viewer_struct *fv, const char *filename
);
int ViewerLoadData(
	file_viewer_struct *fv,
	const void *buf, int buf_len
);
void ViewerUnload(file_viewer_struct *fv);

int ViewerInit(
	file_viewer_struct *fv,
	int x, int y,
	unsigned int width, unsigned int height
);
void ViewerResize(file_viewer_struct *fv);
void ViewerDraw(file_viewer_struct *fv);
int ViewerManage(file_viewer_struct *fv, event_t *event);
void ViewerMap(file_viewer_struct *fv);
void ViewerUnmap(file_viewer_struct *fv);
void ViewerDestroy(file_viewer_struct *fv);


#ifndef MAX
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#endif


/* Do not view files larger than this size (in bytes) */
#define VIEWER_MAX_FILESIZE	100000


/*
 *	Size constants (in pixels):
 */
#define VIEWER_CHAR_WIDTH	7
#define VIEWER_CHAR_HEIGHT	14

#define VIEWER_LINE_SPACING	2

#define VIEWER_MARGIN		10
#define VIEWER_DEF_WIDTH	((90 * VIEWER_CHAR_WIDTH) + (2 * VIEWER_MARGIN))
#define VIEWER_DEF_HEIGHT	((26 * VIEWER_CHAR_HEIGHT) + 100)

#define VIEWER_MIN_WIDTH	336
#define VIEWER_MIN_HEIGHT	240


/* Loading chunk size */
#define FV_LOAD_CHUNK_SIZE      1024


/*
 *	Get number of lines in buf.
 */
static int ViewerCountLines(char *buf, int len)
{
	int i, lines;

	if((buf == NULL) || (len < 1))
	    return(0);

	for(lines = 0, i = 0; i < len; i++)
	{
	    if((buf[i] == '\r') ||
	       (buf[i] == '\n')
	    )
		lines++;
	}

	return(lines);
}


/*
 *	Updates the maximum width of the data in the viewer
 *	for fv's member max_viewer_width.
 */
int ViewerUpdateMaxWidth(file_viewer_struct *fv)
{
	int i, line_len, longest_line;
	win_attr_t wattr;

	if(fv == NULL)
	    return(-1);

	OSWGetWindowAttributes(fv->viewer, &wattr);
	if(fv->buf == NULL)
	{
	    fv->max_viewer_width = wattr.width;
	    return(0);
	}


	/* Calculate longest line */
	line_len = 0;
	longest_line = 0;
	for(i = 0; i < fv->buf_len; i++)
	{
	    /* Carrage return? */
	    if((fv->buf[i] == '\n') ||
	       (fv->buf[i] == '\r')
	    )
	    {
		/* Was this line longer than previous? */
		if(line_len > longest_line)
		    longest_line = line_len;
		line_len = 0;
	    }

	    line_len++;
	}

	/* Update max_viewer_width */
	if(fv->viewer_mode == VIEWER_MODE_HEX)
	    fv->max_viewer_width = (80 * VIEWER_CHAR_WIDTH) +
		(2 * VIEWER_MARGIN) + 32;
	else
	    fv->max_viewer_width = (longest_line * VIEWER_CHAR_WIDTH) +
		(2 * VIEWER_MARGIN) + 32;

	return(0);
}


/*
 *	Get buffer position corresponding to line number.
 */
int GET_LINE_POSITION(char *buf, int len, int line)
{
	int i, lines;

	if((buf == NULL) || (len < 1) || (line < 1))
	    return(0);

	lines = 0;
	for(i = 0; i < len; i++)
	{
	    if(lines >= line)
		break;
	    if(buf[i] == '\0')
		break;

	    if((buf[i] == '\r') || (buf[i] == '\n'))
		lines++;
	}

	/* Sanitize i */
	if(i >= len)
	    i = len - 1;

	return(i);
}


/*
 *	Get length of line number from pos in buf.
 */
int GET_LINE_LENGTH(char *buf, int pos, int len)
{
	int i;
	int len_check;
	int line_length;


	if((buf == NULL) ||
	   (pos < 0) ||
	   (len < 1)
	)
	    return(0);


	/* Calculate length check */
	len_check = len - pos;
	if(len_check < 0)
	    len_check = 0;
 

	line_length = 0;
	for(i = 0; i < len_check; i++)
	{
	    if((buf[i] == '\0') ||
	       (buf[i] == '\r') ||
	       (buf[i] == '\n') 
	    )
		break;

	    line_length++;
	}
	if(line_length >= len)
	    line_length = len - 1;


	return(line_length);
}


/*
 *	Loads the file into the viewer.
 */
int ViewerLoadFile(
	file_viewer_struct *fv, const char *filename
)
{
	off_t filesize;
	off_t memory;
	int i, buf_pos;

	FILE *fp;
	char stringa[256 + PATH_MAX + NAME_MAX];
	char *strptr;
	char need_hex_mode = 0;
	unsigned int width, height;

	if(fv == NULL)
	    return(-1);

	/* Unload */
	ViewerUnload(fv);

	/* Attempt to open file */
	if(filename == NULL)
	{
	    strptr = "Cannot open file with NULL filename.\n";
	    fv->buf = (char *)calloc(1, (strlen(strptr) + 1) * sizeof(char));
	    if(fv->buf != NULL)
	    {
		fv->buf_len = strlen(strptr);
		strcpy(fv->buf, strptr);
		fv->lines = ViewerCountLines(fv->buf, fv->buf_len);
	    }

	    OSWSetWindowTitle(fv->toplevel, "Viewer: Error");

	    return(-2);
	}

	fp = fopen(filename, "rb");
	if(fp == NULL)
	{
	    sprintf(stringa, "Cannot read file: %s\n", filename);
	    strptr = stringa;
	    fv->buf = (char *)calloc(1, (strlen(strptr) + 1) * sizeof(char));
	    if(fv->buf != NULL)
	    {
		fv->buf_len = strlen(strptr);
		strcpy(fv->buf, strptr);
		fv->lines = ViewerCountLines(fv->buf, fv->buf_len);
	    }
		
	    OSWSetWindowTitle(fv->toplevel, "Viewer: Error");

	    return(-1);
	}


	/* Reset values */
	filesize = 0;
	memory = 0;
	buf_pos = 0;

	fv->buf = NULL;		/* Buffer already freed but setting NULL just incase */
	fv->buf_len = 0;
	fv->lines = 0;

	/* Begin reading file */
	while(1)
	{
	    /* Allocate more memory */
	    if(memory <= filesize)
	    {
		memory += FV_LOAD_CHUNK_SIZE;
		fv->buf_len = memory;

		fv->buf = (char *)realloc(fv->buf, fv->buf_len * sizeof(char));
		if(fv->buf == NULL)
		{
		    fv->buf_len = 0;
		    fv->lines = 0;
		    return(-1);
		}
	    }

	    /* Read character */
	    i = fgetc(fp);
	    if(i == EOF)
		break;
	    else
	        fv->buf[buf_pos] = i;

	    /* Check if char is non-ASCII */
	    if(!need_hex_mode)
	    {
		if(!isascii(i))
		    need_hex_mode = 1;
	    }


	    /* Increment values */
	    filesize++;
	    buf_pos++;
	}


	/* Close the file */
	fclose(fp);
	fp = NULL;


	/* Set exact size, and NULL terminate */
	fv->buf_len = filesize;
	fv->buf = (char *)realloc(fv->buf, (fv->buf_len + 1) * sizeof(char));
	if(fv->buf == NULL)
	{
	    fv->buf_len = 0;
	    return(-1);
	}
	fv->buf[fv->buf_len] = '\0';	/* Null terminate */


	/* Count lines */
	fv->lines = ViewerCountLines(fv->buf, fv->buf_len);

	/* Check if hex mode is needed */
	if(need_hex_mode)
	{
	    fv->viewer_mode = VIEWER_MODE_HEX;
	}

	/* Change title */
	sprintf(stringa, "Viewer: %s (%i bytes)\n",
	    filename,
	    (int)fv->buf_len
	);
	OSWSetWindowTitle(fv->toplevel, stringa);

	/* Update max viewer width */
	ViewerUpdateMaxWidth(fv);

	/* Update scroll bars */
	OSWGetWindowGeometry(
	    fv->viewer,
	    NULL, NULL, &width, &height, NULL
	);
	if(fv->viewer_mode == VIEWER_MODE_HEX)
	    SBarQueueResize(
		&fv->sb,
		width, height,
		fv->max_viewer_width,
		((fv->buf_len / 16) + 5) *
		    (VIEWER_CHAR_HEIGHT + VIEWER_LINE_SPACING)
	    );
	else
	    SBarQueueResize(
		&fv->sb,
		width, height,
		fv->max_viewer_width,
		(fv->lines + 5) *
		    (VIEWER_CHAR_HEIGHT + VIEWER_LINE_SPACING)
	    );

	return(0);
}


/*
 *	Load data into file voewer (this function is not complete).
 */
int ViewerLoadData(
	file_viewer_struct *fv,
	const void *buf, int buf_len
)
{
	if((fv == NULL) ||
	   (buf == NULL) ||
	   (buf_len < 1)
	)
	    return(-1);

	/* Unload (free) any contents in viewer first */
	ViewerUnload(fv);

/* TODO */

	return(0);
}


/*
 *	Unload data from file viewer.
 */
void ViewerUnload(
	file_viewer_struct *fv
)
{
	unsigned int width, height;

	if(fv == NULL)
	    return;

	free(fv->buf);
	fv->buf = NULL;

	fv->buf_len = 0;
	fv->lines = 0;

	/* Change title */
	OSWSetWindowTitle(fv->toplevel, "Viewer: Untitled");

	/* Update max viewer width */
	ViewerUpdateMaxWidth(fv);

	/* Update scroll bars */
	OSWGetWindowGeometry(
	    fv->viewer,
	    NULL, NULL, &width, &height, NULL
	);
	if(fv->viewer_mode == VIEWER_MODE_HEX)
	    SBarQueueResize(
		&fv->sb,
		width, height,
		fv->max_viewer_width,
		((fv->buf_len / 16) + 5) *
		    (VIEWER_CHAR_HEIGHT + VIEWER_LINE_SPACING)
	    );
	else  
	    SBarQueueResize(
		&fv->sb,
		width, height,
		fv->max_viewer_width,
		(fv->lines + 5) *
		    (VIEWER_CHAR_HEIGHT + VIEWER_LINE_SPACING)
	    );
}


/*
 *	Initializes file viewer.
 */
int ViewerInit(
	file_viewer_struct *fv,
	int x, int y,
	unsigned int width, unsigned int height
)
{
	keyval_t accel_keys[4];
	key_state_t accel_mods[4];
	push_button_struct *btn;
	osw_gui_struct *gui = &osw_gui[0];
	win_t w, parent = gui->root_win;
	widget_global_struct *wglobal = &widget_global;

	if(fv == NULL)
	    return(-2);

	if(width == 0)
	    width = VIEWER_DEF_WIDTH;
	if(height == 0)
	    height = VIEWER_DEF_HEIGHT;

	fv->type = WIDGET_TYPE_VIEWER;
	fv->map_state = False;
	fv->has_focus = False;
	fv->disabled = False;
	fv->visibility_state = VisibilityFullyObscured;
	fv->x = x;
	fv->y = y;
	fv->width = width;
	fv->height = height;
	fv->font = wglobal->std_font;
	fv->next = NULL;
	fv->prev = NULL;
	fv->data = NULL;

	/* Create toplevel */
	x = fv->x;
	y = fv->y;
	width = fv->width;
	height = fv->height;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	fv->toplevel = w;
	OSWSetWindowInput(w, OSW_EVENTMASK_TOPLEVEL);
	WidgetCenterWindow(w, WIDGET_WINDOW_RELATIVE_ROOT);

	/* WM properties */
	OSWSetWindowWMProperties(
	    w,
	    "Viewer",			/* Title */
	    "Viewer",			/* Icon title */
	    wglobal->std_icon_pm,	/* Icon */
	    False,			/* Let WM set coordinates? */
	    x, y,
	    /* Min width and height */
	    VIEWER_MIN_WIDTH, VIEWER_MIN_HEIGHT,
	    /* Max width and height */
	    gui->display_width, gui->display_height,
	    OSW_WINDOW_FRAME_STANDARD,
	    0, NULL
	);
	parent = w;

	/* Viewer */
	x = 0;
	y = 0;
	width = fv->width;
	height = MAX((int)fv->height - 28 - (2 * VIEWER_MARGIN), 10);
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	fv->viewer = w;
	OSWSetWindowInput(
	    w,
	    ExposureMask | ButtonPressMask | ButtonReleaseMask |
	    PointerMotionMask
	);

	if(OSWCreatePixmap(
	    &fv->viewer_buf, width, height
	))
	    return(-1);

	/* Scroll bars */
	SBarInit(
	    &fv->sb, w,
	    width, height,
	    0, 0
	);

	/* Push buttons */
	btn = &fv->close_btn;
	width = 70;
	height = 28;
	x = VIEWER_MARGIN;
	y = (int)fv->height - (int)height - VIEWER_MARGIN;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Close",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    NULL, fv
	);
	accel_keys[0] = OSW_KEY_c;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	accel_keys[1] = OSW_KEY_w;
	accel_mods[1] = OSW_KEY_STATE_CTRL;
	accel_keys[2] = OSW_KEY_Escape;
	accel_mods[2] = 0;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 3);

	/* ASCII button */
	btn = &fv->ascii_mode_btn;
	x = (int)fv->width - (2 * (int)width) - (2 * VIEWER_MARGIN);
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "ASCII",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    NULL, fv
	);
	accel_keys[0] = OSW_KEY_a;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 1);
	PBtnSetTooltip(
	    btn,
	    "Display in ASCII text format."
	);

	/* Hex */
	btn = &fv->hex_mode_btn;
	x = (int)fv->width - (1 * (int)width) - (1 * VIEWER_MARGIN);
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Hex",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    NULL, fv
	);
	accel_keys[0] = OSW_KEY_h;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 1);
	PBtnSetTooltip(
	    btn,
	    "Display in hexadecimal format."
	);


	fv->filename = NULL;
	fv->buf = NULL;
	fv->buf_len = 0;

	fv->viewer_mode = VIEWER_MODE_ASCII;

	ViewerUpdateMaxWidth(fv);

	/* Add this widget to the widget regeristry */
	WidgetRegister(fv, WIDGET_TYPE_VIEWER);

	return(0);
}

/*
 *	Resizes the file viewer.
 */
void ViewerResize(file_viewer_struct *fv)
{
	win_attr_t wattr;

	if(fv == NULL)
	    return;

	/* Get new sizes */
	OSWGetWindowAttributes(fv->toplevel, &wattr);

	/* Check for no change in size */
	if((fv->width == (unsigned int)wattr.width) &&
	   (fv->height == (unsigned int)wattr.height)
	)
	    return;

	fv->x = wattr.x;
	fv->y = wattr.y;
	fv->width = wattr.width;
	fv->height = wattr.height;


	/* Viewer */
	OSWMoveResizeWindow(
	    fv->viewer,
	    0, 0,
	    fv->width,
	    MAX((int)fv->height - 28 - (2 * VIEWER_MARGIN), 10)
	);
	OSWGetWindowAttributes(fv->viewer, &wattr);

	OSWDestroyPixmap(&fv->viewer_buf);
	if(OSWCreatePixmap(
	    &fv->viewer_buf,
	    wattr.width, wattr.height
	))
	    return;

	ViewerUpdateMaxWidth(fv);

	/* Scroll bars */
	if(fv->viewer_mode == VIEWER_MODE_HEX)
	    SBarResize(
		&fv->sb,
		wattr.width, wattr.height,
		fv->max_viewer_width,
		((fv->buf_len / 16) + 5) * 
		    (VIEWER_CHAR_HEIGHT + VIEWER_LINE_SPACING)
	    );
	else  
	    SBarResize(
		&fv->sb,
		wattr.width, wattr.height,
		fv->max_viewer_width,
		(fv->lines + 5) *
		    (VIEWER_CHAR_HEIGHT + VIEWER_LINE_SPACING)
	    );

	/* Buttons */
	OSWMoveWindow(fv->close_btn.toplevel,
	    VIEWER_MARGIN,
	    (int)fv->height - 28 - VIEWER_MARGIN
	);

	OSWMoveWindow(fv->ascii_mode_btn.toplevel,
	    (int)fv->width - (2 * 70) - (2 * VIEWER_MARGIN),
	    (int)fv->height - 28 - VIEWER_MARGIN
	);

	OSWMoveWindow(fv->hex_mode_btn.toplevel,
	    (int)fv->width - (1 * 70) - (1 * VIEWER_MARGIN),
	    (int)fv->height - 28 - VIEWER_MARGIN
	);
}

/*
 *	Draws the file viewer.
 */
void ViewerDraw(file_viewer_struct *fv)
{
	win_attr_t wattr_toplevel, wattr;

	int i, n, y_pos, start_line;
	int line_count, lines_drawn, lines_visible, line_length;
	int hex_col_count;

	char stringa[256];
	char stringb[64];
	widget_global_struct *wglobal = &widget_global;

	if(fv == NULL)
	    return;

	/* Get attributes of toplevel */
	OSWGetWindowAttributes(fv->toplevel, &wattr_toplevel);
	OSWGetWindowAttributes(fv->viewer, &wattr);

	/* Map as needed */
	if(!fv->map_state)
	{
	    OSWMapRaised(fv->toplevel);
	    fv->map_state = True;
	    fv->visibility_state = VisibilityUnobscured;
	    OSWMapWindow(fv->viewer);
	    SBarDraw(&fv->sb);
	    WidgetMap(&fv->close_btn);
	    WidgetMap(&fv->ascii_mode_btn);
	    WidgetMap(&fv->hex_mode_btn);
	}

	OSWSetFont(fv->font);

	if(fv->viewer_buf == None)
	{
	    OSWGetWindowAttributes(fv->viewer, &wattr);
	    OSWCreatePixmap(
		&fv->viewer_buf, wattr.width, wattr.height
	    );
	}

	/* Toplevel */
	OSWClearWindow(fv->toplevel);

	/* Viewer */
	OSWGetWindowAttributes(fv->viewer, &wattr);
	OSWClearPixmap(
	    fv->viewer_buf, wattr.width, wattr.height,
	    wglobal->surface_editable_pix
	);

	OSWSetForeground(wglobal->editable_text_pix);

	/* Draw data for ASCII mode */
	if((fv->buf != NULL) && (fv->viewer_mode == VIEWER_MODE_ASCII))
	{
	    lines_visible = (int)wattr.height /
		(VIEWER_CHAR_HEIGHT + VIEWER_LINE_SPACING);
	    start_line = (int)fv->sb.y_win_pos /
		(VIEWER_CHAR_HEIGHT + VIEWER_LINE_SPACING);
	    lines_drawn = 0;

	    y_pos = 0;
	    for(line_count = start_line;
		line_count < fv->lines;
		line_count++
	    )
	    {
		if(lines_drawn > lines_visible)
		    break;

	        i = GET_LINE_POSITION(fv->buf, fv->buf_len, line_count);
		line_length = GET_LINE_LENGTH(&(fv->buf[i]), i, fv->buf_len);

		OSWDrawStringLimited(
		    fv->viewer_buf,
		    VIEWER_MARGIN - fv->sb.x_win_pos, y_pos + 18,
		    &(fv->buf[i]), line_length
		);


		lines_drawn++;
		y_pos += VIEWER_CHAR_HEIGHT + VIEWER_LINE_SPACING;
	    }
	}
	/* Draw data for hexadecimal mode */
	if((fv->buf != NULL) && (fv->viewer_mode == VIEWER_MODE_HEX))
	{
	    lines_visible = (int)wattr.height /
		(VIEWER_CHAR_HEIGHT + VIEWER_LINE_SPACING);
	    start_line = (int)fv->sb.y_win_pos /
		(VIEWER_CHAR_HEIGHT + VIEWER_LINE_SPACING);

	    lines_drawn = 0;
	    y_pos = 0;

	    /* Draw each line */
	    for(i = start_line * 16;
		i < (int)fv->buf_len;
		i += 16
	    )
	    {
		/* Do not draw more lines that there are visable */
		if(lines_drawn > lines_visible) break;

		/* Address (line number) */
		sprintf(stringa, "%.8x  ", i);

		/* Hexadecimal colums section */
		for(hex_col_count = 0; hex_col_count < 16; hex_col_count++)
		{
		    n = i + hex_col_count;

		    /* Do not exceed buffer length */
		    if(n >= (int)fv->buf_len) break;

		    /* Hex value */
		    sprintf(stringb, "%.2x ", (unsigned char)fv->buf[n]);

		    /* Strcat stringb containing value to line stringa */
		    strcat(stringa, stringb);

		    /* Strcat delimitations */
		    if(hex_col_count == 3)
			strcat(stringa, " ");
		    if(hex_col_count == 7)
			strcat(stringa, "-- ");
		    if(hex_col_count == 11)
			strcat(stringa, " ");
		}
		OSWDrawString(
		    fv->viewer_buf,
		    VIEWER_MARGIN - fv->sb.x_win_pos, y_pos + 18,
		    stringa
		);

		/* ASCII equivilent section */
		stringa[0] = '\0';
		for(hex_col_count = 0; hex_col_count < 16; hex_col_count++)
		{
		    n = i + hex_col_count;

		    /* Do not exceed buffer length */
		    if(n >= (int)fv->buf_len) break;

		    /* ASCII value */
		    if((0x20 <= (unsigned char)fv->buf[n]) &&
		       ((unsigned char)fv->buf[n] < 0x80)
		    )
		        sprintf(stringb, "%c", (char)fv->buf[n]);
		    else
			sprintf(stringb, " ");

		    /* Strcat stringb containing value to line stringa */
		    strcat(stringa, stringb);
		}
		OSWDrawString(
		    fv->viewer_buf,
		    VIEWER_MARGIN - fv->sb.x_win_pos +
		    (VIEWER_CHAR_WIDTH * 65), y_pos + 18,
		    stringa
		);


		/* Increment positions */
		lines_drawn++;
		y_pos += VIEWER_CHAR_HEIGHT + VIEWER_LINE_SPACING;
	    }
	}

	/* Frame viewer */
	WidgetFrameButtonPixmap(
	    fv->viewer_buf,
	    True,
	    wattr.width,
	    wattr.height,
	    wglobal->surface_highlight_pix,
	    wglobal->surface_shadow_pix
	);

	OSWPutDrawableToWindow(fv->viewer, fv->viewer_buf);
}

/*
 *	Manages the file viewer event.
 */
int ViewerManage(file_viewer_struct *fv, event_t *event)
{
	int events_handled = 0;
	time_t t;
	keycode_t keycode;
	keyval_t keyval;
	event_key_t *key;
	win_t w;

	if((event == NULL) || (fv == NULL))
	    return(events_handled);

	w = event->xany.window;

	switch(event->type)
	{
	  case ConfigureNotify:
	    if((w == fv->toplevel) ||
	       (w == fv->viewer)
	    )
	    {
		ViewerResize(fv);
		WidgetQueueDraw(&fv->sb);
		events_handled++;
	    }   
	    break;

	  case Expose:
	    if((w == fv->toplevel) ||
	       (w == fv->viewer)
	    )
	    {
		events_handled++;
	    }
	    break;

	  case KeyPress:
	    if(!fv->has_focus)
		return(events_handled);

	    key = &event->xkey;
	    keycode = key->keycode;
	    keyval = OSWGetKeyValFromKeyCode(keycode, key->state);

	    /* Space */
	    if(keyval == OSW_KEY_space)
	    {
		scroll_bar_struct *sb = &fv->sb;

		sb->y_win_pos += (int)sb->height;
		if((sb->y_win_pos + (int)sb->height) > (int)sb->total_height)
		    sb->y_win_pos = (int)sb->total_height - (int)sb->height;
		if(sb->y_win_pos < 0)
		    sb->y_win_pos = 0;
		WidgetQueueDraw(sb);

		events_handled++;
	    }
	    /* B or minus */
	    else if((keyval == OSW_KEY_b) ||
		    (keyval == OSW_KEY_minus)
	    )
	    {
		scroll_bar_struct *sb = &fv->sb;

		sb->y_win_pos -= (int)sb->height;
		if((sb->y_win_pos + (int)sb->height) > (int)sb->total_height)
		    sb->y_win_pos = (int)sb->total_height - (int)sb->height;
		if(sb->y_win_pos < 0)
		    sb->y_win_pos = 0;
		WidgetQueueDraw(sb);

		events_handled++;
	    }
	    break;

	  case KeyRelease:
	    if(!fv->has_focus)
		return(events_handled);
	    break;

	  case ButtonPress:
	    if((w == fv->toplevel) ||
	       (w == fv->viewer)
	    )
		fv->has_focus = True;

	    /* Viewer window */
	    if(w == fv->viewer)
	    {
		scroll_bar_struct *sb = &fv->sb;
		sb->has_focus = True;
		events_handled++;
	    }
	    break;

	  case FocusIn:
	    if(w == fv->toplevel)
	    {
		fv->has_focus = True;
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case FocusOut:
	    if(w == fv->toplevel)
	    {
		fv->has_focus = False;
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case MapNotify:
	    if(w == fv->toplevel)
	    {
		if(!fv->map_state)
		{
		    fv->map_state = True;
		}
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case UnmapNotify:
	    if(w == fv->toplevel)
	    {
		if(fv->map_state)
		{
		    WidgetUnmap(&fv->close_btn);
		    WidgetUnmap(&fv->ascii_mode_btn);
		    WidgetUnmap(&fv->hex_mode_btn);

		    OSWDestroyPixmap(&fv->viewer_buf);

		    fv->map_state = False;
		}
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case VisibilityNotify: 
	    if(w == fv->toplevel)
	    {
		fv->visibility_state = event->xvisibility.state;
		events_handled++;
		return(events_handled);
	    }
	    break;

	  case ClientMessage:
	    if(OSWEventIsTakeFocus(fv->toplevel, event, &t))
	    {
		OSWSetInputFocus(fv->toplevel, RevertToParent, t);
		events_handled++;
		return(events_handled);
	    }
	    else if(OSWEventIsDeleteWindow(fv->toplevel, event))
	    {
		ViewerUnmap(fv);
		events_handled++;
		return(events_handled);
	    }
	    break;
	}

	if(events_handled > 0)
	    ViewerDraw(fv);

	/* Manage widgets */

	/* Scroll bar */
	if(events_handled == 0)
	{
	    events_handled += WidgetManage(&fv->sb, event);
	    if(events_handled > 0)
		ViewerDraw(fv);
	}

	/* Close button */
	if(events_handled == 0)
	{
	    events_handled += WidgetManage(
		&fv->close_btn, event
	    );
	    if(((event->type == ButtonRelease) ||
	       (event->type == KeyRelease)) &&
	       (events_handled > 0)
	    )
	    {
		ViewerUnmap(fv);
		return(events_handled);
	    }
	}

	/* ASCII mode button */
	if(events_handled == 0)
	{
	    events_handled += WidgetManage(&fv->ascii_mode_btn, event);
	    if(((event->type == ButtonRelease) ||
	       (event->type == KeyRelease)) &&
	       (events_handled > 0)
	    )
	    {
		scroll_bar_struct *sb = &fv->sb;

		fv->viewer_mode = VIEWER_MODE_ASCII;
		ViewerUpdateMaxWidth(fv);
		ViewerDraw(fv);

		SBarQueueResize(
		    sb,
		    sb->width, sb->height,
		    fv->max_viewer_width,
		    (fv->lines + 5) *
			(VIEWER_CHAR_HEIGHT + VIEWER_LINE_SPACING)
		);
		WidgetQueueDraw(sb);
	    }
	}

	/* Hex mode button */
	if(events_handled == 0)
	{
	    events_handled += WidgetManage(&fv->hex_mode_btn, event);
	    if(((event->type == ButtonRelease) ||
	       (event->type == KeyRelease)) &&
	       (events_handled > 0)
	    )
	    {
		scroll_bar_struct *sb = &fv->sb;

		fv->viewer_mode = VIEWER_MODE_HEX;
		ViewerUpdateMaxWidth(fv);
		ViewerDraw(fv);

		SBarQueueResize(
		    sb,
		    sb->width, sb->height,
		    fv->max_viewer_width,
		    ((fv->buf_len / 16) + 5) *
			(VIEWER_CHAR_HEIGHT + VIEWER_LINE_SPACING)
		);
		WidgetQueueDraw(sb);
	    }
	}

	return(events_handled);
}

/*
 *	Maps the file viewer.
 */
void ViewerMap(file_viewer_struct *fv)
{
	if(fv == NULL)
	    return;

	/* Map */
	fv->map_state = False;
	ViewerDraw(fv);

	/* Grab focus */
	OSWGrabFocus(fv->toplevel);
}

/*
 *	Unmaps the file viewer.
 */
void ViewerUnmap(file_viewer_struct *fv)
{
	if(fv == NULL)
	    return;

	OSWUnmapWindow(fv->toplevel);
}

/*
 *	Destroys the file viewer.
 */
void ViewerDestroy(file_viewer_struct *fv)
{
	if(fv == NULL)
	    return;

	ViewerUnload(fv);

	WidgetDestroy(&fv->sb);
	WidgetDestroy(&fv->close_btn);
	WidgetDestroy(&fv->ascii_mode_btn);
	WidgetDestroy(&fv->hex_mode_btn);

	OSWDestroyPixmap(&fv->viewer_buf);
	OSWDestroyWindow(&fv->viewer);
	OSWDestroyWindow(&fv->toplevel);

	fv->type = WIDGET_TYPE_UNKNOWN;
	fv->map_state = False;
	fv->has_focus = False;
	fv->disabled = False;
	fv->visibility_state = VisibilityFullyObscured;
	fv->x = 0;
	fv->y = 0;
	fv->width = 0;
	fv->height = 0;
	fv->font = NULL;
	fv->next = NULL;
	fv->prev = NULL;

	fv->viewer_mode = VIEWER_MODE_ASCII;
}
